#include "header.hpp"

#include "ppqq20.amplitude.set.hpp"

extern "C" {
  void setup_();
  void set_momenta_(double* P, double* mu_Q, double* m_HQ);
  void callmsqdgav_(double* P, double* m_HQ, double* msqDGav);
}

#ifdef OPENLOOPS

ppqq20_amplitude_OpenLoops_set::ppqq20_amplitude_OpenLoops_set(){
  Logger logger("ppqq20_amplitude_OpenLoops_set::ppqq20_amplitude_OpenLoops_set ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "constructor called..." << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

ppqq20_amplitude_OpenLoops_set::~ppqq20_amplitude_OpenLoops_set(){
  Logger logger("ppqq20_amplitude_OpenLoops_set::~ppqq20_amplitude_OpenLoops_set ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "destructor called..." << endline_flush;

  if (csi->type_contribution == "VT2"){
    delete [] m2cc;
    delete [] m2l1;
    delete [] m2loopcc;
    delete [] m2l1_im;
    delete [] m2loopcc_im;
    delete [] m2l2;
    delete [] loopcc_m2l1_im;
    delete [] loopcc_M2loopcc_im;
    delete [] loopcc_m2l1;
    delete [] loopcc_m2loopcc;
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

void ppqq20_amplitude_OpenLoops_set::initialization_process(){
  Logger logger("ppqq20_amplitude_OpenLoops_set::initialization_process");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  m_HQ = csi->mass_parton[0][3];
  m2_HQ = csi->mass2_parton[0][3];

  process_id = -1;
  if (csi->type_contribution == "VT2"){
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 11);
    // loop-squared amplitudes are needed to run the double-virtual result to m_inv:
    process_id_loopxloop = register_OL_subprocess(0, 12);
    if (process_id_loopxloop == -1){process_id = -1;}
    // imaginary parts of loop*Born amplitudes (and colour correlators) are needed to run the double-virtual result to m_inv:
    munich_ol_setparameter_string("approx", "imag", logger);
    process_id_imag = register_OL_subprocess(0, 11);
    if (process_id_imag == -1){process_id = -1;}
    initialization_grid();
    // polenorm != 0 is not officially supported, and for VT2, no shift is implemented.
    munich_ol_setparameter_int("polenorm", 0, logger);
  }
  else if (csi->type_contribution == "CT"){
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 11);
  }
  else {amplitude_OpenLoops_set::initialization_process();}

  if (csi->coll_choice != 0 && (csi->mass_parton[0][1] != 0. || csi->mass_parton[0][2] != 0.)){
    logger << LOG_WARN << "Incoming parton are not allowed for hadron collisions." << endline_flush;
    process_id = -1;
  }

  if (csi->type_contribution == "VT" || csi->type_contribution == "VT2" || csi->type_contribution == "CT2"){
    // Do not split off the counterterm in general for all qTsubtraction contributions
    //    ol_setparameter_int(OL_ct_on, 1);
    munich_ol_setparameter_int("CT_on", 1, logger);
    // Should not be required, aleady initialized as 0
    osi->VA_X_ME2 = 0.;
  }

  if (csi->type_contribution == "VT2" || csi->type_contribution == "CT2"){
    // initialization for Hayk's calculation of DG and fourcolourcorrelators (etc.)
    setup_();
  }

  if (csi->type_contribution == "VT2"){
    // initialization for required amplitudes
    m0_OL = 0.;
    m2cc = new double[16];
    m2l1 = new double[3];
    m2loopcc = new double[16];
    m2l1_im = new double[3];
    m2loopcc_im = new double[16];
    m2l2 = new double[5];
    loopcc_m2l1_im = new double[3];
    loopcc_M2loopcc_im = new double[n_cc];
    loopcc_m2l1 = new double[3];
    loopcc_m2loopcc = new double[n_cc];
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_OpenLoops_set::calculate_H2_2loop_ingredients(){
  static Logger logger("ppqq20_amplitude_OpenLoops_set::calculate_H2_ingrdients");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  set_phasespacepoint();

  ///  int CT_on = 1; // modification of ...ME2_VA and OpenLoops calls needed !!!
  ///  ol_setparameter_int(stch("ct_on"), CT_on); // modification of ...ME2_VA and OpenLoops calls needed !!!
  set_delta_and_mu_Q();
  set_momenta_(P_OpenLoops, &mu_Q, &m_HQ);

  double dummy = 0.;
  ol_evaluate_ccmatrix(process_id, P_OpenLoops, &m0_OL, m2cc, &acc);
  osi->VA_b_ME2 = m0_OL;
  ol_evaluate_loopccmatrix(process_id, P_OpenLoops, &dummy, m2l1, m2loopcc, &acc);
  osi->VA_V_ME2 = m2l1[0];
  osi->VA_X_ME2 = 0.;
  
  ol_evaluate_loop2(process_id_loopxloop, P_OpenLoops, m2l2, &acc);

  ol_evaluate_loopccmatrix(process_id_imag, P_OpenLoops, &dummy, m2l1_im, m2loopcc_im, &acc);
  for (int i = 0; i < 16; i++){imM1M0cc[i % 4 + 1][i / 4 + 1] = m2loopcc_im[i];}
  ewcc = 0.; // needed ???
  ol_evaluate_loopcc(process_id_imag, P_OpenLoops, &dummy, loopcc_m2l1_im, loopcc_M2loopcc_im, &ewcc);
  for (int i = 1; i < 5; i++){
    for (int j = 1; j < i + 1; j++){
      if (i == j && i < 3){
	imM1M0cc[i][j] = Ci * loopcc_m2l1_im[0];
      }
      else if (i == j && i > 2){
	imM1M0cc[i][j] = C_F * loopcc_m2l1_im[0];
      }
      else {
	imM1M0cc[i][j] = loopcc_M2loopcc_im[(i-1)*(i-2)/2+(j-1)];
	imM1M0cc[j][i] = loopcc_M2loopcc_im[(i-1)*(i-2)/2+(j-1)];
      }
    }
  }

  ol_evaluate_loopcc(process_id, P_OpenLoops, &dummy, loopcc_m2l1, loopcc_m2loopcc, &ewcc);
  for (int i = 1; i < 5; i++){
    for (int j = 1; j < i + 1; j++){
      if (i == j && i < 3){
	M1M0cc[i][j] = Ci * loopcc_m2l1[0];
      }
      else if (i == j && i > 2){
	M1M0cc[i][j] = C_F * loopcc_m2l1[0];
      }
      else {
	M1M0cc[i][j] = loopcc_m2loopcc[(i-1)*(i-2)/2+(j-1)];
	M1M0cc[j][i] = loopcc_m2loopcc[(i-1)*(i-2)/2+(j-1)];
      }
    }
  }

  M0M0 = m0_OL;
  M1M1 = m2l2[0];
  for (int i = 0; i < 16; i++){M0M0cc[i % 4 + 1][i / 4 + 1] = m2cc[i];}

  //C The M1*M0 + cc I can take from OpenLoops, also the cc ones:
  M1M0 = m2l1[0] / (osi->alpha_S * inv2pi);
  imM1M0 = loopcc_m2l1_im[0] / (osi->alpha_S * inv2pi);
  //C The M1*M1 I also take from OL
  M1M1 = M1M1 / pow(osi->alpha_S * inv2pi, 2); 
  for (int i = 1; i < 5; i++){
    for (int j = 1; j < 5; j++){
      M1M0cc[i][j] /= (osi->alpha_S * inv2pi);
      imM1M0cc[i][j] /= (osi->alpha_S * inv2pi);
    }
  }

  ///  CT_on = 0; // modification of ...ME2_VA and OpenLoops calls needed !!!
  ///  ol_setparameter_int(stch("ct_on"), CT_on); // modification of ...ME2_VA and OpenLoops calls needed !!!

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_OpenLoops_set::calculate_ME2_VT2_QCD_H0_doublespinflip(){
  static Logger logger("ppqq20_amplitude_OpenLoops_set::calculate_ME2_VT2_QCD_H0_doublespinflip");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double temp_ME2 = 0.;
  double temp_m2schsf = 0.;
  ol_evaluate_schsf(process_id, P_OpenLoops, &temp_ME2, 1, 2, &temp_m2schsf);
  osi->QT_H0_doublespinflip = temp_m2schsf / temp_ME2;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_OpenLoops_set::calculate_ME2_VT2_QCD_H0_DG(){
  static Logger logger("ppqq20_amplitude_OpenLoops_set::calculate_ME2_VT2_QCD_H0_DG");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // Specific version for QQx production (based on Hayk's code; not directly available from OpenLoops)
  osi->QT_H0_DG = 0.;
  callmsqdgav_(P_OpenLoops, &m_HQ, &osi->QT_H0_DG);
  osi->QT_H0_DG = osi->QT_H0_DG / osi->VA_b_ME2 * pow(osi->alpha_S, 2);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_OpenLoops_set::calculate_ME2_cf_twocolourcorrelators(){
  Logger logger("ppqq20_amplitude_OpenLoops_set::calculate_ME2_cf_twocolourcorrelators");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    if (osi->QT_correlationoperator[i_c].type_combination == 1 ||
	osi->QT_correlationoperator[i_c].type_combination == 4){
      osi->QT_ME2_cf[i_c] = osi->QT_correlationoperator[i_c].charge_factor;
    }
    else {
      osi->QT_ME2_cf[i_c] = M2cc[osi->QT_correlationoperator[i_c].no_BLHA_entry] / osi->ME2;
    }
    logger << LOG_DEBUG_POINT << "OpenLoops:  QT_ME2_cf[" << i_c << "] (em = " << osi->QT_correlationoperator[i_c].no_emitter << ", sp = " << osi->QT_correlationoperator[i_c].no_spectator << ") = " << setw(23) << setprecision(15) << osi->QT_ME2_cf[i_c] << "   charge_factor = " << setw(23) << setprecision(15) << osi->QT_correlationoperator[i_c].charge_factor << "   no_BLHA_entry = " << osi->QT_correlationoperator[i_c].no_BLHA_entry << endline_flush;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_OpenLoops_set::calculate_ME2_loopcf_twocolourcorrelators(){
  Logger logger("ppqq20_amplitude_OpenLoops_set::calculate_ME2_loopcf_twocolourcorrelators");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    if (osi->QT_correlationoperator[i_c].type_combination == 1 ||
	osi->QT_correlationoperator[i_c].type_combination == 4){
      osi->QT_ME2_loopcf[i_c] = osi->QT_correlationoperator[i_c].charge_factor * ME2loop / (osi->alpha_S / pi) / osi->ME2;
    }
    else {
      osi->QT_ME2_loopcf[i_c] = M2loopcc[osi->QT_correlationoperator[i_c].no_BLHA_entry] / (osi->alpha_S / pi) / osi->ME2;
    }
  }
  logger << LOG_DEBUG_POINT << "OL_checksum (1) = " << osi->QT_ME2_loopcf[0] + osi->QT_ME2_loopcf[1] + osi->QT_ME2_loopcf[2] + osi->QT_ME2_loopcf[3] << endline_flush;
  logger << LOG_DEBUG_POINT << "OL_checksum (2) = " << osi->QT_ME2_loopcf[1] + osi->QT_ME2_loopcf[4] + osi->QT_ME2_loopcf[5] + osi->QT_ME2_loopcf[6] << endline_flush;
  logger << LOG_DEBUG_POINT << "OL_checksum (3) = " << osi->QT_ME2_loopcf[2] + osi->QT_ME2_loopcf[5] + osi->QT_ME2_loopcf[7] + osi->QT_ME2_loopcf[8] << endline_flush;
  logger << LOG_DEBUG_POINT << "OL_checksum (4) = " << osi->QT_ME2_loopcf[3] + osi->QT_ME2_loopcf[6] + osi->QT_ME2_loopcf[8] + osi->QT_ME2_loopcf[9] << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_OpenLoops_set::calculate_ME2_VT_QCD(){
  Logger logger("ppqq20_amplitude_OpenLoops_set::calculate_ME2_VT_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (csi->class_contribution_loopinduced){logger << LOG_FATAL << "No explicit separate loop-induced contribution exisits in QQ production." << endline_flush; exit(1);}

  set_phasespacepoint();
  calculate_QT_kinematics();
  set_delta_and_mu_Q();
		   
  if (osi->initial_diag){
    ol_evaluate_full(1, P_OpenLoops, &osi->VA_b_ME2, M2L1, IRL1, M2L2, IRL2, &acc);
    osi->VA_V_ME2 = M2L1[0];
    ///    ol_evaluate_ct(process_id, P_OpenLoops, &M2L0, &osi->VA_X_ME2);
    ol_evaluate_cc(process_id, P_OpenLoops, &osi->ME2, M2cc, &ewcc);
    calculate_ME2_cf_twocolourcorrelators();
    calculate_H1_delta_QCD_from_VX_b();
    calculate_Ft1born();
    //  H1 -> H1 - It1  (It1 = 2 Re < M0 | It1 | M0 >)
    osi->QT_H1_delta_QCD += .5 * osi->Ft1born;
  }
  else {
    // New: H1 and Ft1born are not required for non-diagonal channels.
    ol_evaluate_cc(1, P_OpenLoops, &osi->ME2, M2cc, &ewcc);
    osi->VA_b_ME2 = osi->ME2;
    // Setting these variables to 0 should not be required at every point:
    osi->VA_V_ME2 = 0.;
    osi->VA_X_ME2 = 0.;
    osi->QT_H1_delta_QCD = 0.;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_OpenLoops_set::calculate_ME2_CT_QCD(){
  Logger logger("ppqq20_amplitude_OpenLoops_set::calculate_ME2_CT_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (csi->class_contribution_loopinduced){logger << LOG_FATAL << "No explicit separate loop-induced contribution exisits in QQ production." << endline_flush; exit(1);}

  set_phasespacepoint();
  calculate_QT_kinematics();
  // Why osi->value_ME2term[0], and not simply osi->VA_b_ME2 ???
  ol_evaluate_cc(process_id, P_OpenLoops, &(osi->value_ME2term[0]), M2cc, &ewcc);
  osi->ME2 = osi->value_ME2term[0];
  calculate_ME2_cf_twocolourcorrelators(); // osi->QT_correlationoperator[i_c].type_combination == 4 result changes, but does not contribute here here !!!
  if (osi->initial_pdf_diag){calculate_Gamma1born();}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_OpenLoops_set::calculate_ME2_VT2_QCD(){
  Logger logger("ppqq20_amplitude_OpenLoops_set::calculate_ME2_VT2_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (csi->class_contribution_loopinduced){logger << LOG_FATAL << "No explicit separate loop-induced contribution exists in QQ production." << endline_flush; exit(1);}

  set_phasespacepoint();
  calculate_QT_kinematics();
  
  if (osi->initial_pdf_diag || osi->initial_pdf_gq){
    // maybe exchange osi->ME2 -> osi->VA_b_ME2 , which is required later ???
    ol_evaluate_cc(process_id, P_OpenLoops, &osi->ME2, M2cc, &ewcc);
    calculate_ME2_cf_twocolourcorrelators();
  }

  // actually only required if 1-loop corrections are needed and not calculated together with the 2-loop amplitudes !!!
  set_delta_and_mu_Q();
  ol_evaluate_full(process_id, P_OpenLoops, &osi->VA_b_ME2, M2L1, IRL1, M2L2, IRL2, &acc);
  ///  ol_evaluate_ct(process_id, P_OpenLoops, &M2L0, &osi->VA_X_ME2);
  osi->VA_V_ME2 = M2L1[0];
  if (osi->initial_gg){set_momenta_(P_OpenLoops, &mu_Q, &m_HQ);}

  if (switch_off_H2 || !osi->initial_diag){
    osi->QT_A0 = osi->VA_b_ME2;
    osi->QT_A1 = osi->VA_V_ME2 + osi->VA_X_ME2;
    calculate_H1_delta_QCD_from_VX_b();
    //    osi->QT_H1_delta_QCD = osi->QT_A1 / 2 / (osi->alpha_S * inv2pi * osi->QT_A0);
    osi->QT_A2 = 0.;
    osi->QT_H2_delta_QCD = 0.;
    //osi->QT_H2_delta_QCD = osi->QT_A2 / 4 / (pow(osi->alpha_S * inv2pi, 2) * osi->QT_A0);
  }
  else {
    calculate_fourcolourcorrelators();
    calculate_H2_2loop_ingredients();
    calculate_H2_2loop();
  }

  if (osi->initial_gg){
    calculate_ME2_VT2_QCD_H0_doublespinflip();
    calculate_ME2_VT2_QCD_H0_DG();
  }

  if (osi->initial_pdf_diag || osi->initial_pdf_gq){
    calculate_Ft1born();
    calculate_Gamma1born();
    //  H1 -> H1 - It1  (It1 = 2 Re < M0 | It1 | M0 >)
    osi->QT_H1_delta_QCD += .5 * osi->Ft1born;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_OpenLoops_set::calculate_ME2_CT2_QCD(){
  Logger logger("ppqq20_amplitude_OpenLoops_set::calculate_ME2_CT2_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (csi->class_contribution_loopinduced){logger << LOG_FATAL << "No explicit separate loop-induced contribution exists in QQ production." << endline_flush; exit(1);}

  ///  ol_setparameter_int(OL_ct_on, 1);
  ///  osi->VA_X_ME2 = 0.;
  
  set_phasespacepoint();
  calculate_QT_kinematics();
  set_delta_and_mu_Q();

  // Calculate virtual amplitudes for osi->initial_diag together with loop correlators !!!
  if (osi->initial_pdf_gq && !osi->initial_diag){ // gq channels require only one-loop amplitude, but no corresponding correlators
    ol_evaluate_full(process_id, P_OpenLoops, &osi->VA_b_ME2, M2L1, IRL1, M2L2, IRL2, &acc);
    osi->VA_V_ME2 = M2L1[0];
    //    ol_evaluate_ct(process_id, P_OpenLoops, &M2L0, &osi->VA_X_ME2);
  }
  
  ol_evaluate_cc(process_id, P_OpenLoops, &osi->ME2, M2cc, &ewcc);
  if (!(osi->initial_diag || osi->initial_pdf_gq)){osi->VA_b_ME2 = osi->ME2;}
  calculate_ME2_cf_twocolourcorrelators();

  if (osi->initial_diag){
    //    ol_setparameter_int(OL_ct_on, 1);
    //    Just to check runtime of loopcc !!!
    //    ol_evaluate_full(process_id, P_OpenLoops, &ME2tree, M2L1, IRL1, M2L2, IRL2, &acc);
    ol_evaluate_loopcc(process_id, P_OpenLoops, &ME2tree, M2L1, M2loopcc, &ewcc);
    ME2loop = M2L1[0];
    osi->VA_b_ME2 = ME2tree;
    osi->VA_V_ME2 = M2L1[0];
    //    osi->VA_X_ME2 = 0.; // splitting off counterterm is not required
    //    ol_setparameter_int(OL_ct_on, 0);
    calculate_ME2_loopcf_twocolourcorrelators();
  }
  if (osi->initial_diag_gg){set_momenta_(P_OpenLoops, &mu_Q, &m_HQ);}
  
  calculate_H1_delta_QCD_from_VX_b();
  calculate_Ft1born();
  //  H1 -> H1 - It1  (It1 = 2 Re < M0 | It1 | M0 >)
  osi->QT_H1_delta_QCD += .5 * osi->Ft1born;

  if (osi->initial_diag){
    calculate_fourcolourcorrelators();
    calculate_Ft1born_4correlator();
    for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){osi->QT_ME2_loopcf[i_c] += .5 * osi->Ft1born_4correlator[i_c];}
    logger << LOG_DEBUG_POINT << "shifted: OL_checksum (1) = " << osi->QT_ME2_loopcf[0] + osi->QT_ME2_loopcf[1] + osi->QT_ME2_loopcf[2] + osi->QT_ME2_loopcf[3] << endline_flush;
    logger << LOG_DEBUG_POINT << "shifted: OL_checksum (2) = " << osi->QT_ME2_loopcf[1] + osi->QT_ME2_loopcf[4] + osi->QT_ME2_loopcf[5] + osi->QT_ME2_loopcf[6] << endline_flush;
    logger << LOG_DEBUG_POINT << "shifted: OL_checksum (3) = " << osi->QT_ME2_loopcf[2] + osi->QT_ME2_loopcf[5] + osi->QT_ME2_loopcf[7] + osi->QT_ME2_loopcf[8] << endline_flush;
    logger << LOG_DEBUG_POINT << "shifted: OL_checksum (4) = " << osi->QT_ME2_loopcf[3] + osi->QT_ME2_loopcf[6] + osi->QT_ME2_loopcf[8] + osi->QT_ME2_loopcf[9] << endline_flush;
    calculate_Gamma1loop();
    if (osi->initial_diag_gg){calculate_commutator_Gamma1_Ft1();}
    calculate_Gamma1_squared();
    calculate_Gamma2();
  }
  if (osi->initial_pdf_diag || osi->initial_pdf_gq){calculate_Gamma1born();}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

#endif
