#include "header.hpp"

#include "ppqq20.amplitude.set.hpp"

extern "C" {
  void setup_();
  void set_momenta_(double* P, double* mu_Q, double* m_HQ);
  void callmsqggav_(double* P, double* m_HQ, double* msqGGav);
  void callmsqdgav_(double* P, double* m_HQ, double* msqDGav);
}

#ifdef RECOLA

ppqq20_amplitude_Recola_set::ppqq20_amplitude_Recola_set(){
  Logger logger("ppqq20_amplitude_Recola_set::ppqq20_amplitude_Recola_set ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "constructor called..." << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

ppqq20_amplitude_Recola_set::~ppqq20_amplitude_Recola_set(){
  Logger logger("ppqq20_amplitude_Recola_set::~ppqq20_amplitude_Recola_set ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "destructor called..." << endline_flush;

  if (csi->type_contribution == "VT2" || csi->type_contribution == "VJ2"){
    delete [] P_OLformat;
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::initialization_process(){
  Logger logger("ppqq20_amplitude_Recola_set::initialization_process");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  m_HQ = csi->mass_parton[0][3];
  m2_HQ = csi->mass2_parton[0][3];

  process_id = -1;
  if (csi->type_contribution == "VT2" ||
      csi->type_contribution == "VJ2"){
    // needs to be implemented for Recola as well for 2-loop interface !!!
    amplitude_Recola_set::initialization_process();
    /*
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    logger << LOG_DEBUG << "order_ew   set to   " << csi->contribution_order_alpha_e << endline_flush;
    process_id = register_OL_subprocess(0, 11);
    // loop-squared amplitudes are needed to run the double-virtual result to m_inv:
    int process_id2 = register_OL_subprocess(0, 12);
    if (process_id2 == -1){process_id = -1;}
    // imaginary parts of loop*Born amplitudes (and colour correlators) are needed to run the double-virtual result to m_inv:
    munich_ol_setparameter_string("approx", "imag", logger);
    int process_id3 = register_OL_subprocess(0, 11);
    if (process_id3 == -1){process_id = -1;}
    */
    initialization_grid();
  }
  else {amplitude_Recola_set::initialization_process();}

  if (csi->type_contribution == "VT2" || csi->type_contribution == "CT2"){
    n_momentum = 5 * (csi->n_particle + 2);
    P_OLformat = new double[n_momentum];
    // initialization for Hayk's calculation of DG and fourcolourcorrelators (etc.)
    setup_();
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::set_phasespacepoint_OLformat(){
  static Logger logger("ppqq20_amplitude_Recola_set::set_phasespacepoint_OLformat");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i = 1; i < esi->p_parton[0].size(); i++){
    P_OLformat[5 * (i - 1)]     = esi->p_parton[0][i].x0();
    P_OLformat[5 * (i - 1) + 1] = esi->p_parton[0][i].x1();
    P_OLformat[5 * (i - 1) + 2] = esi->p_parton[0][i].x2();
    P_OLformat[5 * (i - 1) + 3] = esi->p_parton[0][i].x3();
    P_OLformat[5 * (i - 1) + 4] = esi->p_parton[0][i].m();
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::calculate_ME2_cf_twocolourcorrelators(){
  Logger logger("ppqq20_amplitude_Recola_set::calculate_ME2_cf_twocolourcorrelators");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double alpha_S_mu_Q = LHAPDF::alphasPDF(mu_Q);
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    if (osi->QT_correlationoperator[i_c].type_combination == 1 ||
	osi->QT_correlationoperator[i_c].type_combination == 4){
      osi->QT_ME2_cf[i_c] = osi->QT_correlationoperator[i_c].charge_factor;
    }
    else {
      get_colour_correlation_rcl(1, csi->order_alpha_s_born, osi->QT_correlationoperator[i_c].no_emitter, osi->QT_correlationoperator[i_c].no_spectator, osi->QT_ME2_cf[i_c]);
      osi->QT_ME2_cf[i_c] = osi->QT_ME2_cf[i_c] * pow(osi->alpha_S / alpha_S_mu_Q, csi->order_alpha_s_born) * osi->QT_correlationoperator[i_c].colour_factor / osi->ME2;
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::calculate_ME2_loopcf_twocolourcorrelators(){
  Logger logger("ppqq20_amplitude_Recola_set::calculate_ME2_loopcf_twocolourcorrelators");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double alpha_S_mu_Q = LHAPDF::alphasPDF(mu_Q);
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    if (osi->QT_correlationoperator[i_c].type_combination == 1 ||
	osi->QT_correlationoperator[i_c].type_combination == 4){
      osi->QT_ME2_loopcf[i_c] = osi->QT_correlationoperator[i_c].charge_factor * ME2loop / (osi->alpha_S / pi) / osi->ME2;
      logger << LOG_DEBUG_POINT << "QT_ME2_loopcf[" << i_c << "] (em = " << osi->QT_correlationoperator[i_c].no_emitter << ", sp = " << osi->QT_correlationoperator[i_c].no_spectator << ") = " << setw(23) << setprecision(15) << osi->QT_ME2_loopcf[i_c] * osi->VA_b_ME2 << "   no cc from Recola" << endline_flush;
    }
    else {
      // Instead, the RECOLA calls for colour-correlators
      get_colour_correlation_rcl(1, csi->contribution_order_alpha_s - 1, osi->QT_correlationoperator[i_c].no_emitter, osi->QT_correlationoperator[i_c].no_spectator, "NLO", osi->QT_ME2_loopcf[i_c]);
      osi->QT_ME2_loopcf[i_c] = osi->QT_ME2_loopcf[i_c] * pow(osi->alpha_S / alpha_S_mu_Q, csi->contribution_order_alpha_s - 1) * osi->QT_correlationoperator[i_c].colour_factor / (osi->alpha_S / pi) / osi->ME2;

      logger << LOG_DEBUG_POINT << "QT_ME2_loopcf[" << i_c << "] (em = " << osi->QT_correlationoperator[i_c].no_emitter << ", sp = " << osi->QT_correlationoperator[i_c].no_spectator << ") = " << setw(23) << setprecision(15) << osi->QT_ME2_loopcf[i_c] * osi->VA_b_ME2 << "   from Recola" << endline_flush;
      /*
	double temp = osi->QT_ME2_loopcf[i_c];
	// overwrite amplitudes with Hayk's ones !!!
	if (i_c == 1){osi->QT_ME2_loopcf[i_c] = (temp_H1_T34 + (osi->QT_correlationoperator[7].charge_factor - osi->QT_correlationoperator[0].charge_factor) * ME2loop / (osi->alpha_S / pi)) / osi->ME2;}
	if (i_c == 2){osi->QT_ME2_loopcf[i_c] = temp_H1_T13 / osi->ME2;}
	if (i_c == 3){osi->QT_ME2_loopcf[i_c] = -(temp_H1_T13 + temp_H1_T34 + osi->QT_correlationoperator[7].charge_factor * ME2loop / (osi->alpha_S / pi)) / osi->ME2;}
	//	  if (i_c == 3){osi->QT_ME2_loopcf[i_c] = temp_H1_T23 / osi->ME2;}
	if (i_c == 5){osi->QT_ME2_loopcf[i_c] = -(temp_H1_T13 + temp_H1_T34 + osi->QT_correlationoperator[7].charge_factor * ME2loop / (osi->alpha_S / pi)) / osi->ME2;}
	//	  if (i_c == 5){osi->QT_ME2_loopcf[i_c] = temp_H1_T23 / osi->ME2;}
	if (i_c == 6){osi->QT_ME2_loopcf[i_c] = temp_H1_T13 / osi->ME2;}
	if (i_c == 8){osi->QT_ME2_loopcf[i_c] = temp_H1_T34 / osi->ME2;}

	logger << LOG_DEBUG_POINT << "QT_ME2_loopcf[" << i_c << "] (em = " << osi->QT_correlationoperator[i_c].no_emitter << ", sp = " << osi->QT_correlationoperator[i_c].no_spectator << ") = " << setw(23) << setprecision(15) << osi->QT_ME2_loopcf[i_c] * osi->VA_b_ME2 << "   from Hayk" << endline_flush;
	logger << LOG_DEBUG_POINT << "QT_ME2_loopcf[" << i_c << "] (em = " << osi->QT_correlationoperator[i_c].no_emitter << ", sp = " << osi->QT_correlationoperator[i_c].no_spectator << ") = " << osi->QT_ME2_loopcf[i_c] / temp << "   ratio" << endline_flush;
      */
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::calculate_ME2_VT_QCD(){
  Logger logger("ppqq20_amplitude_Recola_set::calculate_ME2_VT_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double P_Recola[esi->p_parton[0].size()][4];
  set_phasespacepoint(P_Recola);
  calculate_QT_kinematics();

  mu_Q = (esi->p_parton[0][1] + esi->p_parton[0][2]).m();
  set_mu_uv_rcl(mu_Q);
  set_mu_ir_rcl(mu_Q);
  double alpha_S_mu_Q = LHAPDF::alphasPDF(mu_Q);
  set_alphas_rcl(alpha_S_mu_Q, mu_Q, N_nondecoupled);
  compute_process_rcl(1, P_Recola, "NLO");
  get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s - 1, "LO", osi->VA_b_ME2);
  osi->VA_b_ME2 = osi->VA_b_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, csi->contribution_order_alpha_s - 1);
  osi->VA_V_ME2 = 0.;
  get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s, "NLO", osi->VA_X_ME2);
  osi->VA_X_ME2 = osi->VA_X_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, csi->contribution_order_alpha_s);
  compute_process_rcl(1, P_Recola, "LO");
  compute_all_colour_correlations_rcl(1, P_Recola);
  // Actually not required: osi->ME2 == osi->VA_b_ME2; to avoid confusion: correct for alpha_S everywhere !!!
  osi->ME2 = osi->VA_b_ME2;
  calculate_ME2_cf_twocolourcorrelators();
  calculate_H1_delta_QCD_from_VX_b();
  calculate_Ft1born();
  //  H1 -> H1 - It1  (It1 = 2 Re < M0 | It1 | M0 >)
  osi->QT_H1_delta_QCD += .5 * osi->Ft1born;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::calculate_ME2_CT_QCD(){
  Logger logger("ppqq20_amplitude_Recola_set::calculate_ME2_CT_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double P_Recola[esi->p_parton[0].size()][4];
  set_phasespacepoint(P_Recola);
  calculate_QT_kinematics();

  mu_Q = (esi->p_parton[0][1] + esi->p_parton[0][2]).m();
  double alpha_S_mu_Q = LHAPDF::alphasPDF(mu_Q);
  set_alphas_rcl(alpha_S_mu_Q, mu_Q, N_nondecoupled);
  osi->ME2 = 0.;
  compute_process_rcl(1, P_Recola, "LO");
  compute_all_colour_correlations_rcl(1, P_Recola);
  get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s - 1, "LO", osi->ME2);
  osi->ME2 = osi->ME2 * pow(osi->alpha_S / alpha_S_mu_Q, csi->contribution_order_alpha_s - 1);
  calculate_ME2_cf_twocolourcorrelators();
  /*
  // alternative version without using mu_Q for alpha_S factor !!!
  osi->ME2 = 0.;
  compute_process_rcl(1, P_Recola, "LO");
  compute_all_colour_correlations_rcl(1, P_Recola);
  get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s - 1, "LO", osi->ME2);
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    if (osi->QT_correlationoperator[i_c].type_combination == 1){
      osi->QT_ME2_cf[i_c] = osi->QT_correlationoperator[i_c].charge_factor;
    }
    else {
      // Instead, the RECOLA calls for colour-correlators
      get_colour_correlation_rcl(1, csi->contribution_order_alpha_s - 1, osi->QT_correlationoperator[i_c].no_emitter, osi->QT_correlationoperator[i_c].no_spectator, osi->QT_ME2_cf[i_c]);
      // * pow(alpha_S / alpha_S_mu_Q, csi->contribution_order_alpha_s - 1)
      osi->QT_ME2_cf[i_c] = osi->QT_ME2_cf[i_c] * osi->QT_correlationoperator[i_c].colour_factor / osi->ME2;
    }
  }
  */
  if (osi->initial_pdf_diag){calculate_Gamma1born();}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::calculate_ME2_VT2_QCD(){
  Logger logger("ppqq20_amplitude_Recola_set::calculate_ME2_VT2_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double P_Recola[esi->p_parton[0].size()][4];
  set_phasespacepoint(P_Recola);
  calculate_QT_kinematics();

  mu_Q = (esi->p_parton[0][1] + esi->p_parton[0][2]).m();
  set_mu_uv_rcl(mu_Q);
  set_mu_ir_rcl(mu_Q);
  double alpha_S_mu_Q = LHAPDF::alphasPDF(mu_Q);
  set_alphas_rcl(alpha_S_mu_Q, mu_Q, N_nondecoupled);
  compute_process_rcl(1, P_Recola, "NLO");
  get_squared_amplitude_rcl(1, osi->csi->contribution_order_alpha_s - 2, "LO", osi->VA_b_ME2);
  osi->VA_b_ME2 = osi->VA_b_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, osi->csi->contribution_order_alpha_s - 2);
  osi->VA_V_ME2 = 0.;
  get_squared_amplitude_rcl(1, osi->csi->contribution_order_alpha_s - 1, "NLO", osi->VA_X_ME2);
  osi->VA_X_ME2 = osi->VA_X_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, osi->csi->contribution_order_alpha_s - 1);

  compute_process_rcl(1, P_Recola, "LO");
  compute_all_colour_correlations_rcl(1, P_Recola);
  osi->ME2 = osi->VA_b_ME2;
  calculate_ME2_cf_twocolourcorrelators();
  if (osi->initial_gg){
    // required for H0_doublespinflip, H0_DG, fourcolourcorrelators, but only for gg channel
    set_phasespacepoint_OLformat();
    set_momenta_(P_OLformat, &mu_Q, &m_HQ);
  }

  if (switch_off_H2 || !osi->initial_diag){
    osi->QT_A0 = osi->VA_b_ME2;
    osi->QT_A1 = osi->VA_V_ME2 + osi->VA_X_ME2;
    calculate_H1_delta_QCD_from_VX_b();
    osi->QT_A2 = 0.;
    osi->QT_H2_delta_QCD = 0.;
    //osi->QT_H2_delta = osi->QT_A2 / 4 / (pow(osi->alpha_S * inv2pi, 2) * osi->QT_A0);
  }
  else {
    calculate_fourcolourcorrelators();
    calculate_H2_2loop_ingredients();
    calculate_H2_2loop();
  }

  if (osi->initial_gg){
    // Use Hayk's implementation for H0_doublespinflip and H0_DG
    calculate_ME2_VT2_QCD_H0_doublespinflip();
    calculate_ME2_VT2_QCD_H0_DG();
  }

  if (osi->initial_pdf_diag || osi->initial_pdf_gq){
    calculate_Ft1born();
    calculate_Gamma1born();
    //  H1 -> H1 - It1  (It1 = 2 Re < M0 | It1 | M0 >)
    osi->QT_H1_delta_QCD += .5 * osi->Ft1born;
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::calculate_ME2_VT2_QCD_H0_doublespinflip(){
  static Logger logger("ppqq20_amplitude_Recola_set::calculate_ME2_VT2_QCD_H0_doublespinflip");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // Specific version for QQx production (based on Hayk's code; not directly available from Recola)
  osi->QT_H0_doublespinflip = 0.;
  callmsqggav_(P_OLformat, &m_HQ, &osi->QT_H0_doublespinflip);
  osi->QT_H0_doublespinflip = osi->QT_H0_doublespinflip / osi->VA_b_ME2 * pow(osi->alpha_S, 2);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::calculate_ME2_VT2_QCD_H0_DG(){
  static Logger logger("ppqq20_amplitude_Recola_set::calculate_ME2_VT2_QCD_H0_DG");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // Specific version for QQx production (based on Hayk's code; not directly available from Recola)
  osi->QT_H0_DG = 0.;
  callmsqdgav_(P_OLformat, &m_HQ, &osi->QT_H0_DG);
  osi->QT_H0_DG = osi->QT_H0_DG / osi->VA_b_ME2 * pow(osi->alpha_S, 2);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::calculate_ME2_CT2_QCD(){
  Logger logger("ppqq20_amplitude_Recola_set::calculate_ME2_CT2_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double P_Recola[esi->p_parton[0].size()][4];
  set_phasespacepoint(P_Recola);
  calculate_QT_kinematics();

  mu_Q = (esi->p_parton[0][1] + esi->p_parton[0][2]).m();
  set_mu_uv_rcl(mu_Q);
  set_mu_ir_rcl(mu_Q);
  double alpha_S_mu_Q = LHAPDF::alphasPDF(mu_Q);
  set_alphas_rcl(alpha_S_mu_Q, mu_Q, N_nondecoupled);
  compute_process_rcl(1, P_Recola, "NLO");
  get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s - 2, "LO", osi->VA_b_ME2);
  osi->VA_b_ME2 = osi->VA_b_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, csi->contribution_order_alpha_s - 2);
  osi->VA_V_ME2 = 0.;
  get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s - 1, "NLO", osi->VA_X_ME2);
  osi->VA_X_ME2 = osi->VA_X_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, csi->contribution_order_alpha_s - 1);

  compute_process_rcl(1, P_Recola, "LO");
  compute_all_colour_correlations_rcl(1, P_Recola);
  // Actually not required: osi->ME2 == osi->VA_b_ME2; to avoid confusion: correct for alpha_S everywhere !!!
  osi->ME2 = osi->VA_b_ME2;
  calculate_ME2_cf_twocolourcorrelators();
  if (osi->initial_diag){
    compute_process_rcl(1, P_Recola, "NLO");
    compute_all_colour_correlations_rcl(1, P_Recola, "NLO");
    ME2loop = osi->VA_X_ME2 + osi->VA_V_ME2;
    calculate_ME2_loopcf_twocolourcorrelators();
  }
  if (osi->initial_diag_gg){
    set_phasespacepoint_OLformat();
    set_momenta_(P_OLformat, &mu_Q, &m_HQ);
  }
  
  // Same as in OpenLoops implementation from here on:
  calculate_H1_delta_QCD_from_VX_b();
  calculate_Ft1born();
  //  H1 -> H1 - It1  (It1 = 2 Re < M0 | It1 | M0 >)
  osi->QT_H1_delta_QCD += .5 * osi->Ft1born;

  if (osi->initial_diag){
    calculate_fourcolourcorrelators();
    calculate_Ft1born_4correlator();
    for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){osi->QT_ME2_loopcf[i_c] += .5 * osi->Ft1born_4correlator[i_c];}
    logger << LOG_DEBUG_POINT << "shifted: OL_checksum (1) = " << osi->QT_ME2_loopcf[0] + osi->QT_ME2_loopcf[1] + osi->QT_ME2_loopcf[2] + osi->QT_ME2_loopcf[3] << endline_flush;
    logger << LOG_DEBUG_POINT << "shifted: OL_checksum (2) = " << osi->QT_ME2_loopcf[1] + osi->QT_ME2_loopcf[4] + osi->QT_ME2_loopcf[5] + osi->QT_ME2_loopcf[6] << endline_flush;
    logger << LOG_DEBUG_POINT << "shifted: OL_checksum (3) = " << osi->QT_ME2_loopcf[2] + osi->QT_ME2_loopcf[5] + osi->QT_ME2_loopcf[7] + osi->QT_ME2_loopcf[8] << endline_flush;
    logger << LOG_DEBUG_POINT << "shifted: OL_checksum (4) = " << osi->QT_ME2_loopcf[3] + osi->QT_ME2_loopcf[6] + osi->QT_ME2_loopcf[8] + osi->QT_ME2_loopcf[9] << endline_flush;
    calculate_Gamma1loop();
    if (osi->initial_diag_gg){calculate_commutator_Gamma1_Ft1();}
    calculate_Gamma1_squared();
    calculate_Gamma2();
  }
  if (osi->initial_pdf_diag || osi->initial_pdf_gq){calculate_Gamma1born();}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_Recola_set::calculate_H2_2loop_ingredients(){
  static Logger logger("ppqq20_amplitude_Recola_set::calculate_H2_ingredients");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double P_rec[osi->esi->p_parton[0].size()][4];
  for (int i = 1; i < osi->esi->p_parton[0].size(); i++){
    P_rec[i - 1][0] = osi->esi->p_parton[0][i].x0();
    P_rec[i - 1][1] = osi->esi->p_parton[0][i].x1();
    P_rec[i - 1][2] = osi->esi->p_parton[0][i].x2();
    P_rec[i - 1][3] = osi->esi->p_parton[0][i].x3();
  }

  for (int i = 1; i < osi->esi->p_parton[0].size(); i++){
    stringstream temp_ss;
    for (int j = 0; j < 4; j++){
      temp_ss << setw(23) << setprecision(15) << P_rec[i - 1][j] << "   ";
    }
    logger << LOG_DEBUG_VERBOSE << "P_rec[" << i << "] = " << temp_ss.str() << endline_flush;
  }
  
  // Renormalization and regularization scales are set to Q here:
  mu_Q = (osi->esi->p_parton[0][1] + osi->esi->p_parton[0][2]).m();
  // osi->var_mu_ren -> mu_Q
  set_mu_uv_rcl(mu_Q);
  logger << LOG_DEBUG_VERBOSE << "set_mu_uv_rcl(" << setprecision(15) << mu_Q << ");" << endline_flush;
  
  set_mu_ir_rcl(mu_Q);
  logger << LOG_DEBUG_VERBOSE << "set_mu_ir_rcl(" << setprecision(15) << mu_Q << ");" << endline_flush;
  
  double alpha_S_mu_Q = LHAPDF::alphasPDF(mu_Q);
  set_alphas_rcl(alpha_S_mu_Q, mu_Q, N_nondecoupled);
  logger << LOG_DEBUG_VERBOSE << "set_alphas_rcl(" << setprecision(15) << osi->var_alpha_S_reference << ", " << setprecision(15) << mu_Q << ", " << N_nondecoupled << ");" << endline_flush;

  /*
  compute_process_rcl(1, P_rec, "NLO");
  logger << LOG_DEBUG_VERBOSE << "compute_process_rcl(1, P_rec, " << char(34) << "NLO" << char(34) << ");" << endline_flush;
  get_squared_amplitude_rcl(1, osi->csi->contribution_order_alpha_s - 2, "LO", osi->VA_b_ME2);
  logger << LOG_DEBUG_POINT << "Recola:     id = 1   ME2_B             = " << osi->VA_b_ME2 << endline_flush;
  osi->VA_b_ME2 = osi->VA_b_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, osi->csi->contribution_order_alpha_s - 2);
  logger << LOG_DEBUG_POINT << "Recola:     id = 1   ME2_B             = " << osi->VA_b_ME2 << "   (corrected for var_rel_alpha_S)" << endline_flush;

  get_squared_amplitude_rcl(1, osi->csi->contribution_order_alpha_s - 1, "NLO", osi->VA_X_ME2);
  logger << LOG_DEBUG_POINT << "Recola:     id = 1   2Re{ME_VX * ME_B} = " << osi->VA_X_ME2 << endline_flush;
  osi->VA_X_ME2 = osi->VA_X_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, osi->csi->contribution_order_alpha_s - 1);
  logger << LOG_DEBUG_POINT << "Recola:     id = 1   2Re{ME_VX * ME_B} = " << osi->VA_X_ME2 << "   (corrected for var_rel_alpha_S)" << endline_flush;
  */
  compute_process_rcl(2, P_rec, "NLO");
  logger << LOG_DEBUG_VERBOSE << "compute_process_rcl(2, P_rec, " << char(34) << "NLO" << char(34) << ");" << endline_flush;
  /*
  osi->VA_V_ME2 = 0.;
  get_squared_amplitude_rcl(2, osi->csi->contribution_order_alpha_s - 2, "LO", osi->VA_b_ME2);
  logger << LOG_DEBUG_POINT << "Recola:     id = 2   ME2_B             = " << osi->VA_b_ME2 << endline_flush;
  osi->VA_b_ME2 = osi->VA_b_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, osi->csi->contribution_order_alpha_s - 2);
  logger << LOG_DEBUG_POINT << "Recola:     id = 2   ME2_B             = " << osi->VA_b_ME2 << "   (corrected for var_rel_alpha_S)" << endline_flush;

  get_squared_amplitude_rcl(2, osi->csi->contribution_order_alpha_s - 1, "NLO", osi->VA_X_ME2);
  logger << LOG_DEBUG_POINT << "Recola:     id = 2   2Re{ME_VX * ME_B} = " << osi->VA_X_ME2 << endline_flush;
  osi->VA_X_ME2 = osi->VA_X_ME2 * pow(osi->alpha_S / alpha_S_mu_Q, osi->csi->contribution_order_alpha_s - 1);
  logger << LOG_DEBUG_POINT << "Recola:     id = 2   2Re{ME_VX * ME_B} = " << osi->VA_X_ME2 << "   (corrected for var_rel_alpha_S)" << endline_flush;
  */
  /*//
  ME2_L_L = 0.;
  ME2_L2_B = 0.;
  get_squared_amplitude_rcl(2, osi->csi->contribution_order_alpha_s, "NLO", ME2_L_L);
  logger << LOG_DEBUG_POINT << "Recola:     ME_VX * ME_VX*    = " << ME2_L_L << endline_flush;
  ME2_L_L = ME2_L_L * pow(osi->alpha_S / alpha_S_mu_Q, osi->csi->contribution_order_alpha_s);
  logger << LOG_DEBUG_POINT << "Recola:     ME_VX * ME_VX*    = " << ME2_L_L <<  "   (corrected for var_rel_alpha_S)" << endline_flush;
  //*/
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

#endif
