#include "header.hpp"

#include "ppqq20.amplitude.set.hpp"

extern "C" {
  void set_momenta_(double* P, double* mu_Q, double* m_t);
  void fourcorrelators_(double* B4, int *i1, int *i2, int *i3, int *i4, int *channel);
  double callmyli3_(double* x, double* result);
  void init_interpolation_t3t4_(double* CA, double* nf, const char *ch_path_Fex2_T3T4_CA, const char *ch_path_Fex2_T3T4_nf, const char *ch_path_Fex11_T3T4, int *len_Fex2_T3T4_CA, int *len_Fex2_T3T4_nf, int *len_Fex11_T3T4);
  void init_interpolation_tttt_(const char *ch_path_Fex1sq_T3T4T3T4, const char *ch_path_Fex1sq_T3T4TITj, const char *ch_path_Fex1sq_TITjTKTl, int *len_Fex1sq_T3T4T3T4, int *len_Fex1sq_T3T4TITj, int *len_Fex1sq_TITjTKTl);
  double h1_tt_(int *i1, int *i2, double* beta, double* cost);
  double h2_tt_(int *i1, int *i2, double* beta, double* cost, double* nf);
  double h2_tttt_(int *i1, int *i2, int *i3, int *i4, double* beta, double* cost);
  double virtgg_(double* beta, double* cost, double* nf, const char * ch_path_gridgg, int * len_path_gridgg);
  double virtqq_(double* beta, double* cost, double* nf, const char * ch_path_gridqq, int * len_path_gridqq);
  double_complex hpl2_(int *i1, int *i2, double_complex *expB34);
  double_complex hpl3_(int *i1, int *i2, int *i3, double_complex *expB34);
}

void ppqq20_amplitude_initialization(munich * MUC){
  static Logger logger("ppqq20_amplitude_initialization");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (MUC->osi->switch_OL){
#ifdef OPENLOOPS
    MUC->asi = new ppqq20_amplitude_OpenLoops_set();
#endif
  }
  else if (MUC->osi->switch_RCL){
#ifdef RECOLA
    MUC->asi = new ppqq20_amplitude_Recola_set();
#endif
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


ppqq20_amplitude_set::ppqq20_amplitude_set(){
  Logger logger("ppqq20_amplitude_set::ppqq20_amplitude_set ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "constructor called..." << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


ppqq20_amplitude_set::~ppqq20_amplitude_set(){
  Logger logger("ppqq20_amplitude_set::~ppqq20_amplitude_set ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "destructor called..." << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_fourcolourcorrelators(){
  static Logger logger("ppqq20_amplitude_set::calculate_fourcolourcorrelators");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    int i1 = osi->QT_correlationoperator[i_c].no_emitter;
    int i2 = osi->QT_correlationoperator[i_c].no_spectator;
    //    for (int j_c = 0; j_c < osi->QT_correlationoperator.size(); j_c++){
    for (int j_c = i_c; j_c < osi->QT_correlationoperator.size(); j_c++){
      int j1 = osi->QT_correlationoperator[j_c].no_emitter;
      int j2 = osi->QT_correlationoperator[j_c].no_spectator;
      double temp_B4c = 0.;
      fourcorrelators_(&temp_B4c, &i1, &i2, &j1, &j2, &osi->initial_channel);
      if (csi->type_contribution == "VT2"){
	temp_B4c *= osi->VA_b_ME2;
	M0M04c[i1][i2][j1][j2] = temp_B4c;
	if (i2 != i1){M0M04c[i2][i1][j1][j2] = temp_B4c;}
	if (j2 != j1){M0M04c[i1][i2][j2][j1] = temp_B4c;}
	if (i2 != i1 && j2 != j1){M0M04c[i2][i1][j2][j1] = temp_B4c;}
	if (i_c != j_c){
	  M0M04c[j1][j2][i1][i2] = temp_B4c;
	  if (i2 != i1){M0M04c[j1][j2][i2][i1] = temp_B4c;}
	  if (j2 != j1){M0M04c[j2][j1][i1][i2] = temp_B4c;}
	  if (i2 != i1 && j2 != j1){M0M04c[j2][j1][i2][i1] = temp_B4c;}
	}
	logger << LOG_DEBUG_POINT << setw(22) << "M0M04c" << "[" << setw(2) << i1 << "][" << setw(2) << i2 << "][" << setw(2) << j1 << "][" << setw(2) << j2 << "]" << " = " << setprecision(15) << setw(23) << M0M04c[i1][i2][j1][j2] << endline_flush;
      }
      else if (csi->type_contribution == "CT2"){
	osi->QT_ME2_cfcf[i_c][j_c] = temp_B4c;
	osi->QT_ME2hc_cfcf[i_c][j_c] = temp_B4c;
	if (i_c != j_c){
	  osi->QT_ME2_cfcf[j_c][i_c] = temp_B4c;
	  osi->QT_ME2hc_cfcf[j_c][i_c] = temp_B4c;
	}
	//	osi->QT_ME2hc_cfcf[i_c][j_c] = 0.;
	//	fourcorrelators_(&osi->QT_ME2hc_cfcf[i_c][j_c], &j1, &j2, &i1, &i2, &osi->initial_channel);
	logger << LOG_DEBUG_POINT << setw(22) << "QT_ME2_cfcf" << "[" << setw(2) << i_c << "][" << setw(2) << j_c << "]" << " = " << setprecision(15) << setw(23) << osi->QT_ME2_cfcf[i_c][j_c] << "   " << setw(22) << "QT_ME2hc_cfcf" << "[" << setw(2) << i_c << "][" << setw(2) << j_c << "]" << " = " << setprecision(15) << setw(23) << osi->QT_ME2hc_cfcf[i_c][j_c] << endline_flush;
      }
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_QT_kinematics(){
  static Logger logger("ppqq20_amplitude_set::calculate_QT_kinematics");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  q2 = (esi->p_parton[0][3] + esi->p_parton[0][4]).m2();
  logger << LOG_DEBUG_POINT << setw(30) << "q2" << " = " << setprecision(15) << setw(23) << q2 << endline_flush;
  // calculate_Gamma1born
  // calculate_Gamma1loop
  // calculate_commutator_Gamma1_Ft1
  // calculate_Gamma2
  // calculate_Gamma1_squared
  
  v34 = sqrt(1. - pow(2 * m2_HQ / (2 * (esi->p_parton[0][3] * esi->p_parton[0][4])), 2));
  logger << LOG_DEBUG_POINT << setw(30) << "v34" << " = " << setprecision(15) << setw(23) << v34 << endline_flush;
  // calculate_Ft1born
  // calculate_Ft1born_4correlator
  // calculate_Gamma1born
  // calculate_Gamma1loop
  // calculate_commutator_Gamma1_Ft1
  // calculate_Gamma2
  // calculate_Gamma1_squared

  frac_v34 = (1. + v34) / (1. - v34);
  logger << LOG_DEBUG_POINT << setw(30) << "frac_v34" << " = " << setprecision(15) << setw(23) << frac_v34 << endline_flush;
  // calculate_Ft1born
  // calculate_Ft1born_4correlator
  // calculate_Gamma1born
  // calculate_Gamma1loop
  // calculate_commutator_Gamma1_Ft1
  // calculate_Gamma2
  // calculate_Gamma1_squared

  log_v34 = log(frac_v34);
  logger << LOG_DEBUG_POINT << setw(30) << "log_v34" << " = " << setprecision(15) << setw(23) << log_v34 << endline_flush;
  // calculate_Ft1born
  // calculate_Ft1born_4correlator
  // calculate_Gamma1born
  // calculate_Gamma1loop
  // calculate_commutator_Gamma1_Ft1
  // calculate_Gamma2
  // calculate_Gamma1_squared

  if (csi->type_contribution != "CT"){
    dilogpT2m2 = gsl_sf_dilog(-esi->p_parton[0][3].pT2() / m2_HQ);
    logger << LOG_DEBUG_POINT << setw(30) << "dilogpT2m2" << " = " << setprecision(15) << setw(23) << dilogpT2m2 << endline_flush;
    // calculate_Ft1born
    // calculate_Ft1born_4correlator
    // calculate_commutator_Gamma1_Ft1

    logmT2m2 = log(esi->p_parton[0][3].ET2() / m2_HQ);
    logger << LOG_DEBUG_POINT << setw(30) << "logmT2m2" << " = " << setprecision(15) << setw(23) << logmT2m2 << endline_flush;
    // calculate_Ft1born
    // calculate_Ft1born_4correlator
    // calculate_commutator_Gamma1_Ft1

    y34 = esi->p_parton[0][3].rapidity() - esi->p_parton[0][4].rapidity();
    logger << LOG_DEBUG_POINT << setw(30) << "y34" << " = " << setprecision(15) << setw(23) << y34 << endline_flush;
    L34 = log_v34 * logmT2m2
      - 2 * gsl_sf_dilog(2 * v34 / (1. + v34))
      - .25 * pow(log_v34, 2)
      + 2 * (gsl_sf_dilog(1. - exp(y34) / sqrt(frac_v34)) + gsl_sf_dilog(1. - exp(-y34) / sqrt(frac_v34)) + .5 * pow(y34, 2));
    logger << LOG_DEBUG_POINT << setw(30) << "L34" << " = " << setprecision(15) << setw(23) << L34 << endline_flush;
    // calculate_Ft1born
    // calculate_Ft1born_4correlator
    // calculate_commutator_Gamma1_Ft1

    if (csi->type_contribution == "CT2"){
      // -> gammacusp2_v
      beta34 = .5 * log_v34;
      logger << LOG_DEBUG_POINT << setw(30) << "beta34" << " = " << setprecision(15) << setw(23) << beta34 << endline_flush;
      arg_1mv34_1pv34 = (1. - v34) / (1. + v34);
      logger << LOG_DEBUG_POINT << setw(30) << "arg_1mv34_1pv34" << " = " << setprecision(15) << setw(23) << arg_1mv34_1pv34 << endline_flush;
      dilog1mv34_1pv34 = gsl_sf_dilog(arg_1mv34_1pv34);
      logger << LOG_DEBUG_POINT << setw(30) << "dilog1mv34_1pv34" << " = " << setprecision(15) << setw(23) << dilog1mv34_1pv34 << endline_flush;
      trilog1mv34_1pv34 = 0.;
      callmyli3_(&arg_1mv34_1pv34, &trilog1mv34_1pv34);
      logger << LOG_DEBUG_POINT << setw(30) << "trilog1mv34_1pv34" << " = " << setprecision(15) << setw(23) << trilog1mv34_1pv34 << endline_flush;
      log_2v34_1pv34 = log(2 * v34 / (1. + v34));
      logger << LOG_DEBUG_POINT << setw(30) << "log_2v34_1pv34" << " = " << setprecision(15) << setw(23) << log_2v34_1pv34 << endline_flush;
      gammacusp2_v = C_A / 2. * (-5 * zeta2 + zeta3 + pow(beta34, 2)
				 + pow(1. / v34, 2) * (trilog1mv34_1pv34 + beta34 * dilog1mv34_1pv34 + pow(beta34, 3) / 3 - (5. / 6) * pi2 * beta34 - zeta3)
				 + (1. / v34) * (dilog1mv34_1pv34 - 2. * beta34 * log_2v34_1pv34 + (5. / 6) * pi2 * beta34 - pow(beta34, 2) + (5. / 6) * pi2 - pow(beta34, 3) / 3));
      logger << LOG_DEBUG_POINT << setw(30) << "gammacusp2_v" << " = " << setprecision(15) << setw(23) << gammacusp2_v << endline_flush;
      // calculate_Gamma2
    }
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_Ft1born(){
  static Logger logger("ppqq20_amplitude_set::calculate_Ft1born");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  osi->Ft1born = 0.;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    if (osi->QT_correlationoperator[i_c].type_combination == 1){ // FS emitter == FS spectator
      osi->Ft1born += (logmT2m2 + dilogpT2m2) * osi->QT_ME2_cf[i_c];
      //      logger << LOG_WARN << "osi->QT_correlationoperator[" << i_c << "]:   type = " << osi->QT_correlationoperator[i_c].type_combination << "   em = " << osi->QT_correlationoperator[i_c].no_emitter << "   type = " << osi->QT_correlationoperator[i_c].no_spectator << "   " << (logmT2m2 + dilogpT2m2) * osi->QT_ME2_cf[i_c] << endline_flush;
      //      logger << LOG_WARN << "osi->QT_correlationoperator[" << i_c << "]:   type = " << osi->QT_correlationoperator[i_c].type_combination << "   em = " << osi->QT_correlationoperator[i_c].no_emitter << "   type = " << osi->QT_correlationoperator[i_c].no_spectator << "   logmT2m2 * osi->QT_ME2_cf[i_c] = " << (logmT2m2) * osi->QT_ME2_cf[i_c] << endline_flush;
    }
    else if (osi->QT_correlationoperator[i_c].type_combination == 2){ // FS emitter != FS spectator
      osi->Ft1born += (2 * dilogpT2m2 + (1. / v34) * L34) * osi->QT_ME2_cf[i_c];
      //      logger << LOG_WARN << "osi->QT_correlationoperator[" << i_c << "]:   type = " << osi->QT_correlationoperator[i_c].type_combination << "   em = " << osi->QT_correlationoperator[i_c].no_emitter << "   type = " << osi->QT_correlationoperator[i_c].no_spectator << "   " << (2 * dilogpT2m2 + (1. / v34) * L34) * osi->QT_ME2_cf[i_c] << endline_flush;
    }
    else {
      //      logger << LOG_WARN << "osi->QT_correlationoperator[" << i_c << "]:   type = " << osi->QT_correlationoperator[i_c].type_combination << "   em = " << osi->QT_correlationoperator[i_c].no_emitter << "   type = " << osi->QT_correlationoperator[i_c].no_spectator << "   no contribution" << endline_flush;
    }
  }

  logger << LOG_DEBUG_POINT << setw(30) << "Ft1born" << " = " <<  setprecision(15) << setw(23) << osi->Ft1born << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_Ft1born_4correlator(){
  static Logger logger("ppqq20_amplitude_set::calculate_Ft1born_4correlator");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    osi->Ft1born_4correlator[i_c] = 0.;
    for (int j_c = 0; j_c < osi->QT_correlationoperator.size(); j_c++){
      if (osi->QT_correlationoperator[j_c].type_combination == 1 || osi->QT_correlationoperator[j_c].type_combination == 2){
	double temp_B4 = osi->QT_ME2_cfcf[i_c][j_c];
	double temp_B4hc = osi->QT_ME2hc_cfcf[i_c][j_c];
	if (osi->QT_correlationoperator[j_c].type_combination == 1){
	  osi->Ft1born_4correlator[i_c] += (logmT2m2 + dilogpT2m2) * .5 * (temp_B4 + temp_B4hc);
	  logger << LOG_DEBUG_POINT << "osi->Ft1born_4correlator[" << i_c << "][" << j_c << "] = " << setprecision(15) << setw(23) << (logmT2m2 + dilogpT2m2) * .5 * (temp_B4 + temp_B4hc) << "   cfcf[" << i_c << "][" << j_c << "] = " << setprecision(15) << setw(23) << temp_B4 << endline_flush;
	}
	else if (osi->QT_correlationoperator[j_c].type_combination == 2){
	  osi->Ft1born_4correlator[i_c] += (2 * dilogpT2m2 + (1. / v34) * L34) * .5 * (temp_B4 + temp_B4hc);
	  logger << LOG_DEBUG_POINT << "osi->Ft1born_4correlator[" << i_c << "][" << j_c << "] = " << setprecision(15) << setw(23) << (2 * dilogpT2m2 + (1. / v34) * L34) * .5 * (temp_B4 + temp_B4hc) << "   cfcf[" << i_c << "][" << j_c << "] = " << setprecision(15) << setw(23) << temp_B4 << endline_flush;
	}
      }
    }
    logger << LOG_DEBUG_POINT << setw(26) << "osi->Ft1born_4correlator" << "[" << setw(2) << i_c << "]" << " = " << setprecision(15) << setw(23) << osi->Ft1born_4correlator[i_c] << endline_flush;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_Gamma1born(){
  static Logger logger("ppqq20_amplitude_set::calculate_Gamma1born");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  osi->Gamma1born = 0.;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    if (osi->QT_correlationoperator[i_c].type_combination == 1){
      osi->Gamma1born += .5 * osi->QT_ME2_cf[i_c];
    }
    else if (osi->QT_correlationoperator[i_c].type_combination == 2){
      osi->Gamma1born += .5 * (1. / v34) * log_v34 * osi->QT_ME2_cf[i_c];
    }
    else if (osi->QT_correlationoperator[i_c].type_combination == 3){
      osi->Gamma1born += .5 * log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ)) * osi->QT_ME2_cf[i_c];
    }
  }
  // Gammat = -2 * Gammat1(Eg.33, arXiv:1408.4564) !!! -> 2 * Re(Gammat1)
  logger << LOG_DEBUG_POINT << setw(30) << "Gamma1born" << " = " << setprecision(15) << setw(23) << osi->Gamma1born << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_Gamma1loop(){
  static Logger logger("ppqq20_amplitude_set::calculate_Gamma1loop");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  osi->Gamma1loop = 0.;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    if (osi->QT_correlationoperator[i_c].type_combination == 1){
      osi->Gamma1loop += .5 * osi->QT_ME2_loopcf[i_c];
      logger << LOG_DEBUG_POINT << "QT_ME2_loopcf[" << i_c << "] = " << setw(23) << setprecision(15) << osi->QT_ME2_loopcf[i_c] << endline_flush;
    }
    else if (osi->QT_correlationoperator[i_c].type_combination == 2){
      osi->Gamma1loop += .5 * (1. / v34) * log_v34 * osi->QT_ME2_loopcf[i_c];
      logger << LOG_DEBUG_POINT << "QT_ME2_loopcf[" << i_c << "] = " << setw(23) << setprecision(15) << osi->QT_ME2_loopcf[i_c] << endline_flush;
    }
    else if (osi->QT_correlationoperator[i_c].type_combination == 3){
      osi->Gamma1loop += .5 * log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ)) * osi->QT_ME2_loopcf[i_c];
      logger << LOG_DEBUG_POINT << "QT_ME2_loopcf[" << i_c << "] = " << setw(23) << setprecision(15) << osi->QT_ME2_loopcf[i_c] << endline_flush;
    }
    else {
      logger << LOG_DEBUG_POINT << "QT_ME2_loopcf[" << i_c << "] = " << setw(23) << setprecision(15) << osi->QT_ME2_loopcf[i_c] << "   (not used)" << endline_flush;
    }
  }
  // Gammat = -2 * Gammat1(Eg.33, arXiv:1408.4564) !!! -> 2 * Re(Gammat1)
  logger << LOG_DEBUG_POINT << setw(30) << "Gamma1loop" << " = " << setprecision(15) << setw(23) << osi->Gamma1loop << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_commutator_Gamma1_Ft1(){
  static Logger logger("ppqq20_amplitude_set::calculate_commutator_Gamma1_Ft1");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  osi->commutator_Gammat1_Ft1 = 0.;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    for (int j_c = 0; j_c < osi->QT_correlationoperator.size(); j_c++){
      double temp_B4 = osi->QT_ME2_cfcf[i_c][j_c];
      double temp_B4ro = osi->QT_ME2hc_cfcf[i_c][j_c];

      if (osi->QT_correlationoperator[i_c].type_combination == 1 && osi->QT_correlationoperator[j_c].type_combination == 1){
	osi->commutator_Gammat1_Ft1 += .5 * (logmT2m2 + dilogpT2m2) * (temp_B4 - temp_B4ro);
	logger << LOG_DEBUG_VERBOSE
	       << " type_combination[" << i_c << "] = " << osi->QT_correlationoperator[i_c].type_combination << "   "
	       << " type_combination[" << j_c << "] = " << osi->QT_correlationoperator[j_c].type_combination << "   "
	       << .5 * (logmT2m2 + dilogpT2m2) * (temp_B4 - temp_B4ro)
	       << endline_flush;
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 1 && osi->QT_correlationoperator[j_c].type_combination == 2){
	osi->commutator_Gammat1_Ft1 += .5 * (2 * dilogpT2m2 + (1. / v34) * L34) * (temp_B4 - temp_B4ro);
	logger << LOG_DEBUG_VERBOSE
	       << " type_combination[" << i_c << "] = " << osi->QT_correlationoperator[i_c].type_combination << "   "
	       << " type_combination[" << j_c << "] = " << osi->QT_correlationoperator[j_c].type_combination << "   "
	       << .5 * (2 * dilogpT2m2 + (1. / v34) * L34) * (temp_B4 - temp_B4ro)
	       << endline_flush;
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 2 && osi->QT_correlationoperator[j_c].type_combination == 1){
	osi->commutator_Gammat1_Ft1 += .5 * (2 * dilogpT2m2 + (1. / v34) * L34) * (temp_B4 - temp_B4ro);
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 2 && osi->QT_correlationoperator[j_c].type_combination == 2){
	osi->commutator_Gammat1_Ft1 += .5 * (1. / v34) * log_v34 * (2 * dilogpT2m2 + (1. / v34) * L34) * (temp_B4 - temp_B4ro);
	logger << LOG_DEBUG_VERBOSE
	       << " type_combination[" << i_c << "] = " << osi->QT_correlationoperator[i_c].type_combination << "   "
	       << " type_combination[" << j_c << "] = " << osi->QT_correlationoperator[j_c].type_combination << "   "
	       << .5 * (1. / v34) * log_v34 * (2 * dilogpT2m2 + (1. / v34) * L34) * (temp_B4 - temp_B4ro)
	       << endline_flush;
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 3 && osi->QT_correlationoperator[j_c].type_combination == 1){
	osi->commutator_Gammat1_Ft1 += .5 * (logmT2m2 + dilogpT2m2)
	  * (log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ))) * (temp_B4 - temp_B4ro);
	logger << LOG_DEBUG_VERBOSE
	       << " type_combination[" << i_c << "] = " << osi->QT_correlationoperator[i_c].type_combination << "   "
	       << " type_combination[" << j_c << "] = " << osi->QT_correlationoperator[j_c].type_combination << "   "
	       << .5 * (logmT2m2 + dilogpT2m2)
	  * (log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ))) * (temp_B4 - temp_B4ro)
	       << endline_flush;
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 3 && osi->QT_correlationoperator[j_c].type_combination == 2){
	osi->commutator_Gammat1_Ft1 += .5 * (2 * dilogpT2m2 + (1. / v34) * L34)
	  * (log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ))) * (temp_B4 - temp_B4ro);
	logger << LOG_DEBUG_VERBOSE
	       << " type_combination[" << i_c << "] = " << osi->QT_correlationoperator[i_c].type_combination << "   "
	       << " type_combination[" << j_c << "] = " << osi->QT_correlationoperator[j_c].type_combination << "   "
	       << .5 * (2 * dilogpT2m2 + (1. / v34) * L34)
	  * (log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ))) * (temp_B4 - temp_B4ro)
	       << endline_flush;
      }
      /*
      else if (osi->QT_correlationoperator[i_c].type_combination == 1 && osi->QT_correlationoperator[j_c].type_combination == 3){
	// already covered in 3 - 1 ???
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 2 && osi->QT_correlationoperator[j_c].type_combination == 3){
	// already covered in 3 - 2 ???
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 3 && osi->QT_correlationoperator[j_c].type_combination == 3){
	// should not exist !!!
      }
      */
    }
  }
  logger << LOG_DEBUG_POINT << setw(30) << "commutator_Gammat1_Ft1" << " = " << setprecision(15) << setw(23) << osi->commutator_Gammat1_Ft1 << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_Gamma2(){
  static Logger logger("ppqq20_amplitude_set::calculate_Gamma2");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

    if (osi->initial_pdf_diag){
      /*
      Gamma2q=0.5d0*gammacusp2q*(
------------------     .      1d0/v34*dlog((1d0+v34)/(1d0-v34))
------------------     .      *Tqq(3,4)
------------------     .      +dlog(4d0*dot(ptrans,1,3)**2/(q2*mt**2))*Tqq(1,3)
------------------     .      +dlog(4d0*dot(ptrans,1,4)**2/(q2*mt**2))*Tqq(1,4)
------------------     .      +dlog(4d0*dot(ptrans,2,3)**2/(q2*mt**2))*Tqq(2,3)
------------------     .      +dlog(4d0*dot(ptrans,2,4)**2/(q2*mt**2))*Tqq(2,4))
------------------     .         +gammacuspprime2q*Tqq(3,4)
------------------     .         -2d0*gammaQ1
------------------     .         +0.5d0*beta0*F1q
      */

      osi->Gamma2 = 0.;
      for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
	if (osi->QT_correlationoperator[i_c].type_combination == 2){
	  osi->Gamma2 += .5 * osi->gammacusp2 * (1. / v34) * log_v34 * osi->QT_ME2_cf[i_c];
	  osi->Gamma2 += .5 * 2 * gammacusp2_v * osi->QT_ME2_cf[i_c];
	}
	else if (osi->QT_correlationoperator[i_c].type_combination == 3){
	  osi->Gamma2 += .5 * osi->gammacusp2 * log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ)) * osi->QT_ME2_cf[i_c];
	}
	/*
	else if (osi->QT_correlationoperator[i_c].type_combination == 1){
	  //	  osi->Gamma2 += .5 * osi->QT_ME2_cf[i_c]; // ???
	}
	else if (osi->QT_correlationoperator[i_c].type_combination == 0){
	}
	*/
	// three-correlator don't contribute for ttbar !!!
      }
      osi->Gamma2 += .5 * (-4 * osi->gamma2Q); // multiplied by 1 (born / born)
      osi->Gamma2 += .5 * osi->beta0 * osi->Ft1born;
      //      osi->Gamma2 += .5 * pi * osi->beta0 * osi->Ft1born; // pi if osi->beta0 is defined with a pi in the denominator !!!

      if (osi->initial_diag_gg){//  [Gamma_t^1, F_t^1] -> only in gg channel !!!
	osi->Gamma2 += .5 * osi->commutator_Gammat1_Ft1;
      }
    }
    /*
    else if (osi->initial_pdf_gq){
      logger << LOG_FATAL << "Gamma²_t does not contribute in the gq channel !!!" << endline_flush;
    }
    */
    
  logger << LOG_DEBUG_POINT << setw(30) << "Gamma2" << " = " << setprecision(15) << setw(23) << osi->Gamma2 << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_Gamma1_squared(){
  static Logger logger("ppqq20_amplitude_set::calculate_Gamma1_squared");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // 4 * Re (-1/4) (-1/4) * {...}
  osi->Gamma1squared = 0.;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    for (int j_c = 0; j_c < osi->QT_correlationoperator.size(); j_c++){
      double temp_B4 = osi->QT_ME2_cfcf[i_c][j_c];

      if (osi->QT_correlationoperator[i_c].type_combination == 1 && osi->QT_correlationoperator[j_c].type_combination == 1){
	osi->Gamma1squared += .25 * temp_B4;
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 1 && osi->QT_correlationoperator[j_c].type_combination == 2){
	osi->Gamma1squared += .25 * temp_B4
	  * (1. / v34) * log_v34;
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 1 && osi->QT_correlationoperator[j_c].type_combination == 3){
	osi->Gamma1squared += .25 * temp_B4
	  * log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[j_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[j_c].no_spectator]), 2) / (q2 * m2_HQ));
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 2 && osi->QT_correlationoperator[j_c].type_combination == 1){
	osi->Gamma1squared += .25 * temp_B4
	  * (1. / v34) * log_v34;
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 2 && osi->QT_correlationoperator[j_c].type_combination == 2){
	osi->Gamma1squared += .25 * temp_B4
	  * pow((1. / v34) * log_v34, 2);
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 2 && osi->QT_correlationoperator[j_c].type_combination == 3){
	osi->Gamma1squared += .25 * temp_B4
	  * (1. / v34) * log_v34
	  * log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[j_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[j_c].no_spectator]), 2) / (q2 * m2_HQ));
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 3 && osi->QT_correlationoperator[j_c].type_combination == 1){
	osi->Gamma1squared += .25 * temp_B4
	  * log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ));
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 3 && osi->QT_correlationoperator[j_c].type_combination == 2){
	osi->Gamma1squared += .25 * temp_B4
	  * (1. / v34) * log_v34
	  * log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ));
      }
      else if (osi->QT_correlationoperator[i_c].type_combination == 3 && osi->QT_correlationoperator[j_c].type_combination == 3){
	osi->Gamma1squared += .25 * temp_B4
	  * log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[i_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[i_c].no_spectator]), 2) / (q2 * m2_HQ))
	  * log(pow(2 * (esi->p_parton[0][osi->QT_correlationoperator[j_c].no_emitter] * esi->p_parton[0][osi->QT_correlationoperator[j_c].no_spectator]), 2) / (q2 * m2_HQ));
      }
    }
  }
  logger << LOG_DEBUG_POINT << setw(30) << "Gamma1squared" << " = " << setprecision(15) << setw(23) << osi->Gamma1squared << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


char * ppqq20_amplitude_set::get_fortran_path(string cpp_path, int & length_path){
  static Logger logger("ppqq20_amplitude_set::get_fortran_path");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  
  logger << LOG_DEBUG << "cpp_path = " << cpp_path << endline_flush;

  char * fortran_path = new char[cpp_path.size() + 1];
  std::copy(cpp_path.begin(), cpp_path.end(), fortran_path);
  length_path = cpp_path.size();
  fortran_path[length_path] = '\0';

  logger << LOG_DEBUG << "fortran_path = " << fortran_path << endline_flush;
  logger << LOG_DEBUG << "length_path = " << length_path << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished - fortran_path = " << fortran_path << endline_flush;
  return fortran_path;  
}

void ppqq20_amplitude_set::initialization_grid(){
  static Logger logger("ppqq20_amplitude_set::initialization_grid");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  CA = C_A;
  nf = double(osi->N_f);
  b0 = 11. / 2 - nf / 3; 
  b1 = 51. / 2 - 19. / 6 * nf;
  
  temp_channel = 0;
  Ci = 0.;
  gama1 = 0.;
  gama2 = 0.;
  
  ch_path_Fex2_T3T4_CA = get_fortran_path(osi->csi->path_MUNICH + "amp/ppqq20/softfunction/grids/Fex2_T3T4_CA.dat", len_Fex2_T3T4_CA);
  logger << LOG_DEBUG << "ch_path_Fex2_T3T4_CA = " << ch_path_Fex2_T3T4_CA << "   (" << len_Fex2_T3T4_CA << ")" << endline_flush;
    
  ch_path_Fex2_T3T4_nf = get_fortran_path(osi->csi->path_MUNICH + "amp/ppqq20/softfunction/grids/Fex2_T3T4_nf.dat", len_Fex2_T3T4_nf);
  logger << LOG_DEBUG << "ch_path_Fex2_T3T4_nf = " << ch_path_Fex2_T3T4_nf << "   (" << len_Fex2_T3T4_nf << ")" << endline_flush;

  ch_path_Fex11_T3T4 = get_fortran_path(osi->csi->path_MUNICH + "amp/ppqq20/softfunction/grids/Fex11_T3T4.dat", len_Fex11_T3T4);
  logger << LOG_DEBUG << "ch_path_Fex11_T3T4 = " << ch_path_Fex11_T3T4 << "   (" << len_Fex11_T3T4 << ")" << endline_flush;
   
  ch_path_Fex1sq_T3T4T3T4 = get_fortran_path(osi->csi->path_MUNICH + "amp/ppqq20/softfunction/grids/Fex1sq_T3T4T3T4.dat", len_Fex1sq_T3T4T3T4);
  logger << LOG_DEBUG << "ch_path_Fex1sq_T3T4T3T4 = " << ch_path_Fex1sq_T3T4T3T4 << "   (" << len_Fex1sq_T3T4T3T4 << ")" << endline_flush;
    
  ch_path_Fex1sq_T3T4TITj = get_fortran_path(osi->csi->path_MUNICH + "amp/ppqq20/softfunction/grids/Fex1sq_T3T4TITj.dat", len_Fex1sq_T3T4TITj);
  logger << LOG_DEBUG << "ch_path_Fex1sq_T3T4TITj = " << ch_path_Fex1sq_T3T4TITj << "   (" << len_Fex1sq_T3T4TITj << ")" << endline_flush;

  ch_path_Fex1sq_TITjTKTl = get_fortran_path(osi->csi->path_MUNICH + "amp/ppqq20/softfunction/grids/Fex1sq_TITjTKTl.dat", len_Fex1sq_TITjTKTl);
  logger << LOG_DEBUG << "ch_path_Fex1sq_TITjTKTl = " << ch_path_Fex1sq_TITjTKTl << "   (" << len_Fex1sq_TITjTKTl << ")" << endline_flush;
   
  string temp_nf = "";
  if (osi->N_f == 4){temp_nf = "_nf4";}
  if (osi->N_f == 3){temp_nf = "_nf3";}

  ch_path_gridgg = get_fortran_path(osi->csi->path_MUNICH + "amp/ppqq20/softfunction/grids/gridgg" + temp_nf + ".dat", len_gridgg);
  logger << LOG_DEBUG << "ch_path_gridgg = " << ch_path_gridgg << "   (" << len_gridgg << ")" << endline_flush;

  ch_path_gridqq = get_fortran_path(osi->csi->path_MUNICH + "amp/ppqq20/softfunction/grids/gridqq" + temp_nf + ".dat", len_gridqq);
  logger << LOG_DEBUG << "ch_path_gridqq = " << ch_path_gridqq << "   (" << len_gridqq << ")" << endline_flush;

  init_interpolation_t3t4_(&CA, &nf, ch_path_Fex2_T3T4_CA, ch_path_Fex2_T3T4_nf, ch_path_Fex11_T3T4, &len_Fex2_T3T4_CA, &len_Fex2_T3T4_nf, &len_Fex11_T3T4);
  init_interpolation_tttt_(ch_path_Fex1sq_T3T4T3T4, ch_path_Fex1sq_T3T4TITj, ch_path_Fex1sq_TITjTKTl, &len_Fex1sq_T3T4T3T4, &len_Fex1sq_T3T4TITj, &len_Fex1sq_TITjTKTl);

  if (osi->initial_gg){
    temp_channel = 1;
    Ci = C_A;
    gama1 = -b0 / 2;
    gama2 = (2 * C_F * nf + (C_A * (256. / 27 - (2 * pi2) / 9) * nf) / 2
	     + C_A*C_A * (-692. / 27 + (11. * pi2) / 18 + 2 * zeta3)) / 16;
  }
  else if (osi->initial_qqx){
    temp_channel = 2; 
    Ci = C_F;
    gama1 = (-3 * C_F) / 4;
    gama2 = ((C_F * (130. / 27 + (2 * pi2) / 3) * nf ) / 2
	     + C_F*C_F * (-3. / 2 + 2 * pi2 - 24 * zeta3) 
	     + C_A * C_F * (-961. / 54 - (11 * pi2) / 6 + 26 * zeta3)) / 16;
  }

  M1M1 = 0.;
  M0M0 = 0.;
  M0M0_3434 = 0.;
  M2M0 = 0.;
  M0M04c.resize(5, vector<vector<vector<double> > > (5, vector<vector<double> > (5, vector<double> (5, 0.))));
  M0M0cc.resize(5, vector<double> (5, 0.));
  M1M0cc.resize(5, vector<double> (5, 0.));
  imM1M0cc.resize(5, vector<double> (5, 0.));
  M1M0 = 0.;
  imM1M0 = 0.;

  m_HQ = osi->msi->M[osi->csi->type_parton[0][3]];
  m2_HQ = m_HQ*m_HQ;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void ppqq20_amplitude_set::calculate_H2_2loop(){
  static Logger logger("ppqq20_amplitude_set::calculate_H2_2loop");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double sn = (osi->esi->p_parton[0][1] + osi->esi->p_parton[0][2]).m2();
  //  double sn = 2 * osi->esi->p_parton[0][1] * osi->esi->p_parton[0][2];
  double tn = (osi->esi->p_parton[0][1] - osi->esi->p_parton[0][3]).m2();
  //  2d0*dot(p,1,3)+mt**2.d0;
  double un = (osi->esi->p_parton[0][1] - osi->esi->p_parton[0][4]).m2();
    //2d0*dot(p,2,3)+mt**2.d0;
  double beta = sqrt(1. - 4 * m2_HQ / sn);
  //  double beta = dsqrt(1.d0-4.d0*mt**2.d0/sn);
  //  double cost = (tn-un)/sn/beta;
  double cost = (tn - un) / sn / beta;
  logger << LOG_DEBUG_POINT << "beta = " << setw(23) << setprecision(17) << showpoint << beta << endline_flush;
  logger << LOG_DEBUG_POINT << "cost = " << setw(23) << setprecision(17) << showpoint << cost << endline_flush;

  /*
  //  only once per run:
  double CA = C_A;
  double nf = double(osi->N_f);
  init_interpolation_t3t4_(&CA, &nf);
  init_interpolation_tttt_();
  */
  int i1 = 1;
  int i2 = 2;
  int i3 = 3;
  int i4 = 4;
  //  double test = 0.;
  /*
  beta = .5;
  cost = .5;
  
  beta = sqrt(1.-0.5);
  cost = 0.974679434480896;
  
  //  test = h1_TT_(&i1, &i2, &beta, &costheta);
  //  test = h1_TT_(&i1, &i2, &beta, &costheta);
  //  test = h1_TTTT_(&i1, &i2, &i3, &i4, &beta, &costheta);
  //  test = virtgg_(&beta, &cost);
  //  cout << "test = " << test << endline_flush;
  //  test = xxxvirtgg_(&beta, &cost);
  //  cout << "h1_tt_(&i3,4) = " << h1_tt_(&beta,&cost) << endline_flush;
  //  double x1 = h1_tt_(&i3,&i4,&beta,&cost);
  cout << "h1_tt_(&i3,4) = " << setw(23) << setprecision(17) << showpoint << h1_tt_(&i3,&i4,&beta,&cost) << endline_flush;
  cout << "h1_tt_(&i3,3) = " << setw(23) << setprecision(17) << showpoint << h1_tt_(&i3,&i3,&beta,&cost) << endline_flush;
  cout << "h2_tt_(&i3,4) = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i3,&i4,&beta,&cost) << endline_flush;
  cout << "h2_tt_(&i3,1) = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i3,&i1,&beta,&cost) << endline_flush;
  cout << "h2_tt_(&i4,2) = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i4,&i2,&beta,&cost) << endline_flush;
  cout << "h2_tt_(&i4,1) = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i4,&i1,&beta,&cost) << endline_flush;
  cout << "h2_tt_(&i3,2) = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i3,&i2,&beta,&cost) << endline_flush;
  cout << "h2_tttt_(&i3,&i4,&i3,4) = " << setw(23) << setprecision(17) << showpoint << h2_tttt_(&i3,&i4,&i3,&i4,&beta,&cost) << endline_flush;
  cout << "h2_tttt_(&i3,&i4,&i3,3) = " << setw(23) << setprecision(17) << showpoint << h2_tttt_(&i3,&i4,&i3,&i3,&beta,&cost) << endline_flush;
  cout << "h2_tttt_(&i3,&i3,&i3,3) = " << setw(23) << setprecision(17) << showpoint << h2_tttt_(&i3,&i3,&i3,&i3,&beta,&cost) << endline_flush;
  for (int i = 2; i < 21; i++){
    beta = 1. / i;
    cout << setw(23) << setprecision(17) << showpoint << beta << "   " << setw(23) << setprecision(17) << showpoint << virtgg_(&beta, &cost) << "   " << setw(23) << setprecision(17) << showpoint << virtqq_(&beta, &cost) << endline_flush;
  }
  */
  //C I evaluate the M2*M0 + cc. Following equation 3.6 of
  //C 1312.6279, this can be obtained from the grid as

  double virtxx = 0.;
  if (osi->initial_gg){
    logger << LOG_DEBUG_POINT << "virtgg = " << setw(23) << setprecision(17) << showpoint << virtxx << endline_flush;
    virtxx = virtgg_(&beta, &cost, &nf, ch_path_gridgg, &len_gridgg);
    logger << LOG_DEBUG_POINT << "virtgg = " << setw(23) << setprecision(17) << showpoint << virtxx << endline_flush;
  }
  else if (osi->initial_qqx){
    logger << LOG_DEBUG_POINT << "virtqq = " << setw(23) << setprecision(17) << showpoint << virtxx << endline_flush;
    virtxx = virtqq_(&beta, &cost, &nf, ch_path_gridqq, &len_gridqq);
    logger << LOG_DEBUG_POINT << "virtqq = " << setw(23) << setprecision(17) << showpoint << virtxx << endline_flush;
  }

  M2M0 = virtxx * 16 * pi / (beta * (1. - pow(beta, 2)));
  /////    M2M0 = virtgg_(&beta, &cost) * 16 * pi / (beta * (1. - pow(beta, 2)));
  /////    M2M0 = virtgg_(&beta, &cost) * 4096 * pi / (beta * (1. - pow(beta, 2)));
  //C where the expansion for M is done in powers of as/(2*pi).
  //C To be consistent with the rest, I add the overall power
  //C of as, which otherwise is missing:
  M2M0 = M2M0 * pow(4 * pi * osi->alpha_S, 2);
  //    logger << LOG_DEBUG_POINT << "4 * pi * osi->alpha_S = " << setw(23) << setprecision(17) << showpoint << 4 * pi * osi->alpha_S << endline_flush;
  
  //C I add factors from avergage of initial state d.o.f
  /////    double facgg = 1. / (8 * 8 * 2 * 2);
  /////    M2M0 = M2M0 * facgg;
  //  double temp_B4;
  /*    
  logger << LOG_DEBUG_POINT << setw(22) << "before M0M04c" << "[" << setw(2) << i3 << "][" << setw(2) << i4 << "][" << setw(2) << i3 << "][" << setw(2) << i4 << "]" << " = " << setprecision(15) << setw(23) << M0M04c[i3][i4][i3][i4] << endline_flush;
  fourcorrelators_(&temp_B4, &i3, &i4, &i3, &i4, &temp_channel);
  // C Finally, the tree-level 4 parton correlator 3434 I take from Hayk
  M0M0_3434 = temp_B4 * M0M0;
  logger << LOG_DEBUG_POINT << setw(38) << "after  M0M0_3434" << " = " << setprecision(15) << setw(23) << M0M0_3434 << endline_flush;
  //  logger << LOG_DEBUG_POINT << "fourcorrelators_(&temp_B4, " << i3 << ", " << i4 << ", " << i3 << ", " << i4 << ", " << temp_channel << "); -> temp_B4 = " << temp_B4 << endline_flush;
  */
  M0M0_3434 = M0M04c[i3][i4][i3][i4];
  //  M0M0_3434 = M0M04c[3][4][3][4];
  
  /////    M0M0 = m0_OL;
  //C The convention from Massimiliano is different from the one in
  //C OpenLoops, so I need to fix that:

  double fac = Ci * pi2_12;
  for (int i = 1; i < 5; i++){
    for (int j = 1; j < 5; j++){
      M1M0cc[i][j] = M1M0cc[i][j] + 2 * fac * M0M0cc[i][j];
    }
  }
  M1M1 = M1M1 + fac * M1M0 + pow(fac, 2) * M0M0;
  M1M0 = M1M0 + 2 * fac * M0M0;

  logger << LOG_DEBUG_POINT << "norm   M0M0 = " << setw(23) << setprecision(17) << showpoint << M0M0 << endline_flush;
  logger << LOG_DEBUG_POINT << "norm   M0M0_3434 = " << setw(23) << setprecision(17) << showpoint << M0M0_3434 << endline_flush;
  logger << LOG_DEBUG_POINT << "norm   imM1M0 = " << setw(23) << setprecision(17) << showpoint << imM1M0 << endline_flush;
  logger << LOG_DEBUG_POINT << "norm   M2M0 = " << setw(23) << setprecision(17) << showpoint << M2M0 << endline_flush;
  logger << LOG_DEBUG_POINT << "shifted   M1M0 = " << setw(23) << setprecision(17) << showpoint << M1M0 << endline_flush;
  logger << LOG_DEBUG_POINT << "shifted   M1M1 = " << setw(23) << setprecision(17) << showpoint << M1M1 << endline_flush;
  for (int i = 0; i < 16; i++){logger << LOG_DEBUG_POINT << "M0M0cc[" << i % 4 + 1 << "][" << i / 4 + 1 << "] = " << M0M0cc[i % 4 + 1][i / 4 + 1] << endline_flush;}
  for (int i = 0; i < 16; i++){logger << LOG_DEBUG_POINT << "shifted   M1M0cc[" << i % 4 + 1 << "][" << i / 4 + 1 << "] = " << M1M0cc[i % 4 + 1][i / 4 + 1] << endline_flush;}
  for (int i = 0; i < 16; i++){logger << LOG_DEBUG_POINT << "norm   imM1M0cc[" << i % 4 + 1 << "][" << i / 4 + 1 << "] = " << imM1M0cc[i % 4 + 1][i / 4 + 1] << endline_flush;}

  osi->QT_H2_delta_QCD =
    M1M1
    + M2M0
    + (M1M0 * (-(Ci * pi2) / 3 + 4 * C_F * h1_tt_(&i3, &i3, &beta, &cost))) / 2
    + M0M0 * ((Ci*Ci * pi4) / 72
	      + (Ci * (C_A * (4856. - 603. * pi2 + 18. * pi4 - 2772. * zeta3)
		       + nf * (-656. + 90. * pi2 + 504. * zeta3)
		       - 216. * C_F * pi2 * h1_tt_(&i3, &i3, &beta, &cost))) / 648
	      + C_F*C_F * (pow(h1_tt_(&i3, &i3, &beta, &cost), 2) 
			   + 2 * h2_tttt_(&i3, &i3, &i3, &i3, &beta, &cost)))
    + (pow(h1_tt_(&i3,&i4, &beta, &cost), 2) 
       + 2 * h2_tttt_(&i3, &i4, &i3, &i4, &beta, &cost)) * M0M0_3434
    + 2 * h2_tt_(&i3, &i1, &beta, &cost, &nf) * M0M0cc[1][3] 
    + 2 * h2_tt_(&i4, &i1, &beta, &cost, &nf) * M0M0cc[1][4]
    + 2 * h2_tt_(&i3, &i2, &beta, &cost, &nf) * M0M0cc[2][3] 
    + 2 * h2_tt_(&i4, &i2, &beta, &cost, &nf) * M0M0cc[2][4]
    + (-(Ci * pi2 * h1_tt_(&i3, &i4, &beta, &cost)) / 3
       + 2 * (h2_tt_(&i3, &i4, &beta, &cost, &nf)
	      + C_F * (h1_tt_(&i3, &i3, &beta, &cost) * h1_tt_(&i3, &i4, &beta, &cost)
		       + h2_tttt_(&i3, &i4, &i3, &i3, &beta, &cost)))) * M0M0cc[3][4]
    + 2 * h1_tt_(&i3, &i4, &beta, &cost) * M1M0cc[3][4];
  
  logger << LOG_DEBUG_POINT << "done" << endline_flush;
  logger << LOG_DEBUG_POINT << "h1_tt_(" << i3 << ", " << i3 << ") = " << setw(23) << setprecision(17) << showpoint << h1_tt_(&i3,&i3,&beta,&cost) << endline_flush;
  logger << LOG_DEBUG_POINT << "h1_tt_(" << i3 << ", " << i4 << ") = " << setw(23) << setprecision(17) << showpoint << h1_tt_(&i3,&i4,&beta,&cost) << endline_flush;
  logger << LOG_DEBUG_POINT << "h2_tt_(" << i3 << ", " << i1 << ") = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i3,&i1,&beta,&cost,&nf) << endline_flush;
  logger << LOG_DEBUG_POINT << "h2_tt_(" << i4 << ", " << i1 << ") = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i4,&i1,&beta,&cost,&nf) << endline_flush;
  logger << LOG_DEBUG_POINT << "h2_tt_(" << i3 << ", " << i2 << ") = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i3,&i2,&beta,&cost,&nf) << endline_flush;
  logger << LOG_DEBUG_POINT << "h2_tt_(" << i4 << ", " << i2 << ") = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i4,&i2,&beta,&cost,&nf) << endline_flush;
  logger << LOG_DEBUG_POINT << "h2_tt_(" << i3 << ", " << i4 << ") = " << setw(23) << setprecision(17) << showpoint << h2_tt_(&i3,&i4,&beta,&cost,&nf) << endline_flush;
  logger << LOG_DEBUG_POINT << "h2_tttt_(" << i3 << ", " << i3 << ", " << i3 << ", " << i3 << ") = " << setw(23) << setprecision(17) << showpoint << h2_tttt_(&i3,&i3,&i3,&i3,&beta,&cost) << endline_flush;
  logger << LOG_DEBUG_POINT << "h2_tttt_(" << i3 << ", " << i4 << ", " << i3 << ", " << i4 << ") = " << setw(23) << setprecision(17) << showpoint << h2_tttt_(&i3,&i4,&i3,&i4,&beta,&cost) << endline_flush;
  logger << LOG_DEBUG_POINT << "h2_tttt_(" << i3 << ", " << i4 << ", " << i3 << ", " << i3 << ") = " << setw(23) << setprecision(17) << showpoint << h2_tttt_(&i3,&i4,&i3,&i3,&beta,&cost) << endline_flush;

  logger << LOG_DEBUG_POINT << "osi->QT_H2_delta_QCD(1) = " << osi->QT_H2_delta_QCD << endline_flush;

  //C I add the term for the scale mismatch between mt and M in 2-loop virtuals
  double logM2mt2 = log(sn / m2_HQ);

  osi->QT_H2_delta_QCD += 2 * logM2mt2 * (b1 * M0M0 + b0 * M1M0 - b0*b0 * M0M0 * logM2mt2);
  
  logger << LOG_DEBUG_POINT << "osi->QT_H2_delta_QCD(2) = " << osi->QT_H2_delta_QCD << endline_flush;
  
  //C I add now the additional term due to the fact that the subtraction
  //C is performed for mu=mt instead of mu=M
  //C I start by defining gama1g and gama2g and K
  
  /////    double gama1 = -b0 / 2;
  /////    double gama2 = (2 * C_F * nf + (C_A * (256. / 27 - (2 * pi2) / 9) * nf) / 2
  /////		      + C_A*C_A * (-692. / 27 + (11. * pi2) / 18 + 2 * zeta3)) / 16;
  double KK = (67. / 18 - pi2_6) * C_A - 5. / 9 * nf;
  double gamaN = (20. * C_F * nf) / 9 + C_A * C_F * (-98. / 9 + (2 * pi2) / 3 - 4 * zeta3);
  
  //C I define v and b34
  ///      double complex b34,expb34,li2b34,li3b34,HPL2,HPL3,CothB34,aux1C,
  ///     &               aux2C,M,Mcc(4,4),aux3C
    
    double vv = sqrt(1. - pow(m2_HQ / (sn / 2 - m2_HQ), 2));
		     //dot(p,3,4)**2.d0);
    double_complex b34 = ri * pi + 1. / 2 * log((1. - vv) / (1. + vv));
    double_complex expB34 = exp(-2. * b34);
    int i0 = 0;
    double_complex li2b34 = hpl2_(&i0, &i1, &expB34);
    double_complex li3b34 = hpl3_(&i0, &i0, &i1, &expB34);
    double_complex CothB34 = (exp(-b34) + exp(b34)) / (-exp(-b34) + exp(b34));
      
    // C I write now all together the terms
    // C -4*b0*logM2mt2*Re(<M0|U1|M0>)+2*Re(<M0|U1|M1>)

    // C I define an auxiliary matrix element:
    // C TODO CHECK IMAGINARY PART
    double_complex M = -4. * b0 * logM2mt2 * M0M0 + M1M0 + ri * imM1M0;
    vector<vector<double_complex> > Mcc(5, vector<double_complex> (5, 0.));
    for (int i = 1; i < 5; i++){
      for (int j = 1; j < 5; j++){
        Mcc[i][j] = -4. * b0 * logM2mt2 * M0M0cc[i][j] + M1M0cc[i][j] + ri * imM1M0cc[i][j];
      }
    }

    double_complex aux1C = (logM2mt2
			    * (-2 * C_F * M + 4 * gama1 * M
			       + Ci * logM2mt2 * M - 2 * Ci * ri * M * pi  
			       - 2. * b34 * CothB34 * Mcc[3][4])) / 2.;
    for (int i = 3; i < 5; i++){
      for (int j = 1; j < 3; j++){
	aux1C += ((logM2mt2 * (-(ri * pi)
			       + log((m2_HQ * sn) / (4. * pow(osi->esi->p_parton[0][i] * osi->esi->p_parton[0][j], 2)))) * Mcc[i][j]) / 2.);
      }
    }
    
    logger << LOG_DEBUG_POINT << "aux1C = " << aux1C << endline_flush;
    logger << LOG_DEBUG_POINT << "real(aux1C) = " << real(aux1C) << endline_flush;

    osi->QT_H2_delta_QCD += real(aux1C);

    logger << LOG_DEBUG_POINT << "osi->QT_H2_delta_QCD(3) = " << osi->QT_H2_delta_QCD << endline_flush;

    //C I write now the term 2*Re(<M0|U2-U1^2|M0>)
    
    /*
    //    double temp_B4;
    for (int i = 3; i < 5; i++){
      for (int j = 1; j < 3; j++){
	logger << LOG_DEBUG_POINT << setw(22) << "trans  M0M04c" << "[" << setw(2) << j << "][" << setw(2) << i << "][" << setw(2) << i3 << "][" << setw(2) << i4 << "]" << " = " << setprecision(15) << setw(23) << M0M04c[j][i][i3][i4] << endline_flush;
	logger << LOG_DEBUG_POINT << setw(22) << "before M0M04c" << "[" << setw(2) << i << "][" << setw(2) << j << "][" << setw(2) << i3 << "][" << setw(2) << i4 << "]" << " = " << setprecision(15) << setw(23) << M0M04c[i][j][i3][i4] << endline_flush;
	temp_B4 = 0.;
	fourcorrelators_(&temp_B4, &j, &i, &i3, &i4, &temp_channel);
	//	logger << LOG_DEBUG_POINT << "fourcorrelators_(&temp_B4, " << j << ", " << i << ", " << i3 << ", " << i4 << ", " << temp_channel << "); -> temp_B4 = " << temp_B4 << endline_flush;
        M0M04c[i][j][3][4] = temp_B4 * M0M0;
	logger << LOG_DEBUG_POINT << setw(22) << "after  M0M04c" << "[" << setw(2) << i << "][" << setw(2) << j << "][" << setw(2) << i3 << "][" << setw(2) << i4 << "]" << " = " << setprecision(15) << setw(23) << M0M04c[i][j][i3][i4] << endline_flush;
	for (int k = 3; k < 5; k++){
	  for (int l = 1; l < 3; l++){
	    logger << LOG_DEBUG_POINT << setw(22) << "trans  M0M04c" << "[" << setw(2) << j << "][" << setw(2) << i << "][" << setw(2) << l << "][" << setw(2) << k << "]" << " = " << setprecision(15) << setw(23) << M0M04c[j][i][l][k] << endline_flush;
	    logger << LOG_DEBUG_POINT << setw(22) << "before M0M04c" << "[" << setw(2) << i << "][" << setw(2) << j << "][" << setw(2) << k << "][" << setw(2) << l << "]" << " = " << setprecision(15) << setw(23) << M0M04c[i][j][k][l] << endline_flush;
	    temp_B4 = 0.;
	    fourcorrelators_(&temp_B4, &j, &i, &l, &k, &temp_channel);
	    //	logger << LOG_DEBUG_POINT << "fourcorrelators_(&temp_B4, " << j << ", " << i << ", " << l << ", " << k << ", " << temp_channel << "); -> temp_B4 = " << temp_B4 << endline_flush;
	    M0M04c[i][j][k][l] = temp_B4 * M0M0;
	    logger << LOG_DEBUG_POINT << setw(22) << "after  M0M04c" << "[" << setw(2) << i << "][" << setw(2) << j << "][" << setw(2) << k << "][" << setw(2) << l << "]" << " = " << setprecision(15) << setw(23) << M0M04c[i][j][k][l] << endline_flush;
	  }
	}
      }
    }
    */
    double_complex aux2C = (logM2mt2
			    * (-36. * b34*b34 * CothB34*CothB34 * logM2mt2*M0M0_3434
			       + 6. * M0M0 * (48. * gama2 + 3. * gamaN - 6. * b0 * C_F * logM2mt2
					      + 12. * b0 * gama1 * logM2mt2 + 6. * Ci * KK * logM2mt2
					      + 4. * b0 * Ci * logM2mt2 * logM2mt2 - 12. * Ci * ri * KK * pi
					      - 6. * b0 * Ci * ri * logM2mt2 * pi)
			       - 9. * logM2mt2 * M0M0 * pow(-2. * C_F + 4. * gama1 + Ci * (logM2mt2 - 2. * ri * pi), 2)
			       - 4. * (-18. * b34*b34 * C_A * (-1. + CothB34)
				       + 6. * pow(b34, 3) * C_A * (-1. + CothB34) * CothB34
				       + b34 * CothB34 * (9. * b0 * logM2mt2 + 18. * C_F * logM2mt2
							  - 36. * gama1 * logM2mt2 - 9. * Ci * logM2mt2 * logM2mt2
							  + 18. * Ci * ri * logM2mt2 * pi
							  + C_A * (67. - 6. * pi2
								   + 3. * CothB34 * (6. * li2b34 + pi2)) - 10. * nf)
				       + 3. * C_A * (pi2 + CothB34 * (6. * li2b34 - pi2)
						     + 6. * CothB34*CothB34 * (li3b34 - zeta3) + 6. * zeta3)
				       - 36. * b34 * C_A * CothB34 * log(1. - expB34)) * M0M0cc[3][4])) / 36.;
      
    for (int i = 3; i < 5; i++){
      for (int j = 1; j < 3; j++){
	aux2C += 
	  (logM2mt2 * (pi * (-9. * b0 * ri * logM2mt2 - 18. * C_F * ri * logM2mt2
			     + 36. * gama1 * ri * logM2mt2 + 9. * Ci * ri * logM2mt2 * logM2mt2
			     + 18. * Ci * logM2mt2 * pi + C_A * ri * (-67. + 3. * pi2)
			     + 10. * ri * nf)
		       + (9. * b0 * logM2mt2 + 18. * C_F * logM2mt2  
			  - 36. * gama1 * logM2mt2 - 9. * Ci * logM2mt2 * logM2mt2  
			  + 18. * Ci * ri * logM2mt2 * pi + C_A * (67. - 3. * pi2)  
			  - 10. * nf) * log((m2_HQ * sn) / (4. * pow(osi->esi->p_parton[0][i] * osi->esi->p_parton[0][j], 2))))
	   * M0M0cc[i][j]) / 18.  
	  + b34 * CothB34 * logM2mt2*logM2mt2
	  * (-(ri * pi) + log((m2_HQ * sn) / (4. * pow(osi->esi->p_parton[0][i] * osi->esi->p_parton[0][j], 2))))
	  * M0M04c[i][j][3][4];
	
	for (int k = 3; k < 5; k++){
	  for (int l = 1; l < 3; l++){
	    aux2C += (
		      (logM2mt2*logM2mt2
		       * (pi2 + log((m2_HQ * sn) / (4. * pow(osi->esi->p_parton[0][i] * osi->esi->p_parton[0][j], 2)))
			  * (2. * ri * pi - log((m2_HQ * sn) / (4. * pow(osi->esi->p_parton[0][k] * osi->esi->p_parton[0][l], 2)))))) / 4.) * M0M04c[i][j][k][l];
	  }
	}
      }
    }
        
    logger << LOG_DEBUG_POINT << "aux2C = " << aux2C << endline_flush;
    logger << LOG_DEBUG_POINT << "real(aux2C) = " << real(aux2C) << endline_flush;

    osi->QT_H2_delta_QCD += real(aux2C);
 
    logger << LOG_DEBUG_POINT << "osi->QT_H2_delta_QCD = " << osi->QT_H2_delta_QCD << endline_flush;
  
  

  
  osi->QT_A0 = osi->VA_b_ME2;
  // Check if this if statements change the "rest" contribution !!!
  // Better: Avoid calculation of the ingredients if not finally needed for the chosen channels !!!
  //  if (osi->initial_diag || osi->initial_pdf_gq){
  osi->QT_A1 = osi->VA_V_ME2 + osi->VA_X_ME2;
  //  osi->QT_A1 = osi->VA_V_ME2;
  osi->QT_H1_delta_QCD = osi->QT_A1 / 2 / (osi->alpha_S * inv2pi * osi->QT_A0);
  //  }
  //  else {osi->QT_H1_delta_QCD = 0.;}
  //  if (osi->initial_diag){
  //  osi->QT_A2 = ME2_L_L + ME2_L2_B;
  osi->QT_H2_delta_QCD = osi->QT_H2_delta_QCD / 4 / (osi->QT_A0);
  logger << LOG_DEBUG_POINT << "osi->QT_H2_delta_QCD / 4 / (osi->QT_A0) = " << osi->QT_H2_delta_QCD << endline_flush;

    //  osi->QT_H2_delta_QCD = osi->QT_A2 / 4 / (osi->QT_A0);
  //pow(osi->alpha_S * inv2pi, 2) * 
  
  //  }
  //  else {osi->QT_H2_delta_QCD = 0.;}
  

  //  osi->QT_H1_delta_QCD = 0.;
  //  osi->QT_H2_delta_QCD = 0.;

  // Catani scheme -> qT scheme
  
  // adapt alpha_S/Pi normalisation
  //  osi->QT_H1_delta_QCD /= 2;
  //  osi->QT_H2_delta_QCD /= 4;
  
  //  osi->QT_H1_delta_QCD /= (inv2pi * osi->alpha_S * osi->QT_A0);
  //  osi->QT_H2_delta_QCD /= (pow(inv2pi * osi->alpha_S, 2) * osi->QT_A0);
  
       
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


string ppqq20_amplitude_set::output_amplitude_VT_QCD(){
  static Logger logger("ppqq20_amplitude_set::output_amplitude_VT_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;
  out_comparison << amplitude_set::output_amplitude_VT_QCD() << endline;
  out_comparison << setw(12) << " Ft1born = " << setprecision(15) << setw(23) << osi->Ft1born << endline;
  out_comparison << endline;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    out_comparison << setw(12) << " QT_ME2_cf[" << i_c << "] = " << setprecision(15) << setw(23) << osi->QT_ME2_cf[i_c] << "   em = " << osi->QT_correlationoperator[i_c].no_emitter << "   sp = " << osi->QT_correlationoperator[i_c].no_spectator << "   (type_combination = " << osi->QT_correlationoperator[i_c].type_combination << ")" << endline;
  }
  out_comparison << endline;

  logger << LOG_DEBUG_VERBOSE << "finished - return (string):" << endline << out_comparison.str() << endline_flush;
  return out_comparison.str();
}


string ppqq20_amplitude_set::output_amplitude_CT_QCD(){
  static Logger logger("ppqq20_amplitude_set::output_amplitude_CT_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;
  out_comparison << amplitude_set::output_amplitude_CT_QCD() << endline;
  //  if (osi->initial_pdf_diag){
  out_comparison << setw(12) << " Gamma1born = " << setprecision(15) << setw(23) << osi->Gamma1born << endline;
  out_comparison << endline;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    out_comparison << setw(12) << " QT_ME2_cf[" << i_c << "] = " << setprecision(15) << setw(23) << osi->QT_ME2_cf[i_c] << "   em = " << osi->QT_correlationoperator[i_c].no_emitter << "   sp = " << osi->QT_correlationoperator[i_c].no_spectator << "   (type_combination = " << osi->QT_correlationoperator[i_c].type_combination << ")" << endline;
  }
  out_comparison << endline;
  //  }

  logger << LOG_DEBUG_VERBOSE << "finished - return (string):" << endline << out_comparison.str() << endline_flush;
  return out_comparison.str();
}


string ppqq20_amplitude_set::output_amplitude_VT2_QCD(){
  static Logger logger("ppqq20_amplitude_set::output_amplitude_VT2_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;
  out_comparison << amplitude_set::output_amplitude_VT2_QCD() << endline;
  out_comparison << setw(25) << "QT_H0_doublespinflip" << " = " << setprecision(15) << setw(23) << osi->QT_H0_doublespinflip << endline;
  out_comparison << setw(25) << "QT_H0_DG" << " = " << setprecision(15) << setw(23) << osi->QT_H0_DG << endline;
 out_comparison << endline;
 //  if (osi->initial_pdf_diag || osi->initial_pdf_gq){
  out_comparison << setw(25) << "Ft1born" << " = " << setprecision(15) << setw(23) << osi->Ft1born << endline;
  out_comparison << endline;
  out_comparison << setw(25) << "Gamma1born" << " = " << setprecision(15) << setw(23) << osi->Gamma1born << endline;
  out_comparison << endline;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    out_comparison << setw(21) << "QT_ME2_cf" << "[" << setw(2) << i_c << "]" << " = " << setprecision(15) << setw(23) << osi->QT_ME2_cf[i_c] << endline;
  }
  out_comparison << endline;
  for (int i_i = 1; i_i < csi->n_particle + 3; i_i++){//if (!is_qcd_charged[i_i]){continue;}
    for (int i_j = i_i + 1; i_j < csi->n_particle + 3; i_j++){//if (!is_qcd_charged[i_j]){continue;}
      for (int i_k = 1; i_k < csi->n_particle + 3; i_k++){//if (!is_qcd_charged[i_k]){continue;}
	for (int i_l = i_k + 1; i_l < csi->n_particle + 3; i_l++){//if (!is_qcd_charged[i_l]){continue;}
	  out_comparison << setw(13) << "M0M04c" << "[" << i_i << "][" << i_j << "][" << i_k << "][" << i_l << "] = " << setprecision(15) << setw(23) << M0M04c[i_i][i_j][i_k][i_l] << endline;
	}
      }
    }
  }
  out_comparison << endline;
  //  }

  logger << LOG_DEBUG_VERBOSE << "finished - return (string):" << endline << out_comparison.str() << endline_flush;
  return out_comparison.str();
}


string ppqq20_amplitude_set::output_amplitude_CT2_QCD(){
  static Logger logger("ppqq20_amplitude_set::output_amplitude_CT2_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;
  out_comparison << amplitude_set::output_amplitude_CT2_QCD() << endline;
  out_comparison << setw(25) << "Ft1born" << " = " << setprecision(15) << setw(23) << osi->Ft1born << endline;
  out_comparison << endline;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    out_comparison << setw(21) << "Ft1born_4correlator" << "[" << setw(2) << i_c << "]" << " = " << setprecision(15) << setw(23) << osi->Ft1born_4correlator[i_c] << endline;
  }
  out_comparison << endline;
  out_comparison << setw(25) << "Gamma1born" << " = " << setprecision(15) << setw(23) << osi->Gamma1born << endline;
  out_comparison << setw(25) << "Gamma1loop" << " = " << setprecision(15) << setw(23) << osi->Gamma1loop << endline;
  out_comparison << setw(25) << "commutator_Gammat1_Ft1" << " = " << setprecision(15) << setw(23) << osi->commutator_Gammat1_Ft1 << endline;
  out_comparison << setw(25) << "Gamma2" << " = " << setprecision(15) << setw(23) << osi->Gamma2 << endline;
  out_comparison << setw(25) << "Gamma1squared" << " = " << setprecision(15) << setw(23) << osi->Gamma1squared << endline;
  out_comparison << endline;

  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    out_comparison << setw(21) << "QT_ME2_cf" << "[" << setw(2) << i_c << "]" << " = " << setprecision(15) << setw(23) << osi->QT_ME2_cf[i_c] << endline;
  }
  out_comparison << endline;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    out_comparison << setw(21) << "QT_ME2_loopcf" << "[" << setw(2) << i_c << "]" << " = " << setprecision(15) << setw(23) << osi->QT_ME2_loopcf[i_c] << endline;
  }
  out_comparison << endline;
  for (int i_c = 0; i_c < osi->QT_correlationoperator.size(); i_c++){
    for (int j_c = i_c; j_c < osi->QT_correlationoperator.size(); j_c++){
      out_comparison << setw(17) << "QT_ME2_cfcf" << "[" << setw(2) << i_c << "][" << setw(2) << j_c << "]" << " = " << setprecision(15) << setw(23) << osi->QT_ME2_cfcf[i_c][j_c] << endline;
    }
  }
  out_comparison << endline;

  logger << LOG_DEBUG_VERBOSE << "finished - return (string):" << endline << out_comparison.str() << endline_flush;
  return out_comparison.str();
}
