#ifndef AMPLITUDESET_H
#define AMPLITUDESET_H

using namespace std;

class phasespace_set;

class amplitude_set {
 private:

 public:
  amplitude_set();
  amplitude_set(observable_set * _osi, contribution_set * _csi, model_set * _msi, event_set * _esi);

  virtual ~amplitude_set(){}

  virtual void set_mu_reg();
  
  virtual void initialization_parameter(){}
  virtual void initialization_process(){}
  virtual void initialization_process_extra(){}
  virtual void output_parameter(string label){}
  virtual void generate_testpoint(){}
  virtual void output_phasespacepoints(){}

  // amplitude/amplitude.initialization.cpp
  virtual void initialization(){cout << "amplitude_set initialization called..." << endl;}
  virtual void initialization(observable_set * _osi, contribution_set * _csi, model_set * _msi, event_set * _esi, inputparameter_set * isi);
  //  virtual void initialization_basic(observable_set * _osi, contribution_set * _csi, model_set * _msi, event_set * _esi, inputparameter_set * isi);
  virtual void initialization_set_default();
  virtual void initialization_input(inputparameter_set * isi);
  virtual void initialization_N_nondecoupled_from_pdf_set();
  virtual void initialization_process_general();

  // amplitude/amplitude.spinorproducts.cpp
  virtual void calculate_spinorproducts(vector<fourvector> & p, vector<vector<double_complex > > & za, vector<vector<double_complex > > & zb, vector<vector <double> > & s);

  // amplitude/amplitude.ME2.born.cpp
  virtual void calculate_ME2_born();
  virtual void calculate_ME2check_born();
  virtual string output_amplitude_born();
  virtual void check_vanishing_ME2_born(){}


  // amplitude/amplitude.dipolesubtraction.ME2.CA.QCD.cpp
  virtual void calculate_ME2_CA_QCD();
  virtual void calculate_ME2check_CA_QCD();
  virtual string output_amplitude_CA_QCD();

  // amplitude/amplitude.dipolesubtraction.ME2.VA.QCD.cpp
  virtual void set_mu_and_delta_VA(){}
  virtual void set_testpoint_alpha_S(const double alpha_S){}
  virtual void set_testpoint_rescale_I_ME2(){}
  virtual void calculate_ME2_ioperator_VA_QCD();
  virtual void calculate_ME2_VA_QCD();
  virtual void calculate_ME2check_VA_QCD();
  virtual string output_amplitude_VA_QCD();

  // amplitude/amplitude.dipolesubtraction.ME2.RA.QCD.cpp
  virtual void calculate_ME2_RA_QCD();
  virtual void calculate_ME2check_RA_QCD(phasespace_set * psi);
  virtual string output_amplitude_RA_QCD();
  virtual void calculate_dipole_Acc_QCD(int x_a);
  virtual void calculate_dipole_Asc_QCD(int x_a, fourvector & Vtensor);
  

  // amplitude/amplitude.dipolesubtraction.ME2.CA.QEW.cpp
  virtual void calculate_ME2_CA_QEW();
  virtual void calculate_ME2check_CA_QEW();
  virtual string output_amplitude_CA_QEW();

  // amplitude/amplitude.dipolesubtraction.ME2.VA.QEW.cpp
  virtual void calculate_ME2_ioperator_VA_QEW();
  virtual void calculate_ME2_VA_QEW();
  virtual void calculate_ME2check_VA_QEW();
  virtual string output_amplitude_VA_QEW();

  // amplitude/amplitude.dipolesubtraction.ME2.RA.QEW.cpp
  virtual void calculate_ME2_RA_QEW();
  virtual void calculate_ME2check_RA_QEW(phasespace_set * psi);
  virtual string output_amplitude_RA_QEW();
  virtual void calculate_dipole_Acc_QEW(int x_a);
  virtual void calculate_dipole_Asc_QEW(int x_a, fourvector & Vtensor);
  
  // amplitude/amplitude.dipolesubtraction.ME2.VA.MIX.cpp
  virtual void calculate_ME2_ioperator_VA_MIX();
  virtual void calculate_ME2_VA_MIX();
  virtual void calculate_ME2check_VA_MIX();
  virtual string output_amplitude_VA_MIX();

  // amplitude/amplitude.dipolesubtraction.ME2.RA.MIX.cpp
  virtual void calculate_ME2_RA_MIX();
  virtual void calculate_ME2check_RA_MIX(phasespace_set * psi);
  virtual string output_amplitude_RA_MIX();

  
  // amplitude/amplitude.qTsubtraction.ME2.CT.QCD.cpp
  virtual void calculate_ME2_CT_QCD();
  virtual void calculate_ME2check_CT_QCD();
  virtual string output_amplitude_CT_QCD();

  // amplitude/amplitude.qTsubtraction.ME2.VT.QCD.cpp
  virtual void calculate_ME2_VT_QCD();
  virtual void calculate_H1_delta_QCD_from_VX_b();
  virtual void calculate_ME2check_VT_QCD();
  virtual string output_amplitude_VT_QCD();


  // amplitude/amplitude.qTsubtraction.ME2.CT2.QCD.cpp
  virtual void calculate_ME2_CT2_QCD();
  virtual void calculate_ME2check_CT2_QCD();
  virtual string output_amplitude_CT2_QCD();

  // amplitude/amplitude.qTsubtraction.ME2.VT2.QCD.cpp
  virtual void calculate_ME2_VT2_QCD();
  virtual void calculate_ME2_VT2_QCD_H0_doublespinflip();
  virtual void calculate_ME2_VT2_QCD_H0_DG();
  virtual void calculate_ME2check_VT2_QCD();
  virtual string output_amplitude_VT2_QCD();

  // amplitude/amplitude.2loop.cpp
  virtual inline void initialization_H2_QCD_2loop(){}
  virtual inline void finalization_H2_QCD_2loop(){}
  virtual void calculate_H2_QCD_2loop();
  virtual inline void initialization_H1gg_QCD_2loop(){}
  virtual inline void finalization_H1gg_QCD_2loop(){}
  virtual void calculate_H1gg_QCD_2loop();

  contribution_set * csi;
  model_set * msi;
  event_set * esi;
  observable_set * osi;

  int switch_amplitude_not_required;

  int switch_off_H1;
  int switch_off_H2;
  int switch_off_H1gg;

  int switch_mu_reg;
  double scale_reg;
  double mu_reg;
  
  double mu_Q;
  double alpha_S_mu_Q;

  int N_quarks;
  int N_nondecoupled;
  int process_id;

  int process_id_QCD;
  int process_id_QEW;

  int allowed_Delta_IRneqUV;

  int failed_evaluation;
};
#endif
