#include "header.hpp"
#ifdef RECOLA

void amplitude_Recola_set::calculate_ME2_ioperator_VA_MIX(){
  Logger logger("amplitude_Recola_set::calculate_ME2_ioperator_VA_MIX");
  logger << LOG_DEBUG_VERBOSE << "called" << flush << endl;

  static int initialization = 1;
  static vector<double> I_ME2_emitter(osi->VA_ioperator->size());
  if (initialization == 1){
    for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
      osi->VA_ME2_cf[i_a].resize((*(osi->VA_ioperator))[i_a].size());
      osi->VA_I_ME2_cf[i_a].resize((*(osi->VA_ioperator))[i_a].size());
    }
    initialization = 0;
  }

  double P_Recola[esi->p_parton[0].size()][4];
  set_phasespacepoint(P_Recola);
  
  set_alphas_rcl(osi->var_alpha_S_reference, osi->var_mu_ren, N_nondecoupled);
  compute_process_rcl(1, P_Recola, "LO");
  compute_all_colour_correlations_rcl(1, P_Recola);

  for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
    if ((*(osi->VA_ioperator))[i_a][0].type_correction() == 1){
      for (int j_a = 0; j_a < (*(osi->VA_ioperator))[i_a].size(); j_a++){
	///        osi->VA_ME2_cf[i_a][j_a] = M2cc[(*(osi->VA_ioperator))[i_a][j_a].no_BLHA_entry];
	get_colour_correlation_rcl(1, csi->contribution_order_alpha_s - 1, (*(osi->VA_ioperator))[i_a][j_a].no_emitter(), (*(osi->VA_ioperator))[i_a][j_a].no_spectator(), "LO", osi->VA_ME2_cf[i_a][j_a]);
	logger << LOG_DEBUG_VERBOSE << "get_colour_correlation_rcl(1, " << csi->contribution_order_alpha_s - 1 << ", " << (*(osi->VA_ioperator))[i_a][j_a].no_emitter() << ", " << (*(osi->VA_ioperator))[i_a][j_a].no_spectator() << ", " << char(34) << "LO" << char(34) << ", " << setprecision(15) << osi->VA_ME2_cf[i_a][j_a] << ");" << endl;
	osi->VA_ME2_cf[i_a][j_a] = osi->VA_ME2_cf[i_a][j_a] * (*(osi->VA_ioperator))[i_a][j_a].colour_factor();
	osi->VA_ME2_cf[i_a][j_a] /= pow(osi->var_rel_alpha_S, 1 - 1. / csi->contribution_order_alpha_s);
	//	double ME2_born_qcd;
	//	get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s - 1, "LO", ME2_born_qcd);
	logger << LOG_DEBUG_POINT << "Recola:   QCD   VA_ME2_cf[" << i_a << "][" << j_a << "] = " << setw(23) << setprecision(15) << osi->VA_ME2_cf[i_a][j_a] << endl;
      }
    }
    else if ((*(osi->VA_ioperator))[i_a][0].type_correction() == 2){
      for (int j_a = 0; j_a < (*(osi->VA_ioperator))[i_a].size(); j_a++){
	///        osi->VA_ME2_cf[i_a][j_a] = (*(osi->VA_ioperator))[i_a][j_a].charge_factor() * osi->VA_b_ME2;
	logger << LOG_DEBUG_VERBOSE << "(*(osi->VA_ioperator))[" << i_a << "][" << j_a << "].charge_factor() = " << (*(osi->VA_ioperator))[i_a][j_a].charge_factor() << endl;
	logger << LOG_DEBUG_VERBOSE << "(*(osi->VA_ioperator))[" << i_a << "][" << j_a << "].type_correction() = " << (*(osi->VA_ioperator))[i_a][j_a].type_correction() << endl;
	get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s, "LO", osi->VA_b_ME2);
	osi->VA_ME2_cf[i_a][j_a] = (*(osi->VA_ioperator))[i_a][j_a].charge_factor() * osi->VA_b_ME2;
	osi->VA_ME2_cf[i_a][j_a] /= osi->var_rel_alpha_S;
	//	double ME2_born_ewk;
	//	get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s, "LO", ME2_born_ewk);
	logger << LOG_DEBUG << "Recola:   QEW   VA_ME2_cf[" << i_a << "][" << j_a << "] = " << setw(23) << setprecision(15) << osi->VA_ME2_cf [i_a][j_a] << endl;
      }
    }
  }
  osi->calculate_ioperator_QCD();
  osi->calculate_ioperator_QEW();

  for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
    I_ME2_emitter[i_a] = accumulate(osi->VA_I_ME2_cf[i_a].begin(), osi->VA_I_ME2_cf[i_a].end(), 0.);
  }
  osi->VA_I_ME2 = accumulate(I_ME2_emitter.begin(), I_ME2_emitter.end(), 0.);

  logger << LOG_DEBUG_VERBOSE << "finished" << flush << endl;
}


void amplitude_Recola_set::calculate_ME2_VA_MIX(){
  Logger logger("amplitude_Recola_set::calculate_ME2_VA_MIX");
  logger << LOG_DEBUG_VERBOSE << "called" << flush << endl;

  if (osi->switch_VI == 0 || osi->switch_VI == 1){
    double P_Recola[esi->p_parton[0].size()][4];
    set_phasespacepoint(P_Recola);
    // Check if mu_reg needs to be re-set for every point !!!
    set_mu_reg();
    set_mu_uv_rcl(mu_reg);
    set_mu_ir_rcl(mu_reg);
    set_alphas_rcl(osi->var_alpha_S_reference, osi->var_mu_ren, N_nondecoupled);
    logger << LOG_DEBUG_VERBOSE << "set_alphas_rcl(" << setprecision(15) << osi->var_alpha_S_reference << ", " << setprecision(15) << osi->var_mu_ren << ", " << N_nondecoupled << ");" << endl;


    compute_process_rcl(1, P_Recola, "NLO");
    logger << LOG_DEBUG_VERBOSE << "compute_process_rcl(1, P_Recola, " << char(34) << "NLO" << char(34) << ");" << endl;
    osi->VA_V_ME2 = 0.;
    get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s, "NLO", osi->VA_X_ME2);
    logger << LOG_DEBUG_VERBOSE << "get_squared_amplitude_rcl(1, " << csi->contribution_order_alpha_s << ", " << char(34) << "NLO" << char(34) << ", " << setprecision(15) << osi->VA_X_ME2 << ");" << endl;
    //      logger << LOG_DEBUG_POINT << "Recola:     ME2_B  = " << setw(23) << setprecision(15) << osi->VA_b_ME2 << endl;
    osi->VA_X_ME2 = osi->VA_X_ME2 / osi->var_rel_alpha_S;
    logger << LOG_DEBUG_POINT << "Recola:     ME2_VX = " << setw(23) << setprecision(15) << osi->VA_V_ME2 + osi->VA_X_ME2 << "   (corrected for var_rel_alpha_S)" << endl;

    if (osi->switch_output_comparison){
      double V_ME2_D4 = 0.;
      get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s, "NLO-D4", V_ME2_D4);
      logger << LOG_DEBUG_VERBOSE << "get_squared_amplitude_rcl(1, " << csi->contribution_order_alpha_s << ", " << char(34) << "NLO-D4" << char(34) << ", " << setprecision(15) << V_ME2_D4 << ");" << endl;
      V_ME2_D4 = V_ME2_D4 / osi->var_rel_alpha_S;
      logger << LOG_DEBUG_POINT << "Recola:     ME2_D4 = " << setw(23) << setprecision(15) << V_ME2_D4 << "   (corrected for var_rel_alpha_S)" << endl;

      double V_ME2_R2 = 0.;
      get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s, "NLO-R2", V_ME2_R2);
      logger << LOG_DEBUG_VERBOSE << "get_squared_amplitude_rcl(1, " << csi->contribution_order_alpha_s << ", " << char(34) << "NLO-R2" << char(34) << ", " << setprecision(15) << V_ME2_R2 << ");" << endl;
      V_ME2_R2 = V_ME2_R2 / osi->var_rel_alpha_S;
      logger << LOG_DEBUG_POINT << "Recola:     ME2_R2 = " << setw(23) << setprecision(15) << V_ME2_R2 << "   (corrected for var_rel_alpha_S)" << endl;

      double V_ME2_CT = 0.;
      get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s, "NLO-CT", V_ME2_CT);
      logger << LOG_DEBUG_VERBOSE << "get_squared_amplitude_rcl(1, " << csi->contribution_order_alpha_s << ", " << char(34) << "NLO-CT" << char(34) << ", " << setprecision(15) << V_ME2_CT << ");" << endl;
      V_ME2_CT = V_ME2_CT / osi->var_rel_alpha_S;
      logger << LOG_DEBUG_POINT << "Recola:     ME2_CT = " << setw(23) << setprecision(15) << V_ME2_CT << "   (corrected for var_rel_alpha_S)" << endl;

      //	logger << LOG_DEBUG << "" << endl;

      // in order to match the OpenLoops output:

      osi->VA_X_ME2 = V_ME2_CT;
      osi->VA_V_ME2 = V_ME2_D4 + V_ME2_R2;

      logger << LOG_DEBUG_POINT << "Recola:     ME2_VX = " << setw(23) << setprecision(15) << osi->VA_V_ME2 + osi->VA_X_ME2 << "   (corrected for var_rel_alpha_S)" << endl;
    }
    osi->VA_I_ME2 = 0.;
    if (osi->switch_VI == 0 || osi->switch_VI == 2){
      calculate_ME2_ioperator_VA_MIX();
    }
    
    if (osi->switch_CV){
      for (int s = 0; s < osi->n_scales_CV; s++){
	set_alphas_rcl(osi->var_alpha_S_CV[s], osi->var_mu_ren_CV[s], N_nondecoupled);
	logger << LOG_DEBUG_VERBOSE << "set_alphas_rcl(" << setprecision(15) << osi->var_alpha_S_CV[s] << ", " << setprecision(15) << osi->var_mu_ren_CV[s] << ", " << N_nondecoupled << ");" << endl;

	rescale_process_rcl(1, "NLO");
	logger << LOG_DEBUG_VERBOSE << "rescale_process_rcl(1, " << char(34) << "NLO" << char(34) << ");" << endl;
	osi->VA_X_ME2_CV[s] = 0.;
	get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s, "NLO", osi->VA_X_ME2_CV[s]);
	logger << LOG_DEBUG_VERBOSE << "get_squared_amplitude_rcl(1, " << csi->contribution_order_alpha_s << ", " << char(34) << "NLO" << char(34) << ", " << setprecision(15) << osi->VA_X_ME2_CV[s] << ");" << endl;
	osi->VA_X_ME2_CV[s] = osi->VA_X_ME2_CV[s] / osi->var_rel_alpha_S_CV[s];
	logger << LOG_DEBUG_POINT << "Recola:     VA_X_ME2_CV[" << s << "] = " << setw(23) << setprecision(15) << osi->VA_X_ME2_CV[s] << "   (after removing alpha_S factor)" << endl;
      }
    }
    
    if (osi->switch_TSV){
      for (int i_v = 0; i_v < osi->max_dyn_ren + 1; i_v++){
	for (int i_r = 0; i_r < osi->n_scale_dyn_ren[i_v]; i_r++){
	  set_alphas_rcl(osi->value_alpha_S_TSV[0][i_v][i_r], osi->value_scale_ren[0][i_v][i_r], N_nondecoupled);
	  logger << LOG_DEBUG_VERBOSE << "set_alphas_rcl(" << setprecision(15) << osi->value_alpha_S_TSV[0][i_v][i_r] << ", " << setprecision(15) << osi->value_scale_ren[0][i_v][i_r] << ", " << N_nondecoupled << ");" << endl;

	  rescale_process_rcl(1, "NLO");
	  logger << LOG_DEBUG_VERBOSE << "rescale_process_rcl(1, " << char(34) << "NLO" << char(34) << ");" << endl;

	  get_squared_amplitude_rcl(1, csi->contribution_order_alpha_s, "NLO", osi->value_ME2term_ren[0][i_v][i_r]);
	  logger << LOG_DEBUG_VERBOSE << "get_squared_amplitude_rcl(1, " << csi->contribution_order_alpha_s << ", " << char(34) << "NLO" << char(34) << ", " << setprecision(15) << osi->value_ME2term_ren[0][i_v][i_r] << ");" << endl;

	  osi->value_ME2term_ren[0][i_v][i_r] = osi->value_ME2term_ren[0][i_v][i_r] / osi->value_relative_factor_alpha_S[0][i_v][i_r];
	  logger << LOG_DEBUG_POINT << "Recola:     ME2_VX_TSV[" << i_v << "][" << i_r << "] = " << setw(23) << setprecision(15) << osi->value_ME2term_ren[0][i_v][i_r] << "   (corrected for var_rel_alpha_S)" << endl;
	  osi->value_ME2term_ren[0][i_v][i_r] += osi->VA_I_ME2;
	}
      }
    }
  }
  else if (osi->switch_VI == 2){
    //  I-operator evaluation -> osi->VA_I_ME2
    calculate_ME2_ioperator_VA_MIX();
    for (int i_v = 0; i_v < osi->max_dyn_ren + 1; i_v++){
      for (int i_r = 0; i_r < osi->n_scale_dyn_ren[i_v]; i_r++){
        osi->value_ME2term_ren[0][i_v][i_r] = osi->VA_I_ME2;
      }
    }
    logger << LOG_DEBUG << "SK: osi->VA_I_ME2 = " << osi->VA_I_ME2 << endl;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << flush << endl;
}

#endif
