#include "header.hpp"

void amplitude_set::initialization(observable_set * _osi, contribution_set * _csi, model_set * _msi, event_set * _esi, inputparameter_set * isi){
  Logger logger("amplitude_set::initialization (osi, csi, msi, esi, isi)");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  osi = _osi;
  csi = _csi;
  msi = _msi;
  esi = _esi;

  initialization_set_default();
  initialization_input(isi);
  set_mu_reg();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

/*
// required ???
void amplitude_set::initialization_basic(observable_set * _osi, contribution_set * _csi, model_set * _msi, event_set * _esi, inputparameter_set * isi){
  Logger logger("amplitude_set::initialization_basic");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  initialization_input(isi);

  osi = _osi;
  csi = _csi;
  msi = _msi;
  esi = _esi;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}
*/

void amplitude_set::initialization_set_default(){
  Logger logger("amplitude_set::initialization_set_default");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  switch_amplitude_not_required = 0;

  switch_off_H1 = 0;
  switch_off_H2 = 0;
  switch_off_H1gg = 0;

  N_quarks = 6;

  switch_mu_reg = 0;
    
  failed_evaluation = 0;
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void amplitude_set::initialization_input(inputparameter_set * isi){
  Logger logger("amplitude_set::initialization_input");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_i = 0; i_i < isi->input_amplitude.size(); i_i++){
    logger << LOG_DEBUG_VERBOSE << "isi->input_amplitude[" << i_i << "] = " << isi->input_amplitude[i_i] << endline_flush;

    if (isi->input_amplitude[i_i].variable == ""){}

    else if (isi->input_amplitude[i_i].variable == "N_quarks"){N_quarks = atoi(isi->input_amplitude[i_i].value.c_str());}
    else if (isi->input_amplitude[i_i].variable == "switch_amplitude_not_required"){switch_amplitude_not_required = atoi(isi->input_amplitude[i_i].value.c_str());}
    else if (isi->input_amplitude[i_i].variable == "switch_off_H1"){switch_off_H1 = atoi(isi->input_amplitude[i_i].value.c_str());}
    else if (isi->input_amplitude[i_i].variable == "switch_off_H2"){switch_off_H2 = atoi(isi->input_amplitude[i_i].value.c_str());}
    else if (isi->input_amplitude[i_i].variable == "switch_off_H1gg"){switch_off_H1gg = atoi(isi->input_amplitude[i_i].value.c_str());}
  // for compatibility with older input files !!!
    else if (isi->input_amplitude[i_i].variable == "switch_H1"){switch_off_H1 = atoi(isi->input_amplitude[i_i].value.c_str());}
    else if (isi->input_amplitude[i_i].variable == "switch_H2"){switch_off_H2 = atoi(isi->input_amplitude[i_i].value.c_str());}
    else if (isi->input_amplitude[i_i].variable == "switch_H1gg"){switch_off_H1gg = atoi(isi->input_amplitude[i_i].value.c_str());}
    else if (isi->input_amplitude[i_i].variable == "switch_mu_reg"){switch_mu_reg = atoi(isi->input_amplitude[i_i].value.c_str());}
    else if (isi->input_amplitude[i_i].variable == "scale_reg"){scale_reg = atof(isi->input_amplitude[i_i].value.c_str());}

  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void amplitude_set::initialization_N_nondecoupled_from_pdf_set(){
  Logger logger("amplitude_set::initialization_N_nondecoupled_from_pdf_set");
  logger << LOG_DEBUG_VERBOSE << "called" << flush << endl;

  ////////////////////////////////////
  //  N_nondecoupled determination  //
  ////////////////////////////////////

#ifdef LHAPDF5
  string LHAPDFversion = "5";
#else
  string LHAPDFversion = "6";
#endif

  logger << LOG_DEBUG << "LHAPDFversion = " << LHAPDFversion << endline_flush;
  logger << LOG_DEBUG << "LHAPDFversion.substr(0, 1) = " << LHAPDFversion.substr(0, 1) << endline_flush;

  int LHAPDF_N_nondecoupled = LHAPDF::getNf();

  if (LHAPDFversion.substr(0, 1) == "6"){
    logger << LOG_DEBUG << "LHAPDF version 6 (" << LHAPDFversion << ") is used." << endline_flush;
    N_nondecoupled = LHAPDF_N_nondecoupled;
    if ((osi->LHAPDFname).substr(0, 10) == "NNPDF21_lo"){N_nondecoupled = 6;}
  }
  else if (LHAPDFversion.substr(0, 1) == "5"){
    logger << LOG_DEBUG << "LHAPDF version 5 (" << LHAPDFversion << ") is used." << endline_flush;

    logger << LOG_DEBUG << "(osi->LHAPDFname) = " << (osi->LHAPDFname) << endline_flush;
    logger << LOG_DEBUG << "LHAPDF_N_nondecoupled = " << LHAPDF_N_nondecoupled << endline_flush;

    if ((osi->LHAPDFname).substr(0, 7) == "NNPDF21"){
      logger << LOG_DEBUG << "(osi->LHAPDFname) = " << (osi->LHAPDFname) << endline_flush;
      N_nondecoupled = 6;
      for (int i_s = (osi->LHAPDFname).size() - 3; i_s >=0; i_s--){
	string temp_Nf = (osi->LHAPDFname).substr(i_s, 3);
	logger << LOG_DEBUG << "(osi->LHAPDFname).substr(" << i_s << ", 3) = " << temp_Nf << endline_flush;
	if (temp_Nf == "NF3"){N_nondecoupled = 3;}
	else if (temp_Nf == "NF4"){N_nondecoupled = 4;}
	else if (temp_Nf == "NF5"){N_nondecoupled = 5;}
      }
    }

    else if ((osi->LHAPDFname).substr(0, 7) == "NNPDF23"){
      logger << LOG_DEBUG << "(osi->LHAPDFname) = " << (osi->LHAPDFname) << endline_flush;
      N_nondecoupled = 6;
      for (int i_s = (osi->LHAPDFname).size() - 3; i_s >=0; i_s--){
	string temp_Nf = (osi->LHAPDFname).substr(i_s, 3);
	logger << LOG_DEBUG << "(osi->LHAPDFname).substr(" << i_s << ", 3) = " << temp_Nf << endline_flush;
	if (temp_Nf == "NF3"){N_nondecoupled = 3;}
	else if (temp_Nf == "NF4"){N_nondecoupled = 4;}
	else if (temp_Nf == "NF5"){N_nondecoupled = 5;}
      }
    }

    else if ((osi->LHAPDFname).substr(0, 6) == "NNPDF3"){
      logger << LOG_DEBUG << "(osi->LHAPDFname) = " << (osi->LHAPDFname) << endline_flush;
      N_nondecoupled = 5;
      for (int i_s = (osi->LHAPDFname).size() - 4; i_s >=0; i_s--){
	string temp_Nf = (osi->LHAPDFname).substr(i_s, 4);
	logger << LOG_DEBUG << "(osi->LHAPDFname).substr(" << i_s << ", 4) = " << temp_Nf << endline_flush;
	if (temp_Nf == "nf_3"){N_nondecoupled = 3;}
	else if (temp_Nf == "nf_4"){N_nondecoupled = 4;}
	else if (temp_Nf == "nf_6"){N_nondecoupled = 6;}
      }
    }

    else if ((osi->LHAPDFname).substr(0, 4) == "CT14"){
      logger << LOG_DEBUG << "(osi->LHAPDFname) = " << (osi->LHAPDFname) << endline_flush;
      N_nondecoupled = 5;
      for (int i_s = (osi->LHAPDFname).size() - 3; i_s >=0; i_s--){
	string temp_Nf = (osi->LHAPDFname).substr(i_s, 3);
	logger << LOG_DEBUG << "(osi->LHAPDFname).substr(" << i_s << ", 3) = " << temp_Nf << endline_flush;
	if (temp_Nf == "NF3"){N_nondecoupled = 3;}
	else if (temp_Nf == "NF4"){N_nondecoupled = 4;}
	else if (temp_Nf == "NF6"){N_nondecoupled = 6;}
      }
    }

    else if ((osi->LHAPDFname).substr(0, 8) == "MMHT2014"){
      logger << LOG_DEBUG << "(osi->LHAPDFname) = " << (osi->LHAPDFname) << endline_flush;
      N_nondecoupled = 5;
      for (int i_s = (osi->LHAPDFname).size() - 3; i_s >=0; i_s--){
	string temp_Nf = (osi->LHAPDFname).substr(i_s, 3);
	logger << LOG_DEBUG << "(osi->LHAPDFname).substr(" << i_s << ", 3) = " << temp_Nf << endline_flush;
	if (temp_Nf == "nf3"){N_nondecoupled = 3;}
	else if (temp_Nf == "nf4"){N_nondecoupled = 4;}
	else if (temp_Nf == "nf6"){N_nondecoupled = 6;}
      }
    }

    else if ((osi->LHAPDFname).substr(0, 9) == "PDF4LHC15"){
      logger << LOG_DEBUG << "(osi->LHAPDFname) = " << (osi->LHAPDFname) << endline_flush;
      N_nondecoupled = 5;
      for (int i_s = (osi->LHAPDFname).size() - 3; i_s >=0; i_s--){
	string temp_Nf = (osi->LHAPDFname).substr(i_s, 3);
	logger << LOG_DEBUG << "(osi->LHAPDFname).substr(" << i_s << ", 3) = " << temp_Nf << endline_flush;
	if (temp_Nf == "nf3"){N_nondecoupled = 3;}
	else if (temp_Nf == "nf4"){N_nondecoupled = 4;}
	else if (temp_Nf == "nf6"){N_nondecoupled = 6;}
      }
    }

    else if ((osi->LHAPDFname).substr(0, 6) == "LUXqed"){
      logger << LOG_DEBUG << "(osi->LHAPDFname) = " << (osi->LHAPDFname) << endline_flush;
      N_nondecoupled = 5;
      for (int i_s = (osi->LHAPDFname).size() - 3; i_s >=0; i_s--){
	string temp_Nf = (osi->LHAPDFname).substr(i_s, 3);
	logger << LOG_DEBUG << "(osi->LHAPDFname).substr(" << i_s << ", 3) = " << temp_Nf << endline_flush;
	if (temp_Nf == "nf3"){N_nondecoupled = 3;}
	else if (temp_Nf == "nf4"){N_nondecoupled = 4;}
	else if (temp_Nf == "nf6"){N_nondecoupled = 6;}
      }
    }
    else {
      N_nondecoupled = LHAPDF_N_nondecoupled;
    }
    logger << LOG_DEBUG << "N_nondecoupled = " << N_nondecoupled << endline_flush;
  }

  if ((osi->LHAPDFname).substr(0, 8) == "MSTW2008" || (osi->LHAPDFname).substr(0, 8) == "MMHT2014"){
    for (int i_s = (osi->LHAPDFname).size() - 6; i_s >=0; i_s--){
      string temp_Nf = (osi->LHAPDFname).substr(i_s, 6);
      logger << LOG_DEBUG << "(osi->LHAPDFname).substr(" << i_s << ", 6) = " << temp_Nf << endline_flush;
      if (temp_Nf == "nf4as5"){N_nondecoupled = 5;}
    }
  }

  if (N_nondecoupled != LHAPDF_N_nondecoupled){
    logger << LOG_WARN << "N_nondecoupled has been changed wrt. LHAPDF_N_nondecoupled in LHAPDF " << LHAPDFversion << endline_flush;
  }

  logger << LOG_DEBUG << "N_nondecoupled = " << N_nondecoupled << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void amplitude_set::initialization_process_general(){
  Logger logger("amplitude_set::initialization_process_general");
  logger << LOG_DEBUG_VERBOSE << "called" << flush << endl;

  if (csi->type_contribution == "VT2" && csi->type_correction == "QCD"){
    initialization_H2_QCD_2loop();
  }
  else if ((csi->type_contribution == "L2VA" || csi->type_contribution == "L2VT") && csi->type_correction == "QCD"){
    initialization_H1gg_QCD_2loop();
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

