#include "header.hpp"

void observable_set::determine_techcut_RA(){
  Logger logger("observable_set::determine_techcut_RA");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  psi->RA_techcut = 0;
  static int num_printouts = 0;
  const int max_printouts = 1000;
  static double severity_threshold = 1.e-12;
  static double impact_threshold = 1.e-3;

  // check singular regions (p_i * p_j / s_hat) < cut_technical
  for (int sr = 0; sr < csi->singular_region_list.size(); sr++){
    int x1 = csi->singular_region_list[sr][0];
    int x2 = csi->singular_region_list[sr][1];
    csi->singular_region[x1][x2] = esi->p_parton[0][x1] * esi->p_parton[0][x2] / psi->xbs_all[0][0];

    if (csi->singular_region[x1][x2] < psi->cut_technical){
      int ccount = 0;
      for (int i_a = 1; i_a < esi->cut_ps.size(); i_a++){if (esi->cut_ps[i_a] == -1){ccount++;}}

      double severity = abs(1. + accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0]);
      //      double severity = 1.;

      // suppress warning if point is (relatively) harmless and output_level is low
      if ((((severity > severity_threshold || Log::getLogThreshold() <= LOG_DEBUG) &&
	    (abs(integrand / sigma_normalization) > impact_threshold || sigma_normalization == 1.) &&
	    num_printouts < max_printouts) ||
	   Log::getLogThreshold() <= LOG_DEBUG_VERBOSE)){
        logger << LOG_DEBUG << endline_flush;
        logger << LOG_DEBUG << "applying technical cut near dipole singularity" << endline_flush;
        logger << LOG_DEBUG << "cut_technical = " << psi->cut_technical << endline_flush;
        for (int i_p = 0; i_p < esi->p_parton[0].size(); i_p++){logger << LOG_DEBUG << "esi->p_parton[0][" << i_p << "] = " << esi->p_parton[0][i_p] << "   " << esi->p_parton[0][i_p].m2() << "   " << sqrt(abs(esi->p_parton[0][i_p].m2())) << endline_flush;}
        logger << LOG_DEBUG << "contributing dipoles:   " << endline_flush;
        for (int i_a = 0; i_a < esi->cut_ps.size(); i_a++){
          if (esi->cut_ps[i_a] != -1){logger << LOG_DEBUG << setw(8) << csi->dipole[i_a].name() << "   " << "RA_ME2[" << i_a << "] = " << RA_ME2[i_a] << endline_flush;}
        }

	if (switch_console_output_techcut_RA){
	  num_printouts++;
	  stringstream temp_ss;
	  temp_ss << "  "
		  << right << setw(12) << psi->i_gen
		  << right << setw(10) << psi->i_acc
		  << "  ("
		  << right << setw(2) << esi->cut_ps[0] << "/"
		  << right << setw(2) << ccount
		  << "  "
		  << csi->singular_region_name[x1][x2] << "/^s = "
		  << right << setw(11) << setprecision(4) << showpoint << csi->singular_region[x1][x2]
		  << ")  int/norm = "
		  << right << setw(15) << setprecision(8) << showpoint << integrand / sigma_normalization;
	  if (accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) != 0. && RA_ME2[0] != 0.){
	    temp_ss << "  A/R = "
		    << setw(23) << setprecision(15) << -accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0]
		    << "  A/R-1 = "
		    << right << scientific << setw(9) << setprecision(2) << -accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0] - 1.
		    << defaultfloat << left;
	  }
	  logger << LOG_INFO << temp_ss.str() << endline_flush;
	}
      }

      /*
      // RA_techcut_integrand could be used to check the dependence on the parameter cut_technical !!!
      if (csi->singular_region[x1][x2] > psi->cut_technical / 10.){RA_techcut_integrand = integrand;}
      else {RA_techcut_integrand = 0.;}
      */
      /*
      if (abs(integrand / sigma_normalization) > 1. || csi->singular_region[x1][x2] < psi->cut_technical / 100.) {
	psi->RA_techcut = 1;
	psi->i_tec++;
	integrand = 0.;
      }
      */
      // new:
      // All points with (p_i * p_j / s_hat) < cut_technical are technically cut.
      // This might require an adaptation of the corresponding values...
      psi->RA_techcut = 1;
      psi->i_tec++;
      integrand = 0.;
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

