#include "header.hpp"

// boosts system to the specified transverse momentum. the azimuth is generated uniformly
void event_set::perform_QT_boost_cms(){
  Logger logger("event_set::perform_QT_boost_cms");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  //  i_a = 0: standard point (CMS), not boosted to 2-dimensional qT
  //  i_a = 1: point with finite 2-dim qT
  for (int i_p = 1; i_p < p_parton[0].size(); i_p++){p_parton[1][i_p] = p_parton[0][i_p].boost(psi->recoil_system.Pinv());}

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
    stringstream temp_ss;
    for (int i_a = 0; i_a < csi->n_ps; i_a++){
      for (int i_p = 1; i_p < p_parton[i_a].size(); i_p++){temp_ss << "p_parton[" << i_a << "][" << i_p << "] = " << p_parton[i_a][i_p] << "   m2 = " << p_parton[i_a][i_p].m2() << endline;}
      temp_ss << endline;
    }
    logger << LOG_DEBUG_POINT << endline << temp_ss.str() << endline_flush;
    logger << LOG_DEBUG_POINT << setw(25) << "psi->recoil_system" << " = " << psi->recoil_system << endline_flush;
    logger << LOG_DEBUG_POINT << setw(25) << "sum CMS FS momenta" << " = " << p_parton[1][3] + p_parton[1][4] << endline_flush;
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void event_set::perform_QT_boost_recoil(){
  Logger logger("event_set::perform_QT_boost_recoil");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double sqrts = 2 * psi->E; 
  double Q2 = psi->xbs_all[0][0];
  //  double Q = sqrt(psi->xbs_all[0][0]);
  double qt2 = psi->recoil_qT2;
  //  double qt = sqrt(qt2);
  double qt = psi->recoil_qT;
  double y = log(sqrt(psi->x_pdf[1] / psi->x_pdf[2]));
  double mt = sqrt(Q2+qt2);
  double qz = 0.5 * mt * (exp(y) - exp(-y));
  double q0 = 0.5 * mt * (exp(y) + exp(-y));
  double phi = psi->recoil_phi; 
    
  fourvector q_CS = fourvector(q0, qt*cos(phi), qt*sin(phi), qz);

  logger << LOG_DEBUG_POINT << "phi = " << phi << endline_flush;
  logger << LOG_DEBUG_POINT << "q_CS (after qT_phi rotation) = " << q_CS << "   y = " << q_CS.rapidity() << "   eta = " << q_CS.eta() << "   m = " << q_CS.m() << endline_flush;
  
  double kt1 = q_CS.x1()/2. ;  
  double kt2 = q_CS.x2()/2. ;   //  CS FRAME PRESCRIPTION
  

  double zeta1 = 1./Q2/2. * ( Q2+ 2.*( q_CS.x1()*kt1+q_CS.x2()*kt2 )
			      + sqrt( pow((Q2 + 2.*( q_CS.x1()*kt1+q_CS.x2()*kt2)),2)
				      - 4.*pow(mt,2)*(pow(kt1,2)+pow(kt2,2)))) ; 

  double qP1 = (q_CS.x0()-q_CS.x3())*sqrts/2.;
  //  double qP2 = (q_CS.x0()+q_CS.x3())*sqrts/2.;

  logger << LOG_DEBUG_POINT << "kt1 = " << kt1 << endline_flush;
  logger << LOG_DEBUG_POINT << "kt2 = " << kt2 << endline_flush;
  
  
  p_parton[1][1] = fourvector( sqrts/2.*( zeta1*Q2/2./qP1 + (pow(kt1,2)+pow(kt2,2))/zeta1*qP1/Q2/pow(sqrts,2)*2.), kt1, kt2,
			       sqrts/2.* (zeta1*Q2/2./qP1 - (pow(kt1,2)+pow(kt2,2))/zeta1*qP1/Q2/pow(sqrts,2)*2.));
  p_parton[1][2] = q_CS - p_parton[1][1];
  /*
  p_parton[1][2] = fourvector( q_CS.x0()-p_parton[1][1].x0(),
			       q_CS.x1()-p_parton[1][1].x1(),
			       q_CS.x2()-p_parton[1][1].x2(),
			       q_CS.x3()-p_parton[1][1].x3() );
  */
  for (int i_p = 3; i_p < p_parton[0].size(); i_p++){p_parton[1][i_p] = p_parton[0][i_p].boostinv(q_CS, Q2);}
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}






/* // old version, adapted from resummation
void event_set::perform_QT_boost_cms(){
  Logger logger("event_set::event_set");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double Q2 = psi->xbs_all[0][0];
  double Q = sqrt(psi->xbs_all[0][0]);
  // ??? //  double Q = psi->xbsqrts_all[0][0];
  double y = log(sqrt(psi->x_pdf[1] / psi->x_pdf[2]));
  logger << LOG_DEBUG_POINT << "Q2 = " << Q2 << endline_flush;
  logger << LOG_DEBUG_POINT << "Q = " << Q << endline_flush;
  logger << LOG_DEBUG_POINT << "y = " << y << endline_flush;
  logger << LOG_DEBUG_POINT << "recoil_qT2 = " << psi->recoil_qT2 << endline_flush;
  logger << LOG_DEBUG_POINT << "recoil_qT = " << psi->recoil_qT << endline_flush;
  // if we are boosting fourvectors in the cms system, we first have to transform to the lab system
  // and then transform back at the very end
  // SK: Does not look very intuitive to me: Why not perform the whole procedure in the CMS system ???
  double boost_cms = (exp(y) - exp(-y)) / (exp(y) + exp(-y));
  logger << LOG_DEBUG_POINT << "boost_cms = " << boost_cms << endline_flush;
  //  boost = (x_pdf[1] - x_pdf[2]) / (x_pdf[1] + x_pdf[2]);

  // boost final state system
  double pt = 0.5 * Q * (exp(y) + exp(-y));
  double pz = 0.5 * Q * (exp(y) - exp(-y));
  double pt2 = pow(pt, 2);
  double pz2 = pow(pz, 2);
  logger << LOG_DEBUG_POINT << "pt = " << setw(23) << setprecision(15) << pt << "   pt2 = " << setw(23) << setprecision(15) << pt2 << endline_flush;
  logger << LOG_DEBUG_POINT << "pz = " << setw(23) << setprecision(15) << pz << "   pz2 = " << setw(23) << setprecision(15) << pz2 << endline_flush;
  fourvector q_lab = fourvector(pt, 0., 0., pz);
  logger << LOG_DEBUG_POINT << "q_lab = " << q_lab << "   y = " << q_lab.rapidity() << "   eta = " << q_lab.eta() << "   m = " << q_lab.m() << endline_flush;
  double mT2 = Q2 + psi->recoil_qT2;
  double mT = sqrt(mT2);

  // boost to the correct z
  double B = pt * pz / (pt2 + mT2 / Q2 * pz2);
  double C = (pz2 - mT2 / Q2 * pz2) / (pt2 + mT2 / Q2 * pz2);
  double beta_z = B - sqrt(B * B - C);
  if (pz < 0.){beta_z = B + sqrt(B * B - C);}
  logger << LOG_DEBUG_POINT << "beta_z = " << beta_z << endline_flush;
  q_lab = q_lab.zboost(beta_z);
  logger << LOG_DEBUG_POINT << "q_lab (after beta_z boost) = " << q_lab << "   y = " << q_lab.rapidity() << "   eta = " << q_lab.eta() << "   m = " << q_lab.m() << endline_flush;

  logger << LOG_DEBUG_POINT << "check:   pz = " << q_lab.x3() << ", " << 0.5 * mT * (exp(y) - exp(-y)) << endline_flush;
  logger << LOG_DEBUG_POINT << "check:   pt = " << q_lab.x0() << ", " << 0.5 * mT * (exp(y) + exp(-y)) << endline_flush;
  logger << LOG_DEBUG_POINT << "correctness : " << 0.5 * mT * (exp(y) - exp(-y)) << ", " << (1.0 / sqrt((1. - beta_z * beta_z))) * (pz - beta_z * pt) << endline_flush;

  // boost to the correct qT
  double beta_x = -sqrt(psi->recoil_qT2 / (q_lab.x0() * q_lab.x0() + psi->recoil_qT2));
  if (q_lab.x0() < 0.){beta_x *= -1.;}
  q_lab = q_lab.xboost(beta_x);
  logger << LOG_DEBUG_POINT << "q_lab (after beta_x boost) = " << q_lab << "   y = " << q_lab.rapidity() << "   eta = " << q_lab.eta() << "   m = " << q_lab.m() << endline_flush;

  // generate angle for qT
  // Move this part to generation of 2-dim. qT in psi !!!
  //  double qT_phi = 2 * pi * double(rand()) / RAND_MAX;
  //  logger << LOG_DEBUG_POINT << "qT_phi = " << qT_phi << endline_flush;
  q_lab = q_lab.rotation(psi->recoil_phi, 1);
  logger << LOG_DEBUG_POINT << "q_lab (after qT_phi rotation) = " << q_lab << "   y = " << q_lab.rapidity() << "   eta = " << q_lab.eta() << "   m = " << q_lab.m() << endline_flush;

  //  i_a = 0: standard point (CMS), not boosted to 2-dim qT
  //  i_a = 1: point with finite 2-dim qT
  stringstream temp_ss;
  temp_ss << "Born phase space point in CM frame" << endline;
  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    for (int i_p = 1; i_p < p_parton[i_a].size(); i_p++){temp_ss << "p_parton[" << i_a << "][" << i_p << "] = " << p_parton[i_a][i_p] << "   m2 = " << p_parton[i_a][i_p].m2() << endline;}
    temp_ss << endline;
  }
  logger << LOG_DEBUG_POINT << endline << temp_ss.str() << endline_flush;
  temp_ss.str("");
  
  for (int i_p = 0; i_p < p_parton[0].size(); i_p++){
    p_parton[1][i_p] = p_parton[0][i_p].zboost(-boost_cms);
  }
  temp_ss << "After boost to LAB frame" << endline;
  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    for (int i_p = 1; i_p < p_parton[i_a].size(); i_p++){temp_ss << "p_parton[" << i_a << "][" << i_p << "] = " << p_parton[i_a][i_p] << "   m2 = " << p_parton[i_a][i_p].m2() << endline;}
    temp_ss << endline;
  }
  logger << LOG_DEBUG_POINT << endline << temp_ss.str() << endline_flush;
  temp_ss.str("");
  
  for (int i_p = 0; i_p < p_parton[0].size(); i_p++){
    p_parton[1][i_p] = p_parton[1][i_p].zboost(beta_z);
  }
  temp_ss << "After boost by beta_z = " << beta_z << endline;
  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    for (int i_p = 1; i_p < p_parton[i_a].size(); i_p++){temp_ss << "p_parton[" << i_a << "][" << i_p << "] = " << p_parton[i_a][i_p] << "   m2 = " << p_parton[i_a][i_p].m2() << endline;}
    temp_ss << endline;
  }
  logger << LOG_DEBUG_POINT << endline << temp_ss.str() << endline_flush;
  temp_ss.str("");

  for (int i_p = 0; i_p < p_parton[0].size(); i_p++){
    p_parton[1][i_p] = p_parton[1][i_p].xboost(beta_x);
  }
  temp_ss << "After boost by beta_x = " << beta_x << endline;
  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    for (int i_p = 1; i_p < p_parton[i_a].size(); i_p++){temp_ss << "p_parton[" << i_a << "][" << i_p << "] = " << p_parton[i_a][i_p] << "   m2 = " << p_parton[i_a][i_p].m2() << endline;}
    temp_ss << endline;
  }
  logger << LOG_DEBUG_POINT << endline << temp_ss.str() << endline_flush;
  temp_ss.str("");

  for (int i_p = 0; i_p < p_parton[0].size(); i_p++){
    p_parton[1][i_p] = p_parton[1][i_p].rotation(psi->recoil_phi, 1.);
  }
  temp_ss << "After rotation by psi->recoil_phi = " << psi->recoil_phi << endline;
  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    for (int i_p = 1; i_p < p_parton[i_a].size(); i_p++){temp_ss << "p_parton[" << i_a << "][" << i_p << "] = " << p_parton[i_a][i_p] << "   m2 = " << p_parton[i_a][i_p].m2() << endline;}
    temp_ss << endline;
  }
  logger << LOG_DEBUG_POINT << endline << temp_ss.str() << endline_flush;
  temp_ss.str("");

  for (int i_p = 0; i_p < p_parton[0].size(); i_p++){
    p_parton[1][i_p] = p_parton[1][i_p].zboost(boost_cms);
  }
  temp_ss << "After back to CM frame" << endline;
  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    for (int i_p = 1; i_p < p_parton[i_a].size(); i_p++){temp_ss << "p_parton[" << i_a << "][" << i_p << "] = " << p_parton[i_a][i_p] << "   m2 = " << p_parton[i_a][i_p].m2() << endline;}
    temp_ss << endline;
  }
  logger << LOG_DEBUG_POINT << endline << temp_ss.str() << endline_flush;
  temp_ss.str("");
  
  //  logger << LOG_INFO << setw(25) << "psi->recoil_qT" << " = " << setw(23) << setprecision(15) << psi->recoil_qT << endline_flush;
  //  logger << LOG_INFO << setw(25) << "calculated recoil_qT" << " = " << setw(23) << setprecision(15) << (p_parton[1][3] + p_parton[1][4]).pT() << endline_flush;
  
  //  stringstream temp_ss;
  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    for (int i_p = 1; i_p < p_parton[i_a].size(); i_p++){temp_ss << "p_parton[" << i_a << "][" << i_p << "] = " << p_parton[i_a][i_p] << "   m2 = " << p_parton[i_a][i_p].m2() << endline;}
    temp_ss << endline;
  }
  logger << LOG_DEBUG_POINT << endline << temp_ss.str() << endline_flush;
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}
*/

