#include "header.hpp"

void event_set::perform_event_selection(){
  static Logger logger("event_set::perform_event_selection");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (switch_output_cutinfo){info_cut << endline << "New PSP:" << "   i_gen = " << setw(12) << psi->i_gen << "   i_acc = " << setw(12) << psi->i_acc << "   i_rej = " << setw(12) << psi->i_rej << "   i_nan = " << setw(4) << psi->i_nan << "   i_tec = " << setw(4) << psi->i_tec << endline_flush;}

  if (switch_output_cutinfo){info_cut << endline << output_particle_definition_selection() << endline;}
  
  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
    logger.newLine(LOG_DEBUG_POINT);
    logger << LOG_DEBUG_POINT << " New PSP: " << "   i_gen = " << setw(12) << psi->i_gen << "   i_acc = " << setw(12) << psi->i_acc << "   i_rej = " << setw(12) << psi->i_rej << "   i_nan = " << setw(4) << psi->i_nan << "   i_tec = " << setw(4) << psi->i_tec << endline_flush;
    logger.newLine(LOG_DEBUG_POINT);

    logger << LOG_DEBUG_POINT << "p_parton.size() = " << p_parton.size() << endline_flush;
    stringstream temp_psp;
    temp_psp << "Final-state parton-level momenta (partonic CMS frame):" << endline;
    for (int i_p = 1; i_p < p_parton[0].size(); i_p++){
      temp_psp << "type_parton[" << setw(2) << 0 << "][" << i_p << "] = " << setw(3) << csi->type_parton[0][i_p] << " -> " << p_parton[0][i_p] << endline;
    }
    temp_psp << endline;
    logger << LOG_DEBUG_POINT << temp_psp.str() << endline_flush;
    logger.newLine(LOG_DEBUG_POINT);
  }

  // standard (ix_y = 0) and rotated (ix_y = 1) events (not implemented yet)
  ix_y = 0;
  
  for (ix_a = 0; ix_a < cut_ps.size(); ix_a++){
    if (csi->class_contribution_QT_powercorrection){
      if (ix_a == 1 && user->switch_value[user->switch_map["PTmode"]] == 1){cut_ps[ix_a] = -1; continue;}
      else if (ix_a == 0 && user->switch_value[user->switch_map["PTmode"]] == -1){cut_ps[ix_a] = -1; continue;}
    }

    // should not be needed, as set to correct value later:
    cut_ps[ix_a] = 0;

    perform_event_selection_phasespace();

    if (switch_output_cutinfo){logger << LOG_DEBUG_POINT << "[" << setw(2) << ix_a << "]" << "   after generic event selection, before user-defined particles   (cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << ")." << endline_flush;}
    if (cut_ps[ix_a] == -1){continue;}
    if (switch_output_cutinfo){info_cut << "convert_event_format_phasespace(" << 0 << ", " << object_list_selection.size() - user->particle_name.size() + 1 << ");" << endline_flush;}
    convert_event_format_phasespace(0, object_list_selection.size() - user->particle_name.size() + 1);
    
    // set user-definded particles and apply defining cuts:
    if (user->particle_name.size() > 0){
      perform_event_selection_user_defined_particles_phasespace();
      if (switch_output_cutinfo){logger << LOG_DEBUG_POINT << "[" << setw(2) << ix_a << "]" << "   after user-defined particles, before general fiducial cuts   (cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << ")." << endline_flush;}
      if (cut_ps[ix_a] == -1){continue;}
    }
    
    // application of general cuts set via fiducialcut interface:
    logger << LOG_DEBUG_POINT << "cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << "   before fiducialcut." << endline_flush;
    for (int i_f = 0; i_f < fiducial_cut.size(); i_f++){
      fiducial_cut[i_f].apply_fiducialcut(ix_a);
      if (switch_output_cutinfo){logger << LOG_DEBUG_POINT << "[" << setw(2) << ix_a << "]" << "   after fiducialcut   " << fiducial_cut[i_f].name << "   (cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << ")." << endline_flush;}
      if (cut_ps[ix_a] == -1){break;}
    }
    if (switch_output_cutinfo){logger << LOG_DEBUG_POINT << "[" << setw(2) << ix_a << "]" << "   after general fiducial cuts, before individual event selection   (cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << ")." << endline_flush;}
    if (cut_ps[ix_a] == -1){continue;}

    if (switch_output_cutinfo){
      for (int i_o = 0; i_o < object_list_selection.size(); i_o++){
	int j_g = access_object_selection[object_list_selection[i_o]];
	info_cut << "[" << setw(2) << ix_a << "]   object_list_selection[" << setw(2) << i_o << "] = " << setw(6) << object_list_selection[i_o] << "   n_object[" << setw(2) << j_g << "][" << setw(2) << ix_a << "] = " << n_object[j_g][ix_a] << "   event_particle_group[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "][" << setw(2) << j_g << "].size() = " << event_particle_group[ix_y][ix_a][j_g].size() << endline;
	for (int i_p = 0; i_p < n_object[j_g][ix_a]; i_p++){
	  info_cut << "       event_particle_group[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "][" << setw(2) << j_g << "][" << setw(2) << i_p << "] = " << event_particle_group[ix_y][ix_a][j_g][i_p]->momentum << endline;
	}
      }
    }

    if (switch_output_cutinfo){info_cut << "convert_event_format_phasespace(" << object_list_selection.size() - user->particle_name.size() + 1 << ", " << object_list_selection.size() << ");" << endline_flush;}
    convert_event_format_phasespace(object_list_selection.size() - user->particle_name.size() + 1, object_list_selection.size());

    cuts(ix_a);

    if (switch_output_cutinfo){info_cut << "[" << setw(2) << ix_a << "]" << "   event selection of user-defined cuts passed and cut_ps[" << ix_a << "] set   (cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << ")." << endline;}
  }
  
  if (switch_output_cutinfo){logger << LOG_INFO << endline << info_cut.str(); info_cut.str("");}
 
  for (int ix_a = 0; ix_a < cut_ps.size(); ix_a++){
    logger << LOG_DEBUG_POINT << "cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << endline_flush;
  }

  logger << LOG_DEBUG_POINT << "switch_output_check_selection = " << switch_output_check_selection << "   psi->i_gen = " << psi->i_gen << endline_flush;
  if (switch_output_check_selection){
    if (psi->i_gen > 0 && psi->i_gen <= switch_output_check_selection){output_phasespacepoints_check_selection();}
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void event_set::perform_event_selection_phasespace(){
  static Logger logger("event_set::perform_event_selection_phasespace");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // index_group_recombination_parton -> indices of partons that could be recombined with photons
  // index_group_photon_parton -> indices of photons
  // index_group_original -> indices of particle groups that partons enter unchanged (potentially after photon recombination)
  // index_group_all_parton -> indices of partons that enter the respective group 
  // index_group_protojet_parton -> indices of QCD partons that enter the jet algorithm
  // index_group_jet -> indices of jet groups
  // index_group_missing_parton -> indices of partons that contribute to missing momentum

  if (csi->class_contribution_QT_powercorrection && ix_a == 1 && !user->switch_value[user->switch_map["recoil"]]){
    // for the recoil kinematics do not boost (use the momenta as they are) 
    for (int i_p = 1; i_p < p_parton[ix_a].size(); i_p++){event_original[0][ix_a][i_p] = particle(p_parton[ix_a][i_p]);}
  }
  else {
    for (int i_p = 1; i_p < p_parton[ix_a].size(); i_p++){event_original[0][ix_a][i_p] = particle(p_parton[ix_a][i_p].zboost(-psi->boost));}
  }
  /*
  // for the recoil kinematics do not boost (use the momenta as they are) 
  if (ix_a==1) {
    for (int i_p = 1; i_p < p_parton[ix_a].size(); i_p++){event_original[0][ix_a][i_p] = particle(p_parton[ix_a][i_p]);}
  }
  */
  /*
  // Introduce different boost for collisions with LAB frame =/= hadronic CM frame !!!
  if (ix_y == 0){for (int i_p = 1; i_p < p_parton[ix_a].size(); i_p++){event_original[0][ix_a][i_p] = particle(p_parton[ix_a][i_p].zboost(-psi->event_boost[0]));}}
  else if (ix_y == 1){for (int i_p = 1; i_p < p_parton[ix_a].size(); i_p++){event_original[1][ix_a][i_p] = (p_parton[ix_a][i_p].zboost(-psi->event_boost[1])).Prot();}
  */
  if (Log::getLogThreshold() <= LOG_DEBUG_POINT && !switch_output_cutinfo){
    output_parton_momenta_LAB();
    logger << LOG_DEBUG_POINT << endline << info_cut.str() << endline_flush;
    info_cut.str("");
    logger.newLine(LOG_DEBUG_POINT);
  }


  ////////////////////////////////////////////////
  //  reset all relevant variables for new PSP  //
  ////////////////////////////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   reset all relevant variables for new PSP" << endline_flush;

  event_photon[ix_y][ix_a].clear();
  event_jet[ix_y][ix_a].clear();
  event_missing[ix_y][ix_a] = particle();

  if (switch_output_cutinfo){output_parton_momenta_LAB();}

  // Make n_object depend on  object_list_selection  instead od  object_list !!!
  for (int i_o = 1; i_o < object_list_selection.size(); i_o++){ // i_o -> counter for objects in relevant_object_list
    int j_g = i_o;
    //    int k_g = equivalent_index[i_o]; // k_g -> index in object_list
    n_object[j_g][ix_a] = 0;
  }
  
  for (int i_o = 1; i_o < object_list_selection.size(); i_o++){event_particle_group[ix_y][ix_a][i_o].clear();} // i_o -> counter for objects in relevant_object_list
  index_unisolated_photon.clear();
  index_unrecombined_photon.clear();

  event_particle_group[ix_y][ix_a][0].resize(event_original[0][ix_a].size());
  for (int i_p = 1; i_p < p_parton[ix_a].size(); i_p++){event_particle_group[ix_y][ix_a][0][i_p] = &event_original[ix_y][ix_a][i_p];}
  
  ////////////////////////////
  //  photon recombination  //
  ////////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   photon recombination" << endline_flush;

  //  Recombinable particles are defined in  photon_recombination_list  (defined via  photon_recombination_selection  and  photon_recombination_disable ).
  //  If a particle is dressed with a photon, the combined momentum is stored at the position of its bare momentum, and the momentum of the recombined photon is empty.
  //  Un-recombined photon indices are stored in  index_unrecombined_photon  such that the emtpy placeholder (leftover from recombined photon) is never accessed.
  //  If no photon recombination is applied,  index_unrecombined_photon  contains indices of all photons.

  if (photon_recombination && index_group_recombination_parton[ix_a].size() > 0 && index_group_photon_parton[ix_a].size() > 0){perform_photon_recombination();}
  else {index_unrecombined_photon = index_group_photon_parton[ix_a];}
  if (switch_output_cutinfo){output_photon_recombination();}

  
  ////////////////////////
  //  runtime original  //
  ////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   runtime original" << endline_flush;

  //  index_group_original   collects partons that enter event selection with their parton-level momenta (after photon dressing if applicable).

  for (size_t i_g = 0; i_g < index_group_original.size(); i_g++){  // i_g -> counter for objects contained in index_group_original
    size_t j_g = index_group_original[i_g];  // j_g -> index in relevant_object_list
    //    size_t k_g = equivalent_index[j_g];  // k_g -> index in object_list
    for (size_t j_p = 0; j_p < index_group_all_parton[ix_a][j_g].size(); j_p++){
      size_t i_p = index_group_all_parton[ix_a][j_g][j_p];  // i_p -> position in  event_particle_group[ix_y][ix_a][0]  of partons that belong to the respective object class (ix_a dependent !!!)
	if (pds[j_g].check_rapidity_eta(&event_original[ix_y][ix_a][i_p])){
	event_particle_group[ix_y][ix_a][j_g].push_back(&event_original[ix_y][ix_a][i_p]);
	if (pds[j_g].check_pT_ET(&event_original[ix_y][ix_a][i_p])){n_object[j_g][ix_a]++;}
      }
    }
    if (switch_output_cutinfo){output_runtime_original(i_g, j_g);}
    
    if (pds[j_g].cut_multiplicity(n_object[j_g][ix_a])){
      cut_ps[ix_a] = -1;
      if (switch_output_cutinfo){info_cut << "[" << right << setw(2) << ix_a << "]" << "   ==========   cut after runtime_original   ==========" << endline << endline;}
      logger << LOG_DEBUG_VERBOSE << "event at [ix_a = " << ix_a << "] cut after runtime_original" << endline_flush;
      return;
    }
    if (n_object[j_g][ix_a] > 1){advanced_sort_by_pT(ix_a, j_g);}
  }
  if (switch_output_cutinfo){output_runtime_original_summary();}


  ///////////////////////
  //  runtime missing  //
  ///////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   runtime missing" << endline_flush;

  if (index_group_missing_parton[ix_a].size() > 0 ){
    for (size_t i_i = 0; i_i < index_group_missing_parton[ix_a].size(); i_i++){
      size_t i_p = index_group_missing_parton[ix_a][i_i];  // i_p -> position in  index_group_missing_parton[ix_a][i_i]  of invisible partons contribute to missing momentum
      event_missing[ix_y][ix_a] += event_original[ix_y][ix_a][i_p];
    }
    event_particle_group[ix_y][ix_a][j_g_missing].push_back(&event_missing[ix_y][ix_a]);
    if (switch_output_cutinfo){output_runtime_missing();}
    if (pds[j_g_missing].check_pT_ET(&event_missing[ix_y][ix_a])){n_object[j_g_missing][ix_a] = 1;}
    else {
      cut_ps[ix_a] = -1;
      if (switch_output_cutinfo){info_cut << "[" << right << setw(2) << ix_a << "]" << "   ==========   cut after runtime_missing   ==========" << endline << endline;}
      logger << LOG_DEBUG_VERBOSE << "event at [ix_a = " << ix_a << "] cut after runtime_missing" << endline_flush;
      return;
    }
    if (switch_output_cutinfo){output_runtime_missing_summary();}
  }


  //////////////////////////////////
  //  determination of protojets  //
  //////////////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   determination of protojets" << endline_flush;

  // temp_protojet  is filled with all QCD partons that enter jet algorithm and are relevant for Frixione isolation (no heavy quarks if not in jet algorithm ???)
  // temp_protojet  does not contain photons here - added later after Frixione isolation.

  temp_protojet.clear();
  for (size_t i_l = 0; i_l < index_group_protojet_parton[ix_a].size(); i_l++){
    size_t i_p = index_group_protojet_parton[ix_a][i_l];
    temp_protojet.push_back(particle_history(event_original[ix_y][ix_a][i_p].momentum, ix_a, i_p, this));
  }
  if (switch_output_cutinfo){output_protojet_determination();}


  //////////////////////////
  //  Frixione isolation  //
  //////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   Frixione isolation" << endline_flush;

  //  Frixione isolation is used to identify isolated (wrt. QCD radiation) photons if (frixione_isolation != 0).
  //  index_unrecombined_photon  contains indices of all photons that have not been recombined with other partons.
  //  index_unisolated_photon  is filled with all photons that are not isolated with these conditions.
  //  If  (photon_photon_recombination)  isolated photons are combined thereafter if close enought to each other.
  
  if (frixione_isolation && j_g_photon){
    for (size_t i_u = 0; i_u < index_unrecombined_photon.size(); i_u++){
      size_t i_p = index_unrecombined_photon[i_u];
      if (perform_frixione_isolation(event_original[ix_y][ix_a][i_p])){event_photon[ix_y][ix_a].push_back(event_original[ix_y][ix_a][i_p]);}
      else {index_unisolated_photon.push_back(i_p);}
    }
    if (switch_output_cutinfo){output_frixione_isolation();}
    bool flag_recombination_photon_photon = false;
    if (photon_photon_recombination && index_unrecombined_photon.size() > 1){flag_recombination_photon_photon = perform_photon_photon_recombination();}
    for (size_t i_i = 0; i_i < event_photon[ix_y][ix_a].size(); i_i++){
      if (pds[j_g_photon].check_rapidity_eta(&event_photon[ix_y][ix_a][i_i])){
	event_particle_group[ix_y][ix_a][j_g_photon].push_back(&event_photon[ix_y][ix_a][i_i]);
	if (pds[j_g_photon].check_pT_ET(&event_photon[ix_y][ix_a][i_i])){n_object[j_g_photon][ix_a]++;}
      }
    }
    if (switch_output_cutinfo){output_frixione_isolation_multiphoton(flag_recombination_photon_photon);}
    if (pds[j_g_photon].cut_multiplicity(n_object[j_g_photon][ix_a])){
      cut_ps[ix_a] = -1;
      if (switch_output_cutinfo){info_cut << "[" << right << setw(2) << ix_a << "]" << "   ==========   cut after runtime_photon_isolation   ==========" << endline << endline;}
      logger << LOG_DEBUG_VERBOSE << "event at [ix_a = " << ix_a << "] cut after runtime_photon_isolation" << endline_flush;
      return;
    }
    if (event_particle_group[ix_y][ix_a][j_g_photon].size() > 1){advanced_sort_by_pT(ix_a, j_g_photon);}
    if (switch_output_cutinfo){output_frixione_isolation_summary();}
  }
  else {index_unisolated_photon = index_unrecombined_photon;}


  /////////////////////////////////////////
  //  treatment of non-isolated photons  //
  /////////////////////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   treatment of non-isolated photons" << endline_flush;

  //  Check if this distinction makes sense in general !!!
  //  At present, only non-isolated photons enter  index_unisolated_photon .
  //  Isolated photons that fail define_y/eta are removed, and those failing define_pT/ET enter event_particle_group[photon], but do not raise n_object[photon].
  //  Non-identified and non-recombined photons are added to protojets (to eventually form "ajet"s).
  if (j_g_photon){
  if (photon_jet_algorithm){
    for (size_t i_u = 0; i_u < index_unisolated_photon.size(); i_u++){
      size_t i_p = index_unrecombined_photon[i_u];
      temp_protojet.push_back(particle_history(event_original[ix_y][ix_a][i_p].momentum, ix_a, i_p, this));
    }
  }
  else {
    // Directly fill "ajet" with left-over photons ??? So far, only addition of photons to jet_algorithm used !!!
    size_t j_g = access_object_selection["ajet"];
    if (j_g > 0){
      //    size_t k_g = observed_object["ajet"];
      logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   index_unisolated_photon.size() = " << index_unisolated_photon.size() << endline_flush;
      for (size_t i_u = 0; i_u < index_unisolated_photon.size(); i_u++){
	size_t i_p = index_unisolated_photon[i_u];
	event_particle_group[ix_y][ix_a][j_g].push_back(&event_original[ix_y][ix_a][i_p]);
      }
      logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   event_particle_group[" << setw(2) << j_g << "].size() = " << event_particle_group[ix_y][ix_a][j_g].size() << endline_flush;
      if (event_particle_group[ix_y][ix_a][j_g].size() > 1){advanced_sort_by_pT(ix_a, j_g);}
    }
  }
  if (switch_output_cutinfo){output_photon_jet_algorithm();}
  }

  //////////////////////////////////////////////////
  //  elimination of protojets too close to beam  //
  //////////////////////////////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   elimination of protojets too close to beam" << endline_flush;

  // Exclude protojets which are too close to the beam (in terms of y or eta)
  
  if (parton_y_max != 0. || parton_eta_max != 0.){
    bool check_removal = false;
    for (size_t j = temp_protojet.size(); j-- > 0;){
      if (abs(temp_protojet[j].rapidityXX()) > parton_y_max || abs(temp_protojet[j].etaXX()) > parton_eta_max){
	temp_protojet.erase(temp_protojet.begin() + j);
	check_removal = true;
      }
    }
    if (switch_output_cutinfo){output_protojet_removal_beam(check_removal);}
  }

  
  /////////////////////
  //  jet algorithm  //
  /////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   jet algorithm" << endline_flush;

  if (switch_output_cutinfo){output_jet_algorithm_protojet();}
  event_jet[0][ix_a].clear();
  jet_algorithm_flavour();
  if (switch_output_cutinfo){output_jet_algorithm_jet();}


  ////////////////////////////////////////////////////////////
  //  Frixione isolation - removal of jets close to photon  //
  ////////////////////////////////////////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   Frixione isolation - removal of jets close to photon" << endline_flush;

  // Remove jets which are too close to an 'isolated photon' according to Frixione cone size
  // might need to be adapted for photon_recombination & frixione_isolation !!!

  if (frixione_isolation && frixione_jet_removal && j_g_photon){ // up to date ??? equivalent_index[access_object_selection[...]] constructions !!!
    if (switch_output_cutinfo){output_frixione_jet_removal();}
    bool check_removal = false;
    for (size_t i_phot = 0; i_phot < n_object[j_g_photon][ix_a]; i_phot++){
      //    for (size_t i_phot = 0; i_phot < n_object[access_object["photon"]][ix_a]; i_phot++){
      for (size_t i_jet = event_jet[0][ix_a].size(); i_jet-- > 0;){
	// basically never tested !!!
	if (sqrt(R2_eta(*event_particle_group[ix_y][ix_a][access_object_selection["photon"]][i_phot], event_jet[0][ix_a][i_jet])) < frixione_delta_0){
	  event_jet[0][ix_a].erase(event_jet[0][ix_a].begin() + i_jet);
	}
      }
    }
    if (switch_output_cutinfo){output_summary_frixione_jet_removal(check_removal);}
  }


  /////////////////////////////////////////////////////////
  //  jet algorithm - interpretation of individual jets  //
  /////////////////////////////////////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   jet algorithm - interpretation of individual jets" << endline_flush;

  sort(event_jet[0][ix_a].begin(), event_jet[0][ix_a].end(), greaterBypT);
  for (int i_j = 0; i_j < event_jet[0][ix_a].size(); i_j++){event_jet[0][ix_a][i_j].determination_identity();}
  if (switch_output_cutinfo){output_jet_interpretation();}

  
  ////////////////////////////////////////////////
  //  jet algorithm - resulting jet containers  //
  ////////////////////////////////////////////////
  logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "]   jet algorithm - resulting jet containers" << endline_flush;

  for (size_t i_g = 0; i_g < index_group_jet.size(); i_g++){
    size_t j_g = index_group_jet[i_g];
    //    size_t k_g = equivalent_index[j_g];
    for (size_t i_j = 0; i_j < event_jet[0][ix_a].size(); i_j++){
      if      (j_g == access_object_selection["jet"] && event_jet[0][ix_a][i_j].identity == 22){continue;}
      else if (j_g == access_object_selection["ajet"] && event_jet[0][ix_a][i_j].identity != 22){continue;}
      else if (j_g == access_object_selection["ljet"] && event_jet[0][ix_a][i_j].identity != 0){continue;}
      else if (j_g == access_object_selection["bjet"] && abs(event_jet[0][ix_a][i_j].identity) != 5){continue;}
      else if (j_g == access_object_selection["bjet_b"] && event_jet[0][ix_a][i_j].identity != +5){continue;}
      else if (j_g == access_object_selection["bjet_bx"] && event_jet[0][ix_a][i_j].identity != -5){continue;}
      if (pds[j_g].check_rapidity_eta(&event_jet[0][ix_a][i_j])){
	event_particle_group[ix_y][ix_a][j_g].push_back(&event_jet[ix_y][ix_a][i_j]);
	if (pds[j_g].check_pT_ET(&event_jet[0][ix_a][i_j])){n_object[j_g][ix_a]++;}
      }
    }
    if (switch_output_cutinfo){output_jet_algorithm_result(i_g, j_g);}
    if (pds[j_g].cut_multiplicity(n_object[j_g][ix_a])){
      cut_ps[ix_a] = -1;
      if (switch_output_cutinfo){info_cut << "[" << right << setw(2) << ix_a << "]" << "   ==========   cut after runtime_jet_recombination   ==========" << endline;}
      logger << LOG_DEBUG_VERBOSE << "event at [ix_a = " << ix_a << "] cut after runtime_jet_recombination" << endline_flush;
      return;
    }
    if (event_particle_group[ix_y][ix_a][j_g].size() > 1){advanced_sort_by_pT(ix_a, j_g);}
  }
  if (switch_output_cutinfo){output_jet_algorithm_result_summary();}

  
  if (switch_output_cutinfo){
    info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   General particle definitions finished and multiplicity cuts passed   (cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << ")." << endline;
  }

  if (csi->class_contribution_QT_powercorrection){cut_ps[ix_a] = psi->recoil_rcut;}
  else if (csi->class_contribution_qTcut_explicit){qT_basic->event_selection_qTcut(ix_a); if (cut_ps[ix_a] == -1){return;}}
  else if (csi->class_contribution_qTcut_implicit){cut_ps[ix_a] = qT_basic->n_qTcut - 1;}
  //  else if (csi->class_contribution_NJcut_explicit){NJ_basic->event_selection_NJcut(ix_a); if (cut_ps[ix_a] == -1){return;}}
  //  else if (csi->class_contribution_NJcut_implicit){cut_ps[ix_a] = NJ_basic->n_NJcut - 1;}
  else {cut_ps[ix_a] = 0;}
  
  
  if (switch_output_cutinfo){info_cut << "[" << right << setw(2) << ix_a << "]" << "   generic event selection passed and cut_ps[" << ix_a << "] set   (cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << ")." << endline;}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void event_set::perform_event_selection_user_defined_particles_phasespace(){
  static Logger logger("event_set::perform_event_selection_user_defined_particles_phasespace");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // User-defined particles adapted to the new format !!! (definition via particle <new_particle> = ... (additional content information ???)
  for (int i_g = 0; i_g < user->particle_name.size(); i_g++){user_particle[i_g][ix_a].clear();}

  particles(ix_a);
      
  if (switch_output_cutinfo){for (int i_g = 0; i_g < user->particle_name.size(); i_g++){for (int i_p = 0; i_p < user_particle[i_g][ix_a].size(); i_p++){info_cut << "user_particle[" << i_g << "][" << ix_a << "][" << i_p << "] = " << user_particle[i_g][ix_a][i_p] << endline;}}}

  // Start at 1 since 0 collect undefined user-defeined particles !!!
  //      for (int i_g = 0; i_g < user->particle_name.size(); i_g++){
  for (int i_g = 1; i_g < user->particle_name.size(); i_g++){
    //	int k_g = access_object[user->particle_name[i_g]];
    int j_g = access_object_selection[user->particle_name[i_g]];
    for (int i_p = 0; i_p < user_particle[i_g][ix_a].size(); i_p++){
      if (pds[j_g].check_rapidity_eta(&user_particle[i_g][ix_a][i_p])){
	event_particle_group[ix_y][ix_a][j_g].push_back(&user_particle[i_g][ix_a][i_p]);
	if (pds[j_g].check_pT_ET(&user_particle[i_g][ix_a][i_p])){n_object[j_g][ix_a]++;}
      }
    }
    if (switch_output_cutinfo){info_cut << "[" << setw(2) << ix_a << "]" << "   " << pds[j_g].n_observed_min << " <= n_object[" << setw(2) << j_g << "][" << setw(2) << ix_a << "] = " << n_object[j_g][ix_a] << " <= " << pds[j_g].n_observed_max << endline;}
    if (pds[j_g].cut_multiplicity(n_object[j_g][ix_a])){
      cut_ps[ix_a] = -1; 
      if (switch_output_cutinfo){info_cut << "[" << setw(2) << ix_a << "]" << "   cut after user_particle[" << i_g << "] " << user->particle_name[i_g] << endline;}
      if (switch_output_cutinfo){logger << LOG_DEBUG << endline << info_cut.str(); }
      // Check why there should be a return here !!! ???
      //	  logger << LOG_WARN << "return statement at strange position !!!" << endline_flush;
      return;
      //      continue;
    }
    //	if (n_object[j_g][ix_a] > 1){sort(event_particle_group[ix_y][ix_a][j_g].begin(), event_particle_group[ix_y][ix_a][j_g].end(), XgreaterBypT);}
    if (n_object[j_g][ix_a] > 1){advanced_sort_by_pT(ix_a, j_g);} 
  }
  if (switch_output_cutinfo){info_cut << "[" << setw(2) << ix_a << "]" << "   event selection of user-defined particles passed and cut_ps[" << ix_a << "] set   (cut_ps[" << ix_a << "] = " << cut_ps[ix_a] << ")." << endline;}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void event_set::convert_event_format_phasespace(int begin, int end){
  static Logger logger("convert_event_format_phasespace");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // Required to keep exactly the particle access format of MATRIX_v2.0.0 in user-defined code !!!
  //  for (int j_g = 0; j_g < object_list_selection.size(); j_g++){
  for (int j_g = begin; j_g < end; j_g++){
    particle_event[ix_y][ix_a][j_g].clear();
    particle_event[ix_y][ix_a][j_g].resize(event_particle_group[ix_y][ix_a][j_g].size());
    for (int i_i = 0; i_i < event_particle_group[ix_y][ix_a][j_g].size(); i_i++){
      if (j_g == 0 && i_i == 0){continue;}
      //      logger << LOG_DEBUG_VERBOSE << "*event_particle_group[" << setw(2) << ix_y << "][" << setw(2) << ix_a << "][" << setw(2) << j_g << "][" << setw(2) << i_i << "] = " << event_particle_group[ix_y][ix_a][j_g][i_i]->momentum << endline_flush;
      event_particle_group[ix_y][ix_a][j_g][i_i]->calculate_all();
      particle_event[ix_y][ix_a][j_g][i_i] = *event_particle_group[ix_y][ix_a][j_g][i_i];
      //      particle_event[ix_y][ix_a][j_g][i_i].calculate_all();
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}
