#include "header.hpp"

fiducialcut::fiducialcut(string & _name, event_set & _esi){
  Logger logger("fiducialcut::fiducialcut - '" + _name + "'");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  name = _name;
  esi = &_esi;

  cut_value_lower = -std::numeric_limits<double>::infinity();
  cut_value_upper = +std::numeric_limits<double>::infinity();

  vector<string> split_name(1, "");
  logger << LOG_DEBUG << "split_name.size() = " << split_name.size() << endline_flush;
  int counter = 0;
  int start = 0;
  for (int i_c = 0; i_c < name.size(); i_c++){
    if (name[i_c] == ' '){start = 0;}
    if (name[i_c] != ' ' && start == 0){
      counter++;
      split_name.push_back("");
      start = 1;
    }
    if (name[i_c] != ' '){
      split_name[counter].push_back(name[i_c]);
    }
  }
  for (int i_s = 0; i_s < split_name.size(); i_s++){
    logger << LOG_DEBUG << "split_name[" << i_s << "] = " << split_name[i_s] << endline_flush;
  }

  vector<int> identification(split_name.size(), 0);
  logger.newLine(LOG_DEBUG);
  logger << LOG_DEBUG << "name = " << name << endline_flush;
  logger.newLine(LOG_DEBUG);

  for (int i_s = 0; i_s < split_name.size(); i_s++){
    if (split_name[i_s] == ""){identification[i_s] = -1; continue;}
    if (identification[i_s - 1] == -1 || identification[i_s - 1] == 4){
      for (int i_l = 0; i_l < esi->fiducial_cut_list.size(); i_l++){
	logger << LOG_DEBUG_VERBOSE << "split_name[" << i_s << "] = " << split_name[i_s] << "  ?=  esi->fiducial_cut_list[" << i_l << "] = " << esi->fiducial_cut_list[i_l] << endline_flush;
	if (split_name[i_s] == esi->fiducial_cut_list[i_l]){
	  logger << LOG_DEBUG_VERBOSE << "split_name[" << i_s << "] = " << split_name[i_s] << " = esi->fiducial_cut_list[" << i_l << "] = " << esi->fiducial_cut_list[i_l] << endline_flush;
	  identification[i_s] = 1;
	  break;
	}
      }
    }

    if (identification[i_s]){continue;}

    if (identification[i_s - 1] == 1 || identification[i_s - 1] == 2 || identification[i_s - 1] == 3){
      for (int i_l = 0; i_l < esi->object_list.size(); i_l++){
	if (split_name[i_s] == esi->object_list[i_l]){
	  logger << LOG_DEBUG_VERBOSE << "split_name[" << i_s << "] = " << split_name[i_s] << " = esi->object_list[" << i_l << "] = " << esi->object_list[i_l] << endline_flush;
	  identification[i_s] = 2;
	  break;
	}
      }
    }

    if (identification[i_s]){continue;}

    if (split_name[i_s] == ">" || split_name[i_s] == "<"){
      identification[i_s] = 4;
    }

    if (identification[i_s]){continue;}

    if (identification[i_s - 1] == 2){
      if ((split_name[i_s][0] == '(' && split_name[i_s][split_name[i_s].size() - 1] == ')') ||
	  (split_name[i_s][0] == '[' && split_name[i_s][split_name[i_s].size() - 1] == ']') ||
	  (split_name[i_s][0] == '{' && split_name[i_s][split_name[i_s].size() - 1] == '}')){
	split_name[i_s].erase(split_name[i_s].size() - 1, 1);
	split_name[i_s].erase(0, 1);
      }
      else {
      }
      identification[i_s] = 3;
    }

    if (identification[i_s]){continue;}

    int check_number = 0;
    for (int i_c = 0; i_c < split_name[i_s].size(); i_c++){
      if (split_name[i_s][i_c] == '+' ||
	  split_name[i_s][i_c] == '-' ||
	  split_name[i_s][i_c] == '.' ||
	  split_name[i_s][i_c] == 'e' ||
	  split_name[i_s][i_c] == '0' ||
	  split_name[i_s][i_c] == '1' ||
	  split_name[i_s][i_c] == '2' ||
	  split_name[i_s][i_c] == '3' ||
	  split_name[i_s][i_c] == '4' ||
	  split_name[i_s][i_c] == '5' ||
	  split_name[i_s][i_c] == '6' ||
	  split_name[i_s][i_c] == '7' ||
	  split_name[i_s][i_c] == '8' ||
	  split_name[i_s][i_c] == '9'){
	check_number++;
      }
    }
    if (split_name[i_s].size() == check_number){
      identification[i_s] = 5;
    }
  }

  for (int i_s = 0; i_s < identification.size(); i_s++){
    logger << LOG_DEBUG_VERBOSE << "split_name[" << i_s << "] = " << setw(10) << split_name[i_s] << "   identification[" << i_s << "] = " << identification[i_s] << endline_flush;
    if (identification[i_s] == 0){
      logger << LOG_FATAL << "Input cannot be interpreted: " << split_name[i_s] << endline_flush;
      exit(1);
    }
  }

  int no_cut = -1;
  // sanity checks and assignment from input to actual cut to be applied:
  for (int i_s = 0; i_s < identification.size(); i_s++){
    if (identification[i_s] == 1 && no_cut == -1){
      no_cut = i_s;
      observable = split_name[i_s];
    }
    else if (identification[i_s] == 1 && no_cut > -1){
      logger << LOG_FATAL << "Two cuts defined -> input cannot be interpreted." << endline;
      logger << LOG_FATAL << "Will be used for alternative cuts (like rapidity gaps)." << endline_flush;
      exit(1);
    }
  }

  logger << LOG_DEBUG << "Defined cut type has been identified." << endline_flush;

  int no_lower = -1;
  int no_upper = -1;

  for (int i_s = 0; i_s < identification.size(); i_s++){
    if (identification[i_s] == 4){
      if ((i_s < no_cut && split_name[i_s] == "<") ||
	  (i_s > no_cut && split_name[i_s] == ">")){
	if (no_lower == -1){
	  no_lower = i_s;
	  if (i_s < no_cut && i_s > 0){
	    if (identification[i_s - 1] == 5){
	      cut_value_lower = atof(split_name[i_s - 1].c_str());
	    }
	  }
	  if (i_s > no_cut && i_s < identification.size() - 1){
	    if (identification[i_s + 1] == 5){
	      cut_value_lower = atof(split_name[i_s + 1].c_str());
	    }
	  }
	}
	else {
	  logger << LOG_FATAL << "Two limits defined -> input cannot be interpreted." << endline_flush;
	  exit(1);
	}
      }

      if ((i_s < no_cut && split_name[i_s] == ">") ||
	  (i_s > no_cut && split_name[i_s] == "<")){
	if (no_upper == -1){
	  no_upper = i_s;
	  if (i_s < no_cut && i_s > 0){
	    if (identification[i_s - 1] == 5){
	      cut_value_upper = atof(split_name[i_s - 1].c_str());
	    }
	  }
	  if (i_s > no_cut && i_s < identification.size() - 1){
	    if (identification[i_s + 1] == 5){
	      cut_value_upper = atof(split_name[i_s + 1].c_str());
	    }
	  }
	}
	else {
	  logger << LOG_FATAL << "Two limits defined -> input cannot be interpreted." << endline_flush;
	  exit(1);
	}
      }
    }
  }

  if (cut_value_lower == -std::numeric_limits<double>::infinity() && cut_value_upper == +std::numeric_limits<double>::infinity()){
    logger << LOG_FATAL << "No cut specified." << endline_flush;
    exit(1);
  }
  else if (cut_value_lower > -std::numeric_limits<double>::infinity() && cut_value_upper == +std::numeric_limits<double>::infinity()){
    type = LOWER;
  }
  else if (cut_value_lower == -std::numeric_limits<double>::infinity() && cut_value_upper < +std::numeric_limits<double>::infinity()){
    type = UPPER;
  }
  else if (cut_value_lower < cut_value_upper){
    type = WINDOW;
  }
  else if (cut_value_lower > cut_value_upper){
    type = GAP;
  }
  else if (cut_value_lower == cut_value_upper){
    logger << LOG_FATAL << "No cut interval specified (lower limit equals upper limit)." << endline_flush;
    exit(1);
  }
  logger << LOG_DEBUG << "Upper and lower limits have been identified." << endline_flush;

  for (int i_s = 0; i_s < identification.size(); i_s++){
    if (identification[i_s] == 2){
      int temp = 0;
      type_particle.push_back(split_name[i_s]);
      if (i_s < identification.size() - 1){
	if (identification[i_s + 1] == 3){
	  logger << LOG_DEBUG << "particle number: split_name[" << i_s + 1 << "] = " << split_name[i_s + 1] << "   atoi -> " << atoi(split_name[i_s + 1].c_str()) << endline_flush;
	  //	  if (split_name[i_s + 1] != "0" && atoi(split_name[i_s + 1].c_str()) == 0){
	  if (atoi(split_name[i_s + 1].c_str()) == 0){
	    logger << LOG_FATAL << "Invalid particle number specified." << endline_flush;
	    exit(1);
	  }
	  // Maybe introduce identifiers like '+' and 'x' here as well ???
	  
	  order_min.push_back(atoi(split_name[i_s + 1].c_str()));
	  order_max.push_back(atoi(split_name[i_s + 1].c_str()));
	  temp = 1;
	}
      }
      if (!temp){
	order_min.push_back(1);
	order_max.push_back(esi->pda[esi->observed_object[type_particle[type_particle.size() - 1]]].n_observed_max);
      }
    }
  }

  logger << LOG_DEBUG << "Involved particle types and specifications have been identified." << endline_flush;

  vector<int> n_particles_in_type(type_particle.size());
  int n_combination_max = 1.;
  for (int i_p = 0; i_p < type_particle.size(); i_p++){
    n_particles_in_type[i_p] = (order_max[i_p] - order_min[i_p] + 1);
    n_combination_max *= (order_max[i_p] - order_min[i_p] + 1);
  }

  logger << LOG_DEBUG_VERBOSE << "type_particle.size() = " << type_particle.size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "order_min.size() = " << order_min.size() << endline_flush;
  for (int i_n = 0; i_n < order_min.size(); i_n++){logger << LOG_DEBUG_VERBOSE << "order_min[" << i_n << "] = " << order_min[i_n] << endline_flush;}
  logger << LOG_DEBUG_VERBOSE << "order_max.size() = " << order_max.size() << endline_flush;
  for (int i_n = 0; i_n < order_max.size(); i_n++){logger << LOG_DEBUG_VERBOSE << "order_max[" << i_n << "] = " << order_max[i_n] << endline_flush;}
  logger << LOG_DEBUG_VERBOSE << "n_particles_in_type.size() = " << n_particles_in_type.size() << endline_flush;
  for (int i_n = 0; i_n < n_particles_in_type.size(); i_n++){logger << LOG_DEBUG_VERBOSE << "n_particles_in_type[" << i_n << "] = " << n_particles_in_type[i_n] << endline_flush;}
  logger << LOG_DEBUG << "n_combination_max = " << n_combination_max << endline_flush;

  for (int i_c = 0; i_c < n_combination_max; i_c++){
    vector<int> temp_entry(type_particle.size());
    int rest = i_c;
    for (int i_p = 0; i_p < type_particle.size(); i_p++){
      temp_entry[i_p] = rest % n_particles_in_type[i_p] + 1;
      logger << LOG_DEBUG_VERBOSE << "i_c = " << i_c << "   rest = " << rest << "   rest - temp_entry[" << i_p << "] = " << rest - temp_entry[i_p] << "   new_rest = " << (rest - temp_entry[i_p]) / n_particles_in_type[i_p] << "   n_particles_in_type[" << i_p << "] = " << n_particles_in_type[i_p] << endline_flush;
      rest = (rest - temp_entry[i_p] + 1) / n_particles_in_type[i_p];
    }
    // n_particles_in_type[i_p] == 1 -> no container, i.e. '==' instead of '>=' in removal procedure:
    // maybe switch meaning of temp_entry here: selected number instead of ordering number (which would always be zero in that case):

    for (int i_p = 0; i_p < type_particle.size(); i_p++){
      if (n_particles_in_type[i_p] == 1){temp_entry[i_p] = order_min[i_p];}
    }

    int remove_combination = 0;
    for (int i_p = 0; i_p < type_particle.size(); i_p++){
      for (int j_p = i_p + 1; j_p < type_particle.size(); j_p++){
	logger << LOG_DEBUG_VERBOSE << "i_c = " << i_c << "   temp_entry[" << i_p << "] = " << temp_entry[i_p] << "   temp_entry[" << j_p << "] = " << temp_entry[j_p] << "     " << (temp_entry[i_p] >= temp_entry[j_p]) << endline_flush;
	logger << LOG_DEBUG << "i_p = " << i_p << "   j_p = " << j_p << "   n_particles_in_type[i_p] = " << n_particles_in_type[i_p] << "   n_particles_in_type[j_p] = " << n_particles_in_type[j_p] << "   type_particle[i_p] = " << type_particle[i_p] << "   type_particle[j_p] = " << type_particle[j_p] << "   temp_entry[i_p] = " << temp_entry[i_p] << "   temp_entry[i_p] = " << temp_entry[i_p] << endline_flush;
	if (type_particle[i_p] == type_particle[j_p] && temp_entry[i_p] >= temp_entry[j_p]){
	  /*
	  if (type_particle[i_p] == type_particle[j_p] &&
	    ((n_particles_in_type[i_p] > 1 && n_particles_in_type[j_p] > 1 && temp_entry[i_p] >= temp_entry[j_p]) ||
	     (n_particles_in_type[i_p] == 1 && n_particles_in_type[j_p] > 1 && temp_entry[i_p] == temp_entry[j_p]) ||
	     //	     (n_particles_in_type[i_p] > 1 && n_particles_in_type[j_p] == 1 && temp_entry[i_p] == temp_entry[j_p]))){
	     //  line added to avoid cuts built upon identical particles
	     // apparently causes problems for dipole kinematics !!! -> find a better solution !!!
	     (n_particles_in_type[i_p] > 1 && n_particles_in_type[j_p] == 1 && temp_entry[i_p] == temp_entry[j_p]) ||
	     (n_particles_in_type[i_p] == 1 && n_particles_in_type[j_p] == 1 && temp_entry[i_p] == temp_entry[j_p]))){
	  */	  
	  // Why so complicated, and not just
	  //	  if (type_particle[i_p] == type_particle[j_p] && n_particles_in_type[i_p] >= 1 && n_particles_in_type[j_p] >= 1 && temp_entry[i_p] >= temp_entry[j_p]){
	  // or simply
	  //	  if (type_particle[i_p] == type_particle[j_p] && temp_entry[i_p] >= temp_entry[j_p]){
	  // ???
	  remove_combination = 1;
	  logger << LOG_DEBUG << "remove_combination = " << remove_combination << endline_flush;
	  break;
	}
      }
    }

    if (!(remove_combination)){
      for (int i_p = 0; i_p < type_particle.size(); i_p++){logger << LOG_DEBUG << "temp_entry[" << i_p << "] = " <<  temp_entry[i_p] << endline_flush;}
      // Check if this is still correct !!!
      // doesn't work so far, if (at least some) particle numbers are given !!!
      // version of temp_entry needed that contains not 0, but the specified ordering number for each group !!!
      // Reduce particle numbers by 1 to simplify container access:
      for (int i_p = 0; i_p < type_particle.size(); i_p++){temp_entry[i_p]--;}
      no_particles.push_back(temp_entry);
    }
  }

  n_combination = no_particles.size();
  cut_combination.resize(n_combination, 0);

  logger.newLine(LOG_DEBUG);


  index_group.resize(type_particle.size());
  index_group_selection.resize(type_particle.size());
  for (int i_i = 0; i_i < type_particle.size(); i_i++){
    index_group[i_i] = esi->access_object[type_particle[i_i]];
    index_group_selection[i_i] = esi->access_object_selection[type_particle[i_i]];
  }

  // More direct determination !!! Use index_particle directly instead of no_particles ???
  index_particle.resize(no_particles.size());
  for (int i_c = 0; i_c < no_particles.size(); i_c++){
    index_particle[i_c].resize(no_particles[i_c].size());
    for (int i_i = 0; i_i < no_particles[i_c].size(); i_i++){
      index_particle[i_c][i_i] = no_particles[i_c][i_i];
    }
  }
  
  // Validity checks for cuts should be implemented !!!

  for (int i_s = 0; i_s < split_name.size(); i_s++){logger << LOG_DEBUG << "split_name[" << i_s << "] = " << setw(10) << split_name[i_s] << "   identification = " << identification[i_s] << endline_flush;}
  logger << LOG_DEBUG << "cut_type = " << split_name[no_cut] << " @ " << no_cut << endline_flush;
  logger << LOG_DEBUG << "cut_value_lower = " << cut_value_lower << " @ " << no_lower << endline_flush;
  logger << LOG_DEBUG << "cut_value_upper = " << cut_value_upper << " @ " << no_upper << endline_flush;
  logger << LOG_DEBUG << "n_combination = " << no_particles.size() << ":" << endline_flush;
  if (no_particles.size() == 0){logger << LOG_DEBUG << "Cut has no effect and can be removed." << endline_flush;}
  for (int i_c = 0; i_c < no_particles.size(); i_c++){
    stringstream temp_no_particles_ss;
    for (int i_p = 0; i_p < type_particle.size(); i_p++){
      temp_no_particles_ss << type_particle[i_p] << " " << no_particles[i_c][i_p] << "   ";
    }
    logger << LOG_DEBUG << "particles = " << temp_no_particles_ss.str() << endline_flush;
  }
  logger.newLine(LOG_DEBUG);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void fiducialcut::apply_fiducialcut(int i_a){
  Logger logger("fiducialcut::apply_fiducialcut");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  cut = 0;
  cut_combination.resize(n_combination, 0);

  for (int i_c = 0; i_c < index_particle.size(); i_c++){
    logger << LOG_DEBUG_VERBOSE << "observable [" << i_c << "]" << endline_flush;

    // Check if  (fail_exist_particle = true) for
    //    index_particle[i_c][i_p] >= esi->event_particle_group[0][i_a][esi->access_object[type_particle[i_p]]].size()
    // or
    //    esi->n_object[esi->access_object_selection[type_particle[i_p]]][i_a]
    // !!!
    // Maybe no need to calculate index for every points ???
    
    // calculate observable

    double fail_exist_particle = false;
    double value_observable = 0.;
    if (observable == "M"){
      fourvector temp_fv;
      for (int i_p = 0; i_p < type_particle.size(); i_p++){
	if (index_particle[i_c][i_p] >= esi->n_object[index_group_selection[i_p]][i_a]){fail_exist_particle = true;}
	else {temp_fv = temp_fv + esi->event_particle_group[esi->ix_y][i_a][index_group_selection[i_p]][index_particle[i_c][i_p]]->momentum;}
      }
      value_observable = temp_fv.m();
    }
    else if (observable == "pT"){
      if (type_particle.size() == 1){
	if (index_particle[i_c][0] >= esi->n_object[index_group_selection[0]][i_a]){fail_exist_particle = true;}
	else {value_observable = esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]]->pTXX();}
      }
      else {
	fourvector temp_fv;
	for (int i_p = 0; i_p < type_particle.size(); i_p++){
	  if (index_particle[i_c][i_p] >= esi->n_object[index_group_selection[i_p]][i_a]){fail_exist_particle = true;}
	  else {temp_fv = temp_fv + esi->event_particle_group[esi->ix_y][i_a][index_group_selection[i_p]][index_particle[i_c][i_p]]->momentum;}
	}
	value_observable = temp_fv.pT();
      }
    }
    else if (observable == "|eta|"){
      if (type_particle.size() == 1){
	if (index_particle[i_c][0] >= esi->n_object[index_group_selection[0]][i_a]){fail_exist_particle = true;}
	else {value_observable = abs(esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]]->etaXX());}
      }
      else {
	fourvector temp_fv;
	for (int i_p = 0; i_p < type_particle.size(); i_p++){
	  if (index_particle[i_c][i_p] >= esi->n_object[index_group_selection[i_p]][i_a]){fail_exist_particle = true;}
	  else {temp_fv = temp_fv + esi->event_particle_group[esi->ix_y][i_a][index_group_selection[i_p]][index_particle[i_c][i_p]]->momentum;}
	}
	value_observable = abs(temp_fv.eta());
      }
    }
    else if (observable == "|y|"){
      if (type_particle.size() == 1){
	if (index_particle[i_c][0] >= esi->n_object[index_group_selection[0]][i_a]){fail_exist_particle = true;}
	else {value_observable = abs(esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]]->rapidityXX());}
      }
      else {
	fourvector temp_fv;
	for (int i_p = 0; i_p < type_particle.size(); i_p++){
	  if (index_particle[i_c][i_p] >= esi->n_object[index_group_selection[i_p]][i_a]){fail_exist_particle = true;}
	  else {temp_fv = temp_fv + esi->event_particle_group[esi->ix_y][i_a][index_group_selection[i_p]][index_particle[i_c][i_p]]->momentum;}
	}
	value_observable = abs(temp_fv.rapidity());
      }
    }
    else if (observable == "|deta|"){
      if (type_particle.size() == 2){
	if (index_particle[i_c][0] >= esi->n_object[index_group_selection[0]][i_a] || index_particle[i_c][1] >= esi->n_object[index_group_selection[1]][i_a]){fail_exist_particle = true;}
	else {value_observable = abs(esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]]->etaXX() - esi->event_particle_group[esi->ix_y][i_a][index_group_selection[1]][index_particle[i_c][1]]->etaXX());}
      }
      else {
	logger << LOG_FATAL << "Fiducial cut for " << observable << " with " << type_particle.size() << " is not defined." << endline_flush;
	exit(1);
      }
    }
    else if (observable == "|dy|"){
      if (type_particle.size() == 2){
	if (index_particle[i_c][0] >= esi->n_object[index_group_selection[0]][i_a] || index_particle[i_c][1] >= esi->n_object[index_group_selection[1]][i_a]){fail_exist_particle = true;}
	else {value_observable = abs(esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]]->rapidityXX() - esi->event_particle_group[esi->ix_y][i_a][index_group_selection[1]][index_particle[i_c][1]]->rapidityXX());}
      }
      else {
	logger << LOG_FATAL << "Fiducial cut for " << observable << " with " << type_particle.size() << " is not defined." << endline_flush;
	exit(1);
      }
    }
    else if (observable == "dReta"){
      if (type_particle.size() == 2){
	if (index_particle[i_c][0] >= esi->n_object[index_group_selection[0]][i_a] || index_particle[i_c][1] >= esi->n_object[index_group_selection[1]][i_a]){fail_exist_particle = true;}
	else {value_observable = R_eta(esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]], esi->event_particle_group[esi->ix_y][i_a][index_group_selection[1]][index_particle[i_c][1]]);}
      }
      else {
	logger << LOG_FATAL << "Fiducial cut for " << observable << " with " << type_particle.size() << " is not defined." << endline_flush;
	exit(1);
      }
    }
    else if (observable == "dRy"){
      if (type_particle.size() == 2){
	if (index_particle[i_c][0] >= esi->n_object[index_group_selection[0]][i_a] || index_particle[i_c][1] >= esi->n_object[index_group_selection[1]][i_a]){fail_exist_particle = true;}
	else {value_observable = R_rapidity(esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]], esi->event_particle_group[esi->ix_y][i_a][index_group_selection[1]][index_particle[i_c][1]]);}
      }
      else {
	logger << LOG_FATAL << "Fiducial cut for " << observable << " with " << type_particle.size() << " is not defined." << endline_flush;
	exit(1);
      }
    }
    else if (observable == "dphi"){
      if (type_particle.size() == 2){
	if (index_particle[i_c][0] >= esi->n_object[index_group_selection[0]][i_a] || index_particle[i_c][1] >= esi->n_object[index_group_selection[1]][i_a]){fail_exist_particle = true;}
	else {
	  value_observable = abs(esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]]->phiXX() - esi->event_particle_group[esi->ix_y][i_a][index_group_selection[1]][index_particle[i_c][1]]->phiXX());
	  if (value_observable > pi){value_observable = 2 * pi - value_observable;}
	}
      }
      else {
	logger << LOG_FATAL << "Fiducial cut for " << observable << " with " << type_particle.size() << " is not defined." << endline_flush;
	exit(1);
      }
    }
    else if (observable == "mT"){
      if (type_particle.size() == 1){
	if (index_particle[i_c][0] >= esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]].size()){fail_exist_particle = true;}
	else {value_observable = esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]]->ETXX();}
      }
      else if (type_particle.size() == 2){
	if (index_particle[i_c][0] >= esi->n_object[index_group_selection[0]][i_a] || index_particle[i_c][1] >= esi->n_object[index_group_selection[1]][i_a]){fail_exist_particle = true;}
	else {
	  // transverse mass, as defined for W production (two particles: lepton + neutrino)
	  double temp_dphi = abs(esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]]->phiXX() - esi->event_particle_group[esi->ix_y][i_a][index_group_selection[1]][index_particle[i_c][1]]->phiXX());
	  if (temp_dphi > pi){temp_dphi = 2 * pi - temp_dphi;}
	  value_observable = sqrt(2 * esi->event_particle_group[esi->ix_y][i_a][index_group_selection[0]][index_particle[i_c][0]]->pTXX() * esi->event_particle_group[esi->ix_y][i_a][index_group_selection[1]][index_particle[i_c][1]]->pTXX() * (1. - cos(temp_dphi)));
	}
      }
      else {
	logger << LOG_FATAL << "Fiducial cut for " << observable << " with " << type_particle.size() << " is not defined." << endline_flush;
	exit(1);
      }
    }
    else {
      logger << LOG_FATAL << "Fiducial cut for " << observable << " is not defined." << endline_flush;
      exit(1);
    }
    logger << LOG_DEBUG_VERBOSE << "value_observable = " << value_observable << endline_flush;

    if (fail_exist_particle){logger << LOG_DEBUG_POINT << "Warning: At least one particle to be cut on has not been found!" << endline_flush;}


    // apply cut on observable
    
    if ((type == UPPER && (value_observable > cut_value_upper)) ||
	(type == LOWER && (value_observable < cut_value_lower)) ||
	(type == WINDOW && (value_observable < cut_value_lower || value_observable > cut_value_upper)) ||
	(type == GAP && (value_observable > cut_value_upper && value_observable < cut_value_lower))){
      logger << LOG_DEBUG_POINT << "i_a = " << setw(2) << i_a << "   Cut applied due to " << name << " [" << i_c << "]: value = " << value_observable << endline_flush;
      esi->cut_ps[i_a] = -1;
    }
    else if (value_observable != value_observable){
      logger << LOG_DEBUG_POINT << "i_a = " << setw(2) << i_a << "   Cut applied due to " << name << " [" << i_c << "]: value = " << value_observable << endline_flush;
      esi->cut_ps[i_a] = -1;
    }
    else {
      logger << LOG_DEBUG_POINT << "i_a = " << setw(2) << i_a << "   Cut passed due to " << name << " [" << i_c << "]: value = " << value_observable << endline_flush;
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


string fiducialcut::output_fiducialcut(size_t indent){
  Logger logger("fiducialcut::output_fiducialcut");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream temp_ss;
  string unit = "";
  if (observable == "M" || observable == "pT"){unit = " GeV";}
  temp_ss << left << setw(35) << name << ": ";
  //<< left << setw(30) 
  //<< "   (" << setw(6) << type << ")" << endline;
  for (int i_c = 0; i_c < no_particles.size(); i_c++){
    stringstream full_observable_ss;
    full_observable_ss << observable << " [ ";
    for (int i_p = 0; i_p < type_particle.size(); i_p++){
      full_observable_ss << type_particle[i_p] << " " << no_particles[i_c][i_p] + 1;
      if (i_p < type_particle.size() - 1){full_observable_ss << "   ";}
    }
    full_observable_ss << " ]";

    if (i_c > 0){temp_ss << setw(indent + 40) << "";}
    if (type == UPPER){
      temp_ss << left << setw(6) << "UPPER" << "   " << full_observable_ss.str() << " < " << cut_value_upper << unit << endline;
    }
    if (type == LOWER){
      temp_ss << setw(6) << "LOWER" << "   " << full_observable_ss.str() << " > " << cut_value_lower << unit << endline;
    }
    if (type == WINDOW){
      temp_ss << setw(6) << "WINDOW" << "   " << cut_value_lower << unit << " < " << full_observable_ss.str() << " < " << cut_value_upper << unit << endline;
    }
    if (type == GAP){
      temp_ss << setw(6) << "GAP" << "   " << full_observable_ss.str() << " < " << cut_value_upper << unit << "   OR   " << full_observable_ss.str() << " > " << cut_value_lower << unit << endline;
    }
  }
  return temp_ss.str();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

  
ostream & operator << (ostream &s, const fiducialcut & fc){
  string unit = "";
  if (fc.observable == "M" || fc.observable == "pT"){unit = " GeV";}
  //    s << setw(30) << fc.name << "   " << setw(6) << fc.type << "   ";
  for (int i_c = 0; i_c < fc.no_particles.size(); i_c++){
    stringstream full_observable_ss;
    full_observable_ss << fc.observable << " [ ";
    for (int i_p = 0; i_p < fc.type_particle.size(); i_p++){
      full_observable_ss << fc.type_particle[i_p] << " " << fc.no_particles[i_c][i_p] + 1;
      if (i_p < fc.type_particle.size() - 1){full_observable_ss << "   ";}
    }
    full_observable_ss << " ]";

    if (fc.type == UPPER){
      s << "UPPER    " << full_observable_ss.str() << " < " << fc.cut_value_upper << unit << endline;
    }
    if (fc.type == LOWER){
      s << "LOWER    " << full_observable_ss.str() << " > " << fc.cut_value_lower << unit << endline;
    }
    if (fc.type == WINDOW){
      s << "WINDOW   " << fc.cut_value_lower << unit << " < " << full_observable_ss.str() << " < " << fc.cut_value_upper << unit << endline;
    }
    if (fc.type == GAP){
      s << "GAP      " << full_observable_ss.str() << " < " << fc.cut_value_upper << unit << "   OR   " << full_observable_ss.str() << " > " << fc.cut_value_lower << unit << endline;
    }
  }
  return s;
}





// examples for generation cuts (unseparated lines correspond to versions of the same cut):

// cut on invariant mass of any lepton(lep)--lepton(lep) pair 
//   fiducial_cut = 50. < M lep lep < 150.
//   fiducial_cut = M lep lep > 50. < 150.

// cut on invariant mass of any charged lepton(lm)--anti-lepton(lp) pair
//   fiducial_cut = 50. < M lm lp < 150.

// cut on invariant mass of any electon(em)--positron(ep) pair
//   fiducial_cut = 50. < M em ep < 150.

// cut on invariant mass of any muon(mum)--anti-muon(mup) pair
//   fiducial_cut = 50. < M mum mup < 150.

// cut on invariant mass of hardest-electon(em)--hardest-positron(ep) pair
//   fiducial_cut = 50. < M em (1) ep (1) < 150.

// cut on invariant mass of any 4-lepton system
//   fiducial_cut = 50. < M lep lep lep lep < 150.

// cut on invariant mass of system of 4 hardest leptons
//   fiducial_cut = 50. < M lep (1) lep (2) lep (3) lep (4) < 150.


// cut on transverse momentum of hardest lepton
//   fiducial_cut = pT lep (1) > 20.

// cut on transverse momentum of 2nd-hardest lepton
//   fiducial_cut = pT lep (2) > 15.

// cut on transverse momentum of 3rd-hardest lepton
//   fiducial_cut = pT lep (3) > 10.

// cut on transverse momentum of hardest-electon(em)--hardest-positron(ep) pair (vectorial sum)
//   fiducial_cut = pT em (1) ep (1) > 20.



// cut on transverse momentum of hardest-electon(em)--hardest-positron(ep) pair (vectorial sum)
//   fiducial_cut = pT em (1) ep (1) > 20.


// cut on absolute rapidity of hardest lepton
//   fiducial_cut = |y| lep (1) < 2.5

// cut on absolute rapidity of hardest lepton (cut ouside of window between 1.5 and 2.5)
//   fiducial_cut = 1.5 < |y| lep (1) < 2.5
//   fiducial_cut = |y| lep (1) > 1.5 < 2.5

// cut on absolute rapidity of hardest lepton (cut only gap between 1.37 and 1.52)
//   fiducial_cut = 1.37 > |y| lep (1) > 1.52
//   fiducial_cut = |y| lep (1) < 1.37 > 1.52

// cut on absolute rapidity of hardest-electon(em)--hardest-positron(ep) pair (vectorial sum)
//   fiducial_cut = |y| em (1) ep (1) < 2.5


// cut on absolute pseudo-rapidity of hardest lepton
//   fiducial_cut = |eta| lep (1) < 2.5


// cut on absolute rapidity difference between hardest electon(em) and hardest positron(ep)
//   fiducial_cut = |dy| em (1) ep (1) > 2.5


// cut on absolute pseudo-rapidity difference between hardest electon(em) and hardest positron(ep)
//   fiducial_cut = |deta| em (1) ep (1) > 2.5


// cut von dR (defined via pseudo-rapidity) between any lepton(lep) and any photon(photon)
// fiducial_cut = dReta lep photon > 0.4

// cut von dR (defined via pseudo-rapidity) between hardest lepton(lep) and hardest photon(photon)
// fiducial_cut = dReta lep (1) photon (1) > 0.4


// cut von dR (defined via rapidity) between any lepton(lep) and any photon(photon)
// fiducial_cut = dRy lep photon > 0.4


// cut von dphi (azimuthal angle) between any lepton(lep) and any photon(photon)
// fiducial_cut = dRphi lep photon > 0.1
