#include "header.hpp"

particle::particle(){
  momentum = fourvector(0., 0., 0., 0.);
  pT2 = 0.;
  pT = 0.;
  ET2 = 0.;
  ET = 0.;
  m2 = 0.;
  m = 0.;
  eta = 0.;
  rapidity = 0.;
  phi = 0.;
}

particle::~particle(){}

particle::particle(fourvector _momentum){
  momentum = _momentum;
  reset_all();
  //  calculate_all();
}

particle::particle(fourvector _momentum, double _M){
  momentum = _momentum;
  is_available_m = true;
  m = _M;
  is_available_m2 = true;
  m2 = pow(m, 2);
  reset_all_but_m();
  //  calculate_all();
}

particle::particle(fourvector _momentum, double _M, double _M2){
  momentum = momentum;
  is_available_m = true;
  m = _M;
  is_available_m2 = true;
  m2 = _M2;
  reset_all_but_m();
  //  calculate_all();
}

void particle::calculate_all(){
  pT2 = pT2XX();
  pT = pTXX();
  ET2 = ET2XX();
  ET = ETXX();
  m2 = m2XX();
  m = mXX();
  eta = etaXX();
  rapidity = rapidityXX();
  phi = phiXX();
}

void particle::reset_all_but_m(){
  is_available_pT = false;
  is_available_pT2 = false;
  is_available_ET = false;
  is_available_ET2 = false;
  is_available_eta = false;
  is_available_rapidity = false;
  is_available_phi = false;
}

void particle::reset_all(){
  is_available_pT = false;
  is_available_pT2 = false;
  is_available_ET = false;
  is_available_ET2 = false;
  is_available_eta = false;
  is_available_rapidity = false;
  is_available_phi = false;
  is_available_m = false;
  is_available_m2 = false;
}

double particle::cosphi() {return cos(phiXX());}
double particle::sinphi() {return sin(phiXX());}

particle particle::operator + (const particle & p2) const{
  particle psum;
  psum.momentum = momentum + p2.momentum;
  psum.reset_all();
  //  psum.calculate_all();
  return psum;
}

particle & particle::operator += (const particle & p2){
  momentum = momentum + p2.momentum;
  reset_all();
  //  calculate_all();
  return *this;
}

double R2_eta(particle & p1, particle & p2){
  double delta_phi = abs(p1.phiXX() - p2.phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return pow(p1.etaXX() - p2.etaXX(), 2) + pow(delta_phi, 2);
}

double R2_rapidity(particle & p1, particle & p2){
  double delta_phi = abs(p1.phiXX() - p2.phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return pow(p1.rapidityXX() - p2.rapidityXX(), 2) + pow(delta_phi, 2);
}

double R2_cosheta_cosphi(particle & p1, particle & p2){
  double delta_phi = abs(p1.phiXX() - p2.phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return 2 * (cosh(p1.etaXX() - p2.etaXX()) - cos(delta_phi));
}

double R2_coshrapidity_cosphi(particle & p1, particle & p2){
  double delta_phi = abs(p1.phiXX() - p2.phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return 2 * (cosh(p1.rapidityXX() - p2.rapidityXX()) - cos(delta_phi));
}


double R2_eta(particle & p1, const fourvector & fv2){
  double delta_phi = abs(p1.phiXX() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return pow(p1.etaXX() - fv2.eta(), 2) + pow(delta_phi, 2);
}

double R2_rapidity(particle & p1, const fourvector & fv2){
  double delta_phi = abs(p1.phiXX() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return pow(p1.rapidityXX() - fv2.rapidity(), 2) + pow(delta_phi, 2);
}

double R2_cosheta_cosphi(particle & p1, const fourvector & fv2){
  double delta_phi = abs(p1.phiXX() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return 2 * (cosh(p1.etaXX() - fv2.eta()) - cos(delta_phi));
}

double R2_coshrapidity_cosphi(particle & p1, const fourvector & fv2){
  double delta_phi = abs(p1.phiXX() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return 2 * (cosh(p1.rapidityXX() - fv2.rapidity()) - cos(delta_phi));
}


double R_eta(particle & p1, particle & p2){
  double delta_phi = abs(p1.phiXX() - p2.phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return sqrt(pow(p1.etaXX() - p2.etaXX(), 2) + pow(delta_phi, 2));
}

double R_rapidity(particle & p1, particle & p2){
  double delta_phi = abs(p1.phiXX() - p2.phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return sqrt(pow(p1.rapidityXX() - p2.rapidityXX(), 2) + pow(delta_phi, 2));
}

double R_eta(particle & p1, const fourvector & fv2){
  double delta_phi = abs(p1.phiXX() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return sqrt(pow(p1.etaXX() - fv2.eta(), 2) + pow(delta_phi, 2));
}

double R_rapidity(particle & p1, const fourvector & fv2){
  double delta_phi = abs(p1.phiXX() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return sqrt(pow(p1.rapidityXX() - fv2.rapidity(), 2) + pow(delta_phi, 2));
}


double R2_eta(particle * p1, particle * p2){
  double delta_phi = abs(p1->phiXX() - p2->phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return pow(p1->etaXX() - p2->etaXX(), 2) + pow(delta_phi, 2);
}

double R2_rapidity(particle * p1, particle * p2){
  double delta_phi = abs(p1->phiXX() - p2->phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return pow(p1->rapidityXX() - p2->rapidityXX(), 2) + pow(delta_phi, 2);
}

double R2_cosheta_cosphi(particle * p1, particle * p2){
  double delta_phi = abs(p1->phiXX() - p2->phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return 2 * (cosh(p1->etaXX() - p2->etaXX()) - cos(delta_phi));
}

double R2_coshrapidity_cosphi(particle * p1, particle * p2){
  double delta_phi = abs(p1->phiXX() - p2->phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return 2 * (cosh(p1->rapidityXX() - p2->rapidityXX()) - cos(delta_phi));
}


double R_eta(particle * p1, particle * p2){
  double delta_phi = abs(p1->phiXX() - p2->phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return sqrt(pow(p1->etaXX() - p2->etaXX(), 2) + pow(delta_phi, 2));
}

double R_rapidity(particle * p1, particle * p2){
  double delta_phi = abs(p1->phiXX() - p2->phiXX());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  return sqrt(pow(p1->rapidityXX() - p2->rapidityXX(), 2) + pow(delta_phi, 2));
}

/*
double particle::pTXX(){
  if (!is_available_pT){
    pT = sqrt(pT2XX());
    is_available_pT = true;
  }
  return pT;
}
double particle::pT2XX(){
  if (!is_available_pT2){
    pT2 = momentum.pT2();
    is_available_pT2 = true;
  }
  return pT2;
}
double particle::ETXX(){
  if (!is_available_ET){
    ET = sqrt(ET2XX());
    is_available_ET = true;
  }
  return ET;
}
double particle::ET2XX(){
  if (!is_available_ET2){
    ET2 = momentum.ET2();
    is_available_ET2 = true;
  }
  return ET2;
}
double particle::mXX(){
  if (!is_available_m){
    m = sqrt(m2XX());
    is_available_m = true;
  }
  return m;
}
double particle::m2XX(){
  if (!is_available_m2){
    m2 = momentum.m2();
    is_available_m2 = true;
  }
  return m2;
}
double particle::etaXX(){
  if (pT2XX() == 0.){eta = 0.;}
  else {eta = momentum.eta();}
  is_available_eta = true;
  return eta;
}
double particle::rapidityXX(){
  if (m2XX() == 0. && pT2XX() == 0.){rapidity = 0.;}
  else {rapidity = momentum.rapidity();}
  is_available_rapidity = true;
  return rapidity;
}
double particle::phiXX(){
  if (pT2XX() == 0.){phi = 0.;} // better 0 than undefined for now
  else {phi = momentum.phi();}
  is_available_phi = true;
  return phi;
}
*/

ostream & operator << (ostream & s, particle & p){
  s << p.momentum << endline << setw(10) << "" << "m = " << setw(21) << setprecision(15) << sqrt(abs(p.m2XX())) << endline << "   pT = " << setw(21) << setprecision(15) << p.pTXX() << "  ET = " << setw(21) << setprecision(15) << p.ETXX() << "  eta = " << setw(21) << setprecision(15) << p.etaXX() << "  y = " << setw(21) << setprecision(15) << p.rapidityXX() << "  phi = " << setw(21) << setprecision(15) << p.phiXX();
  return s;
}


bool lessBypT(particle & p1, particle & p2){return p1.pTXX() < p2.pTXX();}
bool greaterBypT(particle & p1, particle & p2){return p1.pTXX() > p2.pTXX();}

bool XlessBypT(particle * p1, particle * p2){return p1->pTXX() < p2->pTXX();}
bool XgreaterBypT(particle * p1, particle * p2){return p1->pTXX() > p2->pTXX();}
/*
bool idgreaterBypT(particle_id & p1, particle_id & p2){return (*p1.xparticle).pTXX() > (*p2.xparticle).pTXX();}
bool idgreaterByET(particle_id & p1, particle_id & p2){return (*p1.xparticle).ETXX() > (*p2.xparticle).ETXX();}
*/

