#include "header.hpp"

particle_history::particle_history(){}

particle_history::~particle_history(){}

particle_history::particle_history(fourvector _momentum, int i_a, int i_p, event_set * _esi) : particle(_momentum){
  esi = _esi;
  int temp_type = esi->csi->type_parton[i_a][i_p];
  if (abs(esi->csi->type_parton[i_a][i_p]) <= esi->msi->N_f){temp_type = 0;}
  parton_origin = vector<int> (1, i_p);
  flavour = vector<int> (1, temp_type);
}

particle_history::particle_history(fourvector _momentum, double _M, int i_a, int i_p, event_set * _esi) : particle(_momentum, _M){
  esi = _esi;
  int temp_type = esi->csi->type_parton[i_a][i_p];
  if (abs(esi->csi->type_parton[i_a][i_p]) <= esi->msi->N_f){temp_type = 0;}
  parton_origin = vector<int> (1, i_p);
  flavour = vector<int> (1, temp_type);
}

particle_history::particle_history(fourvector _momentum, double _M, double _M2, int i_a, int i_p, event_set * _esi) : particle(momentum, _M, _M2){
  esi = _esi;
  int temp_type = esi->csi->type_parton[i_a][i_p];
  if (abs(esi->csi->type_parton[i_a][i_p]) <= esi->msi->N_f){temp_type = 0;}
  parton_origin = vector<int> (1, i_p);
  flavour = vector<int> (1, temp_type);
}


// Requires an improved implementation !!!
void particle_history::determination_identity(){
  static Logger logger("particle_history::determination_identity");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  identity = accumulate(flavour.begin(), flavour.end(), 0);
  if (esi->msi->M_b > 0. && identity == 0){
    // no IR-safe defintion in case of massless b quarks - b and bx are combined to bottom jet: switch needed !!!
    for (int i_q = 0; i_q < flavour.size(); i_q++){if (abs(flavour[i_q]) == 5){identity = 5; break;}}
  }

  if (identity == 0){} // light jet  or  b-bx jet
  else if (identity == 5 || identity == -5){} // bjet_b  or  bjet_bx
  else {
    double E_phot = 0.;
    double E_had = 0.;
    int count_phot = 0;
    for (int i_l = 0; i_l < flavour.size(); i_l++){
      if (flavour[i_l] == 22){
	E_phot += (esi->event_particle_group[esi->ix_y][esi->ix_a][0][parton_origin[i_l]]->momentum).x0();
	count_phot++;
      }
      else {
	E_had += (esi->event_particle_group[esi->ix_y][esi->ix_a][0][parton_origin[i_l]]->momentum).x0();
      }
    }

    if (esi->switch_output_cutinfo){esi->info_cut << "E_phot = " << E_phot << "   E_had = " << E_had << endline;}

    if (E_phot / (E_phot + E_had) >= esi->photon_E_threshold_ratio){identity = 22;}
    else {identity = identity - count_phot * 22;}
    logger << LOG_DEBUG_VERBOSE << "E_phot = " << E_phot << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "E_had = " << E_had << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "identity = " << identity << endline_flush;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


particle_history particle_history::operator + (const particle_history & p2) const{
  particle_history psum;
  psum.momentum = momentum + p2.momentum;
  psum.reset_all();
  //  psum.calculate_all();
  for (int i_i = 0; i_i < p2.parton_origin.size(); i_i++){psum.parton_origin.push_back(p2.parton_origin[i_i]);}
  for (int i_i = 0; i_i < p2.flavour.size(); i_i++){psum.flavour.push_back(p2.flavour[i_i]);}
  return psum;
}


particle_history & particle_history::operator += (const particle_history & p2) {
  momentum = momentum + p2.momentum;
  reset_all();
  //  calculate_all();
  for (int i_i = 0; i_i < p2.parton_origin.size(); i_i++){parton_origin.push_back(p2.parton_origin[i_i]);}
  for (int i_i = 0; i_i < p2.flavour.size(); i_i++){flavour.push_back(p2.flavour[i_i]);}
  return *this;
}

