#include "header.hpp"

void observable_set::initialization_LHAPDF(){
  static Logger logger("initialization_LHAPDF");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "LHAPDFname = " << LHAPDFname << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "LHAPDFsubset = " << LHAPDFsubset << endline_flush;

  const int SUBSET = LHAPDFsubset;

  logger << LOG_DEBUG_VERBOSE << "LHAPDFname.length() = " << LHAPDFname.length() << endline_flush;
  if (LHAPDFname.length()>=6) {
    if (LHAPDFname.substr(LHAPDFname.size() - 6, 6) == ".LHpdf"){
      const string NAME = LHAPDFname.substr(0, LHAPDFname.size() - 6);
      LHAPDF::initPDFSet(NAME, LHAPDF::LHPDF, SUBSET);
      logger << LOG_DEBUG_VERBOSE << "initialization done for .LHpdf set" << endline_flush;
      return;
    }
  }
  if (LHAPDFname.length()>=7) {
    if (LHAPDFname.substr(LHAPDFname.size() - 7, 7) == ".LHgrid"){
      const string NAME = LHAPDFname.substr(0, LHAPDFname.size() - 7);
      LHAPDF::initPDFSet(NAME, LHAPDF::LHGRID, SUBSET);
      logger << LOG_DEBUG_VERBOSE << "initialization done for .LHgrid set" << endline_flush;
      return;
    }
  }

  const string NAME = LHAPDFname;
  LHAPDF::initPDFSet(NAME, LHAPDF::LHGRID, SUBSET);
  logger << LOG_DEBUG_VERBOSE << "LHAPDF::initPDFSet(NAME = " << NAME << ", LHAPDF::LHGRID, SUBSET = " << SUBSET << ");" << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::initialization_pdf(vector<string> & input_pdf, vector<string> & list_pdf){
  static Logger logger("initialization_pdf");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_d = 0; i_d < input_pdf.size(); i_d++){
    string temp_split = "";
    for (int i_s = 0; i_s < input_pdf[i_d].size(); i_s++){
      if (input_pdf[i_d][i_s] == ' '){continue;}
      else if (input_pdf[i_d][i_s] == '+'){
	if (temp_split != ""){list_pdf.push_back(temp_split); temp_split = "";} // counter++;
      }
      else {temp_split.push_back(input_pdf[i_d][i_s]);}
    }
    if (temp_split != ""){list_pdf.push_back(temp_split);}
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::initialization_pdf_pair(vector<string> & list_pdf, vector<vector<string> > & list_pdf_pair){
  static Logger logger("initialization_pdf_pair");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

 for (int i_d = 0; i_d < list_pdf.size(); i_d++){
    logger << LOG_DEBUG << "list_pdf[" << i_d << "] = " << list_pdf[i_d] << endline_flush;
  }

  list_pdf_pair.resize(list_pdf.size(), vector<string> (2));

  for (int i_p = 0; i_p < list_pdf.size(); i_p++){
    int start = 1;
    for (int i_s = list_pdf[i_p].size() - 1; i_s >= 0; i_s--){
      if (i_s > 0 && start == -1){
	logger << LOG_FATAL << "Wrong input: pdf cannot be " << list_pdf[i_p] << " !" << endline_flush;
	exit(1);
      }
      if (list_pdf[i_p][i_s] == 'x' && i_s > 0){
	list_pdf_pair[i_p][start] = list_pdf[i_p].substr(i_s - 1, 2);
	logger << LOG_DEBUG << "list_pdf_pair[i_p = " << i_p << "][start = " << start << "] = " << list_pdf_pair[i_p][start] << endline_flush;
	start--;
	i_s--;
      }
      else {
	list_pdf_pair[i_p][start] = list_pdf[i_p].substr(i_s, 1);
	logger << LOG_DEBUG << "list_pdf_pair[i_p = " << i_p << "][start = " << start << "] = " << list_pdf_pair[i_p][start] << endline_flush;
	start--;
      }
    }
  }

  for (int i_p = 0; i_p < list_pdf_pair.size(); i_p++){
    stringstream temp;
    temp.str("");
    temp << setw(26) << "list_pdf_pair[" << setw(2) << i_p << "] = ";
    for (int i_s = 0; i_s < list_pdf_pair[i_p].size(); i_s++){
      temp << setw(4) << list_pdf_pair[i_p][i_s];
    }
    logger << LOG_DEBUG << temp.str() << endline_flush;
  }
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::determine_selection_content_pdf(){
  static Logger logger("observable_set::determine_selection_content_pdf");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "N_f_active = " << N_f_active << endline_flush;
  for (int i_p = 0; i_p < pdf_selection.size(); i_p++){
    logger << LOG_DEBUG << "pdf_selection[" << i_p << "] = " << pdf_selection[i_p] << endline_flush;
  }
  for (int i_p = 0; i_p < pdf_restriction.size(); i_p++){
    logger << LOG_DEBUG << "pdf_restriction[" << i_p << "] = " << pdf_restriction[i_p] << endline_flush;
  }
  for (int i_p = 0; i_p < pdf_exclusion.size(); i_p++){
    logger << LOG_DEBUG << "pdf_exclusion[" << i_p << "] = " << pdf_exclusion[i_p] << endline_flush;
  }
  for (int i_d = 0; i_d < pdf_disable.size(); i_d++){
    logger << LOG_DEBUG << "pdf_disable[" << i_d << "] = " << pdf_disable[i_d] << endline_flush;
  }
  vector<int> content_p(14);
  content_p[0] = 1;
  content_p[1] = 2;
  content_p[2] = 3;
  content_p[3] = 4;
  content_p[4] = 5;
  content_p[5] = 6;
  content_p[6] = -1;
  content_p[7] = -2;
  content_p[8] = -3;
  content_p[9] = -4;
  content_p[10] = -5;
  content_p[11] = -6;
  content_p[12] = 0;
  content_p[13] = 7;
  vector<int> content_Q(12);
  content_Q[0] = 1;
  content_Q[1] = 2;
  content_Q[2] = 3;
  content_Q[3] = 4;
  content_Q[4] = 5;
  content_Q[5] = 6;
  content_Q[6] = -1;
  content_Q[7] = -2;
  content_Q[8] = -3;
  content_Q[9] = -4;
  content_Q[10] = -5;
  content_Q[11] = -6;
  vector<int> content_q(6);
  content_q[0] = 1;
  content_q[1] = 2;
  content_q[2] = 3;
  content_q[3] = 4;
  content_q[4] = 5;
  content_q[5] = 6;
  vector<int> content_qx(6);
  content_qx[0] = -1;
  content_qx[1] = -2;
  content_qx[2] = -3;
  content_qx[3] = -4;
  content_qx[4] = -5;
  content_qx[5] = -6;
  vector<int> content_g(1);
  content_g[0] = 0;
  vector<int> content_a(1);
  content_a[0] = 7;
  vector<int> content_D(2);
  content_D[0] = 1;
  content_D[1] = -1;
  vector<int> content_U(2);
  content_U[0] = 2;
  content_U[1] = -2;
  vector<int> content_S(2);
  content_S[0] = 3;
  content_S[1] = -3;
  vector<int> content_C(2);
  content_C[0] = 4;
  content_C[1] = -4;
  vector<int> content_B(2);
  content_B[0] = 5;
  content_B[1] = -5;
  vector<int> content_T(2);
  content_T[0] = 6;
  content_T[1] = -6;
  vector<int> content_d(1);
  content_d[0] = 1;
  vector<int> content_u(1);
  content_u[0] = 2;
  vector<int> content_s(1);
  content_s[0] = 3;
  vector<int> content_c(1);
  content_c[0] = 4;
  vector<int> content_b(1);
  content_b[0] = 5;
  vector<int> content_t(1);
  content_t[0] = 6;
  vector<int> content_dx(1);
  content_dx[0] = -1;
  vector<int> content_ux(1);
  content_ux[0] = -2;
  vector<int> content_sx(1);
  content_sx[0] = -3;
  vector<int> content_cx(1);
  content_cx[0] = -4;
  vector<int> content_bx(1);
  content_bx[0] = -5;
  vector<int> content_tx(1);
  content_tx[0] = -6;

  vector<vector<int> > content_parton(24);
  content_parton[0] = content_p;
  content_parton[1] = content_Q;
  content_parton[2] = content_q;
  content_parton[3] = content_qx;
  content_parton[4] = content_g;
  content_parton[5] = content_a;
  content_parton[6] = content_D;
  content_parton[7] = content_U;
  content_parton[8] = content_S;
  content_parton[9] = content_C;
  content_parton[10] = content_B;
  content_parton[11] = content_T;
  content_parton[12] = content_d;
  content_parton[13] = content_u;
  content_parton[14] = content_s;
  content_parton[15] = content_c;
  content_parton[16] = content_b;
  content_parton[17] = content_t;
  content_parton[18] = content_dx;
  content_parton[19] = content_ux;
  content_parton[20] = content_sx;
  content_parton[21] = content_cx;
  content_parton[22] = content_bx;
  content_parton[23] = content_tx;


  // pdf_disable is a vector of all disabled particles, which may contain "+"'s that are resolved here
  vector<string> temp_pdf_disable;
  for (int i_d = 0; i_d < pdf_disable.size(); i_d++){
    string temp_split = "";
    for (int i_s = 0; i_s < pdf_disable[i_d].size(); i_s++){
      if (pdf_disable[i_d][i_s] == ' '){continue;}
      else if (pdf_disable[i_d][i_s] == '+'){
	if (temp_split != ""){temp_pdf_disable.push_back(temp_split); temp_split = "";} // counter++;
      }
      else {temp_split.push_back(pdf_disable[i_d][i_s]);}
    }
    if (temp_split != ""){temp_pdf_disable.push_back(temp_split);}
  }

  // according to N_f_active, further particles are added to the list of temp_pdf_disable:
  if (N_f_active < 6){temp_pdf_disable.push_back("top");}
  if (N_f_active < 5){temp_pdf_disable.push_back("bottom");}
  if (N_f_active < 4){temp_pdf_disable.push_back("charm");}
  if (N_f_active < 3){temp_pdf_disable.push_back("strange");}
  if (N_f_active < 2){temp_pdf_disable.push_back("up");}
  if (N_f_active < 1){temp_pdf_disable.push_back("down");}

  // according to chosen PDF set, photons are added to the list of temp_pdf_disable:
  if (!(LHAPDF::hasPhoton())){temp_pdf_disable.push_back("photon");}

  logger << LOG_DEBUG << "temp_pdf_disable.size() = " << temp_pdf_disable.size() << endline_flush;
  for (int i_d = 0; i_d < temp_pdf_disable.size(); i_d++){
    logger << LOG_DEBUG << "temp_pdf_disable[" << i_d << "] = " << temp_pdf_disable[i_d] << endline_flush;
  }


  for (int i_d = 0; i_d < temp_pdf_disable.size(); i_d++){
    if (temp_pdf_disable[i_d] == "Q" ||
	temp_pdf_disable[i_d] == "q" ||
	temp_pdf_disable[i_d] == "qx" ||
	temp_pdf_disable[i_d] == "g" ||
	temp_pdf_disable[i_d] == "a" ||
	temp_pdf_disable[i_d] == "D" ||
	temp_pdf_disable[i_d] == "U" ||
	temp_pdf_disable[i_d] == "S" ||
	temp_pdf_disable[i_d] == "C" ||
	temp_pdf_disable[i_d] == "B" ||
	temp_pdf_disable[i_d] == "T" ||
	temp_pdf_disable[i_d] == "d" ||
	temp_pdf_disable[i_d] == "u" ||
	temp_pdf_disable[i_d] == "s" ||
	temp_pdf_disable[i_d] == "c" ||
	temp_pdf_disable[i_d] == "b" ||
	temp_pdf_disable[i_d] == "t" ||
	temp_pdf_disable[i_d] == "dx" ||
	temp_pdf_disable[i_d] == "ux" ||
	temp_pdf_disable[i_d] == "sx" ||
	temp_pdf_disable[i_d] == "cx" ||
	temp_pdf_disable[i_d] == "bx" ||
	temp_pdf_disable[i_d] == "tx"){
      for (int i_c = 0; i_c < content_parton.size(); i_c++){
	for (int i_p = 0; i_p < content_parton[i_c].size(); i_p++){
	  if ((content_parton[i_c][i_p] == 0 && temp_pdf_disable[i_d] == "g") ||
	      (content_parton[i_c][i_p] == 7 && temp_pdf_disable[i_d] == "a") ||
	      (content_parton[i_c][i_p] == 1 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "q" || temp_pdf_disable[i_d] == "d" || temp_pdf_disable[i_d] == "D")) ||
	      (content_parton[i_c][i_p] == 2 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "q" || temp_pdf_disable[i_d] == "u" || temp_pdf_disable[i_d] == "U")) ||
	      (content_parton[i_c][i_p] == 3 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "q" || temp_pdf_disable[i_d] == "s" || temp_pdf_disable[i_d] == "S")) ||
	      (content_parton[i_c][i_p] == 4 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "q" || temp_pdf_disable[i_d] == "c" || temp_pdf_disable[i_d] == "C")) ||
	      (content_parton[i_c][i_p] == 5 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "q" || temp_pdf_disable[i_d] == "b" || temp_pdf_disable[i_d] == "B")) ||
	      (content_parton[i_c][i_p] == 6 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "q" || temp_pdf_disable[i_d] == "t" || temp_pdf_disable[i_d] == "T")) ||
	      (content_parton[i_c][i_p] == -1 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "qx" || temp_pdf_disable[i_d] == "dx" || temp_pdf_disable[i_d] == "D")) ||
	      (content_parton[i_c][i_p] == -2 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "qx" || temp_pdf_disable[i_d] == "ux" || temp_pdf_disable[i_d] == "U")) ||
	      (content_parton[i_c][i_p] == -3 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "qx" || temp_pdf_disable[i_d] == "sx" || temp_pdf_disable[i_d] == "S")) ||
	      (content_parton[i_c][i_p] == -4 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "qx" || temp_pdf_disable[i_d] == "cx" || temp_pdf_disable[i_d] == "C")) ||
	      (content_parton[i_c][i_p] == -5 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "qx" || temp_pdf_disable[i_d] == "bx" || temp_pdf_disable[i_d] == "B")) ||
	      (content_parton[i_c][i_p] == -6 && (temp_pdf_disable[i_d] == "Q" || temp_pdf_disable[i_d] == "qx" || temp_pdf_disable[i_d] == "tx" || temp_pdf_disable[i_d] == "T"))){
	    content_parton[i_c].erase(content_parton[i_c].begin() + i_p);
	    i_p--;
	  }
	}
      }
    }
    else if (temp_pdf_disable[i_d] == "quark" ||
	temp_pdf_disable[i_d] == "antiquark" ||
	temp_pdf_disable[i_d] == "gluon" ||
	temp_pdf_disable[i_d] == "photon" ||
	temp_pdf_disable[i_d] == "down" ||
	temp_pdf_disable[i_d] == "up" ||
	temp_pdf_disable[i_d] == "strange" ||
	temp_pdf_disable[i_d] == "charm" ||
	temp_pdf_disable[i_d] == "bottom" ||
	temp_pdf_disable[i_d] == "top" ||
	temp_pdf_disable[i_d] == "down-quark" ||
	temp_pdf_disable[i_d] == "up-quark" ||
	temp_pdf_disable[i_d] == "strange-quark" ||
	temp_pdf_disable[i_d] == "charm-quark" ||
	temp_pdf_disable[i_d] == "bottom-quark" ||
	temp_pdf_disable[i_d] == "top-quark" ||
	temp_pdf_disable[i_d] == "down-antiquark" ||
	temp_pdf_disable[i_d] == "up-antiquark" ||
	temp_pdf_disable[i_d] == "strange-antiquark" ||
	temp_pdf_disable[i_d] == "charm-antiquark" ||
	temp_pdf_disable[i_d] == "bottom-antiquark" ||
	temp_pdf_disable[i_d] == "top-antiquark"){
      for (int i_c = 0; i_c < content_parton.size(); i_c++){
	for (int i_p = 0; i_p < content_parton[i_c].size(); i_p++){
	  if ((content_parton[i_c][i_p] == 0 && temp_pdf_disable[i_d] == "gluon") ||
	      (content_parton[i_c][i_p] == 7 && temp_pdf_disable[i_d] == "photon") ||
	      (content_parton[i_c][i_p] == 1 && (temp_pdf_disable[i_d] == "quark" || temp_pdf_disable[i_d] == "down" || temp_pdf_disable[i_d] == "down-quark")) ||
	      (content_parton[i_c][i_p] == 2 && (temp_pdf_disable[i_d] == "quark" || temp_pdf_disable[i_d] == "up" || temp_pdf_disable[i_d] == "up-quark")) ||
	      (content_parton[i_c][i_p] == 3 && (temp_pdf_disable[i_d] == "quark" || temp_pdf_disable[i_d] == "strange" || temp_pdf_disable[i_d] == "strange-quark")) ||
	      (content_parton[i_c][i_p] == 4 && (temp_pdf_disable[i_d] == "quark" || temp_pdf_disable[i_d] == "charm" || temp_pdf_disable[i_d] == "charm-quark")) ||
	      (content_parton[i_c][i_p] == 5 && (temp_pdf_disable[i_d] == "quark" || temp_pdf_disable[i_d] == "bottom" || temp_pdf_disable[i_d] == "bottom-quark")) ||
	      (content_parton[i_c][i_p] == 6 && (temp_pdf_disable[i_d] == "quark" || temp_pdf_disable[i_d] == "top" || temp_pdf_disable[i_d] == "top-quark")) ||
	      (content_parton[i_c][i_p] == -1 && (temp_pdf_disable[i_d] == "antiquark" || temp_pdf_disable[i_d] == "down" || temp_pdf_disable[i_d] == "down-antiquark")) ||
	      (content_parton[i_c][i_p] == -2 && (temp_pdf_disable[i_d] == "antiquark" || temp_pdf_disable[i_d] == "up" || temp_pdf_disable[i_d] == "up-antiquark")) ||
	      (content_parton[i_c][i_p] == -3 && (temp_pdf_disable[i_d] == "antiquark" || temp_pdf_disable[i_d] == "strange" || temp_pdf_disable[i_d] == "strange-antiquark")) ||
	      (content_parton[i_c][i_p] == -4 && (temp_pdf_disable[i_d] == "antiquark" || temp_pdf_disable[i_d] == "charm" || temp_pdf_disable[i_d] == "charm-antiquark")) ||
	      (content_parton[i_c][i_p] == -5 && (temp_pdf_disable[i_d] == "antiquark" || temp_pdf_disable[i_d] == "bottom" || temp_pdf_disable[i_d] == "bottom-antiquark")) ||
	      (content_parton[i_c][i_p] == -6 && (temp_pdf_disable[i_d] == "antiquark" || temp_pdf_disable[i_d] == "top" || temp_pdf_disable[i_d] == "top-antiquark"))){
	    content_parton[i_c].erase(content_parton[i_c].begin() + i_p);
	    i_p--;
	  }
	}
      }
    }
    else {
      logger << LOG_FATAL << "Wrong input: pdf_disable cannot be " << temp_pdf_disable[i_d] << " !" << endline_flush;
      exit(1);
    }
  }

  for (int i_c = 0; i_c < content_parton.size(); i_c++){
    stringstream temp;
    temp.str("");
    temp << setw(16) << "content_parton[" << setw(2) << i_c << "] = ";
    for (int i_p = 0; i_p < content_parton[i_c].size(); i_p++){
      temp << setw(4) << content_parton[i_c][i_p];
    }
    logger << LOG_DEBUG << temp.str() << endline_flush;
  }

  map<string, vector<int> > relation_content_parton;
  relation_content_parton["p"] = content_parton[0];
  relation_content_parton["Q"] = content_parton[1];
  relation_content_parton["q"] = content_parton[2];
  relation_content_parton["qx"] = content_parton[3];
  relation_content_parton["g"] = content_parton[4];
  relation_content_parton["a"] = content_parton[5];
  relation_content_parton["D"] = content_parton[6];
  relation_content_parton["U"] = content_parton[7];
  relation_content_parton["S"] = content_parton[8];
  relation_content_parton["C"] = content_parton[9];
  relation_content_parton["B"] = content_parton[10];
  relation_content_parton["T"] = content_parton[11];
  relation_content_parton["d"] = content_parton[12];
  relation_content_parton["u"] = content_parton[13];
  relation_content_parton["s"] = content_parton[14];
  relation_content_parton["c"] = content_parton[15];
  relation_content_parton["b"] = content_parton[16];
  relation_content_parton["t"] = content_parton[17];
  relation_content_parton["dx"] = content_parton[18];
  relation_content_parton["ux"] = content_parton[19];
  relation_content_parton["sx"] = content_parton[20];
  relation_content_parton["cx"] = content_parton[21];
  relation_content_parton["bx"] = content_parton[22];
  relation_content_parton["tx"] = content_parton[23];


  //  pdf_selection
  vector<string> temp_pdf_selection;
  vector<vector<string> > temp_pdf_selection_pair;
  vector<string> temp_pdf_restriction;
  vector<vector<string> > temp_pdf_restriction_pair;
  vector<string> temp_pdf_exclusion;
  vector<vector<string> > temp_pdf_exclusion_pair;

  initialization_pdf(pdf_selection, temp_pdf_selection);
  initialization_pdf(pdf_restriction, temp_pdf_restriction);
  initialization_pdf(pdf_exclusion, temp_pdf_exclusion);

  if (temp_pdf_selection.size() == 0){temp_pdf_selection.push_back("pp");}
  if (temp_pdf_restriction.size() == 0){temp_pdf_restriction.push_back("pp");}

  for (int i_p = 0; i_p < temp_pdf_selection.size(); i_p++){
    logger << LOG_DEBUG << "temp_pdf_selection[i_p = " << i_p << "] = " << temp_pdf_selection[i_p] << endline_flush;
  }
  for (int i_p = 0; i_p < temp_pdf_restriction.size(); i_p++){
    logger << LOG_DEBUG << "temp_pdf_restriction[i_p = " << i_p << "] = " << temp_pdf_restriction[i_p] << endline_flush;
  }
  for (int i_p = 0; i_p < temp_pdf_exclusion.size(); i_p++){
    logger << LOG_DEBUG << "temp_pdf_exclusion[i_p = " << i_p << "] = " << temp_pdf_exclusion[i_p] << endline_flush;
  }
  
  initialization_pdf_pair(temp_pdf_selection, temp_pdf_selection_pair);
  initialization_pdf_pair(temp_pdf_restriction, temp_pdf_restriction_pair);
  initialization_pdf_pair(temp_pdf_exclusion, temp_pdf_exclusion_pair);

  for (int i_p = 0; i_p < temp_pdf_selection_pair.size(); i_p++){
    stringstream temp_ss;
    for (int i_x = 0; i_x < temp_pdf_selection_pair[i_p].size(); i_x++){
      temp_ss << temp_pdf_selection_pair[i_p][i_x] << " ";
    }
    logger << LOG_DEBUG << "temp_pdf_selection_pair[i_p = " << i_p << "] = " << temp_ss.str() << endline_flush;
  }
  for (int i_p = 0; i_p < temp_pdf_restriction_pair.size(); i_p++){
    stringstream temp_ss;
    for (int i_x = 0; i_x < temp_pdf_restriction_pair[i_p].size(); i_x++){
      temp_ss << temp_pdf_restriction_pair[i_p][i_x] << " ";
    }
    logger << LOG_DEBUG << "temp_pdf_restriction_pair[i_p = " << i_p << "] = " << temp_ss.str() << endline_flush;
  }
  for (int i_p = 0; i_p < temp_pdf_exclusion_pair.size(); i_p++){
    stringstream temp_ss;
    for (int i_x = 0; i_x < temp_pdf_exclusion_pair[i_p].size(); i_x++){
      temp_ss << temp_pdf_exclusion_pair[i_p][i_x] << " ";
    }
    logger << LOG_DEBUG << "temp_pdf_exclusion_pair[i_p = " << i_p << "] = " << temp_ss.str() << endline_flush;
  }


  allowed_all_pdf.resize(2);
  vector<int> temp_all_pdf(3);
  int j_x;
  for (int i_x = 0; i_x < 2; i_x++){
    //  for (int i_x = 1; i_x >= -1; i_x -= 2){
    if (i_x == 0){j_x = 1;}
    else if (i_x == 1){j_x = -1;}
    temp_all_pdf[0] = j_x;
    for (int i_p = 0; i_p < temp_pdf_selection_pair.size(); i_p++){
      for (int i_i = 0; i_i < relation_content_parton[temp_pdf_selection_pair[i_p][0]].size(); i_i++){
	for (int i_j = 0; i_j < relation_content_parton[temp_pdf_selection_pair[i_p][1]].size(); i_j++){

	  int flag_selection = false;
	  for (int j_p = 0; j_p < temp_pdf_restriction_pair.size(); j_p++){
	    for (int j_i = 0; j_i < relation_content_parton[temp_pdf_restriction_pair[j_p][0]].size(); j_i++){
	      for (int j_j = 0; j_j < relation_content_parton[temp_pdf_restriction_pair[j_p][1]].size(); j_j++){
		if (relation_content_parton[temp_pdf_selection_pair[i_p][0]][i_i] == relation_content_parton[temp_pdf_restriction_pair[j_p][0]][j_i] &&
		    relation_content_parton[temp_pdf_selection_pair[i_p][1]][i_j] == relation_content_parton[temp_pdf_restriction_pair[j_p][1]][j_j]){flag_selection = true; break;}
	      }
	      if (flag_selection){break;}
	    }
	    if (flag_selection){break;}
	  }
	  if (!flag_selection){continue;}

	  int flag_exclusion = false;
	  for (int j_p = 0; j_p < temp_pdf_exclusion_pair.size(); j_p++){
	    for (int j_i = 0; j_i < relation_content_parton[temp_pdf_exclusion_pair[j_p][0]].size(); j_i++){
	      for (int j_j = 0; j_j < relation_content_parton[temp_pdf_exclusion_pair[j_p][1]].size(); j_j++){
		if (relation_content_parton[temp_pdf_selection_pair[i_p][0]][i_i] == relation_content_parton[temp_pdf_exclusion_pair[j_p][0]][j_i] &&
		    relation_content_parton[temp_pdf_selection_pair[i_p][1]][i_j] == relation_content_parton[temp_pdf_exclusion_pair[j_p][1]][j_j]){flag_exclusion = true; break;}
	      }
	      if (flag_exclusion){break;}
	    }
	    if (flag_exclusion){break;}
	  }
	  if (flag_exclusion){continue;}
      
	  if (j_x == 1){
	    temp_all_pdf[1] = relation_content_parton[temp_pdf_selection_pair[i_p][0]][i_i];
	    temp_all_pdf[2] = relation_content_parton[temp_pdf_selection_pair[i_p][1]][i_j];
	  }
	  else if (j_x == -1){
	    temp_all_pdf[1] = relation_content_parton[temp_pdf_selection_pair[i_p][1]][i_j];
	    temp_all_pdf[2] = relation_content_parton[temp_pdf_selection_pair[i_p][0]][i_i];
	  }
	  allowed_all_pdf[i_x].push_back(temp_all_pdf);
	}
      }
    }
  }

  if (allowed_all_pdf[0].size() == 0){
    logger << LOG_WARN << "No allowed PDFs remain after application of selection criteria." << endline_flush;
  }
  else {
    for (int i_x = 0; i_x < 2; i_x++){
      for (int i_p = 0; i_p < allowed_all_pdf[i_x].size(); i_p++){
	stringstream temp;
	temp.str("");
	temp << setw(26) << "allowed_all_pdf[" << setw(2) << i_x << "][" << setw(2) << i_p << "] = ";
	for (int i_s = 0; i_s < allowed_all_pdf[i_x][i_p].size(); i_s++){
	  temp << setw(4) << allowed_all_pdf[i_x][i_p][i_s];
	}
	logger << LOG_DEBUG << temp.str() << endline_flush;
      }
    }
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::perform_selection_content_pdf(){
  static Logger logger("observable_set::perform_selection_content_pdf");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (!csi->coll_choice){return;}

  determine_selection_content_pdf();
  logger << LOG_DEBUG << "Before perform_selection_content_pdf:" << endline_flush;
  for (int i_x = 0; i_x < csi->combination_pdf.size(); i_x++){
    stringstream temp;
    temp.str("");
    temp << "csi->combination_pdf[" << setw(2) << i_x << "] = ";
    for (int i_y = 0; i_y < csi->combination_pdf[i_x].size(); i_y++){temp << setw(2) << right << csi->combination_pdf[i_x][i_y] << "  ";}
    logger << LOG_DEBUG << temp.str() << endline_flush;
  }

  for (int i_x = 0; i_x < csi->combination_pdf.size(); i_x++){
    int match = 0;
    int x_d = 0;
    if (csi->combination_pdf[i_x][0] == 1){x_d = 0;}
    else if (csi->combination_pdf[i_x][0] == -1){x_d = 1;}
    for (int i_p = 0; i_p < allowed_all_pdf[x_d].size(); i_p++){
      if (csi->combination_pdf[i_x] == allowed_all_pdf[x_d][i_p]){
	match = 1;
	break;
      }
    }
    if (match == 0){
      csi->combination_pdf.erase(csi->combination_pdf.begin() + i_x);
      i_x--;
    }
  }

  logger << LOG_DEBUG << "After perform_selection_content_pdf:" << endline_flush;
  for (int i_x = 0; i_x < csi->combination_pdf.size(); i_x++){
    stringstream temp;
    temp.str("");
    temp << "csi->combination_pdf[" << setw(2) << i_x << "] = ";
    for (int i_y = 0; i_y < csi->combination_pdf[i_x].size(); i_y++){temp << setw(2) << right << csi->combination_pdf[i_x][i_y] << "  ";}
    logger << LOG_DEBUG << temp.str() << endline_flush;
  }
  if (csi->combination_pdf.size() == 0){
    logger << LOG_WARN << "According to pdf-selection criteria, no partonic channels remain: int_end = 1, csi->int_end_pdf_selection = 1 !" << endline_flush;
    int_end = 1;
    csi->int_end_pdf_selection = 1;
  }


  logger << LOG_DEBUG << " Set  list_combination_pdf  and  list_combination_pdf_emission." << endline_flush;
  // new: Set list_combination_pdf in order to use different PDF functions !!!
  list_combination_pdf.resize(1, vector<vector<vector<int> > > (csi->combination_pdf.size(), vector<vector<int> > (1)));
  for (int i_i = 0; i_i < csi->combination_pdf.size(); i_i++){list_combination_pdf[0][i_i][0] = csi->combination_pdf[i_i];}
  list_combination_pdf_emission.resize(1, vector<int> (3, 0));
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::perform_selection_content_pdf_collinear(){
  static Logger logger("observable_set::perform_selection_content_pdf_collinear");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (!csi->coll_choice){
    logger << LOG_WARN << "No contribution due to csi->coll_choice = " << csi->coll_choice << ": int_end = 1 !" << endline_flush;
    int_end = 1; return;
  }

  determine_selection_content_pdf();
  logger << LOG_DEBUG << "Before perform_selection_content_pdf_collinear:" << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "(*CA_collinear).size() = " << (*CA_collinear).size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "CA_combination_pdf.size() = " << CA_combination_pdf.size() << endline_flush;
  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    for (int i_i = 0; i_i < CA_combination_pdf[i_c].size(); i_i++){
      logger << LOG_DEBUG << "pdf contributions: " << setw(15) << left << (*CA_collinear)[i_c][0].all_name()[i_i] << ":   " << endline_flush;
      for (int i_x = 0; i_x < CA_combination_pdf[i_c][i_i].size(); i_x++){
	stringstream temp;
	temp.str("");
	temp << setw(48) << "";
	temp << "CA_combination_pdf[" << setw(2) << i_c << "][" << setw(2) << i_i << "][" << setw(2) << i_x << "] = ";
	for (int i_y = 0; i_y < CA_combination_pdf[i_c][i_i][i_x].size(); i_y++){temp << setw(2) << right << CA_combination_pdf[i_c][i_i][i_x][i_y] << "  ";}
	logger << LOG_DEBUG << temp.str() << endline_flush;
      }
    }
  }
  logger.newLine(LOG_DEBUG);

  for (int i_c = 0; i_c < CA_combination_pdf.size(); i_c++){
    for (int i_i = 0; i_i < CA_combination_pdf[i_c].size(); i_i++){
      for (int i_x = 0; i_x < CA_combination_pdf[i_c][i_i].size(); i_x++){
	int match = 0;
	int x_d = 0;
	if (CA_combination_pdf[i_c][i_i][i_x][0] == 1){x_d = 0;}
	else if (CA_combination_pdf[i_c][i_i][i_x][0] == -1){x_d = 1;}
	for (int i_p = 0; i_p < allowed_all_pdf[x_d].size(); i_p++){
	  if (CA_combination_pdf[i_c][i_i][i_x] == allowed_all_pdf[x_d][i_p]){
	    match = 1;
	    break;
	  }
	}
	if (match == 0){
	  CA_combination_pdf[i_c][i_i].erase(CA_combination_pdf[i_c][i_i].begin() + i_x);
	  CA_Q2f[i_c][i_i].erase(CA_Q2f[i_c][i_i].begin() + i_x);
	  i_x--;
	}
      }
    }
  }

  logger << LOG_DEBUG << "After selection, before removing empty entries, in perform_selection_content_pdf_collinear:" << endline_flush;
  logger.newLine(LOG_DEBUG);
  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    for (int i_i = 0; i_i < CA_combination_pdf[i_c].size(); i_i++){
      logger << LOG_DEBUG << "pdf contributions: " << setw(15) << left << (*CA_collinear)[i_c][0].all_name()[i_i] << ":   " << endline_flush;
      for (int i_x = 0; i_x < CA_combination_pdf[i_c][i_i].size(); i_x++){
	stringstream temp;
	temp.str("");
	temp << setw(48) << "";
	temp << "CA_combination_pdf[" << setw(2) << i_c << "][" << setw(2) << i_i << "][" << setw(2) << i_x << "] = ";
	for (int i_y = 0; i_y < CA_combination_pdf[i_c][i_i][i_x].size(); i_y++){temp << setw(2) << right << CA_combination_pdf[i_c][i_i][i_x][i_y] << "  ";}
	logger << LOG_DEBUG << temp.str() << endline_flush;
      }
    }
  }
  logger.newLine(LOG_DEBUG);

  remove_empty_CA_collinear();

  logger << LOG_DEBUG << "After perform_selection_content_pdf_collinear:" << endline_flush;

  logger.newLine(LOG_DEBUG);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::remove_empty_CA_collinear(){
  Logger logger("observable_set::remove_empty_CA_collinear");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

    logger << LOG_DEBUG << "Remove now those collinear 'dipoles' that do not have any left-over contribution:" << endline_flush;
  logger.newLine(LOG_DEBUG);
  /*
  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    int count_remaining = 0;
    for (int i_i = 0; i_i < CA_combination_pdf[i_c].size(); i_i++){
      count_remaining += CA_combination_pdf[i_c][i_i].size();
    }
    if (count_remaining == 0){
      CA_combination_pdf.erase(CA_combination_pdf.begin() + i_c);
      (*CA_collinear).erase((*CA_collinear).begin() + i_c);
      CA_Q2f.erase(CA_Q2f.begin() + i_c);
      i_c--;
    }
  }
  logger.newLine(LOG_DEBUG);
  */
  // New removal:

  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    size_t x_l = (*CA_collinear)[i_c][0].no_pdf;
    if (list_combination_pdf[x_l].size() == 0){
      logger << LOG_INFO << "Remove collinear dipole   " << (*CA_collinear)[i_c][0].name() << endline_flush;
      (*CA_collinear).erase((*CA_collinear).begin() + i_c);
      CA_combination_pdf.erase(CA_combination_pdf.begin() + i_c);
      CA_Q2f.erase(CA_Q2f.begin() + i_c);
      i_c--;
    }
  }

    
  n_pc = (*CA_collinear).size();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::perform_selection_content_pdf_list(){
  static Logger logger("observable_set::perform_selection_content_pdf_list");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  determine_selection_content_pdf();

  logger.newLine(LOG_DEBUG);
  logger << LOG_DEBUG << "Before perform_selection_content_pdf_list:   list_combination_pdf.size() = " << list_combination_pdf.size() << endline_flush;
  logger.newLine(LOG_DEBUG);
  for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
    logger << LOG_DEBUG << "list_combination_pdf[" << setw(2) << i_l << "]:" << endline_flush;
    for (int i_i = 0; i_i < list_combination_pdf[i_l].size(); i_i++){
      for (int i_z = 0; i_z < list_combination_pdf[i_l][i_i].size(); i_z++){
	stringstream temp;
	temp.str("");
	temp << setw(32) << "";
	temp << "list_combination_pdf[" << setw(2) << i_l << "][" << setw(2) << i_i << "][" << setw(2) << i_z << "] = ";
	for (int i_y = 0; i_y < list_combination_pdf[i_l][i_i][i_z].size(); i_y++){temp << setw(2) << right << list_combination_pdf[i_l][i_i][i_z][i_y] << "  ";}
	logger << LOG_DEBUG << temp.str() << endline_flush;
      }
    }
  }
  logger.newLine(LOG_DEBUG);

  for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
    for (int i_i = 0; i_i < list_combination_pdf[i_l].size(); i_i++){
      for (int i_x = 0; i_x < list_combination_pdf[i_l][i_i].size(); i_x++){
	int match = 0;
	int x_d = 0;
	if (list_combination_pdf[i_l][i_i][i_x][0] == 1){x_d = 0;}
	else if (list_combination_pdf[i_l][i_i][i_x][0] == -1){x_d = 1;}
	for (int i_p = 0; i_p < allowed_all_pdf[x_d].size(); i_p++){
	  if (list_combination_pdf[i_l][i_i][i_x] == allowed_all_pdf[x_d][i_p]){
	    match = 1;
	    break;
	  }
	}
	if (match == 0){
	  list_combination_pdf[i_l][i_i].erase(list_combination_pdf[i_l][i_i].begin() + i_x);
	  //new:
	  QT_Q2f[i_l][i_i].erase(QT_Q2f[i_l][i_i].begin() + i_x);
	  i_x--;
	}
      }
    }
  }

  logger << LOG_DEBUG << "Remove now those entries of 'list_combination_pdf[i_l]' that do not have any left-over contribution." << endline_flush;
  logger.newLine(LOG_DEBUG);

  for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
    for (int i_i = list_combination_pdf[i_l].size() - 1; i_i >= 0; i_i--){
      if (list_combination_pdf[i_l][i_i].size() == 0){
	list_combination_pdf[i_l].erase(list_combination_pdf[i_l].begin() + i_i);
      }
    }
  }

  logger << LOG_DEBUG << "After perform_selection_content_pdf_list:   list_combination_pdf.size() = " << list_combination_pdf.size() << endline_flush;
  for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
    logger << LOG_DEBUG << "list_combination_pdf[" << setw(2) << i_l << "]:" << endline_flush;
    for (int i_i = 0; i_i < list_combination_pdf[i_l].size(); i_i++){
      for (int i_z = 0; i_z < list_combination_pdf[i_l][i_i].size(); i_z++){
	stringstream temp;
	temp.str("");
	temp << setw(32) << "";
	temp << "list_combination_pdf[" << setw(2) << i_l << "][" << setw(2) << i_i << "][" << setw(2) << i_z << "] = ";
	for (int i_y = 0; i_y < list_combination_pdf[i_l][i_i][i_z].size(); i_y++){temp << setw(2) << right << list_combination_pdf[i_l][i_i][i_z][i_y] << "  ";}
	logger << LOG_DEBUG << temp.str() << endline_flush;
      }
    }
  }
  logger.newLine(LOG_DEBUG);

  // NEW !!! Check if contribution vanishes due to PDF content:
  bool empty = true;
  for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){if (list_combination_pdf[i_l].size() > 0){empty = false; break;}}
  if (empty){csi->int_end_pdf_selection = 1; logger << LOG_WARN << "Contribution vanishes due to PDF content: csi->int_end_pdf_selection = 1 !" << endline_flush;}
  
  // Should be called from elsewhere !!!
  //  remove_empty_ncollinear();
  //  determine_initial_channel();
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


// Could be shifted into qTsubtraction initialization (used in particular to reduce complexity of calculation for selected PDF channels) !!!

void observable_set::determine_initial_channel(){
  Logger logger("observable_set::determine_initial_channel");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;


  // !!! Has nothing to do with pdfs: Only relevant on Born phase-space (e.g. CT[2], VT[2])
  // Calculate charge of final state:
  double charge_IS = 0.;
  for (int i_p = 1; i_p < 3; i_p++){charge_IS += csi->charge_parton[0][i_p];}
  for (double x_charge = -1.; x_charge <= 1.; x_charge += 1.){if (abs(charge_IS - x_charge) < 1.e-13){charge_IS = x_charge; break;}}
  logger << LOG_DEBUG << "charge_IS = " << charge_IS << endline_flush;
  /*
  double charge_FS = 0.;
  for (int i_p = 3; i_p < csi->n_particle + 3; i_p++){charge_FS += csi->charge_parton[0][i_p];}
  for (double x_charge = -1.; x_charge <= 1.; x_charge += 1.){if (abs(charge_FS - x_charge) < 1.e-13){charge_FS = x_charge; break;}
  logger << LOG_DEBUG << "charge_FS = " << charge_FS << endline_flush;
  */
    
  initial_gg = 0;
  initial_qqx = 0;
  initial_aa = 0;
  initial_ga = 0;
  if (csi->type_parton[0][1] == 0 &&
      csi->type_parton[0][2] == 0){initial_gg = 1;}  // hard process: IS: gg
  else if (csi->type_parton[0][1] == 22 &&
      csi->type_parton[0][2] == 22){initial_aa = 1;}  // hard process: IS: aa
  else if ((csi->type_parton[0][1] == 22 && csi->type_parton[0][2] == 0) || (csi->type_parton[0][1] == 0 && csi->type_parton[0][2] == 22)){initial_ga = 1;}  // hard process: IS: ga
  else if (csi->type_parton[0][1] != 0 &&
	   csi->type_parton[0][1] != 22 &&
	   csi->type_parton[0][1] == -csi->type_parton[0][2] &&
	   charge_IS == 0.){initial_qqx = 1;}  // hard process: IS: qqx (SF, neutral case)
  else if (csi->type_parton[0][1] != 0 &&
	   csi->type_parton[0][1] != 22 &&
	   csi->type_parton[0][1] * csi->type_parton[0][2] < 0 &&
	   abs(charge_IS) == 1. &&
	   (abs(msi->charge_pdg[csi->type_parton[0][1]] + msi->charge_pdg[csi->type_parton[0][2]] - charge_IS) < 1.e-13)){initial_qqx = 2;}  // hard process: IS: qqx (charged FS case)
  //	   (abs(csi->type_parton[0][1]) + abs(csi->type_parton[0][2])) % 2 == 1){initial_qqx = 2;}  // hard process: IS: qqx (all flavour combinations, charged case) - might need refinement !!!)

  initial_pdf_gg = 0;
  initial_pdf_qqx = 0;
  initial_pdf_aa = 0;
  initial_pdf_gq = 0;
  initial_pdf_ga = 0;
  initial_pdf_rest = 0;
  initial_pdf_diag = 0;

  for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
    for (int i_i = 0; i_i < list_combination_pdf[i_l].size(); i_i++){
      for (int i_z = 0; i_z < list_combination_pdf[i_l][i_i].size(); i_z++){
	if (list_combination_pdf[i_l][i_i][i_z][1] == 0 &&
	    list_combination_pdf[i_l][i_i][i_z][2] == 0){initial_pdf_gg = 1;}  // PDF initial state: gg
	else if ((list_combination_pdf[i_l][i_i][i_z][1] == 0 &&
		  list_combination_pdf[i_l][i_i][i_z][2] != 0 &&
		  list_combination_pdf[i_l][i_i][i_z][2] != 7) ||
		 (list_combination_pdf[i_l][i_i][i_z][2] == 0 &&
		  list_combination_pdf[i_l][i_i][i_z][1] != 0 &&
		  list_combination_pdf[i_l][i_i][i_z][1] != 7)){initial_pdf_gq = 1;}  // PDF initial state: gQ or Qg
	else if ((list_combination_pdf[i_l][i_i][i_z][1] == 7 &&
		  list_combination_pdf[i_l][i_i][i_z][2] != 0 &&
		  list_combination_pdf[i_l][i_i][i_z][2] != 7) ||
		 (list_combination_pdf[i_l][i_i][i_z][2] == 7 &&
		  list_combination_pdf[i_l][i_i][i_z][1] != 0 &&
		  list_combination_pdf[i_l][i_i][i_z][1] != 7)){initial_pdf_aq = 1;}  // PDF initial state: aQ or Qa
	else if (list_combination_pdf[i_l][i_i][i_z][1] != 0 &&
		 list_combination_pdf[i_l][i_i][i_z][1] != 7 &&
		 list_combination_pdf[i_l][i_i][i_z][1] == -list_combination_pdf[i_l][i_i][i_z][2] &&
		 charge_IS == 0.){initial_pdf_qqx = 1;}  // PDF initial state: qqx (SF only: might need refinement for charged case !!! -> see below ) 
// The qqx case could be refined, depending on the charge of the produced FS !!!
// In practise, this is only called with the right quark flavours for LO subprocesses.
//
	//(abs(list_combination_pdf[i_l][i_i][i_z][1]) + abs(list_combination_pdf[i_l][i_i][i_z][2])) % 2 == 1  
	else if (list_combination_pdf[i_l][i_i][i_z][1] != 0 &&
		 list_combination_pdf[i_l][i_i][i_z][1] != 7 &&
		 list_combination_pdf[i_l][i_i][i_z][1] * list_combination_pdf[i_l][i_i][i_z][2] < 0 &&
		 abs(charge_IS) == 1. &&
		 (abs(msi->charge_pdg[list_combination_pdf[i_l][i_i][i_z][1]] + msi->charge_pdg[list_combination_pdf[i_l][i_i][i_z][2]] - charge_IS) < 1.e-13)){initial_pdf_qqx = 2;}  // PDF initial state: qqx charged case	
	else if (list_combination_pdf[i_l][i_i][i_z][1] == 7 &&
		 list_combination_pdf[i_l][i_i][i_z][2] == 7){initial_pdf_aa = 1;}  // PDF initial state: aa
	else if ((list_combination_pdf[i_l][i_i][i_z][1] == 7 && 
		  list_combination_pdf[i_l][i_i][i_z][2] == 0) ||
		 (list_combination_pdf[i_l][i_i][i_z][1] == 0 && 
		  list_combination_pdf[i_l][i_i][i_z][2] == 7)){initial_pdf_ga = 1;}  // PDF initial state: ga
	else {initial_pdf_rest = 1;}  // PDF initial state: all other combinations, i.e. qq, qxqx, qqx (DF only: might need refinement for charged case !!!)

	// This check is not sufficient if e.g. ssx is calculated (without ddx) in the ddx channel
	if (list_combination_pdf[i_l][i_i][i_z][1] == csi->type_parton[0][1] &&
	    list_combination_pdf[i_l][i_i][i_z][2] == csi->type_parton[0][2]){initial_pdf_diag = 1;}  // PDF IS equal to hard-process IS: relevant for e.g. if H2 is needed, etc.
	if (list_combination_pdf[i_l][i_i][i_z][1] != 0 &&
	    list_combination_pdf[i_l][i_i][i_z][1] != 7 &&
	    list_combination_pdf[i_l][i_i][i_z][1] == -list_combination_pdf[i_l][i_i][i_z][2] &&
	    initial_qqx == 1 &&
	    list_combination_pdf[i_l][i_i][i_z][1] % 2 == list_combination_pdf[i_l][i_i][i_z][1] % 2){initial_pdf_diag = 1; logger << LOG_DEBUG << "initial_pdf_diag = 1 set due to new condition." << endline_flush;}
      }
    }
  }
  // Check precise meaning of  initial_pdf_diag  !!!
  // (seems to be related to the ssx - ddx situation ...)

  initial_diag = 0;
  initial_diag_gg = 0;
  initial_diag_qqx = 0;
  initial_diag_aa = 0;
  initial_diag_ga = 0;

  if (initial_gg && initial_pdf_gg){
    initial_diag_gg = 1;
    initial_pdf_diag = 1;
    initial_diag = 1;
  }
  if (initial_qqx && initial_pdf_qqx){
    initial_diag_qqx = 1;
    initial_pdf_diag = 1;
    initial_diag = 2;
  }
  if (initial_aa && initial_pdf_aa){
    initial_diag_aa = 1;
    initial_pdf_diag = 1;
    initial_diag = 3;
  }
  if (initial_ga && initial_pdf_ga){
    initial_diag_ga = 1;
    initial_pdf_diag = 1;
    initial_diag = 4;
  }
  
  initial_channel = 0;
  if (initial_gg){initial_channel = 1;}
  else if (initial_qqx){initial_channel = 2;}
  //  else {initial_channel = 0;}
  // Check if other values of initial_channel (e.g. aa, ga) would be useful !!!


  /*
  if (initial_gg && initial_pdf_gg){initial_diag = 1;}
  if (initial_qqx && initial_pdf_qqx){initial_diag = 2;}
  //  if (initial_channel){initial_diag = 1;}
  */

  logger << LOG_DEBUG << "initial_channel  = " << initial_channel << endline_flush;

  logger << LOG_DEBUG << "initial_gg       = " << initial_gg << endline_flush;
  logger << LOG_DEBUG << "initial_qqx      = " << initial_qqx << endline_flush;
  logger << LOG_DEBUG << "initial_aa       = " << initial_aa << endline_flush;
  logger << LOG_DEBUG << "initial_ga       = " << initial_ga << endline_flush;

  logger << LOG_DEBUG << "initial_pdf_gg   = " << initial_pdf_gg << endline_flush;
  logger << LOG_DEBUG << "initial_pdf_aa   = " << initial_pdf_aa << endline_flush;
  logger << LOG_DEBUG << "initial_pdf_gq   = " << initial_pdf_gq << endline_flush;
  logger << LOG_DEBUG << "initial_pdf_aq   = " << initial_pdf_aq << endline_flush;
  logger << LOG_DEBUG << "initial_pdf_qqx  = " << initial_pdf_qqx << endline_flush;
  logger << LOG_DEBUG << "initial_pdf_ga   = " << initial_pdf_ga << endline_flush;
  logger << LOG_DEBUG << "initial_pdf_rest = " << initial_pdf_rest << endline_flush;

  logger << LOG_DEBUG << "initial_pdf_diag = " << initial_pdf_diag << endline_flush;

  logger << LOG_DEBUG << "initial_diag     = " << initial_diag << endline_flush;
  logger << LOG_DEBUG << "initial_diag_gg  = " << initial_diag_gg << endline_flush;
  logger << LOG_DEBUG << "initial_diag_aa  = " << initial_diag_aa << endline_flush;
  logger << LOG_DEBUG << "initial_diag_qqx = " << initial_diag_qqx << endline_flush;
  logger << LOG_DEBUG << "initial_diag_ga  = " << initial_diag_ga << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


string observable_set::output_list_combination_pdf(){
  Logger logger("observable_set::output_list_combination_pdf");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream temp_ss;
  //  logger.newLine(LOG_DEBUG);
  temp_ss << "pdf contributions:   list_combination_pdf, list_combination_pdf_emission" << endline;
  temp_ss << endline;
  for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
    stringstream temp_em;
    for (int i_e = 1; i_e < 3; i_e++){
      temp_em << "   ";
      if (list_combination_pdf_emission[i_l][i_e] == 1){temp_em << "x_" << i_e << " / z_" << i_e;}
      else if (list_combination_pdf_emission[i_l][i_e] == 0){temp_em << "x_" << i_e;}
      else {logger << LOG_FATAL << temp_em.str() << " is not a valid value." << endline_flush; exit(1);}
    }
    temp_ss << "pdf contributions [" << setw(2) << i_l << "]   @" << temp_em.str() << endline;// << setw(15) << left << ncollinear[i_c].all_name[i_i] << ":   " << endline;

    for (int i_i = 0; i_i < list_combination_pdf[i_l].size(); i_i++){
       for (int i_x = 0; i_x < list_combination_pdf[i_l][i_i].size(); i_x++){
	stringstream temp;
	temp.str("");
	temp << setw(32) << "";
	temp << "list_combination_pdf[" << setw(2) << i_l << "][" << setw(2) << i_i << "][" << setw(2) << i_x << "] = ";
	for (int i_y = 0; i_y < list_combination_pdf[i_l][i_i][i_x].size(); i_y++){temp << setw(2) << right << list_combination_pdf[i_l][i_i][i_x][i_y] << "  ";}
	temp << "   ";
	temp << "QT_Q2f = " << QT_Q2f[i_l][i_i][i_x];
	temp_ss << temp.str() << endline;
      }
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished - return" << endline_flush;
  return temp_ss.str();
}



