#include "header.hpp"

observable_set::observable_set(){
  Logger logger("observable_set::observable_set ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "constructor called..." << endl;

  int_end = 0;
  max_dyn_ren = 0;
  max_dyn_fact = 0;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

observable_set::observable_set(int _n_set_TSV){
  Logger logger("observable_set::observable_set (n_set_TSV)");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "constructor called..." << endl;

  n_set_TSV = _n_set_TSV;
  int_end = 0;
  max_dyn_ren = 0;
  max_dyn_fact = 0;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

observable_set::observable_set(inputparameter_set * isi, contribution_set * _csi, model_set * _msi, event_set * _esi, phasespace_set * _psi, user_defined * _user){
  Logger logger("observable_set::observable_set (isi, csi, msi, esi, psi, user)");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "constructor called..." << endl;

  initialization(isi, _csi, _msi, _esi, _psi, _user);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

observable_set::~observable_set(){
  Logger logger("observable_set::~observable_set ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  logger << LOG_DEBUG << "destructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}



void observable_set::calculate_optimization_modifier(){
  Logger logger("observable_set::calculate_optimization_modifier");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (selection_optimization_modifier == "tau"){
    optimization_modifier = pow(psi->x_pdf[0], optimization_modifier_exponent);
  }
  else if (selection_optimization_modifier == "offshellZ"){
    double temp_m2_ll = 0.;
    if (csi->n_ps == 1){temp_m2_ll = (esi->event_particle_group[0][0][esi->access_object_selection["lep"]][0]->momentum + esi->event_particle_group[0][0][esi->access_object_selection["lep"]][1]->momentum).m2();}
    else {
      
      // version where, among accepted dipole kinematics, the m_ll values closest to the resonance is used for bias function
      temp_m2_ll = 1.e99;
      for (int i_a = 0; i_a < csi->n_ps; i_a++){
	if (esi->cut_ps[i_a] == -1){continue;}
	double this_temp_m2_ll = (esi->event_particle_group[0][i_a][esi->access_object_selection["lep"]][0]->momentum + esi->event_particle_group[0][i_a][esi->access_object_selection["lep"]][1]->momentum).m2();
	if (abs(this_temp_m2_ll - msi->M2_Z) < abs(temp_m2_ll - msi->M2_Z)){temp_m2_ll = this_temp_m2_ll;}
      }
      
      /*
      // version where, among accepted dipole kinematics, the m_ll values farthest from the resonance is used for bias function
      temp_m2_ll = msi->M2_Z;
      for (int i_a = 0; i_a < csi->n_ps; i_a++){
	if (esi->cut_ps[i_a] == -1){continue;}
	double this_temp_m2_ll = (esi->event_particle_group[0][i_a][esi->access_object_selection["lep"]][0]->momentum + esi->event_particle_group[0][i_a][esi->access_object_selection["lep"]][1]->momentum).m2();
	if (abs(this_temp_m2_ll - msi->M2_Z) > abs(temp_m2_ll - msi->M2_Z)){temp_m2_ll = this_temp_m2_ll;}
      }
      */
      /*
      // version where the averaged m_ll value among accepted dipole kinematics is used for bias function
      int n_ps_acc = 0;
      for (int i_a = 0; i_a < csi->n_ps; i_a++){
	if (esi->cut_ps[i_a] == -1){continue;}
	n_ps_acc++;
	temp_m2_ll += (esi->event_particle_group[0][i_a][esi->access_object_selection["lep"]][0]->momentum + esi->event_particle_group[0][i_a][esi->access_object_selection["lep"]][1]->momentum).m();
      }
      temp_m2_ll = pow(temp_m2_ll / n_ps_acc, 2);
      */
    }
    optimization_modifier = pow(pow(temp_m2_ll - msi->M2_Z, 2) + msi->M2_Z * pow(msi->Gamma_Z, 2), optimization_modifier_exponent);
  }
  else if (selection_optimization_modifier == "offshellW"){
    double temp_m2_ln = 0.;
    if (csi->n_ps == 1){temp_m2_ln = (esi->event_particle_group[0][0][esi->access_object_selection["lep"]][0]->momentum + esi->event_particle_group[0][0][esi->access_object_selection["nua"]][0]->momentum).m2();}
    else {
      
      // version where, among accepted dipole kinematics, the m_ll values closest to the resonance is used for bias function
      temp_m2_ln = 1.e99;
      for (int i_a = 0; i_a < csi->n_ps; i_a++){
	if (esi->cut_ps[i_a] == -1){continue;}
	double this_temp_m2_ln = (esi->event_particle_group[0][i_a][esi->access_object_selection["lep"]][0]->momentum + esi->event_particle_group[0][i_a][esi->access_object_selection["nua"]][0]->momentum).m2();
	if (abs(this_temp_m2_ln - msi->M2_W) < abs(temp_m2_ln - msi->M2_W)){temp_m2_ln = this_temp_m2_ln;}
      }
      
      /*
      // version where, among accepted dipole kinematics, the m_ll values farthest from the resonance is used for bias function
      temp_m2_ln = msi->M2_W;
      for (int i_a = 0; i_a < csi->n_ps; i_a++){
	if (esi->cut_ps[i_a] == -1){continue;}
	double this_temp_m2_ln = (esi->event_particle_group[0][i_a][esi->access_object_selection["lep"]][0]->momentum + esi->event_particle_group[0][i_a][esi->access_object_selection["nua"]][0]->momentum).m2();
	if (abs(this_temp_m2_ln - msi->M2_W) > abs(temp_m2_ln - msi->M2_W)){temp_m2_ln = this_temp_m2_ln;}
      }
      */
      /*
      // version where the averaged m_ll value among accepted dipole kinematics is used for bias function
      int n_ps_acc = 0;
      for (int i_a = 0; i_a < csi->n_ps; i_a++){
	if (esi->cut_ps[i_a] == -1){continue;}
	n_ps_acc++;
	temp_m2_ln += (esi->event_particle_group[0][i_a][esi->access_object_selection["lep"]][0]->momentum + esi->event_particle_group[0][i_a][esi->access_object_selection["nua"]][0]->momentum).m();
      }
      temp_m2_ln = pow(temp_m2_ln / n_ps_acc, 2);
      */
    }
    optimization_modifier = pow(pow(temp_m2_ln - msi->M2_W, 2) + msi->M2_W * pow(msi->Gamma_W, 2), optimization_modifier_exponent);
  }
  else {
    logger << LOG_WARN << "No allowed modifier selected." << endline_flush; exit(1);
  }
    
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::calculate_intermediate_result(){
  Logger logger("observable_set::calculate_intermediate_result");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  calculate_Xsection(psi->i_gen, Xsection, Xsection_delta, full_sum_weight, full_sum_weight2, step_sum_weight, step_sum_weight2);

  logger << LOG_DEBUG << "full_sum_weight = " << full_sum_weight << endline_flush;

  if (switch_CV){
    for (int i_c = 0; i_c < n_qTcut; i_c++){
      for (int i_s = 0; i_s < n_scales_CV; i_s++){
	calculate_Xsection(psi->i_gen, Xsection_CV[i_c][i_s], Xsection_delta_CV[i_c][i_s], full_sum_weight_CV[i_c][i_s], full_sum_weight2_CV[i_c][i_s], step_sum_weight_CV[i_c][i_s], step_sum_weight2_CV[i_c][i_s]);
      }
    }
  }

  if (switch_TSV){
    for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
      for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	  for (int i_q = 0; i_q < output_n_qTcut; i_q++){
	    if (active_qTcut){
	      logger << LOG_DEBUG_VERBOSE << "    sum_weight_qTcut_TSV[" << i_s << "][" << i_q << "][" << i_r << "][" << i_f << "] = " << sum_weight_qTcut_TSV[i_s][i_q][i_r][i_f] << endline_flush;
	      calculate_Xsection(psi->i_gen, Xsection_TSV[i_s][i_q][i_r][i_f], Xsection_delta_TSV[i_s][i_q][i_r][i_f], fullsum_weight_qTcut_TSV[i_s][i_q][i_r][i_f], fullsum_weight2_qTcut_TSV[i_s][i_q][i_r][i_f], sum_weight_qTcut_TSV[i_s][i_q][i_r][i_f], sum_weight2_qTcut_TSV[i_s][i_q][i_r][i_f]);
	      logger << LOG_DEBUG_VERBOSE << "fullsum_weight_qTcut_TSV[" << i_s << "][" << i_q << "][" << i_r << "][" << i_f << "] = " << fullsum_weight_qTcut_TSV[i_s][i_q][i_r][i_f] << endline_flush;
	      logger << LOG_DEBUG_VERBOSE << "      Xsection_TSV[" << i_s << "][" << i_q << "][" << i_r << "][" << i_f << "] = " << Xsection_TSV[i_s][i_q][i_r][i_f] << endline_flush;
	    }
	    else {
	      logger << LOG_DEBUG_VERBOSE << "    sum_weight_TSV[" << i_s << "][" << i_r << "][" << i_f << "] = " << sum_weight_TSV[i_s][i_r][i_f] << endline_flush;
	      calculate_Xsection(psi->i_gen, Xsection_TSV[i_s][i_q][i_r][i_f], Xsection_delta_TSV[i_s][i_q][i_r][i_f], fullsum_weight_TSV[i_s][i_r][i_f], fullsum_weight2_TSV[i_s][i_r][i_f], sum_weight_TSV[i_s][i_r][i_f], sum_weight2_TSV[i_s][i_r][i_f]);
	      logger << LOG_DEBUG_VERBOSE << "fullsum_weight_TSV[" << i_s << "][" << i_r << "][" << i_f << "] = " << fullsum_weight_TSV[i_s][i_r][i_f] << endline_flush;
	      logger << LOG_DEBUG_VERBOSE << "      Xsection_TSV[" << i_s << "][" << i_q << "][" << i_r << "][" << i_f << "] = " << Xsection_TSV[i_s][i_q][i_r][i_f] << endline_flush;
	    }
	  }
	}
      }
    }
  }

  if (switch_moment){
    for (int i_m = 0; i_m < n_moments; i_m++){
      calculate_Xsection(psi->i_gen, Xsection_moment[i_m], Xsection_moment_delta[i_m], full_sum_moment[i_m], full_sum_moment2[i_m], step_sum_moment[i_m], step_sum_moment2[i_m]);

      if (switch_CV){
	for (int i_c = 0; i_c < n_qTcut; i_c++){
	  for (int i_s = 0; i_s < n_scales_CV; i_s++){
	    calculate_Xsection(psi->i_gen, Xsection_moment_CV[i_m][i_c][i_s], Xsection_moment_delta_CV[i_m][i_c][i_s], full_sum_moment_CV[i_m][i_c][i_s], full_sum_moment2_CV[i_m][i_c][i_s], step_sum_moment_CV[i_m][i_c][i_s], step_sum_moment2_CV[i_m][i_c][i_s]);
	  }
	}
      }

      if (switch_TSV){
	for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
	  if (!switch_moment_TSV[i_s]){continue;}
	  for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	    for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	      for (int i_q = 0; i_q < output_n_qTcut; i_q++){
		if (active_qTcut){
		  logger << LOG_DEBUG_VERBOSE << "    sum_moment_qTcut_TSV[" << i_s << "][" << i_m << "][" << i_q << "][" << i_r << "][" << i_f << "] = " << sum_moment_qTcut_TSV[i_s][i_m][i_q][i_r][i_f] << endline_flush;
		  calculate_Xsection(psi->i_gen, Xsection_moment_TSV[i_s][i_m][i_q][i_r][i_f], Xsection_moment_delta_TSV[i_s][i_m][i_q][i_r][i_f], fullsum_moment_qTcut_TSV[i_s][i_m][i_q][i_r][i_f], fullsum_moment2_qTcut_TSV[i_s][i_m][i_q][i_r][i_f], sum_moment_qTcut_TSV[i_s][i_m][i_q][i_r][i_f], sum_moment2_qTcut_TSV[i_s][i_m][i_q][i_r][i_f]);
		  logger << LOG_DEBUG_VERBOSE << "fullsum_moment_qTcut_TSV[" << i_s << "][" << i_m << "][" << i_q << "][" << i_r << "][" << i_f << "] = " << fullsum_moment_qTcut_TSV[i_s][i_m][i_q][i_r][i_f] << endline_flush;
		  logger << LOG_DEBUG_VERBOSE << "     Xsection_moment_TSV[" << i_s << "][" << i_m << "][" << i_q << "][" << i_r << "][" << i_f << "] = " << Xsection_moment_TSV[i_s][i_m][i_q][i_r][i_f] << endline_flush;
		}
		else {
		  logger << LOG_DEBUG_VERBOSE << "     sum_moment_TSV[" << i_s << "][" << i_m << "][" << i_r << "][" << i_f << "] = " << sum_moment_TSV[i_s][i_m][i_r][i_f] << endline_flush;
		  calculate_Xsection(psi->i_gen, Xsection_moment_TSV[i_s][i_m][i_q][i_r][i_f], Xsection_moment_delta_TSV[i_s][i_m][i_q][i_r][i_f], fullsum_moment_TSV[i_s][i_m][i_r][i_f], fullsum_moment2_TSV[i_s][i_m][i_r][i_f], sum_moment_TSV[i_s][i_m][i_r][i_f], sum_moment2_TSV[i_s][i_m][i_r][i_f]);
		  logger << LOG_DEBUG_VERBOSE << " fullsum_moment_TSV[" << i_s << "][" << i_m << "][" << i_r << "][" << i_f << "] = " << fullsum_moment_TSV[i_s][i_m][i_r][i_f] << endline_flush;
		  logger << LOG_DEBUG_VERBOSE << "Xsection_moment_TSV[" << i_s << "][" << i_m << "][" << i_q << "][" << i_r << "][" << i_f << "] = " << Xsection_moment_TSV[i_s][i_m][i_q][i_r][i_f] << endline_flush;
		}
	      }
	    }
	  }
	}
      }
    }
  }

  if ((psi->i_acc == psi->n_events_max) || (psi->i_acc >= psi->n_events_min && abs(Xsection_delta / sigma_normalization) < sigma_normalization_deviation)){
    logger << LOG_WARN << "Integration finshed due to psi->i_acc == psi->n_events_max or alike: int_end = 1 !" << endline_flush;
    int_end = 1;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::calculate_Xsection(long long i_gen, double & this_Xsection, double & this_Xsection_delta, double & this_sum_weights, double & this_sum_weights2, double & this_temp_sum_weights, double & this_temp_sum_weights2){
  Logger logger("observable_set::calculate_Xsection");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  this_sum_weights += this_temp_sum_weights;
  this_sum_weights2 += this_temp_sum_weights2;
  this_temp_sum_weights = 0.;
  this_temp_sum_weights2 = 0.;
  this_Xsection_delta = sqrt((this_sum_weights2 - pow(this_sum_weights, 2) / i_gen)) / (i_gen - 1);
  this_Xsection = this_sum_weights / i_gen;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::initialization_runtime(){
  Logger logger("observable_set::initialization_runtime");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  start = clock();
  start_step = clock();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::determine_runtime(){
  Logger logger("observable_set::determine_runtime");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG_POINT << left << "PSP finished:" << "   i_gen = " << setw(12) << psi->i_gen << "   i_acc = " << setw(12) << psi->i_acc << "   i_rej = " << setw(12) << psi->i_rej << "   i_nan = " << setw(4) << psi->i_nan << "   i_tec = " << setw(4) << psi->i_tec << endline_flush;

  if (start > clock()){start = clock(); time_counter = 0;}
  if (start < 0 && clock() > 0){start = clock(); time_counter = 0;}

  int seconds_passed = clock()/CLOCKS_PER_SEC - start/CLOCKS_PER_SEC + sec_import;

  int old_h = h;
  int old_min = min;

  if (seconds_passed > 3600) {
    h = int(seconds_passed / 3600);
    min = 0;
    seconds_passed -= h * 3600;
  }
  if (seconds_passed > 60) {
    min = int(seconds_passed / 60);
    seconds_passed -= min * 60;
  }
  sec = seconds_passed;

  if ((switch_console_output_runtime == 1 && ((min > old_min && h == 0) || h > old_h)) ||
      (switch_console_output_runtime == 2 && (min > old_min || h > old_h))){
    // complete efficiency, not only that of the last interval
    double efficiency = double(psi->i_acc) / psi->i_gen;

    logger << LOG_INFO << setw(3) << right << h << " hours" << setw(3) << right << min << " minutes needed for " << psi->i_acc << " accepted (" << psi->i_gen << " generated) events (efficiency = " << efficiency << ")." << endline_flush;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


// contains only TSV information !!!
ostream & operator << (ostream & s, const observable_set & osi){
  s << "observable_set independent stuff:" << endline;
  s << setw(38) << right << "switch_qTcut" << setw(5) << "" << "=" << setw(5) << "" << osi.switch_qTcut << endline;
  s << setw(38) << right << "n_qTcut" << setw(5) << "" << "=" << setw(5) << "" << osi.n_qTcut << endline;
  s << setw(38) << right << "min_qTcut" << setw(5) << "" << "=" << setw(5) << "" << osi.min_qTcut << endline;
  s << setw(38) << right << "step_qTcut" << setw(5) << "" << "=" << setw(5) << "" << osi.step_qTcut << endline;
  s << setw(38) << right << "switch_distribution_at_all_TSV" << setw(5) << "" << "=" << setw(5) << "" << osi.switch_distribution_at_all_TSV << endline;

    s << "observable_set output:   n_set_TSV = " << osi.n_set_TSV << endline << endline;

  if (osi.switch_TSV == 0){
    s << "observable_set output:   TSV has been switched off." << endline << endline;
  }
  else {
    s << "observable_set output:   n_set_TSV = " << osi.n_set_TSV << endline << endline;
    for (int i_s = 0; i_s < osi.n_set_TSV; i_s++){
      s << setw(35) << right << "name_set_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.name_set_TSV[i_s] << endline;
      s << setw(35) << right << "dynamic_scale_ren_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.dynamic_scale_ren_TSV[i_s] << endline;
      s << setw(35) << right << "central_scale_ren_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.central_scale_ren_TSV[i_s] << endline;
      s << setw(35) << right << "relative_central_scale_ren_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.relative_central_scale_ren_TSV[i_s] << endline;
      s << setw(35) << right << "n_scale_ren_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.n_scale_ren_TSV[i_s] << endline;
      s << setw(35) << right << "factor_scale_ren_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.factor_scale_ren_TSV[i_s] << endline;
      s << setw(35) << right << "dynamic_scale_fact_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.dynamic_scale_fact_TSV[i_s] << endline;
      s << setw(35) << right << "central_scale_fact_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.central_scale_fact_TSV[i_s] << endline;
      s << setw(35) << right << "relative_central_scale_fact_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.relative_central_scale_fact_TSV[i_s] << endline;
      s << setw(35) << right << "n_scale_fact_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.n_scale_fact_TSV[i_s] << endline;
      s << setw(35) << right << "factor_scale_fact_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.factor_scale_fact_TSV[i_s] << endline;
      s << setw(35) << right << "min_qTcut_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.min_qTcut_TSV[i_s] << endline;
      s << setw(35) << right << "max_qTcut_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.max_qTcut_TSV[i_s] << endline;
      s << setw(35) << right << "switch_distribution_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.switch_distribution_TSV[i_s] << endline;
      s << setw(35) << right << "max_n_integrand_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.max_n_integrand_TSV[i_s] << endline;
      s << setw(35) << right << "min_qTcut_distribution_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.min_qTcut_distribution_TSV[i_s] << endline;
      s << setw(35) << right << "max_qTcut_distribution_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.max_qTcut_distribution_TSV[i_s] << endline;
      s << setw(35) << right << "switch_moment_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.switch_moment_TSV[i_s] << endline;
      s << setw(35) << right << "filename_integration_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.filename_integration_TSV[i_s] << endline;
      s << setw(35) << right << "filename_result_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.filename_result_TSV[i_s] << endline;
      s << setw(35) << right << "filename_distribution_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.filename_distribution_TSV[i_s] << endline;
      s << setw(35) << right << "filename_moment_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.filename_moment_TSV[i_s] << endline;
      s << endline;
    }
    for (int i_s = 0; i_s < osi.n_extended_set_TSV; i_s++){
      s << setw(35) << right << "name_extended_set_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.name_extended_set_TSV[i_s] << endline;
      s << setw(35) << right << "n_scale_ren_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.n_scale_ren_TSV[i_s] << endline;
      s << setw(35) << right << "n_scale_fact_TSV[" << setw(2) << i_s << "]" << setw(5) << "" << "=" << setw(5) << "" << osi.n_scale_fact_TSV[i_s] << endline;
    }

    s << "observable_set output:   n_set_TSV = " << osi.n_set_TSV << endline << endline;
  }
  return s;
}


