#include "header.hpp"

// Why not references in many arguments - or inline functions ??? !!!

/////////////////////
//  angle mapping  //
/////////////////////

double phasespace_set::c_phi(double r1){
  static Logger logger("phasespace_set::c_phi");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double phi = 2 * pi * r1;

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << phi << endline_flush;
  return phi;
}

double phasespace_set::inv_phi(double & phi){
  static Logger logger("phasespace_set::inv_phi");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double r_phi = phi / (2 * pi);

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << r_phi << endline_flush;
  return r_phi;
}

double phasespace_set::c_costheta(double & r1){
  static Logger logger("phasespace_set::c_costheta");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double costheta = 2 * r1 - 1.;

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << costheta << endline_flush;
  return costheta;
}

double phasespace_set::inv_costheta(double & costheta){
  static Logger logger("phasespace_set::inv_costheta");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double r_costheta = (costheta + 1.) / 2;

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << r_costheta << endline_flush;
  return r_costheta;
}


double phasespace_set::c_costheta_pTcut(double & r1, double & costheta_pTcut, double n){
  static Logger logger("phasespace_set::c_costheta");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double costheta;
  if (r1 < 0.25){
    costheta = -costheta_pTcut - pow(1. - 4 * r1, n) * (1. - costheta_pTcut);
  }
  else if (r1 < 0.5){
    costheta = -costheta_pTcut + pow(4 * r1 - 1., n) * costheta_pTcut;
  }
  else if (r1 < 0.75){
    costheta = costheta_pTcut - pow(3. - 4 * r1, n) * costheta_pTcut;
  }
  else {
    costheta = costheta_pTcut + pow(4 * r1 - 3., n) * (1. - costheta_pTcut);
  }

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << costheta << endline_flush;
  return costheta;
}


double phasespace_set::g_costheta_pTcut(double & costheta, double & costheta_pTcut, double n){
  static Logger logger("phasespace_set::c_costheta");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double g_costheta;
  if (costheta < -costheta_pTcut){
    g_costheta = 1. / (4 * n * pow(1. - costheta_pTcut, 1. / n) * pow(-(costheta + costheta_pTcut), 1. - 1. / n));
  }
  else if (costheta < 0.){
    g_costheta = 1. / (4 * n * pow(costheta_pTcut, 1. / n) * pow(costheta_pTcut + costheta, 1. - 1. / n));
  }
  else if (costheta < costheta_pTcut){
    g_costheta = 1. / (4 * n * pow(costheta_pTcut, 1. / n) * pow(costheta_pTcut - costheta, 1. - 1. / n));
  }
  else {
    g_costheta = 1. / (4 * n * pow(1. - costheta_pTcut, 1. / n) * pow(costheta - costheta_pTcut, 1. - 1. / n));
  }

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << costheta << endline_flush;
  return g_costheta;
}


double phasespace_set::inv_costheta_pTcut(double & costheta, double & costheta_pTcut, double n){
  static Logger logger("phasespace_set::inv_costheta");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double r_costheta;
  if (costheta < -costheta_pTcut){
    r_costheta = 0.25 - 0.25 * pow(-(costheta + costheta_pTcut) / (1. - costheta_pTcut), 1. / n);
  }
  else if (costheta < 0.){
    r_costheta = 0.25 + 0.25 * pow((costheta + costheta_pTcut) / costheta_pTcut, 1. / n);
  }
  else if (costheta < costheta_pTcut){
    r_costheta = 0.75 - 0.25 * pow((costheta_pTcut - costheta) / costheta_pTcut, 1. / n);
  }
  else {
    r_costheta = 0.75 + 0.25 * pow((costheta - costheta_pTcut) / (1. - costheta_pTcut), 1. / n);
  }

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << r_costheta << endline_flush;
  return r_costheta;
}


//////////////////////////
//  propagator mapping  //
//////////////////////////

void phasespace_set::c_propagator(int i_a, int m, int out, double & smin, double & smax, double r1){
  static Logger logger("phasespace_set::c_propagator");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "s[" << out << "] with m = " << m << "    map_Gamma = " << map_Gamma[abs(m)] << "    M2 = " << M2[abs(m)] << "   smin = " << smin << "   smax = " << smax << endline_flush;

  // Just for compatibility; could be removed now, I guess !!!
  //  m = abs(m);

  if (map_Gamma[m] == 0){
    double mass2;
    if (M2[m] == 0.){mass2 = mass0;}
    else {mass2 = M2[m] + mass0;}
    xbs_all[i_a][out] = c_propagator_vanishing_width(r1, mass2, smin, smax, nuxs);
  }
  else {
    xbs_all[i_a][out] = c_propagator_Breit_Wigner(r1, m, smin, smax);
  }
  xbsqrts_all[i_a][out] = sqrt(xbs_all[i_a][out]);

  if (munich_isnan(xbsqrts_all[i_a][out])) {
    logger << LOG_WARN << "xbsqrts_all[" << i_a << "][" << out << "] is NaN (r1 =" << r1 << ", m = " << m << ", smin = " << smin << ", smax = " << smax << ", map_Gamma[" << m << "] = " << map_Gamma[m] << ", map_Gamma[" << m << "] = " << map_Gamma[m] << endline_flush;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


double phasespace_set::g_propagator(int i_a, int m, int out, double smin, double smax){
  Logger logger("phasespace_set::g_propagator");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "s[" << out << "] with m = " << m << "    map_Gamma = " << map_Gamma[abs(m)] << "    M2 = " << M2[abs(m)] << endline_flush;

  // !!! What should happen in this case ???
  if (switch_console_output_phasespace_issue){
    if (smax <= smin){
      logger << LOG_WARN << "smin = " << smin << "   smax = " << smax << ", m = " << m << ", out = " << out << endline_flush;
      for (int i = 0; i < xbs_all[i_a].size(); i++){
	if (xbs_all[i_a][i] != 0.){
	  logger << LOG_WARN << "xbs_all[" << i_a << "][" << i << "] = " << xbs_all[i_a][i] << endline_flush;
	}
      }
    }
  }

  double gs;

  // Just for compatibility; could be removed now, I guess !!!
  //  m = abs(m);

  if (map_Gamma[m] == 0){
    double mass2;
    if (M2[m] == 0.){mass2 = mass0;}
    else {mass2 = M2[m] + mass0;}
    gs = g_propagator_vanishing_width(xbs_all[i_a][out], mass2, smin, smax, nuxs);
  }
  else {
    gs = g_propagator_Breit_Wigner(xbs_all[i_a][out], m, smin, smax);
  }

  if (munich_isnan(gs)){logger << LOG_WARN << "smin = " << smin << "   s[" << out << "] = " << xbs_all[i_a][out] << "   smax = " << smax << "   gs = " << gs << "   gtot = nan from s-channel mapping" << endline_flush;}
  if (gs < 0.){logger << LOG_WARN << "gs < 0.   smin = " << smin << "   s[" << out << "] = " << xbs_all[i_a][out] << "   smax = " << smax << "   gs = " << gs << "   M[" << m << "] = " << M[m] << "   map_Gamma[" << m << "] = " << map_Gamma[m] << endline_flush;}

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << gs << endline_flush;
  return gs;
}

void phasespace_set::inv_propagator(int i_a, int m, int out, double smin, double smax, double & r1){
  static Logger logger("phasespace_set::inv_propagator");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // Just for compatibility; could be removed now, I guess !!!
  //  m = abs(m);

  if (map_Gamma[m] == 0){
    double mass2;
    if (M2[m] == 0.){mass2 = mass0;}
    else {mass2 = M2[m] + mass0;}
    r1 = inv_propagator_vanishing_width(xbs_all[i_a][out], mass2, smin, smax, nuxs);
  }
  else {
    r1 = inv_propagator_Breit_Wigner(xbs_all[i_a][out], m, smin, smax);
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


///////////////////////////////////////
//  Breit-Wigner propagator mapping  //
///////////////////////////////////////

double phasespace_set::c_propagator_Breit_Wigner(double & r1, int m, double & smin, double & smax){
  Logger logger("phasespace_set::c_propagator_Breit_Wigner");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // see Bronstein, Sec. 2.8.7, Eq. (2.7)
  double arg1 = (smin - M2[m]) / (M[m] * map_Gamma[m]);
  double arg2 = (smax - M2[m]) / (M[m] * map_Gamma[m]);
  double delta_y;
  if (arg1 * arg2 > -1.){delta_y = atan((arg2 - arg1) / (1 + arg1 * arg2));}
  else if (arg2 > 0.){delta_y = M_PI + atan((arg2 - arg1) / (1. + arg1 * arg2));}
  else {delta_y = -M_PI + atan((arg2 - arg1) / (1 + arg1 * arg2));}
  // see Bronstein, Sec. 2.7.2.3, Eq. (2.87)
  double s = M[m] * map_Gamma[m] * (arg1 + tan(r1 * delta_y)) / (1 - arg1 * tan(r1 * delta_y)) + M2[m];

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << s << endline_flush;
  return s;
}

double phasespace_set::g_propagator_Breit_Wigner(double & s, int m, double & smin, double & smax){
  Logger logger("phasespace_set::g_propagator_Breit_Wigner");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double arg1 = (smin - M2[m]) / (M[m] * map_Gamma[m]);
  double arg2 = (smax - M2[m]) / (M[m] * map_Gamma[m]);
  double delta_y;
  if (arg1 * arg2 > -1.){delta_y = atan((arg2 - arg1)/(1 + arg1 * arg2));}
  else if (arg2 > 0.){delta_y = M_PI + atan((arg2 - arg1) / (1 + arg1 * arg2));}
  else {delta_y = -M_PI + atan((arg2 - arg1) / (1 + arg1 * arg2));}
  double gs = (M[m] * map_Gamma[m]) / (delta_y * (pow(s - M2[m], 2) + M2[m] * pow(map_Gamma[m], 2)));

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << gs << endline_flush;
  return gs;
}

double phasespace_set::inv_propagator_Breit_Wigner(double s, int m, double & smin, double & smax){
  Logger logger("phasespace_set::inv_propagator_Breit_Wigner");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // TODO: micro optimize as above !!!
  double arg1 = atan((smin - M2[m]) / (M[m] * map_Gamma[m]));
  double arg2 = atan((smax - M2[m]) / (M[m] * map_Gamma[m]));
  double r = (atan((s - M2[m]) / (M[m] * map_Gamma[m])) - arg1) / (arg2 - arg1);
  if (munich_isnan(r)) {
    logger << LOG_DEBUG << "r is NaN in inv_propagator_Breit_Wigner!" << endline_flush;
    logger << LOG_DEBUG << "    r = " << setw(23) << setprecision(15) << r << ";            smin = " << setw(23) << setprecision(15) << smin << ";      s = " << setw(23) << setprecision(15) << s << ";   smax = " << setw(23) << setprecision(15) << smax << endline_flush;
    logger << LOG_DEBUG << "M[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << M[m] << ";   map_Gamma[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << map_Gamma[m] << ";   arg1 = " << setw(23) << setprecision(15) << arg1 << ";   arg2 = " << setw(23) << setprecision(15) << arg2 << endline_flush;
    if (s <= smin){r = map_technical_s;}
    else if (s >= smax){r = 1. - map_technical_s;}
    else {
      if (switch_console_output_phasespace_issue){
        logger << LOG_WARN << "   smin = " << smin << "   s = " << s << "   smax = " << smax << endline;
	logger << LOG_WARN << "   arg1 = " << arg1 << "   arg2 = " << arg2 << endline;
	logger << LOG_WARN << "   M[" << m << "] = " << M[m] << "   M2[" << m << "] = " << M2[m] << "   map_Gamma[" << m << "] = " << map_Gamma[m] << endline_flush;
        logger << LOG_ERROR << "r remains NaN in inv_propagator_Breit_Wigner!" << endline_flush;
      }
      r = map_technical_s;
      // Otherwise program might stop !!! Reason for "nan" should be clarified !!!
    }
  }
  else if (r <= 0.) {
    logger << LOG_DEBUG << "r <= 0 in inv_propagator_Breit_Wigner!" << endline_flush;
    logger << LOG_DEBUG << "    r = " << setw(23) << setprecision(15) << r << ";            smin = " << setw(23) << setprecision(15) << smin << ";      s = " << setw(23) << setprecision(15) << s << ";   smax = " << setw(23) << setprecision(15) << smax << endline_flush;
    logger << LOG_DEBUG << "M[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << M[m] << ";   map_Gamma[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << map_Gamma[m] << ";   arg1 = " << setw(23) << setprecision(15) << arg1 << ";   arg2 = " << setw(23) << setprecision(15) << arg2 << endline_flush;
    r = map_technical_s;
  }
  else if (r >= 1.) {
    logger << LOG_DEBUG << "r >= 1 in inv_propagator_Breit_Wigner!" << endline_flush;
    logger << LOG_DEBUG << "    r = " << setw(23) << setprecision(15) << r << ";            smin = " << setw(23) << setprecision(15) << smin << ";      s = " << setw(23) << setprecision(15) << s << ";   smax = " << setw(23) << setprecision(15) << smax << endline_flush;
    logger << LOG_DEBUG << "M[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << M[m] << ";   map_Gamma[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << map_Gamma[m] << ";   arg1 = " << setw(23) << setprecision(15) << arg1 << ";   arg2 = " << setw(23) << setprecision(15) << arg2 << endline_flush;
    r = 1. - map_technical_s;
  }

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << r << endline_flush;
  return r;
}


///////////////////////////////////////
//  narrow-width propagator mapping  //
///////////////////////////////////////

double phasespace_set::c_propagator_narrow_width(){return 0.;}

double phasespace_set::g_propagator_narrow_width(int m){
  Logger logger("phasespace_set::g_propagator_narrow_width");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double gs = 1. / (pi * M[m] * map_Gamma[m]);

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << gs << endline_flush;
  if (gs < 0.){logger << LOG_DEBUG << "gs < 0.   gs = " << gs << "   M[" << m << "] = " << M[m] << "   map_Gamma[" << m << "] = " << map_Gamma[m] << endline_flush;}
  return gs;
}

double phasespace_set::inv_propagator_narrow_width(){return 0.;}


//////////////////////////////////////////
//  vanishing-width propagator mapping  //
//////////////////////////////////////////

double phasespace_set::c_propagator_vanishing_width(double & r, double & mass2, double & smin, double & smax, double & nu){
  Logger logger("c_propagator_vanishing_width");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double s;
  if (nu != 1.){s = pow(r * pow(smax - mass2, 1. - nu) + (1. - r) * pow(smin - mass2, 1. - nu), (1. / (1. - nu))) + mass2; }
  else {s = exp(r * log(smax - mass2) + (1. - r) * log(smin - mass2)) + mass2;}
  logger << LOG_DEBUG_VERBOSE << "s = " << s << endline_flush;
  if (munich_isnan(s)) {
    logger << LOG_DEBUG << "s      = " << s << endline_flush;
    logger << LOG_DEBUG << "smin   = " << smin << endline_flush;
    logger << LOG_DEBUG << "smax   = " << smax << endline_flush;
    logger << LOG_DEBUG << "mass2  = " << mass2 << endline_flush;
    logger << LOG_DEBUG << "nu     = " << nu << endline_flush;
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << s << endline_flush;
  return s;
}

double phasespace_set::g_propagator_vanishing_width(double s, double mass2, double smin, double smax, double nu){
  Logger logger("g_propagator_vanishing_width");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "smin = " << smin << "  s = " << s << "   smax = " << smax << "   mass2 = " << mass2 << "   smin - mass0 = " << smin - mass2 << " ^(1-nu = " << nu << ") = " << pow(smin - mass2, 1. - nu) << endline_flush;
  double gs;
  if (smax <= smin) {
    if (switch_console_output_phasespace_issue){
      logger << LOG_WARN << "smin = " << smin << "  s = " << s << "   smax = " << smax << "   mass2 = " << mass2 << "   smin - mass0 = " << smin-mass2 << " ^(1-nu = " << nu << ") = " << pow(smin-mass2,1-nu) << endline_flush;
      logger << LOG_ERROR << "gs is set to infinity." << endline_flush;
    }
    gs = std::numeric_limits<double>::infinity();
    //    gs = 1. / 0.; //1.e99;
  }
  else {
    if (nu != 1.){gs = (1. - nu) / ((pow(smax - mass2, 1. - nu) - pow(smin - mass2, 1. - nu)) * pow(s - mass2, nu));}
    else {gs = 1. / ((log(smax - mass2) - log(smin - mass2)) * (s - mass2));}
  }

  if (gs < 0.){logger << LOG_DEBUG << "gs < 0.   smin = " << smin << "   s = " << s << "   smax = " << smax << "   gs = " << gs << "   mass2 = " << mass2 << endline_flush;}

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << gs << endline_flush;
  return gs;
}

double phasespace_set::inv_propagator_vanishing_width(double s, double mass2, double smin, double smax, double nu){
  Logger logger("inv_propagator_vanishing_width");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double r;
  if (nu != 1.){
    double sminm0 = pow(smin - mass2, 1. - nu);
    r = (pow(s - mass2, 1. - nu) - sminm0) / (pow(smax - mass2, 1. - nu) - sminm0);
  }
  else {r = log((s - mass2) / (smin - mass2)) / log((smax - mass2) / (smin - mass2));}

  if (munich_isnan(r)) {
    logger << LOG_DEBUG << "smin   = " << smin << endline_flush;
    logger << LOG_DEBUG << "sminm0 = " << pow(smin - mass2, 1. - nu) << endline_flush;
    logger << LOG_DEBUG << "smax   = " << smax << endline_flush;
    logger << LOG_DEBUG << "mass2  = " << mass2 << endline_flush;
    logger << LOG_DEBUG << "s      = " << s << endline_flush;
    logger << LOG_DEBUG << "nu     = " << nu << endline_flush;
    logger << LOG_DEBUG << "r is NaN in inv_propagator_vanishing_width!" << endline_flush;
    logger << LOG_DEBUG << "r = " << r << "; smin, s, smax=" << smin << ", " << s << ", " << smax << endline_flush;
    if (s <= smin){r = map_technical_r;}
    else if (s >= smax){r = 1. - map_technical_r;}
    else {
      if (switch_console_output_phasespace_issue){
        logger << LOG_ERROR << "r remains NaN in inv_propagator_vanishing_width!" << endline_flush;
	logger << LOG_WARN << "r = " << r << "; smin = " << smin << ", s = " << s << ", smax = " << smax << "   mass2 = " << mass2 << "   nu = " << nu << endline_flush;
      }
      r = map_technical_r;
      // Otherwise program might stop !!! Reason for "nan" should be clarified !!!
    }
  }
  else if (r <= 0.){
    logger << LOG_DEBUG << "r <= 0 in inv_propagator_vanishing_width!" << endline_flush;
    logger << LOG_DEBUG << "r = " << r << "; smin = " << smin << ", s = " << s << ", smax = " << smax << "   mass2 = " << mass2 << "   nu = " << nu << endline_flush;
    r = map_technical_r;
  }
  if (r >= 1.){
    logger << LOG_DEBUG << "r >= 1 in inv_propagator_vanishing_width!" << endline_flush;
    logger << LOG_DEBUG << "r = " << r << "; smin = " << smin << ", s = " << s << ", smax = " << smax << "   mass2 = " << mass2 << "   nu = " << nu << endline_flush;
    r = 1. - map_technical_r;
  }

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << r << endline_flush;
  return r;
}


////////////////////////////////////////
//  no-propagator-equivalent mapping  //
////////////////////////////////////////

void phasespace_set::c_timelikeinvariant(int i_a, int out, double & smin, double & smax, double r1){
  static Logger logger("phasespace_set::c_timelikeinvariant");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  xbs_all[i_a][out] = r1 * smax + (1. - r1) * smin;
  xbsqrts_all[i_a][out] = sqrt(xbs_all[i_a][out]);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

double phasespace_set::g_timelikeinvariant(double smin, double smax){
  static Logger logger("phasespace_set::g_timelikeinvariant");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double gs = 1. / (smax - smin);

  if (gs < 0.){logger << LOG_DEBUG << "gs < 0.   smin = " << smin << "   smax = " << smax << "   gs = " << gs << endline_flush;}

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << gs << endline_flush;
  return gs;
}

void phasespace_set::inv_timelikeinvariant(int i_a, int out, double smin, double smax, double & r1){
  static Logger logger("inv_timelikeinvariant");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  r1 = (xbs_all[i_a][out] - smin) / (smax - smin);

  //  static double temp_map_technical_r = 1.e-8; -> map_technical_r
  if (munich_isnan(r1)) {
    logger << LOG_DEBUG << "r1 is NaN in inv_timelikeinvariant!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; smin, s, smax=" << smin << ", " << xbs_all[i_a][out] << ", " << smax << endline_flush;
    if (xbs_all[i_a][out] <= smin){r1 = map_technical_r;}
    else if (xbs_all[i_a][out] >= smax){r1 = 1. - map_technical_r;}
    else {
      if (switch_console_output_phasespace_issue){
        logger << LOG_ERROR << "r1 remains NaN in inv_timelikeinvariant!" << endline_flush;
      }
      r1 = map_technical_r;
      // Otherwise program might stop !!! Reason for "nan" should be clarified !!!
    }
  }
  else if (r1 <= 0.){
    logger << LOG_DEBUG << "r1 < 0 in inv_timelikeinvariant!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; smin, s, smax=" << smin << ", " << xbs_all[i_a][out] << ", " << smin << endline_flush;
    r1 = map_technical_r;
  }
  if (r1 >= 1.){
    logger << LOG_DEBUG << "r1 > 1 in inv_timelikeinvariant!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; smin, s, smax=" << smin << ", " << xbs_all[i_a][out] << ", " << smin << endline_flush;
    r1 = 1. - map_technical_r;
  }
}


////////////////////////////////////
//  t-channel propagator mapping  //
////////////////////////////////////

double phasespace_set::c_t_propagator(double & r1, int m, double smin, double smax){
  static Logger logger("c_t_propagator");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double s, mass2, nuexp;
  if (M2[m] == 0.){
    mass2 = mass0;
    nuexp = nuxt;
  }
  else {
    mass2 = -M2[m];// + mass0;
    nuexp = nuxt;
  }
  if (nuexp != 1.){s = pow(r1 * pow(smax - mass2, 1 - nuexp) + (1 - r1) * pow(smin - mass2, 1. - nuexp), (1. / (1. - nuexp))) + mass2;}
  else {s = exp(r1 * log(smax - mass2) + (1. - r1) * log(smin - mass2)) + mass2;}

  if (s < 0.){logger << LOG_DEBUG << "s < 0.   smin = " << smin << "   s = " << s << "   smax = " << smax << endline_flush;}

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << s << endline_flush;
  return s;
}

double phasespace_set::g_t_propagator(double s, int m, double smin, double smax){
  static Logger logger("g_t_propagator");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double gs, mass2, nuexp;
  if (M2[m] == 0.){
    mass2 = mass0;
    nuexp = nuxt;
  }
  else {
    mass2 = -M2[m];// + mass0;
    nuexp = nuxt;
  }
  if (nuexp != 1.){gs = (1. - nuexp) / ((pow(smax - mass2, 1. - nuexp) - pow(smin - mass2, 1. - nuexp)) * pow(s - mass2, nuexp));}
  else {gs = 1. / ((log(smax - mass2) - log(smin - mass2)) * (s - mass2));}

  if (munich_isnan(gs)) {
    logger << LOG_DEBUG << "gs == nan in g_t_propagator!   -tmax = " << setprecision(15) << setw(23) << smin << "   -t = " << setprecision(15) << setw(23) << s << "   -tmin = " << setprecision(15) << setw(23) << smax << endline_flush;
  }

  if (gs > 0.){logger << LOG_DEBUG << "gs > 0.   smin = " << smin << "   s = " << s << "   smax = " << smax << "   gs = " << gs << "   gtot = nan from s-channel mapping" << endline_flush;}

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << gs << endline_flush;
  return gs;
}

double phasespace_set::inv_t_propagator(double s, int m, double smin, double smax){
  static Logger logger("inv_t_propagator");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double r, mass2, nuexp;
  if (M2[m] == 0.){
    mass2 = mass0;
    nuexp = nuxt;
  }
  else {
    mass2 = -M2[m];// + mass0;
    nuexp = nuxt;
  }
  if (nuexp != 1.) {
    double sminm0 = pow(smin - mass2, 1. - nuexp);
    r = (pow(s - mass2, 1. - nuexp) - sminm0) / (pow(smax - mass2, 1. - nuexp) - sminm0);
  }
  else {r = log((s - mass2) / (smin - mass2)) / log((smax - mass2) / (smin - mass2));}

  //  static double temp_map_technical_r = 1.e-8; -> map_technical_t
  if (munich_isnan(r)) {
    logger << LOG_DEBUG << "r is NaN in inv_t_propagator!  r = " << setprecision(15) << setw(23) << r << "   smin = " << setprecision(15) << setw(23) << smin << "   s = " << setprecision(15) << setw(23) << s << "   smax = " << setprecision(15) << setw(23) << smax << endline_flush;
    if (s <= smin){r = map_technical_t;}
    else if (s >= smax){r = 1. - map_technical_t;}
    else {
      if (switch_console_output_phasespace_issue){
	logger << LOG_ERROR << "r remains NaN in inv_t_propagator!" << endline_flush;
      }
      r = map_technical_t;
      // Otherwise program might stop !!! Reason for "nan" should be clarified !!!
    }
  }
  else if (r <= 0.){
    logger << LOG_DEBUG << "r <= 0 in inv_t_propagator!  r = " << setprecision(15) << setw(23) << r << "   -tmax = " << setprecision(15) << setw(23) << smin << "   -t = " << setprecision(15) << setw(23) << s << "   -tmin = " << setprecision(15) << setw(23) << smax << endline_flush;
    r = map_technical_t;
  }
  else if (r >= 1.){
    logger << LOG_DEBUG << "r >= 1 in inv_t_propagator!  r = " << setprecision(15) << setw(23) << r << "   -tmax = " << setprecision(15) << setw(23) << smin << "   -t = " << setprecision(15) << setw(23) << s << "   -tmin = " << setprecision(15) << setw(23) << smax << endline_flush;
    r = 1. - map_technical_t;
  }

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << r << endline_flush;
  return r;
}


/////////////////////
//  decay mapping  //
/////////////////////

void phasespace_set::c_decay(int i_a, int in, int out1, int out2, double r1, double r2){
  Logger logger("phasespace_set::c_decay");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double costheta;
  if (csi->class_contribution_QT_powercorrection && user->double_value[user->double_map["exp_cth"]] != 0.){
    //    double pT_max = (sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])) / (2 * xbsqrts_all[i_a][in]);
    double pT2_cut = pow(user->double_value[user->double_map["threshold_pT"]], 2);
    double pT2_max = lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2]) / (4 * xbs_all[i_a][in]);
    logger << LOG_DEBUG_POINT << "pT2_max = " << setw(25) << setprecision(15) << pT2_max << "   " << "pT2_cut = " << setw(25) << setprecision(15) << pT2_cut << endline_flush;
    if (pT2_cut <= pT2_max){
      double costheta_pTcut = sqrt(1. - pT2_cut / pT2_max);
      costheta = c_costheta_pTcut(r1, costheta_pTcut, user->double_value[user->double_map["exp_cth"]]);
      logger << LOG_DEBUG_POINT << "costheta = " << setw(25) << setprecision(15) << costheta << "   " << "costheta_pTcut = " << setw(25) << setprecision(15) << costheta_pTcut << endline_flush;
    }
    else {costheta = c_costheta(r1);}
  }
  else {costheta = c_costheta(r1);}
  //  double costheta = c_costheta(r1);
  double phi = c_phi(r2);
  fourvector k10(((xbs_all[i_a][in] + xbs_all[i_a][out1] - xbs_all[i_a][out2]) / (2 * xbsqrts_all[i_a][in])), 0., 0., (sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])) / (2 * xbsqrts_all[i_a][in])));
  xbp_all[i_a][out1] = k10.rotation(phi, costheta);
  xbp_all[i_a][out1] = xbp_all[i_a][out1].boost(xbp_all[i_a][in].Pinv(), xbs_all[i_a][in]);
  xbp_all[i_a][out2] = xbp_all[i_a][in] - xbp_all[i_a][out1];
  logger << LOG_DEBUG_POINT << "xbp_all[" << i_a << "][" << out1 << "].pT() = " << setw(25) << setprecision(15) << xbp_all[i_a][out1].pT() << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

double phasespace_set::g_decay(int i_a, int in, int out1, int out2){
  Logger logger("phasespace_set::g_decay");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double gd;
  if (csi->class_contribution_QT_powercorrection && user->double_value[user->double_map["exp_cth"]] != 0.){
    double pT2_cut = pow(user->double_value[user->double_map["threshold_pT"]], 2);
    double pT2_max = lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2]) / (4 * xbs_all[i_a][in]);
    if (pT2_cut <= pT2_max){
      fourvector k10 = xbp_all[i_a][out1].boost(xbp_all[i_a][in], xbs_all[i_a][in]);
      double costheta = k10.x3() / k10.r();
      double costheta_pTcut = sqrt(1. - pT2_cut / pT2_max);
      double g_costheta = g_costheta_pTcut(costheta, costheta_pTcut, user->double_value[user->double_map["exp_cth"]]);
      gd = g_costheta * (4 * xbs_all[i_a][in]) / (pi * sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])));
    }
    else {
      gd = (2 * xbs_all[i_a][in]) / (pi * sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])));
    }
  }
  else {
    gd = (2 * xbs_all[i_a][in]) / (pi * sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])));
  }
  //  double gd = (2 * xbs_all[i_a][in]) / (pi * sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])));
  if (gd < 0.){logger << LOG_DEBUG << "gd < 0.   in = " << in << "   out1 = " << out1 << "   out2 = " << out2 << "   xbs_all[" << i_a << "][" << in << "] = " << xbs_all[i_a][in] << endline_flush;}

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << gd << endline_flush;
  return gd;
}


void phasespace_set::inv_decay(int i_a, int in, int out1, int out2, double & r1, double & r2){
  static Logger logger("phasespace_set::inv_decay");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  fourvector k10 = xbp_all[i_a][out1].boost(xbp_all[i_a][in], xbs_all[i_a][in]);

  double phi = 0.;
  double costheta = 0.;
  if (k10.x1() != 0 || k10.x2() != 0) {
    phi = atan2(k10.x2(), k10.x1());
    if (phi < 0.){phi += 2 * pi;}
  }
  else {
    logger << LOG_DEBUG << "x1=x2=0 in inv_decay" << endline_flush;
    phi = atan2(1., 1.);
  }

  //  costheta=k10.x3()/(sqrt(lambda(s[in], s[out1], s[out2])) / (2 * sqrt(s[in])));
  costheta = k10.x3() / k10.r();

  r2 = inv_phi(phi);

  if (csi->class_contribution_QT_powercorrection && user->double_value[user->double_map["exp_cth"]] != 0.){
    double pT2_cut = pow(user->double_value[user->double_map["threshold_pT"]], 2);
    double pT2_max = lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2]) / (4 * xbs_all[i_a][in]);
    if (pT2_cut <= pT2_max){
      double costheta_pTcut = sqrt(1. - pT2_cut / pT2_max);
      r1 = inv_costheta_pTcut(costheta, costheta_pTcut, user->double_value[user->double_map["exp_cth"]]);
    }
    else {r1 = inv_costheta(costheta);}
  }
  else {r1 = inv_costheta(costheta);}
  // r1 = inv_costheta(costheta);

  // Would this work as well ???
  //  r2 = inv_phi(k10.phi());
  //  r1 = inv_costheta(k10.costheta());

  if (munich_isnan(r1)) {
    logger << LOG_DEBUG << "r1 is NaN in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; k10.x, k10.y = " << k10.x1() << ", " << k10.x2() << endline_flush;
    r1 = 0.5; // why ???
  }
  else if (r1 <= 0.){
    logger << LOG_DEBUG << "r1 < 0 in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; k10.x, k10.y = " << k10.x1() << ", " << k10.x2() << endline_flush;
    r1 = map_technical_r;
  }
  else if (r1 >= 1.){
    logger << LOG_DEBUG << "r1 > 1 in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 << "; k10.x, k10.y=" << k10.x1() << ", " << k10.x2() << endline_flush;
    r1 = 1. - map_technical_r;
  }

  if (munich_isnan(r2)){
    logger << LOG_DEBUG << "r2 is NaN in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r2 = " << r2 <<"; k10.z, r = " << k10.x3() << ", " << k10.r() << endline_flush;
    logger << LOG_DEBUG << xbp_all[i_a][in] << endline_flush;
    r2 = 0.5;
  }
  else if (r2 <= 0.){
    logger << LOG_DEBUG << "r2 < 0 in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r2 = " << r2 <<"; k10.z, r = " << k10.x3() << ", " << k10.r() << endline_flush;
    logger << LOG_DEBUG << xbp_all[i_a][in] << endline_flush;
    r2 = map_technical_r;
  }
  else if (r2 >= 1.){
    logger << LOG_DEBUG << "r2 > 1 in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r2 = " << r2 <<"; k10.z, r=" << k10.x3() << ", " << k10.r() << endline_flush;
    logger << LOG_DEBUG << xbp_all[i_a][in] << endline_flush;
    r2 = 1. - map_technical_r;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


/////////////////////////
//  t-channel mapping  //
/////////////////////////

// Unify name : 'in' or 'out' !!!

void phasespace_set::c_tchannel(int i_a, int m, int in, int in1, int in2, int out1, int out2, double r1, double r2){
  static Logger logger("phasespace_set::c_tchannel");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double term1 = (xbs_all[i_a][in] + xbs_all[i_a][out1] - xbs_all[i_a][out2]) * (xbs_all[i_a][in] + xbs_all[i_a][in1] - xbs_all[i_a][in2]);
  double term2 = sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])) * sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][in1], xbs_all[i_a][in2]));
  double term3 = (2 * sqrt(xbs_all[i_a][in])); // !!! why not xbsqrts_all[i_a][in] ??? - check if calculated ...
  double tmin = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 + term2) / (2 * xbs_all[i_a][in]);
  double tmax = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 - term2) / (2 * xbs_all[i_a][in]);
  xbs_all[i_a][in1 + out1] = -c_t_propagator(r1, m, -tmin, -tmax);
  double phi = c_phi(r2);
  double costheta = ((2 * xbs_all[i_a][in] * (xbs_all[i_a][in1 + out1] - xbs_all[i_a][out1] - xbs_all[i_a][in1])) + term1) / term2;
  fourvector p10(((xbs_all[i_a][in] + xbs_all[i_a][in1] - xbs_all[i_a][in2]) / term3), 0, 0, (sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][in1], xbs_all[i_a][in2])) / term3));
  fourvector p20(((xbs_all[i_a][in] - xbs_all[i_a][in1] + xbs_all[i_a][in2]) / term3), 0, 0, -(sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][in1], xbs_all[i_a][in2])) / term3));
  fourvector k10(((xbs_all[i_a][in] + xbs_all[i_a][out1] - xbs_all[i_a][out2]) / term3), 0, 0, (sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])) / term3));
  k10 = k10.rotation_beam(phi, costheta);
  fourvector k20 = p10 + p20 - k10;
  fourvector p1b = xbp_all[i_a][in1].boost(xbp_all[i_a][in], xbs_all[i_a][in]);
  xbp_all[i_a][out1] = (k10.rotation(p1b)).boost(xbp_all[i_a][in].Pinv(), xbs_all[i_a][in]);
  xbp_all[i_a][out2] = (k20.rotation(p1b)).boost(xbp_all[i_a][in].Pinv(), xbs_all[i_a][in]);
  xbp_all[i_a][in1 + out1] = xbp_all[i_a][in1] - xbp_all[i_a][out1];

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

double phasespace_set::phasespace_set::g_tchannel(int i_a, int m, int in, int in1, int in2, int out1, int out2){
  static Logger logger("phasespace_set::g_tchannel");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double term1 = (xbs_all[i_a][in] + xbs_all[i_a][out1] - xbs_all[i_a][out2]) * (xbs_all[i_a][in] + xbs_all[i_a][in1] - xbs_all[i_a][in2]);
  double term2 = sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])) * sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][in1], xbs_all[i_a][in2]));
  double tmin = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 + term2) / (2 * xbs_all[i_a][in]);
  double tmax = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 - term2) / (2 * xbs_all[i_a][in]);
  if (tmin > 0.){logger << LOG_DEBUG << "tmin = " << tmin << endline_flush;}
  if (tmax > 0.){logger << LOG_DEBUG << "tmax = " << tmax << endline_flush; tmax = 0.;}
  double gp = -g_t_propagator(-xbs_all[i_a][in1 + out1], m, -tmin, -tmax) * (2 * sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][in1], xbs_all[i_a][in2]))) / pi;

  if (gp < 0.){logger << LOG_DEBUG << "gp < 0." << "   in = " << in << "   in1 = " << in1 << "   in2 = " << in2 << "   out1 = " << out1 << "   out2 = " << out2 << endline_flush; gp = 0.;}

  logger << LOG_DEBUG_VERBOSE << "finished - return " << setw(23) << setprecision(15) << gp << endline_flush;
  return gp;
}

void phasespace_set::inv_tchannel(int i_a, int m, int in, int in1, int in2, int out1, int out2, double & r1, double & r2){
  static Logger logger("phasespace_set::inv_tchannel");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double term1 = (xbs_all[i_a][in] + xbs_all[i_a][out1] - xbs_all[i_a][out2]) * (xbs_all[i_a][in] + xbs_all[i_a][in1] - xbs_all[i_a][in2]);
  double term2 = sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][out1], xbs_all[i_a][out2])) * sqrt(lambda(xbs_all[i_a][in], xbs_all[i_a][in1], xbs_all[i_a][in2]));
  double tmin = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 + term2) / (2 * xbs_all[i_a][in]);
  double tmax = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 - term2) / (2 * xbs_all[i_a][in]);

  fourvector p1b = xbp_all[i_a][in1].boost(xbp_all[i_a][in], xbs_all[i_a][in]);

  // invert second part
  fourvector k10 = xbp_all[i_a][out1].boost(xbp_all[i_a][in], xbs_all[i_a][in]);
  k10 = k10.rotation_inverse(p1b);

  // Would this work as well ???
  //  fourvector k10 = (xbp_all[i_a][out1].boost(xbp_all[i_a][in], xbs_all[i_a][in]).rotation_inverse(p1b));

  double phi = 0.;
  if (k10.x1()!=0 || k10.x2()!=0) {
    phi = atan2(k10.x2(),k10.x1());
    if (phi<0)
      phi+=2*M_PI;
  }
  else
    logger << LOG_DEBUG << "x1==x20 in inv_tchannel" << endline_flush;

  r2=phi/(2*M_PI);

  // Would this work as well ???
  //  r2 = inv_phi(k10.phi());
  
  // Set r2 to  map_technical_r  and  1. - map_technical_r  instead ???
  //  if (r2 <= 0){
  if (!(r2>=0)) {
    logger << LOG_DEBUG << "r2<0 in inv_tchannel!" << endline_flush;
    logger << LOG_DEBUG << "r2=" << r2 <<"; k10.x, k10.y=" << k10.x1() << ", " << k10.x2() << endline_flush;
    logger << LOG_DEBUG << xbp_all[i_a][in] << endline_flush;
    r2=0;
  }
  if (r2>1) {
    logger << LOG_DEBUG << "r2>1 in inv_tchannel!" << endline_flush;
    logger << LOG_DEBUG << "r2=" << r2 <<"; k10.x, k10.y=" << k10.x1() << ", " << k10.x2() << endline_flush;
    logger << LOG_DEBUG << xbp_all[i_a][in] << endline_flush;
    r2=1;
  }

  // invert first part
  r1 = inv_t_propagator(-xbs_all[i_a][in1 + out1], m, -tmin, -tmax);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void phasespace_set::c_tchannel_opt(int i_a, int m, int out, int in1, int in2, int out1, int out2, double r1, double r2){
  static Logger logger("phasespace_set::c_tchannel_opt");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  double term1 = (xbs_all[i_a][out] + xbs_all[i_a][out1] - xbs_all[i_a][out2]) * (xbs_all[i_a][out] + xbs_all[i_a][in1] - xbs_all[i_a][in2]);
  double term2 = sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][out1], xbs_all[i_a][out2])) * sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2]));
  double term3 = (2 * sqrt(xbs_all[i_a][out])); // !!! why not xbsqrts_all[i_a][out] ??? - check if calculated ...

  double tmin = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 + term2) / (2 * xbs_all[i_a][out]);
  double tmax = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 - term2) / (2 * xbs_all[i_a][out]);
  double tmax_cut = 0.;
  if (MC_optswitch > 0 && xbs_all[i_a][out1] == 0.){
    vector<int> temp_out_particle = vectorint_from_binary(out1);
    if (temp_out_particle.size() == 1){
      vector<int> temp_smin_out = vectorbinary_from_binary(out - out1); // all outgoing ones, but out1
      double smin_out = 0.;
      for (int i = 0; i < temp_smin_out.size(); i++){smin_out += xbsqrts_all[i_a][temp_smin_out[i]];}
      smin_out = pow(smin_out, 2);
      tmax_cut = .5 *(xbs_all[i_a][0] - smin_out) * (sqrt(1. - (4. * xbs_all[i_a][0] * mapping_cut_pT[abs(csi->type_parton[0][temp_out_particle[0]])]) / pow((xbs_all[i_a][0] - smin_out), 2.)) - 1.);
    }
  }
  else if (MC_optswitch > 0 && xbs_all[i_a][out1] > 0.){
    vector<int> temp_out_particle = vectorint_from_binary(out1);
    if (temp_out_particle.size() == 1){
      vector<int> temp_smin_out = vectorbinary_from_binary(out - out1); // all outgoing ones, but out1
      double smin_out = 0.;
      for (int i = 0; i < temp_smin_out.size(); i++){smin_out += xbsqrts_all[i_a][temp_smin_out[i]];}
      smin_out = pow(smin_out, 2);

      tmax_cut = M2[abs(csi->type_parton[0][temp_out_particle[0]])] + .5 * (xbs_all[i_a][0] + M2[abs(csi->type_parton[0][temp_out_particle[0]])] - smin_out) * (sqrt(1. - (4. * xbs_all[i_a][0] * (M2[abs(csi->type_parton[0][temp_out_particle[0]])] + mapping_cut_pT[abs(csi->type_parton[0][temp_out_particle[0]])]) / pow((xbs_all[i_a][0] + M2[abs(csi->type_parton[0][temp_out_particle[0]])] - smin_out), 2.))) - 1.);
    }
  }

  if (MC_optswitch < 2){
    if ((MC_optswitch == 1) && (tmax_cut < tmax) && (tmin < tmax_cut)){tmax = tmax_cut;}
    xbs_all[i_a][in1 + out1] = -c_t_propagator(r1, m, -tmin, -tmax);
  }
  else {
    // This parametrization is incomplete and has never been used - remove !!!
    logger << LOG_FATAL << "This parametrization is incomplete and has never been used - remove !!!" << endline_flush;
    exit(1);
    if ((tmax_cut < tmax) && (tmin < tmax_cut)){
      logger << LOG_DEBUG << "(tmax_cut < tmax) && (tmin < tmax_cut):   " << "(" << tmax_cut << " < " << tmax << ") && (" << tmin << " < " << tmax_cut << ")" << endline_flush;
      double r0 = 0.95;
      if (r1 < r0){
	double r_temp = r1 / r0;
	xbs_all[i_a][in1 + out1] = -c_t_propagator(r_temp, m, -tmin, -tmax_cut);
      }
      else {

      }
    }
    else {
      xbs_all[i_a][in1 + out1] = -c_t_propagator(r1, m, -tmin, -tmax);
    }
  }
  double phi = c_phi(r2);
  double costheta = ((2 * xbs_all[i_a][out] * (xbs_all[i_a][in1 + out1] - xbs_all[i_a][out1] - xbs_all[i_a][in1])) + term1) / (term2);
  fourvector p10(((xbs_all[i_a][out] + xbs_all[i_a][in1] - xbs_all[i_a][in2]) / term3), 0, 0, (sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2])) / term3));
  fourvector p20(((xbs_all[i_a][out] - xbs_all[i_a][in1] + xbs_all[i_a][in2]) / term3), 0, 0, -(sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2])) / term3));
  fourvector k10(((xbs_all[i_a][out] + xbs_all[i_a][out1] - xbs_all[i_a][out2]) / term3), 0, 0, (sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][out1], xbs_all[i_a][out2])) / term3));
  k10 = k10.rotation_beam(phi, costheta);
  fourvector k20 = p10 + p20 - k10;
  fourvector p1b = xbp_all[i_a][in1].boost(xbp_all[i_a][out], xbs_all[i_a][out]);
  xbp_all[i_a][out1] = (k10.rotation(p1b)).boost(xbp_all[i_a][out].Pinv(), xbs_all[i_a][out]);
  xbp_all[i_a][out2] = (k20.rotation(p1b)).boost(xbp_all[i_a][out].Pinv(), xbs_all[i_a][out]);
  xbp_all[i_a][in1 + out1] = xbp_all[i_a][in1] - xbp_all[i_a][out1];

  if (munich_isnan(xbp_all[i_a])){
    logger << LOG_DEBUG << "Debug output if point creation fails!" << endline_flush;
    logger << LOG_DEBUG << "xbs_all[" << i_a << "][" << out << "] = " << xbs_all[i_a][out] << endline_flush;
    logger << LOG_DEBUG << "xbs_all[" << i_a << "][" << out1 << "] = " << xbs_all[i_a][out1] << endline_flush;
    logger << LOG_DEBUG << "xbs_all[" << i_a << "][" << out2 << "] = " << xbs_all[i_a][out2] << endline_flush;
    logger << LOG_DEBUG << "xbs_all[" << i_a << "][" << in1 << "] = " << xbs_all[i_a][in1] << endline_flush;
    logger << LOG_DEBUG << "xbs_all[" << i_a << "][" << in2 << "] = " << xbs_all[i_a][in2] << endline_flush;
    logger << LOG_DEBUG << "xbs_all[" << i_a << "][" << in1 + out1 << "] = " << xbs_all[i_a][in1 + out1] << endline_flush;
    logger << LOG_DEBUG << "tmin = " << tmin << endline_flush;
    logger << LOG_DEBUG << "tmax = " << tmax << endline_flush;
    logger << LOG_DEBUG << "tmax_cut = " << tmax_cut << endline_flush;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

double phasespace_set::g_tchannel_opt(int i_a, int m, int out, int in1, int in2, int out1, int out2){
  static Logger logger("phasespace_set::g_tchannel_opt");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  double gp = 0.;
  double term1 = (xbs_all[i_a][out] + xbs_all[i_a][out1] - xbs_all[i_a][out2]) * (xbs_all[i_a][out] + xbs_all[i_a][in1] - xbs_all[i_a][in2]);
  double term2 = sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][out1], xbs_all[i_a][out2])) * sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2]));
  double tmin = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 + term2) / (2 * xbs_all[i_a][out]);
  double tmax = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 - term2) / (2 * xbs_all[i_a][out]);
  if (tmin > 0.){logger << LOG_DEBUG << "tmin = " << tmin << endline_flush;}
  if (tmax > 0.){logger << LOG_DEBUG << "tmax = " << tmax << endline_flush; tmax = 0.;}
  double tmax_cut = 0.;

  if (MC_optswitch > 0 && xbs_all[i_a][out1] == 0.){
    vector<int> temp_smin_out = vectorbinary_from_binary(out - out1); // all outgoing ones, but out1
    double smin_out = 0.;
    for (int i = 0; i < temp_smin_out.size(); i++){smin_out += xbsqrts_all[i_a][temp_smin_out[i]];}
    smin_out = pow(smin_out, 2);
    vector<int> temp_out_particle = vectorint_from_binary(out1);
    if (temp_out_particle.size() == 1){
      tmax_cut = .5 *(xbs_all[i_a][0] - smin_out) * (sqrt(1. - (4. * xbs_all[i_a][0] * mapping_cut_pT[abs(csi->type_parton[0][temp_out_particle[0]])]) / pow((xbs_all[i_a][0] - smin_out), 2.)) - 1.);
    }
  }
  else if (MC_optswitch > 0 && xbs_all[i_a][out1] > 0.){
    vector<int> temp_smin_out = vectorbinary_from_binary(out - out1); // all outgoing ones, but out1
    double smin_out = 0.;
    for (int i = 0; i < temp_smin_out.size(); i++){smin_out += xbsqrts_all[i_a][temp_smin_out[i]];}
    smin_out = pow(smin_out, 2);
    vector<int> temp_out_particle = vectorint_from_binary(out1);
    if (temp_out_particle.size() == 1){
      tmax_cut = M2[abs(csi->type_parton[0][temp_out_particle[0]])] + .5 * (xbs_all[i_a][0] + M2[abs(csi->type_parton[0][temp_out_particle[0]])] - smin_out) * (sqrt(1. - (4. * xbs_all[i_a][0] * (M2[abs(csi->type_parton[0][temp_out_particle[0]])] + mapping_cut_pT[abs(csi->type_parton[0][temp_out_particle[0]])]) / pow((xbs_all[i_a][0] + M2[abs(csi->type_parton[0][temp_out_particle[0]])] - smin_out), 2.))) - 1.);
    }
  }

  if (MC_optswitch < 2){
    if ((MC_optswitch == 1) && (tmax_cut < tmax) && (tmin < tmax_cut)){tmax = tmax_cut;}
    gp = -g_t_propagator(-xbs_all[i_a][in1 + out1], m, -tmin, -tmax) * (2 * sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2]))) / pi;
  }
  else {
    // This parametrization is incomplete and has never been used - remove !!!
    logger << LOG_FATAL << "This parametrization is incomplete and has never been used - remove !!!" << endline_flush;
    exit(1);
    if ((tmax_cut < tmax) && (tmin < tmax_cut)){

    }
    else {
      gp = -g_t_propagator(-xbs_all[i_a][in1 + out1], m, -tmin, -tmax) * (2 * sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2]))) / pi;
    }
  }

  if (munich_isnan(gp)){
    logger << LOG_DEBUG << "error output:  g_t  is  NaN !" << endline_flush;
    for (int ip = 1; ip < xbp_all[i_a].size(); ip*=2){
      logger << LOG_DEBUG << "xbp_all[i_a][" << setw(3) << ip << "] = " << xbp_all[i_a][ip] << endline_flush;
    }
    logger << LOG_DEBUG << "m = " << m << endline_flush;
    logger << LOG_DEBUG << "out = " << out << endline_flush;
    logger << LOG_DEBUG << "in1 = " << in1 << endline_flush;
    logger << LOG_DEBUG << "in2 = " << in2 << endline_flush;
    logger << LOG_DEBUG << "out1 = " << out1 << endline_flush;
    logger << LOG_DEBUG << "out2 = " << out2 << endline_flush;
    logger << LOG_DEBUG << "term1 = " << term1 << endline_flush;
    logger << LOG_DEBUG << "term2 = " << term2 << endline_flush;
    logger << LOG_DEBUG << "tmin = " << tmin << endline_flush;
    logger << LOG_DEBUG << "tmax = " << tmax << endline_flush;
    logger << LOG_DEBUG << "tmax_cut = " << tmax_cut << endline_flush;
    logger << LOG_DEBUG << "-xbs_all[i_a][in1 + out1] = " << -xbs_all[i_a][in1 + out1] << endline_flush;
    logger << LOG_DEBUG << "g_t_propagator(-xbs_all[i_a][in1 + out1], m, -tmin, -tmax) = " << g_t_propagator(-xbs_all[i_a][in1 + out1], m, -tmin, -tmax) << endline_flush;
    logger << LOG_DEBUG << "lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2]) = " << lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2]) << endline_flush;
    logger << LOG_DEBUG << "sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2])) = " << sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2])) << endline_flush;
    logger << LOG_DEBUG << "lambda(xbs_all[i_a][out], xbs_all[i_a][out1], xbs_all[i_a][out2]) = " << lambda(xbs_all[i_a][out], xbs_all[i_a][out1], xbs_all[i_a][out2]) << endline_flush;
    logger << LOG_DEBUG << "sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][out1], xbs_all[i_a][out2])) = " << sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][out1], xbs_all[i_a][out2])) << endline_flush;
    logger << LOG_DEBUG << "xbs_all[i_a][out = " << out << "] = " << xbs_all[i_a][out] << endline_flush;
    logger << LOG_DEBUG << "xbs_all[i_a][out1 = " << out1 << "] = " << xbs_all[i_a][out1] << endline_flush;
    logger << LOG_DEBUG << "xbs_all[i_a][out2 = " << out2 << "] = " << xbs_all[i_a][out2] << endline_flush;
  }
  if (gp < 0.){logger << LOG_DEBUG << "gp < 0." << "   out = " << out << "   in1 = " << in1 << "   in2 = " << in2 << "   out1 = " << out1 << "   out2 = " << out2 << endline_flush; gp = 0.;}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
  return gp;
}

void phasespace_set::inv_tchannel_opt(int i_a, int m, int out, int in1, int in2, int out1, int out2, double & r1, double & r2){
  static Logger logger("phasespace_set::inv_tchannel_opt");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  double term1 = (xbs_all[i_a][out] + xbs_all[i_a][out1] - xbs_all[i_a][out2]) * (xbs_all[i_a][out] + xbs_all[i_a][in1] - xbs_all[i_a][in2]);
  double term2 = sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][out1], xbs_all[i_a][out2])) * sqrt(lambda(xbs_all[i_a][out], xbs_all[i_a][in1], xbs_all[i_a][in2]));
  double tmin = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 + term2) / (2 * xbs_all[i_a][out]);
  double tmax = xbs_all[i_a][out1] + xbs_all[i_a][in1] - (term1 - term2) / (2 * xbs_all[i_a][out]);
  double tmax_cut = 0.;
  if (MC_optswitch > 0 && xbs_all[i_a][out1] == 0.){
    vector<int> temp_out_particle = vectorint_from_binary(out1);
    if (temp_out_particle.size() == 1){
      vector<int> temp_smin_out = vectorbinary_from_binary(out - out1); // all outgoing ones, but out1
      double smin_out = 0.;
      for (int i = 0; i < temp_smin_out.size(); i++){smin_out += xbsqrts_all[i_a][temp_smin_out[i]];}// cout << i << "   " << temp_smin_out[i] << "   " << xbsqrts_all[i_a][temp_smin_out[i]] << endline_flush;}
      smin_out = pow(smin_out, 2);
      tmax_cut = .5 *(xbs_all[i_a][0] - smin_out) * (sqrt(1. - (4. * xbs_all[i_a][0] * mapping_cut_pT[abs(csi->type_parton[0][temp_out_particle[0]])]) / pow((xbs_all[i_a][0] - smin_out), 2.)) - 1.);
    }
  }
  else if (MC_optswitch > 0 && xbs_all[i_a][out1] > 0.){
    vector<int> temp_out_particle = vectorint_from_binary(out1);
    if (temp_out_particle.size() == 1){
      vector<int> temp_smin_out = vectorbinary_from_binary(out - out1); // all outgoing ones, but out1
      double smin_out = 0.;
      for (int i = 0; i < temp_smin_out.size(); i++){smin_out += xbsqrts_all[i_a][temp_smin_out[i]];}
      smin_out = pow(smin_out, 2);
      tmax_cut = M2[abs(csi->type_parton[0][temp_out_particle[0]])] + .5 * (xbs_all[i_a][0] + M2[abs(csi->type_parton[0][temp_out_particle[0]])] - smin_out) * (sqrt(1. - (4. * xbs_all[i_a][0] * (M2[abs(csi->type_parton[0][temp_out_particle[0]])] + mapping_cut_pT[abs(csi->type_parton[0][temp_out_particle[0]])]) / pow((xbs_all[i_a][0] + M2[abs(csi->type_parton[0][temp_out_particle[0]])] - smin_out), 2.))) - 1.);
    }
  }

  fourvector p1b = xbp_all[i_a][in1].boost(xbp_all[i_a][out], xbs_all[i_a][out]);

  // invert second part
  fourvector k10 = xbp_all[i_a][out1].boost(xbp_all[i_a][out], xbs_all[i_a][out]);
  k10 = k10.rotation_inverse(p1b);

  // Would this work as well ???
  //  fourvector k10 = (xbp_all[i_a][out1].boost(xbp_all[i_a][out], xbs_all[i_a][out]).rotation_inverse(p1b));

  double phi = 0.;
  if (k10.x1()!=0 || k10.x2()!=0) {
    phi = atan2(k10.x2(),k10.x1());
    if (phi<0)
      phi+=2*M_PI;
  }
  else {
    logger << LOG_DEBUG << "x1=x2=0 in inv_tchannel_opt!" << endline_flush;
    phi=0;
  }

  r2=phi/(2*M_PI);

  // Would this work as well ???
  //  r2 = inv_phi(k10.phi());
  
  // Set r2 to  map_technical_r  and  1. - map_technical_r  instead ???
  //  if (r2 <= 0){
  if (!(r2>=0)) {
    logger << LOG_DEBUG << "r2<0 in inv_tchannel_opt!" << endline_flush;
    logger << LOG_DEBUG << "r2=" << r2 <<"; k10.x, k10.y=" << k10.x1() << ", " << k10.x2() << endline_flush;
    r2 = 0;
  }
  if ((r2>1)) {
    logger << LOG_DEBUG << "r2>1 in inv_tchannel_opt!" << endline_flush;
    logger << LOG_DEBUG << "r2=" << r2 <<"; k10.x, k10.y=" << k10.x1() << ", " << k10.x2() << endline_flush;
    r2=1;
  }

  // invert first part
  if (MC_optswitch < 2){
    if ((MC_optswitch == 1) && (tmax_cut < tmax) && (tmin < tmax_cut)){tmax = tmax_cut;}
    r1 = inv_t_propagator(-xbs_all[i_a][in1 + out1], m, -tmin, -tmax);
  }
  else {
    // This parametrization is incomplete and has never been used - remove !!!
    logger << LOG_FATAL << "This parametrization is incomplete and has never been used - remove !!!" << endline_flush;
    exit(1);
    if ((tmax_cut < tmax) && (tmin < tmax_cut)){
      logger << LOG_DEBUG << "(tmax_cut < tmax) && (tmin < tmax_cut): " << "(" << tmax_cut << " < " << tmax << ") && (" << tmin << " < " << tmax_cut << ")" << endline_flush;
      double r0 = 0.95;
      double temp_r = -xbs_all[i_a][in1 + out1];
      r1 = r0 * c_t_propagator(temp_r, m, -tmin, -tmax_cut);
    }
    else {
      r1 = inv_t_propagator(-xbs_all[i_a][in1 + out1], m, -tmin, -tmax);
    }
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

