#include "header.hpp"

void observable_set::remove_empty_ncollinear(){
  Logger logger("observable_set::remove_empty_ncollinear");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger.newLine(LOG_DEBUG);
  logger << LOG_DEBUG << "Remove now those entries of 'ncollinear' that do not have any left-over contribution:" << endline_flush;
  logger.newLine(LOG_DEBUG);
  int counter_ncollinear = 0;
  for (int i_c = ncollinear.size() - 1; i_c >= 0; i_c--){
    logger << LOG_DEBUG_VERBOSE << "ncollinear[i_c].no_pdf = " << ncollinear[i_c].no_pdf << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "list_combination_pdf[ncollinear[i_c].no_pdf].size() = " << list_combination_pdf[ncollinear[i_c].no_pdf].size() << endline_flush;
    for (int i_i = 0; i_i < list_combination_pdf[ncollinear[i_c].no_pdf].size(); i_i++){
      logger << LOG_DEBUG_VERBOSE << "list_combination_pdf[ncollinear[i_c].no_pdf][i_i].size() = " << list_combination_pdf[ncollinear[i_c].no_pdf][i_i].size() << endline_flush;
    }

    if (list_combination_pdf[ncollinear[i_c].no_pdf].size() == 0){
      logger << LOG_DEBUG << ncollinear[i_c].name << " [" << setw(2) << i_c << "] is removed due to parton content." << endline_flush;

      //      ncollinear.erase(ncollinear.begin() + i_c); // !!! check if this works -> doesn't work yet if i_c == 0 is directly accessed !!!

      counter_ncollinear++;
    }
  }

  if (csi->class_contribution_QT_collinear || csi->class_contribution_QT_virtual || csi->class_contribution_QT_powercorrection){
  if (ncollinear.size() == 0){
    logger << LOG_WARN << "According to pdf-selection criteria, no partonic channels remain (ncollinear.size() = " << ncollinear.size() << "): int_end = 1 !" << endline_flush;
    int_end = 1;
  }
  //
  if (counter_ncollinear == ncollinear.size()){
    logger << LOG_WARN << "According to pdf-selection criteria, no partonic channels remain (counter_ncollinear = " << counter_ncollinear << " = ncollinear.size()): int_end = 1 !" << endline_flush;
    int_end = 1;
  }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::determine_QT_list_combination_pdf(){
  Logger logger("observable_set::determine_QT_list_combination_pdf");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // direct determination of  list_combination_pdf  and  list_combination_pdf_emission  from ncollinear (including list_combination_pdf_emission):

  list_combination_pdf.resize(ncollinear.size(), vector<vector<vector<int> > > (csi->combination_pdf.size()));
  list_combination_pdf_emission.resize(ncollinear.size());

  for (int i_c = 0; i_c < ncollinear.size(); i_c++){
    list_combination_pdf_emission[i_c] = ncollinear[i_c].emission;
    ncollinear[i_c].no_pdf = i_c;
    for (int i_i = 0; i_i < list_combination_pdf[i_c].size(); i_i++){
      // list_combination_pdf[i_c] contains a vector, which contains the usual osi_csi->combination_pdf(3) with 0 -> direction (1, -1), 1 -> parton with x1 (in hadron 1/2 for +1/-1), 2 -> parton with x2 (in hadron 2/1 for +1/-1)
      if ((ncollinear[i_c].pdf[i_i][1] == 10) && (ncollinear[i_c].pdf[i_i][2] == 10)){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	    if (j_q == 0){continue;}
	    vector<int> new_temp_combination_pdf = ncollinear[i_c].pdf[i_i];
	    new_temp_combination_pdf[1] = i_q;
	    new_temp_combination_pdf[2] = j_q;
	    list_combination_pdf[i_c][i_i].push_back(new_temp_combination_pdf);
	  }
	}
      }
      else if (ncollinear[i_c].pdf[i_i][1] == 10){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = ncollinear[i_c].pdf[i_i];
	  new_temp_combination_pdf[1] = i_q;
	  list_combination_pdf[i_c][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else if (ncollinear[i_c].pdf[i_i][2] == 10){
	for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	  if (j_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = ncollinear[i_c].pdf[i_i];
	  new_temp_combination_pdf[2] = j_q;
	  list_combination_pdf[i_c][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else {
	list_combination_pdf[i_c][i_i].push_back(ncollinear[i_c].pdf[i_i]);
      }
    }
  }
  logger << LOG_DEBUG << "list_combination_pdf determined " << endl << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


  // xz_pdf: 0 - 0 : empty - could be filled with tau = x1 * x2
  //         1 - 0 : x1 -> no collinear emission from 1st parton
  //         2 - 0 : x2 -> no collinear emission from 2nd parton
  //         0 - 1 : empty
  //         1 - 1 : x1 / z1 -> collinear emission from 1st parton
  //         2 - 1 : x2 / z2 -> collinear emission from 2nd parton
/*
void observable_set::calculate_IS_CX(){
  Logger logger("observable_set::calculate_IS_CX");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_FATAL << "Should not be called any more !!!" << endline_flush; exit(1);
  /*
  logger << LOG_DEBUG_VERBOSE << "psi->x_pdf.size()  = " << psi->x_pdf.size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "psi->z_coll.size() = " << psi->z_coll.size() << endline_flush;
  //  logger << LOG_DEBUG_VERBOSE << "x_pdf.size()  = " << x_pdf.size() << endline_flush;
  //  logger << LOG_DEBUG_VERBOSE << "z_coll.size() = " << z_coll.size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "xz_pdf.size() = " << xz_pdf.size() << endline_flush;
  for (int i_x = 0; i_x < 3; i_x++){
    logger << LOG_DEBUG_VERBOSE << "xz_pdf[" << i_x << "].size() = " << xz_pdf[i_x].size() << endline_flush;
  }

  for (int i_x = 0; i_x < 3; i_x++){xz_pdf[i_x][0] = psi->x_pdf[i_x];}
  for (int i_x = 1; i_x < 3; i_x++){xz_pdf[i_x][1] = psi->x_pdf[i_x] / psi->z_coll[i_x];}
*//*
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}
*/

string observable_set::output_ncollinear(multicollinear_set & this_ncollinear){
  Logger logger("observable_set::output_ncollinear");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  //  vector<string> pa_name(csi->type_parton[0].size(), "");
  vector<string> pa_name(3, "");
  if (csi->type_parton[0][1] >= -10 && csi->type_parton[0][1] <= 10){pa_name[1] = "a";}
  if (csi->type_parton[0][2] >= -10 && csi->type_parton[0][2] <= 10){pa_name[2] = "b";}
  //  for (int i_p = 3; i_p < pa_name.size(); i_p++){if (csi->type_parton[0][i_p] >= -10 && csi->type_parton[0][i_p] <= 10){pa_name[i_p] = char(105 + i_p - 3);}}

  logger.newLine(LOG_DEBUG_VERBOSE);
  logger << LOG_DEBUG_VERBOSE << "ncollinear.size() = " << ncollinear.size() << endline_flush;
  logger.newLine(LOG_DEBUG_VERBOSE);

  stringstream temp;
  int n_name = this_ncollinear.name.size();
  temp << "new name convention:    " << setw(n_name) << left << this_ncollinear.name << endl;
  //    logger << LOG_DEBUG << temp.str() << endline_flush;
  temp << "old name convention:    " << setw(n_name) << left << this_ncollinear.oldname << endl;
  //<< this_ncollinear.n_emission[0] << ", "
  //    logger << LOG_DEBUG << temp.str() << endline_flush;
  //    temp.str("");
  temp << setw(n_name + 22) << "";
  temp << "csi->type_parton = (";
  for (int i_p = 1; i_p < csi->type_parton[0].size(); i_p++){
    temp << setw(4) << right << this_ncollinear.type_parton[i_p];
    if (i_p == 2){temp << " ->";}
  }
  temp << ")   ";
  temp << "no_prc = " << this_ncollinear.no_prc << "   " << endl;
  //    temp << "ps = " << this_ncollinear.previous_splitting[this_ncollinear.n_emission[0]] << "   ";
  //    logger << LOG_DEBUG << temp.str() << endline_flush;

  vector<string> name_emission(3);
  name_emission[0] = "n_emission_leg_type    ";
  name_emission[1] = "n_emission_leg_type_QCD";
  name_emission[2] = "n_emission_leg_type_QEW";
  for (int i_t = 0; i_t < 3; i_t++){
    //      temp.str("");
    temp << setw(n_name + 22) << "";
    temp << name_emission[i_t] << " = ";
    temp << "(";
    for (int i_e = 0; i_e < 3; i_e++){
      temp << this_ncollinear.n_emission_leg_type[i_e][i_t];
      if (i_e < 2){temp << " ";}
    }
    temp << ")";
    temp << endl;
    /*
    if (i_t == 0){
      for (int i_e = 1; i_e < 3; i_e++){
	//	  if (this_ncollinear.n_emission_leg_type[i_e][i_t] == 0){continue;}
	temp << "   " << pa_name[i_e] << " -> " << this_ncollinear.type_splitting_full[i_e];
      }
    }
    temp << endl;
    */
    //      logger << LOG_DEBUG << temp.str() << endline_flush;
  }
  //    temp.str("");
  if (this_ncollinear.no_pdf != -1){
    for (int i_e = 1; i_e < 3; i_e++){
      temp << setw(n_name + 22) << "";
      //	  if (this_ncollinear.n_emission_leg_type[i_e][i_t] == 0){continue;}
      temp << pa_name[i_e] << " -> " << "type_splitting_full[" << i_e << "] = " << this_ncollinear.type_splitting_full[i_e] << endl;
    }
 
    temp << setw(n_name + 22) << "";
    temp << "emission         -> no_pdf";
    if (this_ncollinear.no_pdf == -1){temp << " to be determined later" << endl;}
    else {temp << " = " << setw(2) << right << this_ncollinear.no_pdf << endl;}
    //    logger << LOG_DEBUG << temp.str() << endline_flush;
    for (int i_e = 0; i_e < 4; i_e++){
      //      temp.str("");
      temp << setw(n_name + 22) << "";
      if (this_ncollinear.no_endpoint[i_e] == -1){
	temp << "endpoint " << i_e << " -> " << setw(2) << "--" << " -> no_pdf = "  << setw(2) << "--" << endl;
      }
      else {
	temp << "endpoint " << i_e << " -> " << setw(2) << this_ncollinear.no_endpoint[i_e] << " -> no_pdf";
	if (this_ncollinear.no_pdf == -1){temp << " to be determined later" << endl;}
	else {temp << " = " << setw(2) << right << ncollinear[this_ncollinear.no_endpoint[i_e]].no_pdf << endl;}
      }
      //      temp << "endpoint " << i_e << " -> " << this_ncollinear.no_endpoint[i_e];
      //      temp << endl;
      //      logger << LOG_DEBUG << temp.str() << endline_flush;
    }
  }

  temp << setw(n_name + 22) << "";
  temp << "n_emission = ";
  temp << "(";
  for (int i_z = 0; i_z < 3; i_z++){
    temp << this_ncollinear.n_emission[i_z];
    if (i_z < 2){temp << " ";}
  }
  temp << ")" << endl;
  //  logger << LOG_DEBUG << temp.str() << endline_flush;
  //  temp.str("");

  for (int j_e = 1; j_e < this_ncollinear.n_emission[0] + 1; j_e++){
    temp << setw(n_name + 22) << "";
    temp << "emission " << j_e << " :   ";
    if      (this_ncollinear.type_correction[j_e] == 1){temp << "QCD   ";}
    else if (this_ncollinear.type_correction[j_e] == 2){temp << "QEW   ";}
    temp << "no_emitter = ";
    temp << this_ncollinear.no_emitter[j_e] << "   ";
    temp << "type_splitting = ";
    temp << this_ncollinear.type_splitting[j_e] << "   ";
    /*
      temp << "charge_factor = ";
      temp << setw(15) << setprecision(8) << this_ncollinear.charge_factor[j_e] << " ";
      temp << "  ";
    */
    temp << "in_collinear = ";
    temp << "(";
    for (int i_z = 0; i_z < 3; i_z++){
      temp << this_ncollinear.in_collinear[j_e][i_z];
      if (i_z < 2){temp << " ";}
    }
    temp << ")" << endl;
    //	logger << LOG_DEBUG << temp.str() << endline_flush;
    //	temp.str("");
  }

  for (int i_x = 0; i_x < this_ncollinear.all_name.size(); i_x++){
    //      temp.str("");
    temp << setw(17) << "";
    temp << setw(n_name) << left << this_ncollinear.all_name[i_x] << " :   " << "pdf =     (";
    for (int i_z = 0; i_z < 3; i_z++){temp << setw(4) << right << this_ncollinear.pdf[i_x][i_z];}
    temp << ")" << endl;
    //      logger << LOG_DEBUG << temp.str() << endline_flush;
  }
  //    logger.newLine(LOG_DEBUG);
  //  }
    
  logger << LOG_DEBUG_VERBOSE << "finished - return (string)" << temp.str() << endline_flush;
  return temp.str();
}


