#include "header.hpp"

#include <cstdio>
#include <iostream>
#include <memory>
#include <stdexcept>
#include <array>

void system_execute(Logger & logger, string xorder){
  logger << LOG_DEBUG_VERBOSE << "system_execute(logger, xorder) called" << endline_flush;

  logger << LOG_DEBUG << "xorder = " << xorder << endline_flush;

  int isystem = 0;
  system_execute(logger, xorder, isystem);

  logger << LOG_DEBUG_VERBOSE << "xorder = " << xorder << "   execution status: " << isystem << "." << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "system_execute(logger, xorder) finished" << endline_flush;
}


void system_execute(Logger & logger, string xorder, int isystem){
  logger << LOG_DEBUG_VERBOSE << "system_execute called" << endline_flush;

  logger << LOG_DEBUG << "xorder = " << xorder << endline_flush;
  logger << LOG_DEBUG << "system_execute xorder = " << xorder << endline_flush;

  string order = xorder.substr(0, 5);
  if (order == "mkdir"){
    DIR *pDir;
    string directory = xorder.substr(6, xorder.size() - 6);
    logger << LOG_DEBUG_VERBOSE << "xorder = " << xorder << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "directory = " << directory << endline_flush;
    pDir = opendir (directory.c_str());
    if (pDir == NULL) {
      logger << LOG_DEBUG << "pDir is NULL." << endline_flush;
      vector<string> subdirectory;
      for (int i_c = 0; i_c < directory.size(); i_c++){
	if (directory[i_c] == '/'){subdirectory.push_back(directory.substr(0, i_c));}
      }
      subdirectory.push_back(directory);
      for (int i_s = 0; i_s < subdirectory.size(); i_s++){
	logger << LOG_DEBUG << "Open subdirectory[" << i_s << "] = " << subdirectory[i_s] << " !" << endline_flush;
	pDir = opendir (subdirectory[i_s].c_str());

	if (pDir == NULL) {
	  //	  closedir (pDir);
	  //	  logger << LOG_DEBUG << "after if (pDir == NULL) {" << endline_flush;
	  xorder = "mkdir " + subdirectory[i_s];
	  logger << LOG_DEBUG << "xorder = " << xorder << endline_flush;
	  isystem = system(xorder.c_str());
	  ///	  isystem = mkdir(directory.c_str(), 0777);
	  //cloud:	  isystem = mkdir(subdirectory[i_s].c_str(), 0777);
	  logger << LOG_DEBUG << "isystem = " << isystem << endline_flush;
	  ///	  if (isystem){logger << LOG_ERROR << directory << " could not be created!" << endline_flush; exit(1);}
	}
	else {
	  logger << LOG_DEBUG << "Close subdirectory[" << i_s << "] = " << subdirectory[i_s] << " !" << endline_flush;
	  closedir (pDir);
	}
      }
    }
    else {
      closedir (pDir);
      logger << LOG_DEBUG << "directory = " << directory << " is not NULL." << endline_flush;
    }
  }
  /*
  else if (order == "chmod"){
    logger << LOG_DEBUG << "Check if chmod causes code to extremely slow down (no chmod executed)." << endline_flush;
  }
  */
  else {
    isystem = system(xorder.c_str());
  }
  logger << LOG_DEBUG << "xorder = " << xorder << "   execution status: " << isystem << "." << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "system_execute(logger, xorder, isystem) finished" << endline_flush;
}


/*
* Look for lines in the procfile contents like: 
* VmRSS:         5560 kB
* VmSize:         5560 kB
*
* Grab the number between the whitespace and the "kB"
* If 1 is returned in the end, there was a serious problem 
* (we could not find one of the memory usages)
*/
void get_memory_usage_kb(Logger & logger, string label_position, long * vmrss_kb, long * vmsize_kb){
  logger << LOG_DEBUG_VERBOSE << "get_memory_usage_kb(logger, label_position, vmrss_kb, vmsize_kb) called" << endline_flush;

  /* Get the the current process' status file from the proc filesystem */
  FILE* procfile = fopen("/proc/self/status", "r");
  long to_read = 8192;
  char buffer[to_read];
  int read = fread(buffer, sizeof(char), to_read, procfile);
  logger << LOG_DEBUG << "read = " << read << endline_flush;
  fclose(procfile);
  short found_vmrss = 0;
  short found_vmsize = 0;
  char* search_result;
  /* Look through proc status contents line by line */
  char delims[] = "\n";
  char* line = strtok(buffer, delims);
  while (line != NULL && (found_vmrss == 0 || found_vmsize == 0)){
    search_result = strstr(line, "VmRSS:");
    if (search_result != NULL){
      sscanf(line, "%*s %ld", vmrss_kb);
      found_vmrss = 1;
    }
    search_result = strstr(line, "VmSize:");
    if (search_result != NULL){
      sscanf(line, "%*s %ld", vmsize_kb);
      found_vmsize = 1;
    }
    line = strtok(NULL, delims);
  }
  logger << LOG_WARN << "VmRSS = " << setw(12) << *vmrss_kb << " KB, VmSize = " << setw(12) << *vmsize_kb << " KB" << "   (" << label_position << ")" << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "get_memory_usage_kb(logger, label_position, vmrss_kb, vmsize_kb) finished" << endline_flush;
}


void output_memory_usage_kb(Logger & logger, string label_position){
  logger << LOG_DEBUG_VERBOSE << "output_memory_usage_kb(logger, label_position) called" << endline_flush;

  long vmrss_kb;
  long vmsize_kb;
  get_memory_usage_kb(logger, label_position, &vmrss_kb, &vmsize_kb);

  logger << LOG_DEBUG_VERBOSE << "output_memory_usage_kb(logger, label_position) finished" << endline_flush;
}


// Should be replaced by the above functions !!!
void output_memory_consumption(const string identifier) {
  static Logger logger("output_memory_consumption");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  string result;
  pid_t pid = getpid();
  stringstream temp_pid;
  temp_pid << "cat /proc/" << pid << "/status | grep VmRSS";
  string temp_s_pid = temp_pid.str();
  char * cmd = new char[temp_s_pid.size() + 1];
  std::copy(temp_s_pid.begin(), temp_s_pid.end(), cmd);
  array<char, 128> buffer;
  unique_ptr<FILE, decltype(&pclose)> pipe(popen(cmd, "r"), pclose);
  if (!pipe) {
    logger << LOG_DEBUG << left << setw(18) << "VmRSS: no output (popen() failed)" << "      " << " - " << setw(50) << identifier << endline_flush;//" - " << temp_s
    //    throw runtime_error("popen() failed!");
    return;
  }
  while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr) {
    result += buffer.data();
  }
  delete [] cmd;

  if (result.size() > 0){
    // replace '\t' by <space> ???
    logger << LOG_DEBUG << left << setw(18) << result.substr(0, 18) << " - " << setw(50) << identifier << endline_flush;//" - " << temp_s_pid << endline_flush;
  }
  else {
    logger << LOG_DEBUG << left << setw(18) << "VmRSS: no output" << "      " << " - " << setw(50) << identifier << endline_flush;//" - " << temp_s
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


string get_path_executable(){
  static Logger logger("get_path_executable");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  char exepath[PATH_MAX + 1] = {0};

  int check = readlink("/proc/self/exe", exepath, 1024);
  logger << LOG_DEBUG_VERBOSE << "check = " << check << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "exepath = " << exepath << endline_flush;
  string s_exepath = string(exepath);
  string path_MUNICH;
  int counter = 0;
  for (int i_s = s_exepath.size() - 1; i_s >=0; i_s--){
    if (s_exepath[i_s] == '/'){counter++;}
    if (counter == 2){
      path_MUNICH = s_exepath.substr(0, i_s + 1);
      break;
    }
  }

  logger << LOG_DEBUG << "finished - " << path_MUNICH << endline_flush;
  return path_MUNICH;
}


string get_current_dir(){
  static Logger logger("get_current_dir");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  char buff[FILENAME_MAX]; //create string buffer to hold path
   //   GetCurrentDir
   char * test = getcwd( buff, FILENAME_MAX );
   logger << LOG_DEBUG << test << endline_flush;
   string current_working_dir(buff);
   logger << LOG_DEBUG << current_working_dir << endline_flush;
   int i_s4 = -1;
   int i_s5 = -1;
   size_t counter = 0;
   for (int i_s = 0; i_s < current_working_dir.size(); i_s++){
     int j_s = current_working_dir.size() - i_s;
     if (current_working_dir[j_s] == '/'){
       counter++;
       if (counter == 4){i_s4 = j_s;}
       if (counter == 5){i_s5 = j_s; break;}
     }
   }
   string mainrundirectory = current_working_dir.substr(i_s5 + 1, i_s4 - i_s5 - 1);
   logger << LOG_DEBUG << "mainrundirectory = " << mainrundirectory << endline_flush;
   
  logger << LOG_DEBUG << "finished - " << current_working_dir << endline_flush;
   return current_working_dir;
}
