#include "header.hpp"

void summary_contribution::readin_runtime_contribution(){
  Logger logger("summary_contribution::readin_runtime_contribution");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  int default_size_CV = 1 + 2;
  if (active_qTcut){default_size_CV += 2 * osi->n_qTcut * osi->n_scales_CV;}
  else {default_size_CV += 2 * osi->n_scales_CV;}
  logger << LOG_DEBUG_VERBOSE << "default_size_CV = " << default_size_CV << endline_flush;

  for (int i_p = 1; i_p < subprocess.size(); i_p++){xsubprocess[i_p]->readin_runtime("result");}

  // reset to zero for [0] component !!!
  xsubprocess[0]->initialization_runtime();

  for (int i_p = 1; i_p < subprocess.size(); i_p++){
    logger << LOG_DEBUG << "CROSSCHECK   " << xsubprocess[i_p]->name << "   xsubprocess[" << i_p << "].error2_time = " << xsubprocess[i_p]->error2_time << endline_flush;

    // Check if this should be error2_time or sqrt(error2_time) !!!
    xsubprocess[0]->error2_time += xsubprocess[i_p]->error2_time;
    xsubprocess[0]->used_runtime += xsubprocess[i_p]->used_runtime;
    xsubprocess[0]->used_n_event += xsubprocess[i_p]->used_n_event;
  }

  for (int i_o = 0; i_o < extended_directory.size(); i_o++){
    logger << LOG_DEBUG << setw(10) << infix_order_contribution << "   used_runtime_group[" << i_o << "] = " << xsubprocess[0]->used_runtime_group[i_o] << "   error2_time_group[" << i_o << "] = " << xsubprocess[0]->error2_time_group[i_o] << endline_flush;
    xsubprocess[0]->error2_time_group[i_o] = 0.;
    xsubprocess[0]->used_runtime_group[i_o] = 0.;
    xsubprocess[0]->used_n_event_group[i_o] = 0;
    
    for (int i_p = 1; i_p < subprocess.size(); i_p++){
      logger << LOG_DEBUG << "CROSSCHECK   " << xsubprocess[i_p]->name << "   xsubprocess[" << i_p << "].error2_time_group[" << i_o << "] = " << xsubprocess[i_p]->error2_time_group[i_o] << endline_flush;
      
      // Check if this should be error2_time or sqrt(error2_time) !!!
      xsubprocess[0]->error2_time_group[i_o] += xsubprocess[i_p]->error2_time_group[i_o];
      xsubprocess[0]->used_runtime_group[i_o] += xsubprocess[i_p]->used_runtime_group[i_o];
      xsubprocess[0]->used_n_event_group[i_o] += xsubprocess[i_p]->used_n_event_group[i_o];

    }
    //    xsubprocess[0]->time_per_event_group[i_o] = xsubprocess[0]->used_runtime_group[i_o] / xsubprocess[0]->used_n_event_group[i_o];

    logger << LOG_DEBUG << setw(10) << infix_order_contribution << "   used_runtime_group[" << i_o << "] = " << xsubprocess[0]->used_runtime_group[i_o] << "   error2_time_group[" << i_o << "] = " << xsubprocess[0]->error2_time_group[i_o] << endline_flush;
  }

  logger << LOG_INFO << "Used runtimes in  " << resultdirectory << " - " << infix_order_contribution << " (output_runtime_contribution)" << endline << endline << output_runtime_contribution() << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


string summary_contribution::output_runtime_contribution(){
  Logger logger("summary_contribution::output_runtime_contribution");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream temp_ss;
  temp_ss << left << setw(20) << "subprocess" << "   ";
  temp_ss << left << setw(15) << "optimization" << "   ";
  temp_ss << right << setw(15) << "n_acc" << "   ";
  temp_ss << right << setw(15) << "runtime (s)" << "   ";
  //  temp_ss << left << setw(15) << "time / acc event" << "   ";
  temp_ss << left << setw(15) << "runtime / n_acc" << "   ";
  temp_ss << right << setw(15) << "n_gen" << "   ";
  temp_ss << right << setw(15) << "error2_time" << "   ";
  temp_ss << endline;
  for (int i_s = 0; i_s < 128; i_s++){temp_ss << "-";}
  temp_ss << endline;
  for (int i_p = 0; i_p < subprocess.size(); i_p++){
    for (int i_o = 0; i_o < extended_directory.size(); i_o++){
      if (i_o == 0){temp_ss << left << setw(20) << xsubprocess[i_p]->name << "   ";}
      else {temp_ss << setw(23) << "";}
      //      temp_ss << setw(15) << extended_directory[i_o] << "   ";
      temp_ss << setw(15) << left << i_o << "   ";
      if (xsubprocess[i_p]->used_runtime_group[i_o] == 0.){temp_ss << right << setw(15) << 0 << "   ";}
      else if (i_p == 0){temp_ss << right << setw(15) << "---" << "   ";}
      else {temp_ss << right << setw(15) << long(xsubprocess[i_p]->used_runtime_group[i_o] / xsubprocess[i_p]->time_per_event_group[i_o]) << "   ";}
      //  temp_ss << setw(25) << xsubprocess[i_p]->used_n_event_group[i_o] << "   ";
      temp_ss << setw(15) << long(xsubprocess[i_p]->used_runtime_group[i_o]) << "   ";
      if (xsubprocess[i_p]->used_runtime_group[i_o] == 0.){temp_ss << right << setw(15) << "---" << "   ";}
      else if (i_p == 0){temp_ss << right << setw(15) << "---" << "   ";}
      else {temp_ss << right << setw(15) << xsubprocess[i_p]->time_per_event_group[i_o] << "   ";}
      temp_ss << right << setw(15) << xsubprocess[i_p]->used_n_event_group[i_o] << "   ";
      temp_ss << right << setw(15) << xsubprocess[i_p]->error2_time_group[i_o] << endline;
    }
    if (i_p == 0){temp_ss << endline;}
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
  return temp_ss.str();
}


void summary_contribution::output_extrapolated_runtime_directories(int x_cs, int x_m, int & list_contribution_counter){
  Logger logger("summary_contribution:output_extrapolated_runtime_directories");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  //////////////////////////////////////////////////
  //  extrapolated runtime - new run directories  //
  //////////////////////////////////////////////////

  max_number_of_jobs_for_one_channel = ygeneric->run_min_number_of_jobs_for_one_channel;
  for (int i_p = 1; i_p < xsubprocess.size(); i_p++){
    xsubprocess[i_p]->max_number_of_jobs_for_one_channel = ygeneric->run_min_number_of_jobs_for_one_channel;
    double temp_njc = xsubprocess[i_p]->extrapolated_runtime / ygeneric->run_max_time_per_job;
    if (temp_njc > ygeneric->run_min_number_of_jobs_for_one_channel){xsubprocess[i_p]->max_number_of_jobs_for_one_channel = int(temp_njc) + 1;}
    else {xsubprocess[i_p]->max_number_of_jobs_for_one_channel = ygeneric->run_min_number_of_jobs_for_one_channel;}
    if (xsubprocess[i_p]->max_number_of_jobs_for_one_channel > max_number_of_jobs_for_one_channel){max_number_of_jobs_for_one_channel = xsubprocess[i_p]->max_number_of_jobs_for_one_channel;}
  }

  stringstream temp_photon_induced;
  if (ylist->photon_induced){temp_photon_induced << "a";}

  // actually needed only once:
  /*
  stringstream list_path;
  list_path << ygeneric->final_resultdirectory + "/newrundir";
  list_path << "/" << ylist->type_perturbative_order;
  if (ylist->type_subtraction_method != "---"){list_path << "." << ylist->type_subtraction_method;}
  system_execute(logger, "mkdir " + list_path.str());
  logger << LOG_INFO << "list_path = " << list_path.str() << endline_flush;
  stringstream temp_photon_induced;
  if (ylist->photon_induced){temp_photon_induced << "a";}
  list_path << "/" << temp_photon_induced.str() << ylist->in_contribution_order_alpha_s << ylist->in_contribution_order_alpha_e;
  system_execute(logger, "mkdir " + list_path.str());
  logger << LOG_INFO << "list_path = " << list_path.str() << endline_flush;
  */
 

  // Create directories in channel / optimization specific  newrundir  folder
  
  stringstream list_path_start;
  list_path_start << ygeneric->final_resultdirectory + "/" + ygeneric->newrundir_channel_phasespace_optimization[x_cs][x_m];
  list_path_start << "/" << ylist->type_perturbative_order;
  if (ylist->type_subtraction_method != "---"){list_path_start << "." << ylist->type_subtraction_method;}
  system_execute(logger, "mkdir " + list_path_start.str());
  logger << LOG_INFO << "list_path_start = " << list_path_start.str() << endline_flush;
  list_path_start << "/" << temp_photon_induced.str() << ylist->in_contribution_order_alpha_s << ylist->in_contribution_order_alpha_e;
  system_execute(logger, "mkdir " + list_path_start.str());

  string runscript_name = ygeneric->final_resultdirectory + "/" + ygeneric->newrundir_channel_phasespace_optimization[x_cs][x_m] + "/start" + "/MUNICH." + infix_order_contribution + dot_extra_selection_contribution + ".start.sh";
  ofstream out_runscript_contribution;
  out_runscript_contribution.open(runscript_name.c_str(), ofstream::out | ofstream::trunc);
  out_runscript_contribution << "#!/bin/bash" << endline;
  out_runscript_contribution << "echo 'Contribution " << infix_path_contribution << dot_extra_selection_contribution << "'" << endline;
  out_runscript_contribution << "echo" << endline;
  out_runscript_contribution << "cd " << infix_path_contribution << dot_extra_selection_contribution << endline;
  out_runscript_contribution << endline;
  out_runscript_contribution << "for i in {0.." << max_number_of_jobs_for_one_channel - 1 << "}" << endline;
  out_runscript_contribution << "do" << endline;
  out_runscript_contribution << "   echo " << char(34) << infix_path_contribution << dot_extra_selection_contribution << "/run.$i" << char(34) << endline;
  out_runscript_contribution << "   cd run.$i" << endline;
  out_runscript_contribution << "   ../../../../MUNICH.start.sh" << endline;
  out_runscript_contribution << "   cd .." << endline;
  out_runscript_contribution << "done" << endline;
  out_runscript_contribution << "cd ../../.." << endline;
  out_runscript_contribution << "echo" << endline;
  out_runscript_contribution.close();
  system_execute(logger, "chmod 744 " + runscript_name);

  ofstream out_runlist_contribution;
  string runlist_name = ygeneric->final_resultdirectory + "/" + ygeneric->newrundir_channel_phasespace_optimization[x_cs][x_m] + "/start" + "/list.run.subprocesses." + infix_order_contribution + dot_extra_selection_contribution + ".txt";
  out_runlist_contribution.open(runlist_name.c_str(), ofstream::out | ofstream::trunc);

  ofstream out_plainrunlist_contribution_start;
  string plainrunlist_name_start = ygeneric->final_resultdirectory + "/" + ygeneric->newrundir_channel_phasespace_optimization[x_cs][x_m] + "/result/list.runs" + "/runlist." + infix_order_contribution + dot_extra_selection_contribution + ".dat";
  out_plainrunlist_contribution_start.open(plainrunlist_name_start.c_str(), ofstream::out | ofstream::trunc);


  // Create directories in main  newrundir  folder

  string runscript_name_main = ygeneric->final_resultdirectory + "/newrundir/start" + ygeneric->slash_channel_selection[x_cs] + ygeneric->slash_phasespace_optimization[x_m] + "/MUNICH." + infix_order_contribution + dot_extra_selection_contribution + ".start.sh";
  ofstream out_runscript_contribution_main;
  out_runscript_contribution_main.open(runscript_name_main.c_str(), ofstream::out | ofstream::trunc);
  out_runscript_contribution_main << "#!/bin/bash" << endline;
  //  out_runscript_contribution_main << "echo 'Contribution " << infix_path_contribution << ygeneric->phasespace_optimization[x_m] << "'" << endline;
  out_runscript_contribution_main << "echo 'Contribution " << infix_path_contribution << dot_extra_selection_contribution << "'" << endline;
  out_runscript_contribution_main << "echo" << endline;

  //  out_runscript_contribution_main << "cd ../" << ygeneric->name_mainrundirectory << ygeneric->dot_channel_selection[x_cs] << ygeneric->dot_phasespace_optimization[x_m] << endline;
  out_runscript_contribution_main << "cd ." << ygeneric->slash_channel_selection[x_cs] << ygeneric->slash_phasespace_optimization[x_m] << endline;

  out_runscript_contribution_main << "cd " << infix_path_contribution << dot_extra_selection_contribution << endline;
  out_runscript_contribution_main << endline;
  out_runscript_contribution_main << "for i in {0.." << max_number_of_jobs_for_one_channel - 1 << "}" << endline;
  out_runscript_contribution_main << "do" << endline;
  out_runscript_contribution_main << "   echo " << char(34) << infix_path_contribution << dot_extra_selection_contribution << "/run.$i" << char(34) << endline;
  out_runscript_contribution_main << "   cd run.$i" << endline;
  out_runscript_contribution_main << "   ../../../../MUNICH.start.sh" << endline;
  out_runscript_contribution_main << "   cd .." << endline;
  out_runscript_contribution_main << "done" << endline;
  out_runscript_contribution_main << "cd ../../.." << endline;
  //  out_runscript_contribution_main << "cd ../" << ygeneric->name_mainrundirectory << endline;
  if (ygeneric->slash_channel_selection[x_cs] != "" && ygeneric->slash_phasespace_optimization[x_m] != ""){out_runscript_contribution_main << "cd ../.." << endline;}
  else if (ygeneric->slash_channel_selection[x_cs] != "" || ygeneric->slash_phasespace_optimization[x_m] != ""){out_runscript_contribution_main << "cd .." << endline;}
  else {out_runscript_contribution_main << "cd ." << endline;}
  
  out_runscript_contribution_main << "echo" << endline;
  out_runscript_contribution_main.close();
  system_execute(logger, "chmod 744 " + runscript_name_main);


  ofstream out_plainrunlist_contribution;
  //  string plainrunlist_name = ygeneric->final_resultdirectory + "/newrundir/" + ygeneric->start_channel_phasespace_optimization[x_cs][x_m] + "/runlist." + infix_order_contribution + ".dat";
  string plainrunlist_name = ygeneric->final_resultdirectory + "/newrundir/" + ygeneric->runlist_channel_phasespace_optimization[x_cs][x_m] + "/runlist." + infix_order_contribution + dot_extra_selection_contribution + ".dat";
  out_plainrunlist_contribution.open(plainrunlist_name.c_str(), ofstream::out | ofstream::trunc);

  stringstream temp_path;
  temp_path << list_path_start.str() << "/" << infix_contribution + dot_extra_selection_contribution;
  system_execute(logger, "mkdir " + temp_path.str());

  //  run.<xxx>/file_parameter.dat
  system_execute(logger, "mkdir " + temp_path.str() + "/log");
  for (int i_r = 0; i_r < max_number_of_jobs_for_one_channel; i_r++){
    stringstream thisrun_path;
    thisrun_path << temp_path.str() << "/run." << i_r;
    system_execute(logger, "mkdir " + thisrun_path.str());
    system_execute(logger, "mkdir " + thisrun_path.str() + "/log");
    ofstream out_file_parameter_run;
    string name_file_parameter_run = thisrun_path.str() + "/file_parameter.dat";
    out_file_parameter_run.open(name_file_parameter_run.c_str(), ofstream::out | ofstream::trunc);
    write_infile_int(out_file_parameter_run, "zwahl", i_r);
    out_file_parameter_run.close();
  }

  //  <runscript> and <processlist> of contribution
  stringstream ss_runlist_contribution;
  for (int i_r = 0; i_r < max_number_of_jobs_for_one_channel; i_r++){
    for (int i_p = 1; i_p < xsubprocess.size(); i_p++){
      if (xsubprocess[i_p]->extrapolated_sigma_normalization == 0.){continue;}
      if (i_r < xsubprocess[i_p]->max_number_of_jobs_for_one_channel){
	ss_runlist_contribution << right << setw(5) << ++list_contribution_counter << "   " << left << setw(25) << infix_path_contribution << "   ";
	ss_runlist_contribution << setw(5) << ylist->channel_selection << "   ";
	ss_runlist_contribution << setw(10) << extra_selection_contribution << "   ";
	ss_runlist_contribution << "run." << left << setw(4) << i_r << "   " << setw(20) << xsubprocess[i_p]->name << endline_flush;
      }
    }
  }
  //  system_execute(logger, "chmod 744 " + runscript_name);

  int counter_separation = 8 + 30 + 3 + 5 + 3 + 10 + 5 + 5 + 3 + 5 + 2;
  out_runlist_contribution << right << setw(8) << "" << setw(30) << left << infix_order_contribution << "     " << right << setw(5) << list_contribution_counter << " ( " << setw(5) << max_number_of_jobs_for_one_channel << " )" << endline;
  for (int i_s = 0; i_s < counter_separation; i_s++){out_runlist_contribution << "=";} out_runlist_contribution << endline;
  //  out_runlist_contribution << right << setw(8) << "========" << setw(30) << "==============================" << "=====" << "=====" << "===" << "=====" << "==" << endline;
  for (int i_p = 1; i_p < xsubprocess.size(); i_p++){
    out_runlist_contribution << right << setw(8) << "" << setw(30) << left << xsubprocess[i_p]->name << "     ";
    out_runlist_contribution << setw(5) << ylist->channel_selection << "   ";
    out_runlist_contribution << setw(10) << extra_selection_contribution << "     ";
    out_runlist_contribution << right << setw(5) << xsubprocess[i_p]->max_number_of_jobs_for_one_channel << " ( " << setw(5) << xsubprocess[i_p]->max_number_of_jobs_for_one_channel << " )" << endline;
  }
  out_runlist_contribution << endline;
    out_runlist_contribution << right << setw(5) << "no" << "   " << left << setw(30) << "path to contribution" << "   " << setw(5) << "" << "   " << setw(10) << "" << "   " << "run " << left << setw(5) << "" << "  " << setw(20) << "subprocess" << endline;
  for (int i_s = 0; i_s < counter_separation; i_s++){out_runlist_contribution << "=";} out_runlist_contribution << endline;
  //  out_runlist_contribution << right << setw(8) << "========" << setw(30) << "==============================" << "===" << "=====" << "=====" << "==" << "====================" << endline;
  out_runlist_contribution << ss_runlist_contribution.str();
  ss_runlist_contribution.clear();
  out_runlist_contribution.close();

  // dependent on x_cs ???
  // xcontribution[x_cs]->
  size_t temp_n_runs = max_number_of_jobs_for_one_channel;
  if (extended_directory[x_m].size() > temp_n_runs){temp_n_runs = extended_directory[x_m].size();}
  for (int i_r = 0; i_r < temp_n_runs ; i_r++){
    out_plainrunlist_contribution_start << infix_path_contribution << dot_extra_selection_contribution << "/run." << i_r << endline;
  }
  out_plainrunlist_contribution_start.close();

  logger << LOG_DEBUG << "ygeneric->name_mainrundirectory = " << ygeneric->name_mainrundirectory << endline_flush;

  for (int i_r = 0; i_r < max_number_of_jobs_for_one_channel; i_r++){
    //    if (x_cs > 0 || x_m > 0){out_plainrunlist_contribution << ygeneric->runlist_channel_phasespace_optimization[x_cs][x_m] << "/";}
    //    if (x_cs > 0 || x_m > 0){out_plainrunlist_contribution << ygeneric->start_channel_phasespace_optimization[x_cs][x_m] << "/";}
    //    if (x_cs > 0 || x_m > 0){out_plainrunlist_contribution << "../" << ygeneric->name_mainrundirectory << ygeneric->dot_channel_selection[x_cs] << ygeneric->dot_phasespace_optimization[x_m] << "/";}
    //    if (ygeneric->channel_selection[x_cs] != "" || x_m > 0){out_plainrunlist_contribution << "../" << ygeneric->name_mainrundirectory << ygeneric->dot_channel_selection[x_cs] << ygeneric->dot_phasespace_optimization[x_m] << "/";}
    if (ygeneric->channel_selection[x_cs] != "" || x_m > 0){out_plainrunlist_contribution << "." << ygeneric->slash_channel_selection[x_cs] << ygeneric->slash_phasespace_optimization[x_m] << "/";}
    out_plainrunlist_contribution << infix_path_contribution << dot_extra_selection_contribution << "/run." << i_r << endline;
  }
  out_plainrunlist_contribution.close();

  //  run.<xxx>/subprocesslist.dat
  ofstream out_subprocesslist;
  for (int i_r = 0; i_r < max_number_of_jobs_for_one_channel; i_r++){
    stringstream thisrun_path;
    thisrun_path << temp_path.str() << "/run." << i_r;
    string subprocesslist_name = thisrun_path.str() + "/subprocesslist.dat";
    out_subprocesslist.open(subprocesslist_name.c_str(), ofstream::out | ofstream::trunc);
    for (int i_p = 1; i_p < xsubprocess.size(); i_p++){
      if (xsubprocess[i_p]->extrapolated_sigma_normalization == 0.){continue;}
      if (i_r < xsubprocess[i_p]->max_number_of_jobs_for_one_channel){
	out_subprocesslist << xsubprocess[i_p]->name << endline;
      }
    }
    out_subprocesslist.close();
  }

  for (int i_p = 1; i_p < xsubprocess.size(); i_p++){
    long long n_event_per_job = (long long)(xsubprocess[i_p]->extrapolated_n_event) / xsubprocess[i_p]->max_number_of_jobs_for_one_channel + 1;
    double sigma_normalization_deviation_per_job = xsubprocess[i_p]->extrapolated_sigma_normalization_deviation * sqrt(xsubprocess[i_p]->max_number_of_jobs_for_one_channel);
    for (int i_r = -1; i_r < xsubprocess[i_p]->max_number_of_jobs_for_one_channel; i_r++){
      stringstream thisrun_path;
      if (i_r == -1){ thisrun_path << temp_path.str() << "/log";}
      else {thisrun_path << temp_path.str() << "/run." << i_r << "/log";}

      if (xsubprocess[i_p]->extrapolated_sigma_normalization == 0.){continue;}

      //  run.<xxx>/log/file_parameter.<process>.dat
      ofstream out_file_parameter_process;
      string name_file_parameter_process = thisrun_path.str() + "/file_parameter." + xsubprocess[i_p]->name + ".dat";
      out_file_parameter_process.open(name_file_parameter_process.c_str(), ofstream::out | ofstream::trunc);

      stringstream temp_ext_runtime;
      int length_sec = int(log10(xsubprocess[i_p]->extrapolated_runtime));
      temp_ext_runtime << setw(4 + length_sec) << showpoint << xsubprocess[i_p]->extrapolated_runtime;
      out_file_parameter_process << "# runtime = " << temp_ext_runtime.str() << endline;

      long long rounded_n = n_event_per_job + ygeneric->run_min_n_step - n_event_per_job % ygeneric->run_min_n_step;
      long long selected_n = rounded_n;

      if (ygeneric->run_min_n_event > rounded_n){
	logger.newLine(LOG_DEBUG);
	logger << LOG_DEBUG << setw(50) << "i_r" << " = " << i_r << endline_flush;
	logger << LOG_DEBUG << setw(50) << "ygeneric->run_min_n_event" << " = " << ygeneric->run_min_n_event << endline_flush;
	logger << LOG_DEBUG << setw(50) << "n_event_per_job" << " = " << n_event_per_job << endline_flush;
	logger << LOG_DEBUG << setw(50) << "ygeneric->run_min_n_step" << " = " << ygeneric->run_min_n_step << endline_flush;
	logger << LOG_DEBUG << setw(50) << "n_event_per_job % ygeneric->run_min_n_step" << " = " << n_event_per_job % ygeneric->run_min_n_step << endline_flush;
	logger << LOG_DEBUG << setw(50) << "rounded_n" << " = " << rounded_n << endline_flush;
	logger << LOG_DEBUG << setw(50) << "xsubprocess[i_p]->extrapolated_runtime " << " = " << xsubprocess[i_p]->extrapolated_runtime  << endline_flush;
	logger << LOG_DEBUG << setw(50) << "xsubprocess[i_p]->max_number_of_jobs_for_one_channel" << " = " << xsubprocess[i_p]->max_number_of_jobs_for_one_channel << endline_flush;
	logger << LOG_DEBUG << setw(50) << "ygeneric->run_max_time_per_job" << " = " << ygeneric->run_max_time_per_job << endline_flush;
	//	logger << LOG_DEBUG << setw(50) << "" << " = " <<  << endline_flush;
	
	if (double(ygeneric->run_min_n_event) / rounded_n * (xsubprocess[i_p]->extrapolated_runtime / xsubprocess[i_p]->max_number_of_jobs_for_one_channel) > ygeneric->run_max_time_per_job){
	  logger << LOG_INFO << "Estimate for too high increased runtime: " << double(ygeneric->run_min_n_event) / rounded_n * (xsubprocess[i_p]->extrapolated_runtime / xsubprocess[i_p]->max_number_of_jobs_for_one_channel) << " s > " << ygeneric->run_max_time_per_job << " s" << endline;
	  // Set n such that each run takes approx ygeneric->run_max_time_per_job
	  long long this_n_event_per_job = rounded_n * ygeneric->run_max_time_per_job / (xsubprocess[i_p]->extrapolated_runtime / xsubprocess[i_p]->max_number_of_jobs_for_one_channel);
	  rounded_n = this_n_event_per_job + ygeneric->run_min_n_step - this_n_event_per_job % ygeneric->run_min_n_step;
	  selected_n = rounded_n;
	  logger << LOG_DEBUG << setw(50) << "this_n_event_per_job" << " = " << this_n_event_per_job << endline_flush;
	  logger << LOG_DEBUG << setw(50) << "n_event_per_job" << " = " << n_event_per_job << endline_flush;
	  logger << LOG_DEBUG << setw(50) << "rounded_n" << " = " << rounded_n << endline_flush;
	  logger << LOG_DEBUG << setw(50) << "selected_n" << " = " << selected_n << endline_flush;
	}
	else {selected_n = ygeneric->run_min_n_event;}
      }
      /*
      if (type_contribution != "VT2" &&
	  type_contribution != "RVA" &&
	  type_contribution != "L2VA" &&
	  type_contribution != "L2VT"){
	if (ygeneric->run_min_n_event > rounded_n){selected_n = ygeneric->run_min_n_event;}
      }
      */   
     
      if (selected_n < xsubprocess[i_p]->N_group_max[0]){selected_n = xsubprocess[i_p]->N_group_max[0];}
	
      write_infile_long_long(out_file_parameter_process, "n_step", selected_n / ygeneric->run_min_n_step);
      write_infile_long_long(out_file_parameter_process, "n_events_min", selected_n);
      write_infile_long_long(out_file_parameter_process, "n_events_max", ygeneric->run_factor_max * selected_n);
      write_infile_double(out_file_parameter_process, "sigma_normalization", xsubprocess[i_p]->extrapolated_sigma_normalization);
      write_infile_double(out_file_parameter_process, "sigma_normalization_deviation", sigma_normalization_deviation_per_job);

      out_file_parameter_process.close();
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}





