#include "header.hpp"

summary_order::summary_order(){
  Logger logger("summary_order::summary_order ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "constructor called..." << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

summary_order::summary_order(string _resultdirectory, summary_generic * _generic){
  Logger logger("summary_order::summary_order (resultdirectory, generic)");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "constructor called..." << endline_flush;

  initialization(_resultdirectory, _generic);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

summary_order::~summary_order(){
  Logger logger("summary_order::~summary_order");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "destructor called..." << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void summary_order::initialization(string _resultdirectory, summary_generic * _generic){
  Logger logger("summary_order::initialization");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ygeneric = _generic;
  osi = ygeneric->osi;

  resultdirectory = _resultdirectory;
  contribution_file = vector<string> ();
  combination.resize(1);
  combination_type.resize(1, 0);
  accuracy_relative = 0.;
  accuracy_normalization = "";

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void summary_order::calculate_combination_result(vector<double> & combination_result, vector<double> & combination_deviation2, double & result, double & deviation){
  Logger logger("summary_order::calculate_combination_result");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (size_t i_c = 0; i_c < combination_control.size(); i_c++){
    if (combination_control[i_c] == 1){
      result += combination_result[i_c];
      deviation += combination_deviation2[i_c];
    }
    else if (combination_control[i_c] == 2){ // by definition identical to addition
      result +=combination_result[i_c] + combination_result[i_c + 1];
      //      result +=combination_result[i_c] * (1. + combination_result[i_c + 1] / combination_result[i_c]);
      deviation += combination_deviation2[i_c] + combination_deviation2[i_c + 1];
      i_c++;
    }
    else if (combination_control[i_c] == 3){
      // exceptional treatment if the factorized prefactor (typically LO cross section) vanishes): add other orders to reproduce additive result (and avoid nan output):
      if (combination_result[i_c] != 0){
	result +=combination_result[i_c] * (1. + combination_result[i_c + 1] / combination_result[i_c]) * (1. + combination_result[i_c + 2] / combination_result[i_c]);
	deviation += combination_deviation2[i_c] * pow(1. - combination_result[i_c + 1] * combination_result[i_c + 2] / pow(combination_result[i_c], 2), 2);
	deviation += combination_deviation2[i_c + 1] * pow(1. + combination_result[i_c + 2] / combination_result[i_c], 2);
	deviation += combination_deviation2[i_c + 2] * pow(1. + combination_result[i_c + 1] / combination_result[i_c], 2);
      }
      else {
	result += combination_result[i_c + 1] + combination_result[i_c + 2];
	deviation += combination_deviation2[i_c + 1] + combination_deviation2[i_c + 2];
      }
      i_c += 2;
    }
    else if (combination_control[i_c] > 1){
      logger << LOG_WARN << "combination_control[" << i_c << "] > 3  should actually not be required at present." << endline_flush;
      double temp_result = combination_result[i_c];
      double temp_deviation_term1 = 1.;
      double temp_deviation_term2 = 0.;
      for (int j_c = i_c + 1; j_c < i_c + combination_control[i_c]; j_c++){
	temp_result *= (1. + combination_result[j_c] / combination_result[i_c]);
	temp_deviation_term1 *= (1. + combination_result[j_c] / combination_result[i_c]);
	double temp_factor = 1.;
	for (int k_c = i_c + 1; k_c < i_c + combination_control[i_c]; k_c++){if (k_c != j_c){temp_factor *= (1. + combination_result[k_c] / combination_result[i_c]);}}
	temp_deviation_term2 += -combination_result[j_c] / combination_result[i_c] * temp_factor;
	deviation += pow(temp_factor, 2) * combination_deviation2[j_c];
      }
      result += temp_result;
      deviation += pow(temp_deviation_term1 + temp_deviation_term2, 2) * combination_deviation2[i_c];
      i_c += combination_control[i_c] - 1;
    }
    else if (combination_control[i_c] == -1){
      result -= combination_result[i_c];
      deviation += combination_deviation2[i_c];
    }
  }
  deviation = sqrt(deviation);
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

/*
void summary_order::calculate_multiplicative_result(vector<double> & combination_result, vector<double> & combination_deviation2, double & multiplicative_result, double & multiplicative_deviation){
  Logger logger("summary_order::calculate_multiplicative_result");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // working version with correct error evaluation for multiplicative combination - equivalent, but less transparent
  double finalized_result = 0.;
  double finalized_deviation2 = 0.;
  double temp_result = 0.;
  double temp_deviation2 = 0.;
  logger << LOG_INFO << "resultdirectory = " << resultdirectory << endline_flush;
  for (size_t i_c = 0; i_c < combination.size() + 1; i_c++){
    if (combination_index[i_c] == i_c){
      if (i_c > 0){
	finalized_result += temp_result;
	finalized_deviation2 += temp_deviation2;
	if (i_c == combination.size()){break;}
      }
      if (combination_type[i_c] == 0){temp_result = combination_result[i_c];}
      else if (combination_type[i_c] == 2){temp_result = -combination_result[i_c];}
      temp_deviation2 = combination_deviation2[i_c];
    }
    else {
      size_t index_next = i_c;
      for (int j_c = i_c; j_c < combination.size() + 1; j_c++){if (combination_index[j_c] == j_c){index_next = j_c; break;}}
      if (combination_type[i_c - 1] != 1){
	double temp_factor = 1.;
	for (int j_c = i_c; j_c < index_next; j_c++){temp_factor *= (1. + combination_result[j_c] / combination_result[combination_index[i_c]]);}
	for (int j_c = i_c; j_c < index_next; j_c++){
	  double temp_term = -combination_result[j_c] / combination_result[combination_index[i_c]];
	  for (int k_c = i_c; k_c < index_next; k_c++){if (k_c != j_c){temp_term *= (1. + combination_result[k_c] / combination_result[combination_index[i_c]]);}}
	  temp_factor += temp_term;
	}
	temp_deviation2 *= pow(temp_factor, 2);
      }
      temp_result *= (1. + combination_result[i_c] / combination_result[combination_index[i_c]]);
      double temp_factor = 1.;
      for (int j_c = combination_index[i_c] + 1; j_c < index_next; j_c++){if (j_c != i_c){temp_factor *= (1. + combination_result[j_c] / combination_result[combination_index[i_c]]);}}
      temp_deviation2 += pow(temp_factor, 2) * combination_deviation2[i_c];
    }
  }
  multiplicative_result = finalized_result;
  multiplicative_deviation = sqrt(finalized_deviation2);
  logger << LOG_INFO << setw(25) << "sigma (general)" << " = " << setw(23) << setprecision(15) << multiplicative_result << " +- " << setw(23) << setprecision(15) << multiplicative_deviation << endline_flush;

  double B = combination_result[0];
  double QCD = combination_result[1];
  double EW = combination_result[2];
  double dB2 = combination_deviation2[0];
  double dQCD2 = combination_deviation2[1];
  double dEW2 = combination_deviation2[2];

  logger << LOG_INFO << "B   = " << B << "   dB2   = " << dB2 << " * " << pow(1. - QCD * EW / pow(B, 2), 2) << " = " << pow(1. - QCD * EW / pow(B, 2), 2) * dB2 << " -> " << sqrt(pow(1. - QCD * EW / pow(B, 2), 2) * dB2) << endline_flush;
  logger << LOG_INFO << "QCD = " << QCD << "   dQCD2 = " << dQCD2 << " * " << pow(1. + EW / B, 2) << " = " << dQCD2 * pow(1. + EW / B, 2) << " -> " << sqrt(dQCD2 * pow(1. + EW / B, 2)) << endline_flush;
  logger << LOG_INFO << "EW  = " << EW << "   dEW2  = " << dEW2 << " * " << pow(1. + QCD / B, 2) << " = " << dEW2 * pow(1. + QCD / B, 2) << " -> " << sqrt(dEW2 * pow(1. + QCD / B, 2)) << endline_flush;
  logger << LOG_INFO << setw(25) << "sigma (direct)" << " = " << setw(23) << setprecision(15) << B * (1. + QCD / B) * (1. + EW / B) << " +- " << setw(23) << setprecision(15) << sqrt(pow(1. - QCD * EW / pow(B, 2), 2) * dB2  +  dQCD2 * pow(1. + EW / B, 2)  +  dEW2 * pow(1. + QCD / B, 2)) << endline_flush;
  logger.newLine(LOG_INFO);

  /*  
  double finalized_result = 0.;
  double finalized_deviation2 = 0.;
  double temp_result = 0.;
  double temp_deviation2 = 0.;
  logger << LOG_WARN << "resultdirectory = " << resultdirectory << endline_flush;
  for (size_t i_c = 0; i_c < combination.size() + 1; i_c++){
    //    logger << LOG_WARN << "[" << i_c << "] = " << i_c << endline_flush;
    if (combination_index[i_c] == i_c){
      if (i_c > 0){
	if (combination_type[i_c - 1] == 1){temp_deviation2 *= pow(temp_result, 2);}
	finalized_result += temp_result;
	finalized_deviation2 += temp_deviation2;
	//	logger << LOG_WARN << "[" << i_c << "] = " << i_c << "   finalized_result     += " << setw(23) << setprecision(15) << temp_result << "   finalized_result     = " << setw(23) << setprecision(15) << finalized_result << endline_flush;
	//	logger << LOG_WARN << "[" << i_c << "] = " << i_c << "   finalized_deviation2 += " << setw(23) << setprecision(15) << temp_deviation2 << "   finalized_deviation2 = " << setw(23) << setprecision(15) << finalized_deviation2 << endline_flush;
	if (i_c == combination.size()){break;}
      }
      if (combination_type[i_c] == 0){temp_result = combination_result[i_c];}
      else if (combination_type[i_c] == 2){temp_result = -combination_result[i_c];}
      temp_deviation2 = combination_deviation2[i_c];
      //      logger << LOG_WARN << "[" << i_c << "] = " << i_c << "   temp_result      = " << setw(23) << setprecision(15) << temp_result << endline_flush;
      //      logger << LOG_WARN << "[" << i_c << "] = " << i_c << "   temp_deviation2  = " << setw(23) << setprecision(15) << temp_deviation2 << endline_flush;
    }
    else {
      if (combination_type[i_c - 1] != 1){
	temp_deviation2 /= pow(temp_result, 2);
	//	logger << LOG_WARN << "Delta(B)^2 = " << temp_deviation2 << endline_flush;
      }
      temp_result *= (1. + combination_result[i_c] / combination_result[combination_index[i_c]]);
      //
      temp_deviation2 += (pow(combination_result[combination_index[i_c]], 2) * combination_deviation2[i_c] + pow(combination_result[i_c], 2) * combination_deviation2[combination_index[i_c]]) / pow(combination_result[combination_index[i_c]], 4) / pow((1. + combination_result[i_c] / combination_result[combination_index[i_c]]), 2);
      //      temp_deviation2 += combination_deviation2[i_c] / pow(combination_result[combination_index[i_c]], 2) / pow((1. + combination_result[i_c] / combination_result[combination_index[i_c]]), 2);
      //      logger << LOG_WARN << "[" << i_c << "] = " << i_c << "   temp_result     *= " << setw(23) << setprecision(15) << (1. + combination_result[i_c] / combination_result[combination_index[i_c]]) << endline_flush;
      //      logger << LOG_WARN << "[" << i_c << "] = " << i_c << "   temp_result      = " << setw(23) << setprecision(15) << temp_result << endline_flush;
      //      logger << LOG_WARN << "[" << i_c << "] = " << i_c << "   temp_deviation2 += " << setw(23) << setprecision(15) << combination_deviation2[i_c] / pow(combination_result[combination_index[i_c]], 2) / pow((1. + combination_result[i_c] / combination_result[combination_index[i_c]]), 2) << endline_flush;
      //      logger << LOG_WARN << "[" << i_c << "] = " << i_c << "   temp_deviation2  = " << setw(23) << setprecision(15) << temp_deviation2 << endline_flush;
    }
  }
  multiplicative_result = finalized_result;
  multiplicative_deviation = sqrt(finalized_deviation2);
  logger << LOG_WARN << setw(25) << "multiplicative_result" << " = " << setw(23) << setprecision(15) << multiplicative_result << " +- " << setw(23) << setprecision(15) << multiplicative_deviation << endline_flush;
  
  double B = combination_result[0];
  double QCD = combination_result[1];
  double EW = combination_result[2];
  double dB2 = combination_deviation2[0];
  double dQCD2 = combination_deviation2[1];
  double dEW2 = combination_deviation2[2];
  //  logger << LOG_WARN << "B   = " << B << "   dB2   = " << dB2 << " * (" << (pow(1. + pow(QCD / B, 2)), 2) << " * " << (pow(1. + pow(EW / B, 2)), 2) << endline_flush;
  //  logger << LOG_WARN << "delta(B)^2 = " <<  << endline_flush;

  logger << LOG_WARN << "B   = " << B << "   dB2   = " << dB2
	 << " * (" << pow(1. + QCD / B, 2) * pow(1. + EW / B, 2)
	 << " + " << pow(QCD / B, 2) * pow(1. + EW / B, 2)
	 << " + " << pow(EW / B, 2) * pow(1. + QCD / B, 2)
	 << " = " << pow(1. + QCD / B, 2) * pow(1. + EW / B, 2) + pow(QCD / B, 2) * pow(1. + EW / B, 2) + pow(EW / B, 2) * pow(1. + QCD / B, 2)
	 << endline_flush;
  logger << LOG_WARN << "QCD = " << QCD << "   dQCD2 = " << dQCD2 << " * " << (1. + pow(EW / B, 2)) << endline_flush;
  logger << LOG_WARN << "EW  = " << EW << "   dEW2  = " << dEW2 << " * " << (1. + pow(QCD / B, 2)) << endline_flush;
  logger << LOG_WARN << setw(25) << "sigma " << " = " << B * (1. + QCD / B) * (1. + EW / B) << " +- " << sqrt(dB2 * (pow(1. + QCD / B, 2) * pow(1. + EW / B, 2) + pow(QCD / B, 2) * pow(1. + EW / B, 2) + pow(EW / B, 2) * pow(1. + QCD / B, 2))  +  dQCD2 * pow(1. + EW / B, 2) +  dEW2 * pow(1. + QCD / B, 2)) << endline_flush;
*//*
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}
*/
