#include <iostream>
#include <iomanip>
#include <cmath>
#include <cstdio>
#include <ginac/ginac.h>
#include "ggvvamp.h"
#include "ggvvampprec.h"

using namespace std;
#include "ampggVVprime.h"
#include "ampggVVprime_p.h"

using namespace GiNaC;

ggVVprime::ggVVprime() {
  mData = new ggVVprimePrivate();
}

ggVVprime::~ggVVprime() {
  delete mData;
}

void ggVVprime::initAmplitude(int Nf) {
  mData->amp = new ggvvampprec(Nf, IRschemeQt);
  mData->amp->maxrel = 1e-6;
}

void ggVVprime::computeAmplitude(double s, double t, double ma2, double mb2) {
  mData->amp->compute(s,t,ma2,mb2);
}

// l = number of loops
// j = helicity 0=LL, 1=LR; see eqn. (3.12) E_i^{h1,h2}, h1,h2=LL or h1,h2=LR as independent configurations
// i = index of coefficient; i in E_i^{h1,h2} of eqn. (3.12)
void ggVVprime::getAmplitude(int l, int i, int j, double &Er, double &Ei) {
//   cout << i << ", " << j << ", " << l << endl;
  Er = mData->amp->E[j][i][l][0];
  Ei = mData->amp->E[j][i][l][1];
}
// 
//   int Nf = 5;
//   ggvvampprec amp(Nf, IRschemeQt);
//   amp.maxrel = 1e-6;
// 
// // alternatively use fixed precision evaluation:
// //  typedef quadtype T;
// //  typedef cln::cl_F T;
// //  typedef double T;
// //  GiNaC::Digits = 32;
// //  ggvvamp<T> amp(Nf, IRschemeQt);
// 
//   // read from command line
//   double s, t, ma2, mb2;
//   if (argc == 2 && strcmp(argv[1],"--benchmark") == 0) {
//     // benchmark point from 1408.6409
//     s = 15625.;
//     t = -2384.89727896833;
//     ma2 = 6467.215561;
//     mb2 = 625.;
//   } else if (argc != 1 + 4 ||
//       (sscanf(argv[1], "%lf", &s) != 1) || (sscanf(argv[2], "%lf", &t) != 1) ||
//       (sscanf(argv[3], "%lf", &ma2) != 1) || (sscanf(argv[4], "%lf", &mb2) != 1)) {
//     usage(argv[0]);
//     return 1;
//   }
//   
//   // for checks on crossings
//   //swap(ma2,mb2); // crossing x12x34
//   //t = ma2 + mb2 - s - t; // crossing x12
// 
//   amp.compute(s,t,ma2,mb2);
// 
//   cout << setprecision(15);
//   for (int l = 0 ; l < nloop; ++l) {
//     for (int i = 1 ; i <= nEcoeff; ++i) {
//       cout << l << " \t" << i;
//       for (int j = 0 ; j < ntype; ++j) {
//          cout << " \t" << amp.E[i][j][l][0];
//          cout << " \t" << amp.E[i][j][l][1];
//       }
//       cout << "\n";
//     }
//   }
//   cout << endl;
//   
//   return 0;  
// }
