#include <iostream>
#include <iomanip>
#include <complex>

#include "ME2/qqyyy_h2.h"

/**
 * Here we demonstrate how to use the library to evaluate the squared helicity/color summed finite remainder for the qq->yyy process.
 */

int main () {

    using namespace FivePointAmplitudes;

    // Perform evaluations in double precision in this program.
    // Can be switched to dd_real if the support of higher precision is compiled
    using T = double;

    //Construct H2, including the two available contributions: H2(2,0) and H2(2,Nf)
    qqyyy_h2<T> H2({true,false,true,false});


    // Now evaluate

    // Define kinematics by the values of {s12, s23, s34, s45, s15}

    std::array<T, 5> kin = {
        1.322499999999999795718963468971204450660825281407116617285652827,   -0.9941094982857642509598316394306294510649507329718547518082950804,
        0.2644715912591248749171852420867085059089492634565178442467326276,  0.2671260487335744070776761035463255162438335749571773920325166998,
        -0.8837952297598047718249741030261595723499834269647052179121872201,
    };

    std::cout << std::scientific << std::setprecision(17);

    // Evaluate H2, the second argument is the renormalization scale. Here we set it to s12 as an example.
    // The output is gauge-invaraint components of H2 as defined in the publication.
    // It is left to the user to combine it with appropriate color factors.
    auto h2_val = H2(kin, kin[0]);

    std::cout << "H(1) : " << h2_val[1][0] << std::endl;
    std::cout << "H(2,0) : " << h2_val[2][0] << std::endl;
    std::cout << "H(2,Nf) : " << h2_val[2][2] << std::endl;

    return 0;
}
