#pragma once
#include "2q3y_2l_pmmpp_nf_12345_coefficients.hpp"

template <typename T> class remainder_2q3y_2l_pmmpp_nf_12345 {
	static constexpr size_t NCOEFFS = 57;
	static constexpr size_t NFUNCS = 338;
	static constexpr size_t NLETTERS = 26;

	using TC = std::complex<T>;
	using RCoeffsType = T(*)(const std::array<T,5>&, const std::array<T,NLETTERS>&);
        using FLC = std::vector<std::pair<int,T>>;
        using TCoeffList = std::array<std::pair<FLC, FLC>, NFUNCS>;

	std::shared_ptr<FunctionsManager> f_manager;

	const std::array<RCoeffsType, NCOEFFS> coeff_basis;
	std::array<std::vector<std::pair<f_type,int>>, NFUNCS> functions_basis;
	// {index of coefficient, index of function, matrix element}
	TCoeffList coefficients;

  public:
	remainder_2q3y_2l_pmmpp_nf_12345();
	remainder_2q3y_2l_pmmpp_nf_12345(std::shared_ptr<FunctionsManager>);
	
	TC eval(const std::array<T,5>&, int, T);

  private:
        void construct_functions_basis();
        void construct_coefficient_list();
};

template <typename T> remainder_2q3y_2l_pmmpp_nf_12345<T>::remainder_2q3y_2l_pmmpp_nf_12345(std::shared_ptr<FunctionsManager> mng) : f_manager(mng),
	coeff_basis{
{eval_coef_2q3y_2l_pmmpp_nf_12345_1<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_2<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_3<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_4<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_5<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_6<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_7<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_8<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_9<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_10<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_11<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_12<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_13<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_14<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_15<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_16<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_17<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_18<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_19<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_20<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_21<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_22<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_23<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_24<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_25<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_26<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_27<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_28<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_29<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_30<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_31<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_32<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_33<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_34<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_35<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_36<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_37<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_38<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_39<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_40<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_41<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_42<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_43<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_44<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_45<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_46<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_47<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_48<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_49<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_50<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_51<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_52<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_53<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_54<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_55<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_56<T>,
eval_coef_2q3y_2l_pmmpp_nf_12345_57<T>}
	}
{
    construct_functions_basis();
    construct_coefficient_list();
}

template <typename T> remainder_2q3y_2l_pmmpp_nf_12345<T>::remainder_2q3y_2l_pmmpp_nf_12345() : remainder_2q3y_2l_pmmpp_nf_12345(std::make_shared<FunctionsManager>()) {}

template <typename T> void remainder_2q3y_2l_pmmpp_nf_12345<T>::construct_functions_basis() {
    auto input = MathList(read_file(std::string(DATADIR)+"/2q3y_2l_pmmpp_nf_12345_fmonomials.dat"));

    size_t n = 0;

    for (MathList mm : input) {
        for (MathList pp : mm) {
            f_type t = from_string(pp.head);
            switch (t) {
                case f_type::one:
                case f_type::str5: functions_basis.at(n).emplace_back(t, 1); break;
                case f_type::bc:
                case f_type::bci: functions_basis.at(n).emplace_back(t, std::stoul(pp[0])); break;
                case f_type::f: {
                    int id = 0;
                    if (pp.size() >= 3) { id = f_manager->add({std::stoi(pp[0]), std::stoi(pp[1]), std::stoi(pp[2])}); }
                    else if (pp.size() == 2) {
                        id = f_manager->add({std::stoi(pp[0]), std::stoi(pp[1])});
                    }
                    functions_basis.at(n).emplace_back(t, id);
                }
            }
        }
        ++n;
    }
}

template <typename T> void remainder_2q3y_2l_pmmpp_nf_12345<T>::construct_coefficient_list() {
    auto input = MathList(read_file(std::string(DATADIR)+"/2q3y_2l_pmmpp_nf_12345_mmatrix.dat"));

    size_t n = 0;

    for (MathList ppair : input) {
        std::vector<FLC> nentry;
        for (MathList lc : ppair) {
            if (lc.size() == 0) {
                nentry.push_back({});
            }
            else{
                FLC coeff;
                for(MathList prod : lc ){
                    coeff.emplace_back(std::stoi(prod[0]), get_rational_number<T>(prod[1]));
                }
                nentry.push_back(coeff);
            }
        }
        coefficients.at(n).first = nentry[0];
        coefficients.at(n).second = nentry[1];
        ++n;
    }
}

template <typename T> std::complex<T> remainder_2q3y_2l_pmmpp_nf_12345<T>::eval(const std::array<T,5>& v_in, int str5, T musq) {
    return eval_remainder<NLETTERS>(v_in, str5, musq, coeff_basis, functions_basis, coefficients, f_manager, v_in[0]);
}

extern template class remainder_2q3y_2l_pmmpp_nf_12345<double>;
#ifdef FivePointAmplitudes_QD_ENABLED 
extern template class remainder_2q3y_2l_pmmpp_nf_12345<FunctionsManager::RHP>;
#endif
