#pragma once
#include "2q3y_2l_pmppp_12345_coefficients.hpp"

template <typename T> class remainder_2q3y_2l_pmppp_12345 {
	static constexpr size_t NCOEFFS = 62;
	static constexpr size_t NFUNCS = 86;
	static constexpr size_t NLETTERS = 26;

	using TC = std::complex<T>;
	using RCoeffsType = T(*)(const std::array<T,5>&, const std::array<T,NLETTERS>&);
        using FLC = std::vector<std::pair<int,T>>;
        using TCoeffList = std::array<std::pair<FLC, FLC>, NFUNCS>;

	std::shared_ptr<FunctionsManager> f_manager;

	const std::array<RCoeffsType, NCOEFFS> coeff_basis;
	const std::array<std::vector<std::pair<f_type,int>>, NFUNCS> functions_basis;
	// {index of coefficient, index of function, matrix element}
	const TCoeffList coefficients;

  public:
	remainder_2q3y_2l_pmppp_12345();
	remainder_2q3y_2l_pmppp_12345(std::shared_ptr<FunctionsManager>);
	
	TC eval(const std::array<T,5>&, int, T);

  private:
        TCoeffList construct_coefficient_list();
};

template <typename T> remainder_2q3y_2l_pmppp_12345<T>::remainder_2q3y_2l_pmppp_12345(std::shared_ptr<FunctionsManager> mng) : f_manager(mng),
	coeff_basis{
{eval_coef_2q3y_2l_pmppp_12345_1<T>,
eval_coef_2q3y_2l_pmppp_12345_2<T>,
eval_coef_2q3y_2l_pmppp_12345_3<T>,
eval_coef_2q3y_2l_pmppp_12345_4<T>,
eval_coef_2q3y_2l_pmppp_12345_5<T>,
eval_coef_2q3y_2l_pmppp_12345_6<T>,
eval_coef_2q3y_2l_pmppp_12345_7<T>,
eval_coef_2q3y_2l_pmppp_12345_8<T>,
eval_coef_2q3y_2l_pmppp_12345_9<T>,
eval_coef_2q3y_2l_pmppp_12345_10<T>,
eval_coef_2q3y_2l_pmppp_12345_11<T>,
eval_coef_2q3y_2l_pmppp_12345_12<T>,
eval_coef_2q3y_2l_pmppp_12345_13<T>,
eval_coef_2q3y_2l_pmppp_12345_14<T>,
eval_coef_2q3y_2l_pmppp_12345_15<T>,
eval_coef_2q3y_2l_pmppp_12345_16<T>,
eval_coef_2q3y_2l_pmppp_12345_17<T>,
eval_coef_2q3y_2l_pmppp_12345_18<T>,
eval_coef_2q3y_2l_pmppp_12345_19<T>,
eval_coef_2q3y_2l_pmppp_12345_20<T>,
eval_coef_2q3y_2l_pmppp_12345_21<T>,
eval_coef_2q3y_2l_pmppp_12345_22<T>,
eval_coef_2q3y_2l_pmppp_12345_23<T>,
eval_coef_2q3y_2l_pmppp_12345_24<T>,
eval_coef_2q3y_2l_pmppp_12345_25<T>,
eval_coef_2q3y_2l_pmppp_12345_26<T>,
eval_coef_2q3y_2l_pmppp_12345_27<T>,
eval_coef_2q3y_2l_pmppp_12345_28<T>,
eval_coef_2q3y_2l_pmppp_12345_29<T>,
eval_coef_2q3y_2l_pmppp_12345_30<T>,
eval_coef_2q3y_2l_pmppp_12345_31<T>,
eval_coef_2q3y_2l_pmppp_12345_32<T>,
eval_coef_2q3y_2l_pmppp_12345_33<T>,
eval_coef_2q3y_2l_pmppp_12345_34<T>,
eval_coef_2q3y_2l_pmppp_12345_35<T>,
eval_coef_2q3y_2l_pmppp_12345_36<T>,
eval_coef_2q3y_2l_pmppp_12345_37<T>,
eval_coef_2q3y_2l_pmppp_12345_38<T>,
eval_coef_2q3y_2l_pmppp_12345_39<T>,
eval_coef_2q3y_2l_pmppp_12345_40<T>,
eval_coef_2q3y_2l_pmppp_12345_41<T>,
eval_coef_2q3y_2l_pmppp_12345_42<T>,
eval_coef_2q3y_2l_pmppp_12345_43<T>,
eval_coef_2q3y_2l_pmppp_12345_44<T>,
eval_coef_2q3y_2l_pmppp_12345_45<T>,
eval_coef_2q3y_2l_pmppp_12345_46<T>,
eval_coef_2q3y_2l_pmppp_12345_47<T>,
eval_coef_2q3y_2l_pmppp_12345_48<T>,
eval_coef_2q3y_2l_pmppp_12345_49<T>,
eval_coef_2q3y_2l_pmppp_12345_50<T>,
eval_coef_2q3y_2l_pmppp_12345_51<T>,
eval_coef_2q3y_2l_pmppp_12345_52<T>,
eval_coef_2q3y_2l_pmppp_12345_53<T>,
eval_coef_2q3y_2l_pmppp_12345_54<T>,
eval_coef_2q3y_2l_pmppp_12345_55<T>,
eval_coef_2q3y_2l_pmppp_12345_56<T>,
eval_coef_2q3y_2l_pmppp_12345_57<T>,
eval_coef_2q3y_2l_pmppp_12345_58<T>,
eval_coef_2q3y_2l_pmppp_12345_59<T>,
eval_coef_2q3y_2l_pmppp_12345_60<T>,
eval_coef_2q3y_2l_pmppp_12345_61<T>,
eval_coef_2q3y_2l_pmppp_12345_62<T>}
	},

	functions_basis{
{{{f_type::one,1}},
{{f_type::bci,0}},
{{f_type::f,f_manager->add({1,1,1})}},
{{f_type::f,f_manager->add({1,1,2})}},
{{f_type::f,f_manager->add({1,1,3})}},
{{f_type::f,f_manager->add({1,1,4})}},
{{f_type::f,f_manager->add({1,1,5})}},
{{f_type::f,f_manager->add({1,1,6})}},
{{f_type::f,f_manager->add({1,1,7})}},
{{f_type::f,f_manager->add({1,1,8})}},
{{f_type::f,f_manager->add({1,1,9})}},
{{f_type::f,f_manager->add({1,1,10})}},
{{f_type::f,f_manager->add({2,1,1})}},
{{f_type::f,f_manager->add({2,1,2})}},
{{f_type::f,f_manager->add({2,1,3})}},
{{f_type::f,f_manager->add({2,1,4})}},
{{f_type::f,f_manager->add({2,1,5})}},
{{f_type::f,f_manager->add({2,1,6})}},
{{f_type::f,f_manager->add({2,1,7})}},
{{f_type::f,f_manager->add({2,1,8})}},
{{f_type::f,f_manager->add({2,1,9})}},
{{f_type::f,f_manager->add({2,1,10})}},
{{f_type::f,f_manager->add({2,1,11})}},
{{f_type::f,f_manager->add({2,1,12})}},
{{f_type::f,f_manager->add({2,1,13})}},
{{f_type::f,f_manager->add({2,1,14})}},
{{f_type::f,f_manager->add({2,1,15})}},
{{f_type::bci,0}, {f_type::bci,0}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,1})}},
{{f_type::f,f_manager->add({1,1,2})}, {f_type::f,f_manager->add({1,1,2})}},
{{f_type::f,f_manager->add({1,1,3})}, {f_type::f,f_manager->add({1,1,3})}},
{{f_type::f,f_manager->add({1,1,4})}, {f_type::f,f_manager->add({1,1,4})}},
{{f_type::f,f_manager->add({1,1,5})}, {f_type::f,f_manager->add({1,1,5})}},
{{f_type::f,f_manager->add({1,1,6})}, {f_type::f,f_manager->add({1,1,6})}},
{{f_type::f,f_manager->add({1,1,7})}, {f_type::f,f_manager->add({1,1,7})}},
{{f_type::f,f_manager->add({1,1,8})}, {f_type::f,f_manager->add({1,1,8})}},
{{f_type::f,f_manager->add({1,1,9})}, {f_type::f,f_manager->add({1,1,9})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,1,1})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,1,2})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,1,5})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,1,6})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,1,7})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,1,9})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,1,10})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,2,2})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,2,3})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,2,6})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,2,7})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,2,8})}},
{{f_type::bci,0}, {f_type::f,f_manager->add({1,2,9})}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,2})}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,3})}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,4})}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,5})}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,6})}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,7})}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,8})}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,9})}},
{{f_type::f,f_manager->add({1,1,1})}, {f_type::f,f_manager->add({1,1,10})}},
{{f_type::f,f_manager->add({1,1,2})}, {f_type::f,f_manager->add({1,1,3})}},
{{f_type::f,f_manager->add({1,1,2})}, {f_type::f,f_manager->add({1,1,4})}},
{{f_type::f,f_manager->add({1,1,2})}, {f_type::f,f_manager->add({1,1,5})}},
{{f_type::f,f_manager->add({1,1,2})}, {f_type::f,f_manager->add({1,1,8})}},
{{f_type::f,f_manager->add({1,1,2})}, {f_type::f,f_manager->add({1,1,9})}},
{{f_type::f,f_manager->add({1,1,3})}, {f_type::f,f_manager->add({1,1,4})}},
{{f_type::f,f_manager->add({1,1,3})}, {f_type::f,f_manager->add({1,1,5})}},
{{f_type::f,f_manager->add({1,1,3})}, {f_type::f,f_manager->add({1,1,6})}},
{{f_type::f,f_manager->add({1,1,3})}, {f_type::f,f_manager->add({1,1,7})}},
{{f_type::f,f_manager->add({1,1,3})}, {f_type::f,f_manager->add({1,1,8})}},
{{f_type::f,f_manager->add({1,1,3})}, {f_type::f,f_manager->add({1,1,9})}},
{{f_type::f,f_manager->add({1,1,3})}, {f_type::f,f_manager->add({1,1,10})}},
{{f_type::f,f_manager->add({1,1,4})}, {f_type::f,f_manager->add({1,1,5})}},
{{f_type::f,f_manager->add({1,1,4})}, {f_type::f,f_manager->add({1,1,6})}},
{{f_type::f,f_manager->add({1,1,4})}, {f_type::f,f_manager->add({1,1,7})}},
{{f_type::f,f_manager->add({1,1,4})}, {f_type::f,f_manager->add({1,1,8})}},
{{f_type::f,f_manager->add({1,1,4})}, {f_type::f,f_manager->add({1,1,9})}},
{{f_type::f,f_manager->add({1,1,4})}, {f_type::f,f_manager->add({1,1,10})}},
{{f_type::f,f_manager->add({1,1,5})}, {f_type::f,f_manager->add({1,1,7})}},
{{f_type::f,f_manager->add({1,1,5})}, {f_type::f,f_manager->add({1,1,8})}},
{{f_type::f,f_manager->add({1,1,6})}, {f_type::f,f_manager->add({1,1,7})}},
{{f_type::f,f_manager->add({1,1,6})}, {f_type::f,f_manager->add({1,1,8})}},
{{f_type::f,f_manager->add({1,1,6})}, {f_type::f,f_manager->add({1,1,10})}},
{{f_type::f,f_manager->add({1,1,7})}, {f_type::f,f_manager->add({1,1,8})}},
{{f_type::f,f_manager->add({1,1,8})}, {f_type::f,f_manager->add({1,1,9})}},
{{f_type::f,f_manager->add({1,1,8})}, {f_type::f,f_manager->add({1,1,10})}},
{{f_type::f,f_manager->add({1,1,9})}, {f_type::f,f_manager->add({1,1,10})}}}
	},

	coefficients{ construct_coefficient_list() }
{}

template <typename T> remainder_2q3y_2l_pmppp_12345<T>::remainder_2q3y_2l_pmppp_12345() : remainder_2q3y_2l_pmppp_12345(std::make_shared<FunctionsManager>()) {}

template <typename T> typename remainder_2q3y_2l_pmppp_12345<T>::TCoeffList remainder_2q3y_2l_pmppp_12345<T>::construct_coefficient_list() {
    TCoeffList toret;

toret[0] = {{{61, 1}},{{60, 1}}};
toret[1] = {{{5, -1},
{13, -1},
{14, -1},
{42, -1}},{{0, -1},
{3, -1},
{4, -1},
{27, -1}}};
toret[2] = {{{42, 1}},{{27, 1}}};
toret[3] = {{{55, 1}},{{49, 1}}};
toret[4] = {{{13, 1}},{{3, 1}}};
toret[5] = {{{5, 1}},{{0, 1}}};
toret[6] = {{{56, 1}},{{51, 1}}};
toret[7] = {{{5, -1},
{13, -1},
{14, -1},
{42, T(13) / T(9)},
{55, -1},
{56, -1},
{57, -1},
{58, -1},
{59, -1}},{{0, -1},
{3, -1},
{4, -1},
{27, T(13) / T(9)},
{49, -1},
{51, -1},
{52, -1},
{53, -1},
{54, -1}}};
toret[8] = {{{57, 1}},{{52, 1}}};
toret[9] = {{{14, 1}},{{4, 1}}};
toret[10] = {{{58, 1}},{{53, 1}}};
toret[11] = {{{59, 1}},{{54, 1}}};
toret[12] = {{{39, -1},
{44, -1},
{46, 1},
{47, 1}},{{20, -1},
{29, -1},
{33, 1},
{35, 1}}};
toret[13] = {{{28, -1},
{40, -1},
{41, -1},
{46, 1},
{48, 1},
{50, -2}},{{12, -1},
{23, -1},
{24, -1},
{33, 1},
{36, 1},
{43, -2}}};
toret[14] = {{{6, -1},
{7, -1},
{28, -1}},{{1, -1},
{2, -1},
{12, -1}}};
toret[15] = {{{6, -1},
{31, -1}},{{1, -1},
{15, -1}}};
toret[16] = {{{31, -1},
{34, -1},
{44, 1}},{{15, -1},
{17, -1},
{29, 1}}};
toret[17] = {{{50, 2}},{{43, 2}}};
toret[18] = {{{25, -1},
{34, 1}},{{10, -1},
{17, 1}}};
toret[19] = {{{26, -1},
{40, 1}},{{11, -1},
{23, 1}}};
toret[20] = {{{38, -1},
{45, -1},
{47, 1},
{48, 1}},{{19, -1},
{30, -1},
{35, 1},
{36, 1}}};
toret[21] = {{{25, 1},
{26, 1},
{38, -1}},{{10, 1},
{11, 1},
{19, -1}}};
toret[22] = {{{7, -1},
{32, -1}},{{2, -1},
{16, -1}}};
toret[23] = {{{32, -1},
{37, -1},
{45, 1}},{{16, -1},
{18, -1},
{30, 1}}};
toret[24] = {{{21, -1},
{22, -1},
{39, 1}},{{8, -1},
{9, -1},
{20, 1}}};
toret[25] = {{{21, -1},
{41, 1}},{{8, -1},
{24, 1}}};
toret[26] = {{{22, -1},
{37, 1}},{{9, -1},
{18, 1}}};
toret[27] = {{{21, T(-1) / T(6)},
{22, T(-1) / T(6)},
{28, T(-1) / T(6)},
{31, T(-1) / T(6)},
{32, T(-1) / T(6)},
{34, T(-1) / T(6)},
{37, T(-1) / T(6)},
{38, T(-1) / T(6)},
{40, T(-1) / T(6)},
{41, T(-1) / T(6)},
{46, T(1) / T(6)},
{47, T(1) / T(6)},
{48, T(1) / T(6)}},{{8, T(-1) / T(6)},
{9, T(-1) / T(6)},
{12, T(-1) / T(6)},
{15, T(-1) / T(6)},
{16, T(-1) / T(6)},
{17, T(-1) / T(6)},
{18, T(-1) / T(6)},
{19, T(-1) / T(6)},
{23, T(-1) / T(6)},
{24, T(-1) / T(6)},
{33, T(1) / T(6)},
{35, T(1) / T(6)},
{36, T(1) / T(6)}}};
toret[28] = {{{28, T(-1) / T(2)},
{38, T(-1) / T(2)},
{39, T(-1) / T(2)},
{40, T(-1) / T(2)},
{41, T(-1) / T(2)},
{44, T(-1) / T(2)},
{45, T(-1) / T(2)},
{46, 1},
{47, 1},
{48, 1},
{50, -1}},{{12, T(-1) / T(2)},
{19, T(-1) / T(2)},
{20, T(-1) / T(2)},
{23, T(-1) / T(2)},
{24, T(-1) / T(2)},
{29, T(-1) / T(2)},
{30, T(-1) / T(2)},
{33, 1},
{35, 1},
{36, 1},
{43, -1}}};
toret[29] = {{{6, -1},
{7, -1},
{28, T(-1) / T(2)},
{31, T(-1) / T(2)},
{32, T(-1) / T(2)}},{{1, -1},
{2, -1},
{12, T(-1) / T(2)},
{15, T(-1) / T(2)},
{16, T(-1) / T(2)}}};
toret[30] = {{{21, T(-1) / T(2)},
{22, T(-1) / T(2)},
{31, T(-1) / T(2)},
{34, T(-1) / T(2)},
{39, T(1) / T(2)},
{44, T(1) / T(2)}},{{8, T(-1) / T(2)},
{9, T(-1) / T(2)},
{15, T(-1) / T(2)},
{17, T(-1) / T(2)},
{20, T(1) / T(2)},
{29, T(1) / T(2)}}};
toret[31] = {{{50, 1}},{{43, 1}}};
toret[32] = {{{25, T(-1) / T(2)},
{34, T(1) / T(2)}},{{10, T(-1) / T(2)},
{17, T(1) / T(2)}}};
toret[33] = {{{21, T(-1) / T(2)},
{26, T(-1) / T(2)},
{40, T(1) / T(2)},
{41, T(1) / T(2)}},{{8, T(-1) / T(2)},
{11, T(-1) / T(2)},
{23, T(1) / T(2)},
{24, T(1) / T(2)}}};
toret[34] = {{{25, T(1) / T(2)},
{26, T(1) / T(2)},
{38, T(-1) / T(2)}},{{10, T(1) / T(2)},
{11, T(1) / T(2)},
{19, T(-1) / T(2)}}};
toret[35] = {{{32, T(-1) / T(2)},
{37, T(-1) / T(2)},
{45, T(1) / T(2)}},{{16, T(-1) / T(2)},
{18, T(-1) / T(2)},
{30, T(1) / T(2)}}};
toret[36] = {{{22, T(-1) / T(2)},
{37, T(1) / T(2)}},{{9, T(-1) / T(2)},
{18, T(1) / T(2)}}};
toret[37] = {{{28, 1},
{38, 1},
{39, 1},
{40, 1},
{41, 1},
{44, 1},
{45, 1},
{50, 2}},{{12, 1},
{19, 1},
{20, 1},
{23, 1},
{24, 1},
{29, 1},
{30, 1},
{43, 2}}};
toret[38] = {{{31, -1},
{32, -1}},{{15, -1},
{16, -1}}};
toret[39] = {{{6, 1},
{25, -1}},{{1, 1},
{10, -1}}};
toret[40] = {{{21, -1},
{26, -1}},{{8, -1},
{11, -1}}};
toret[41] = {{{34, -1},
{40, -1}},{{17, -1},
{23, -1}}};
toret[42] = {{{7, 1},
{22, -1}},{{2, 1},
{9, -1}}};
toret[43] = {{{37, -1},
{41, -1}},{{18, -1},
{24, -1}}};
toret[44] = {{{21, 1},
{22, 1},
{39, -1}},{{8, 1},
{9, 1},
{20, -1}}};
toret[45] = {{{50, -2}},{{43, -2}}};
toret[46] = {{{6, -1},
{7, -1},
{28, -1}},{{1, -1},
{2, -1},
{12, -1}}};
toret[47] = {{{31, 1},
{34, 1},
{44, -1}},{{15, 1},
{17, 1},
{29, -1}}};
toret[48] = {{{25, 1},
{26, 1},
{38, -1}},{{10, 1},
{11, 1},
{19, -1}}};
toret[49] = {{{32, 1},
{37, 1},
{45, -1}},{{16, 1},
{18, 1},
{30, -1}}};
toret[50] = {{{28, 1}},{{12, 1}}};
toret[51] = {{{46, -1},
{47, -1}},{{33, -1},
{35, -1}}};
toret[52] = {{{46, -1},
{48, -1}},{{33, -1},
{36, -1}}};
toret[53] = {{{44, 1}},{{29, 1}}};
toret[54] = {{{40, 1},
{41, 1},
{50, 2}},{{23, 1},
{24, 1},
{43, 2}}};
toret[55] = {{{38, 1}},{{19, 1}}};
toret[56] = {{{47, -1},
{48, -1}},{{35, -1},
{36, -1}}};
toret[57] = {{{45, 1}},{{30, 1}}};
toret[58] = {{{39, 1}},{{20, 1}}};
toret[59] = {{{31, 1}},{{15, 1}}};
toret[60] = {{{6, 1},
{7, 1}},{{1, 1},
{2, 1}}};
toret[61] = {{{6, 1}},{{1, 1}}};
toret[62] = {{{32, 1}},{{16, 1}}};
toret[63] = {{{7, 1}},{{2, 1}}};
toret[64] = {{{46, 1}},{{33, 1}}};
toret[65] = {{{44, -1}},{{29, -1}}};
toret[66] = {{{21, 1}},{{8, 1}}};
toret[67] = {{{34, 1}},{{17, 1}}};
toret[68] = {{{47, 1}},{{35, 1}}};
toret[69] = {{{22, 1}},{{9, 1}}};
toret[70] = {{{39, -1}},{{20, -1}}};
toret[71] = {{{6, -1}},{{1, -1}}};
toret[72] = {{{40, -1},
{41, -1},
{50, -2}},{{23, -1},
{24, -1},
{43, -2}}};
toret[73] = {{{40, 1}},{{23, 1}}};
toret[74] = {{{48, 1}},{{36, 1}}};
toret[75] = {{{7, -1}},{{2, -1}}};
toret[76] = {{{41, 1}},{{24, 1}}};
toret[77] = {{{34, -1}},{{17, -1}}};
toret[78] = {{{25, 1}},{{10, 1}}};
toret[79] = {{{40, -1}},{{23, -1}}};
toret[80] = {{{26, 1}},{{11, 1}}};
toret[81] = {{{41, -1}},{{24, -1}}};
toret[82] = {{{25, -1},
{26, -1}},{{10, -1},
{11, -1}}};
toret[83] = {{{45, -1}},{{30, -1}}};
toret[84] = {{{37, 1}},{{18, 1}}};
toret[85] = {{{37, -1}},{{18, -1}}}
    ;

    return toret;
}

template <typename T> std::complex<T> remainder_2q3y_2l_pmppp_12345<T>::eval(const std::array<T,5>& v_in, int str5, T musq) {
    return eval_remainder<NLETTERS>(v_in, str5, musq, coeff_basis, functions_basis, coefficients, f_manager, v_in[0]);
}

extern template class remainder_2q3y_2l_pmppp_12345<double>;
#ifdef FivePointAmplitudes_QD_ENABLED 
extern template class remainder_2q3y_2l_pmppp_12345<FunctionsManager::RHP>;
#endif
